/* \make blend

see `blendpar.h' for VERSION number !

Read the MACSIMUS manual first!

This is the main for `blend'.  There is no `blend.h'.

To compile the standard version, do 
  makemake linux lj gcc
  make blend
otherwise see `metamake'

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%% */

#include "ground.h"
#include "alloc.h"

#include "vector.h"
#include "intrapot.h"
#include "sitesite.h"
#include "blendpar.h"
#include "blendmed.h"
#include "blendedt.h"
#include "blendgen.h"
#include "options.h"
#include "rndgen.h"

#ifdef X11
#include "draw.h"
#endif

#ifdef POLAR
#define THISVERSION VERSION"-POLAR"
#else
#define THISVERSION VERSION
#endif

#ifdef TINY
#define HDR "\
! BLEND "VERSION" TINY\n"
#else
#define HDR "\
! ------ MACSIMUS / BLEND " THISVERSION " ------ (c) J.Kolafa 1993-2012 ------\n"
#endif

int measure; /* to make intrapot.c happy */

/***
list of options:
  use x for undefined option, default value for valid option.
options are not case sensitive!
-X -X+ -X1 set option X to value 1
-X- -X0    clears option X to value 0
***/

#define x badoption
#define Q -9 /* inactive option */
int optionlist[32] =
/*   `  a   b  c d e f g h i j k   l    m n o p   q r s t u v w   x y z { |   }   ~ ? */
  {100,100,100,0,0,0,0,0,0,5,0,3,256,-100,0,0,1,100,0,0,0,0,0,0,100,3,0,0,Q,100,100,6};
/*   @  A   B  C D E F G H I J K   L    M N O P   Q R S T U V W   X Y Z [ \   ]   ^ _ */
#undef x

#ifndef TINY
static void AnotA(char *arg)
{
  if (arg[2]) ERROR(("option %s : arg not allowed",arg))
}
#endif

static int iframe(char *arg)
{
  int i=atoi(arg);

  if (i==0) ERROR(("no frame 0: 1st frame is 1"))
  else if (i>0) i--;

  return i;
}

static void mwzero(void)
{
  if (option('w')) prt("! option -w set to zero");
  if (option('m')) prt("! option -m set to zero");
  if (option('p')) prt("! option -p set to zero");
  option('p')=option('w')=option('m')=0;
}

int main(int narg,char **arg) /*************************************** main */
{
  int i,nmol=0,che,output_file=0;
  species_t *spec=NULL;
  char *outfn=NULL;
  int plbframe=0; /* frame # for plb read  (option -r4), see also Xopt */
    /* NOTE:  1=1st, ...  -1=last, ...; 0=1st */

#ifndef TINY
#include "blendhlp.c"
#endif

#if CHECKHEAP==-2
  AllocRange=64;
#endif

#if !defined(DOS) && !defined(X11) && !defined(EMX)
  option('g')=badoption;
#endif

#ifdef X11
  {
    static char title[82]="blend";
    static char icon[32];
    int i;
    char *t=title+5,*c;

#include "blendbit.c"
#ifdef GUI
  gui.exe="gshow";
  gui.name="blend";
#endif

    xwindowhints.icowidth=blendbit_width;
    xwindowhints.icoheight=blendbit_height;
    xwindowhints.icobits=blendbit_bits;
  
    /* args pre-parsing: -o name and geometry */
    loop (i,1,narg) {
      /* default -v is -v3 with -o, otherwise -v0 */
      if (arg[i][0]=='-' && arg[i][1]=='o') option('v')=3;
      if (arg[i][0]!='-' || arg[i][1]=='o') {
        *t++=' '; c=arg[i];
        while (t<title+80 && *c) *t++=*c++; }
      else if (arg[i][1]=='g')
        if ( (c=strpbrk(arg[i]," _")) ) xwindowhints.geometry=c+1; }

    *t=0;
  
    if (t==title+80) copy(t-3,"..",3);
    xwindowhints.winname=title;
    copy(icon,title,32);
    if (icon[31]!=0) copy(icon+29,"..",3);
    xwindowhints.iconame=icon;
  }
#endif

  initscroll(0);
  rndinit(7,0);

  fputs(HDR,stderr);
  
  /*** print help if no argument ***/
  if (narg<2) {
#ifdef TINY
    return 1;
#else
    prts(usage);
  
    for (;;) {
      char s[8];
  
      prts_(help);
      fgets(s,8,stdin);
      switch (s[0]) {
        case 'g': prts(gen_opt); break;
        case 'p': prts(par_opt); break;
        case 'm': prts(mol_opt); break;
        case 'x': prts(extra_mol_opt); break;
        case 'f': prts(mol_files); break;
        default: return 0; } }
#endif
  }

/*** argument parsing + reading mol+cfg ***/

  loop (i,1,narg) {

    if (arg[i][0]=='-') {

#ifndef TINY
      if (isupper(arg[i][1])) switch (arg[i][1]) {
        case 'A': Xopt.A=1; AnotA(arg[i]); 
#if 0 /* OBSOLETE */
                  if (option('d')) {
                    WARNING(("all_dihedrals (option -d) changed to 0 because of option -A"))
                    option('d')=0; }
#endif
                  mwzero();
                  break;
        case 'E': Xopt.E=toupper(arg[i][2]);
                  if (!Xopt.E) Xopt.E=' ';
                  if (!strchr("ABCH",Xopt.E)) ERROR(("bad -E%c option",Xopt.E))
                  mwzero();
                  break;
        case 'G': Xopt.G=1; 
                  mwzero();
                  break;
        case 'I': Xopt.I=atoi(arg[i]+2); break;
        case 'J': Xopt.Jeps=fabs(atof(arg[i]+2));
                  if (Xopt.Jeps==0) Xopt.Jeps=1e-11;
                  if (option('v')&4) Xopt.Jeps=-Xopt.Jeps;
                  break;
        case 'N': Xopt.N=1;
                  Xopt.dr=atof(arg[i]+2); if (Xopt.dr==0) Xopt.dr=1e-5; break;
        case 'D': Xopt.D=1; AnotA(arg[i]); 
                  mwzero();
                  break;
        case 'P': Xopt.P=atoi(arg[i]+2); break;
        case 'F': Xopt.F=atoi(arg[i]+2);
                  if (abs(Xopt.F)<2) ERROR(("-F%d too small",Xopt.F))
                  break;
        case 'M': Xopt.amplitude=atof(arg[i]+2);
                  if (Xopt.amplitude==0) Xopt.amplitude=1; break;
        case 'C': Xopt.C=atoi(arg[i]+2); break;
        case 'V': Xopt.V=atoi(arg[i]+2); break;
        case 'T': Xopt.T=atof(arg[i]+2); break;
        case 'R': Xopt.core=atof(arg[i]+2); break;
        default: ERROR(("%s is bad option",arg[i])) }
      else
#endif

      getoption(arg[i]) /* see options.h */

      switch (arg[i][1]) {
        case 'q': 
          chargewarned=1; 
          break;
        case 's':
          if (output_file) ERROR(("-s misplaced or used with -o"))
          initscroll((option('s')+(option('s')==1)*30)*1024);
          break;
        case 'o':
          if (output_file) ERROR(("misplaced or duplicate -o"))
          ble_file=output_file=option('o')=1;
          if (option('s')) ERROR(("-o misplaced or used with -s"))
          if (arg[i][2]) {
            alloc(outfn,strlen(arg[i])+3); /* + extension maybe */
            strcpy(outfn,arg[i]+2); }
          else if (arg[++i]) {
            alloc(outfn,strlen(arg[i])+5); /* + extension maybe */
            strcpy(outfn,arg[i]); }
          else
            ERROR(("option -o without file name"))
          break;
        case 'z':
          rndinit(7,option('z'));
          break;
        case 'r': {
          char *ch=strchr(arg[i],':');
          if (ch) {
            plbframe=iframe(++ch);
#ifndef TINY
     	    Xopt.ropt=0;
            Xopt.bdihedral[0]=atof(ch);
            ch=strchr(ch,':');
            if (ch) {
	      Xopt.ropt=1;
              Xopt.toframe=iframe(++ch);
	      Xopt.bdihedral[1]=atof(ch);
              if ( (ch=strchr(ch,':')) ) {
                Xopt.byframe=Xopt.bdihedral[2]=atof(++ch);
                if (option('r')<5) if (Xopt.byframe<=0)
                  ERROR(("illegal by frame %d",Xopt.byframe))
                if ( (ch=strchr(ch,':')) ) Xopt.fn=ch+1; } }
#endif
          } }
#ifndef TINY
          break;
        case 'w': {
          char *ch=strchr(arg[i],':');
          if (ch) PDBstyle=atoi(ch+1); }
#endif
      } 
    }

    else /* no -option ==> file name */ {
      char *fn,*c;
      enum { NONE,PAR,BIN,CHE,MOL,PCH,REA } ext;
  
      if (option('o')) {
        /* this means that previous option was `-oNAME' or `-o NAME': open .ble */
        char *c;
        int hasext=1;
        for (c=outfn; *c; c++)
          if (*c==*SLASH) hasext=1;   /* if the directory name contains '.' */
          else if (*c=='.') hasext=0; /* extension */
        if (hasext) strcat(outfn,".ble");
        out=fopen(outfn,"wt");
        prts(HDR); }
      option('o')=0;
      if (out==NULL) ERROR(("cannot write to %s",outfn?outfn:"stdout"))
      output_file=1; /* output file MUST be opened here */
  
      /* making a copy: some functions alter extension, 
         some functions however use arg[i] directly */
      alloc(fn,strlen(arg[i])+5);
      strcpy(fn,arg[i]);
      /* file name extension: assumed 3 letters long */
      c=fn+strlen(fn)-4;
  
      ext=NONE;
      if (c>fn)
        if      (!strcmp(c,".par")) ext=PAR;
        else if (!strcmp(c,".bin")) ext=BIN;
        else if (!strcmp(c,".che")) ext=CHE;
        else if (!strcmp(c,".mol")) ext=MOL;
        else if (!strcmp(c,".pch")) ext=PCH;
        else if (!strcmp(c,".rea")) ext=REA;
  
      che=0;
      switch (ext) {
  
        case BIN:
          if (!rwBIN(fn,NULL)) /* read bin */
            ERROR(("parameter file %s not found (check BLENDPATH)",fn))
          break;
  
        case PAR:
          readPAR(fn);
          if (option('b')==1)
            rwBIN(fn,c); /* make bin */
          break;
  
        case PCH:
          if (PCHname)
            ERROR(("%s %s multiply defined partial charge file",PCHname,arg[i]))
          PCHname=arg[i];
          break;
  
        case REA:
          if (REAname)
            ERROR(("%s %s multiply defined reaction file",REAname,arg[i]))
          REAname=arg[i];
          break;
  
        case MOL:
          spec=readMOL(fn);
          if (!spec) ERROR(("mol-file %s not found",fn))
          break;
  
        case NONE:
          strcat(fn,".mol");
          spec=readMOL(fn);
          if (spec) break;
          strcpy(fn,arg[i]); strcat(fn,".che");
        case CHE:
          spec=readCHE(fn);
          che++;
          if (!spec) ERROR(("che-file %s not found",fn))
          break;
        }

#ifdef POLAR
      scalepolarizabilities(option('~')*0.01,option(']')*0.01);
#endif

      scalebonds();
      scaleangles();
  
      if (spec) {
        int or=abs(option('r'));
  
        nmol++;
        spec->frame=plbframe;
        switch (or) {
          case 0: if (read3D(spec,1)) break; /* try .plb; new in V2.1a */
                  if (read3D(spec,3)) break; /* try .pla; new in V2.1a */
                  if (read3D(spec,0)) break; /* try .3db (deprecated) */
                  if (read3D(spec,2)) break; /* try .3dt (deprecated) */
                  if (che) { che=0; break; }
                  if (read2D(spec,"rb")) break; /* try .2db (deprecated) */
                  if (read2D(spec,"rt")) break; /* try .2db (deprecated) */
          case 1: randomcfg(spec); che=0; break;
#ifndef TINY
          case 5: alphahelixold(spec); che=0; break;
          case 6: alphahelix(spec); che=0; break;
          case 7: makechain(spec); che=0; break;
#endif
          case 2: 
          case 12: if (!che) if (!read2D(spec,or==2?"rb":"rt")) goto rerr;
                  che=0; break;
          case 3: 
          case 4: 
         case 13: if (read3D(spec,or==3?0:or==4?1:2)) break;
            rerr: ERROR(("no %s",spec->fn))
                  break;
          default: ERROR(("option -r%d bad value",option('r')))
          }
        if (che>0) WARNING(("%s: 3D cfg with *.che is suspicious",fn))
  
        autobonds(spec,option('d'));
  
        chiralities(spec,option('c'));
        partialcharges(spec);
        
        if (spec->wrmol<0) {
          if (!spec->opt_w) ERROR(("\
%s: mol-file has been renumbered and will be (re)written\n\
but no configuration will be saved because of option -w0",spec->fn))
          else if (abs(spec->opt_w)>1) WARNING(("\
%s: mol-file has been renumbered and will be (re)written\n\
but only text (.3dt) configuration will be saved: use -r13 next time",spec->fn))

          } } /* if (spec) */

      free(fn); } }
      

  if (option('b')!=1) {
    if (nmol<=0) WARNING(("no molecular file specified - nothing to do"))
    if (nspec<=0) WARNING(("no molecule found - nothing to do")) }
  if (nspec<=0) return 0;

  _n

  if (abs(option('_'))<4 || abs(option('_'))>12)
    WARNING(("option -_%d not in recommended range [4..12]",option('_')))
  fillprec();

  /* list all sites */
  listofsites();

  if (Xopt.V) {
    switch (nspec) {
      case 1: newemptyspec(); appendspec(spec0->next,spec0);
              *(spec0->ext)=0; prt("%s duplicated",spec0->fn);
      case 2: newemptyspec(); appendspec(spec0->next->next,spec0); appendspec(spec0->next->next,spec0->next);
              *(spec0->ext)=0; prt("%s+%s merged",spec0->fn,spec0->next->fn);
              break;
      case 3: prt("virial warning: the 3rd species specified explicitly, 3=1+2 assumed");
              break;
      default: ERROR(("virial: bad number of species")); }
    if (nspec!=3) ERROR(("nspec !=3"))
    spec0->Xopt.V=0;
    spec0->next->Xopt.V=0; 
    spec0->next->next->opt_m=0; }

  /* build force field for all species (molecules) */
  for (spec=spec0; spec; spec=spec->next) {
    react(spec,REAname);
    if (spec->edit) moledit(spec);

    build(spec);
    
    if (option('i')) {
      prt("! max angle energy = %g kcal/mol",Uanglemax);
      if (Uanglemax>option('i')) WARNING(("too large angle energy (%g kcal/mol) around site %d\n\
*** (check, e.g., pyramidal error of tetrahedral angles)",Uanglemax,ianglemax)) }      

    /* force fix charge if not rounded and option -_ */
    if (option('_')<0 && fabs(roundedcharge(spec,-1))>1e-13) spec->wrmol++;

    if (spec->wrmol) {
      if (spec->opt_p || spec->opt_w) writeMOL(spec); 
      else 
#ifndef TINY
          if (!spec->probe.ns)
#endif
          WARNING(("%s: mol-file not updated because neither -w nor -p",spec->fn)) }
    if (spec->opt_p) write3D(spec,3);
    if (spec->opt_w) write3D(spec,0); }
    
  if (option('s')) {
    fprintf(stderr,"quit by \';\' or scroll (\'$?\' for help)\n");
    getdata enddata }
  
  _n
  if (nspec!=newnspec) prt("! molecule(s) split: new nspec=%d",newnspec);
  fclose(out);
  
  /* this is really very nasty patch: */
  if (ble_file) if (nspec!=newnspec) {
    char s[256];
    FILE *oout;
    
    /* # of species changed because of molecule split: rewrite it in .ble file */
    if (rename(outfn,tempname()) | !(oout=fopen(tempname(),"rt")))
      ERROR(("cannot rename %s to %s",outfn,tempname()))
    out=fopen(outfn,"wt");
    if (!out) ERROR(("cannot rewrite %s",outfn))
    while (fgets(s,256,oout)) {
      /* rewrite the correct # of species */
      if (!memcmp("nspec=",s,6)) sprintf(s+6,"%d !fixed\n",newnspec);
      prts_(s); }
    fclose(out);
    unlink(tempname()); }

  return 0;
}
