/*.....#include "cpmark.h"*/

void inertiamatrix(species_t *spec) /************************* inertiamatrix */
{
  int ns=spec->ns,i,j,k;
  FILE *cp,*plb;
  float rcp[7];
  unsigned it;
  double M,m;
  vector CM,IM[3],R[3],R0[3];
  static int ord0[3]={0,1,2};
  double *passIM[3],*passR[3];
  
  if (sizeof(double)*3 != sizeof(vector)) ERROR(("vector bad length"))
  
  site=spec->site;
  if (abs(option('r'))!=4) ERROR(("-r4 or -r-4 required"))
  
  loop (j,0,3) {
    passIM[j]=IM[j];
    passR[j]=R[j];
    loop (k,0,3) R0[j][k]=j==k; }
  
  cp=fopen("inertia.cp","wt");
  rcp[0]=CPmark;
  *(int4*)(&rcp[1])=7;
  /*          3---4---5---6---7--- */
  copy(rcp+2,"I_zzRg_xRg_yRg_zRgyr",(7-2)*4);
  fwrite(rcp,4,7,cp);
  
  plb=fopen("inertia.mol","wt");
  fputs("for inertia axes: see blend -G\n\
\n\
number_of_atoms = 6\n\
\n\
atoms\n\
! i  atom-id  a-type charge chir nbonds bound_atoms\n\
  0 X1         N     0.0000  0    1  1\n\
  1 X2         N     0.0000  0    1  0\n\
  2 Y1         CT    0.0000  0    1  3\n\
  3 Y2         CT    0.0000  0    1  2\n\
  4 Z1         ST    0.0000  0    1  5\n\
  5 Z2         ST    0.0000  0    1  4\n\
\n",plb);
  fclose(plb);
  prt("! inertia.mol created");
  
  plb=fopen("inertia.plb","wb");
  rcp[0]=6;
  rcp[1]=0;
  fwrite(rcp,4,2,plb);
  
  prt("! inertia matrix analysis");
  
  for (it=0;;it++) {
    if (spec->frame>spec->Xopt.toframe) break;
    if (!read3D(spec,-1)) break;
    spec->frame+=spec->Xopt.byframe;
    M=0;
    VO(CM,=0)
    memset(IM,0,sizeof(IM));
    loop (i,0,ns) {
      m=atom[site[i].type].mass;
      M+=m;
      VV(CM,+=m*site[i].r) }
    VO(CM,/=M)
    loop (i,0,ns) {
      vector dr;
      m=atom[site[i].type].mass;
      VVV(dr,=site[i].r,-CM)
      loop (j,0,3) loop (k,0,3) IM[j][k]+=m*dr[j]*dr[k]; }
    m=Jacobi(3,passIM,passR,spec->Xopt.Jeps);
  
    /* principal axes multiplied by "partial radii of gyration"
       then, the following Cartesian cross or `molecule':
  
            O
            |  O
            | /
            |/
       O----+----O
           /|
          / |
         O  |
            O
  
       where the masses of `atoms' O are 1/2 and the lengths of
       radiusvectors are R_x, R_y, R_z, has the same tensor (matrix) of
       inertia as the original molecule */
  
    prt("! diagonalized inertia matrix in g/mol*A^2: %g %g %g",IM[0][0],IM[1][1],IM[2][2]);
    prt("! principal moments of inertia in kg*m^2:   %g %g %g",
        (IM[2][2]+IM[1][1])*1.66053878e-47,
        (IM[0][0]+IM[2][2])*1.66053878e-47,
        (IM[0][0]+IM[1][1])*1.66053878e-47);
  
    loop (i,0,3)
      loop (j,0,3) R[i][j]*=sqrt(fmax(0,IM[i][i])/M);
  
    { /* this code reorders the indices so that the above mentioned
         cross is as similar as possible as the cross from the previous
         step.  Hence, the principal axes do not swap from step to step
         (if possible, of course, if two R_i become almost the same,
         this is impossible in principle)
         BUG: the sign of both vectors in each O---O `bond' is undefined.
            This does not cause any problems while `show'-ing, but might
            cause problems in further calculations */
         
      double s,ss=-9;
      static int ord[6][3]={{0,1,2},{0,2,1},{1,2,0},{1,0,2},{2,0,1},{2,1,0}};
      int iord=-1;
  
      loop (i,0,6) {
        s=0;
        loop (j,0,3) s+=fabs(SCAL(R0[ord0[j]],R[ord[i][j]]));
        if (s>ss) iord=i,ss=s; }
  
      if (iord<0) ERROR(("intertia matrix: numeric problem"))
      copy(ord0,ord[iord],sizeof(ord0)); }
  
    m=IM[0][0]+IM[1][1]+IM[2][2];
    loop (i,0,3) {
      double Mdiag=IM[ord0[i]][ord0[i]];
  
      if (Mdiag<0) Mdiag=0;
      rcp[i]=m-Mdiag;
      rcp[i+3]=sqrt(Mdiag/M); }
    rcp[6]=sqrt(m/M);
    fwrite(rcp,4,7,cp);
  
    loop (i,0,3) {
      loop (j,0,3)
        rcp[j+3]=-(rcp[j]=R[ord0[i]][j]);
      fwrite(rcp,4,6,plb); }
  
    copy(R0,R,sizeof(R)); }
  
  fclose(cp);
  fclose(plb);
  
  prt("! %d frames analyzed, inertia.cp and inertia.plb closed",it);
}
