/*
  inefficiency in blendmin.c:
  evaluation of Upot() is sometimes unnecessarily duplicated
  should store individual terms (dihedrals, ...) and print them separately
  arrays r,f,... unnecessarily allocated
  backcopy from rotated (by show) site[].r to r0 inefficient and cumbersome
*/

#define xCONTOURPLOT
/*
special patch for contour plot */

/* debug forces and cutoff */
#define xDEBUG

/* keep-keep omission, not supported under !CUTOFF */
/*.....#define OMITKEPT*/

#ifdef POLAR
#ifdef CUTOFF
/* WARNING CUTOFF turned off because of POLAR */
#undef CUTOFF
#endif /*# CUTOFF */
#else /*# POLAR */
#define CUTOFF /* cutting off the potential enabled */
#endif /*# POLAR */

/* %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%% BLENDMIN.C

This module evaluates energy and forces for a given configuration,
performs energy minimization, and in DOS/Turbo C and X11 versions
shows the molecule graphically.

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%% */

#define TRACE(X)
/*.....#define TRACE(X) prtc(X);*/

#define eyedist 0.35 /* stereo: ratio of the distance of eyes/screen height */

#include "ground.h"
#include "options.h"
#include "alloc.h"
#include "vector.h"
#include "intrapot.h"
#include "sitesite.h"
#include "blendpar.h"
#include "blendmed.h"
#include "blendmin.h"
#include "rndgen.h"
#include "units.h"


#define fullangle 0 /* 0:dihedral potential in (-180,180); 1: (0,360) */
typedef vector fvector;
#include "dihangle.c"

#if defined(DOS) || defined(X11) || defined(EMX)
#define GRAPHICS
#include "draw.h"
#include "mouselib.h"
#include "mousekey.c"
#endif /*# defined(DOS) || defined(X11) || defined(EMX) */

#ifdef EMX
int emxgrmode=1;
void mystartgraph(int dummy)
{
startgraph(emxgrmode);
}
#define startgraph mystartgraph
#endif /*# EMX */

#ifdef DOS
#include <conio.h>
#include "eprtscr.c"
#endif /*# DOS */
#include <time.h>

#ifdef POLAR
#include "log1.c"
static vector *mu; /* dipole moments */
#endif /*# POLAR */

static double Upot(vector *r,vector *f,species_t *spec,int toprint);

extern int measure;

static int keep;
static int testout=1; /* used by Upot and virial */

volatile int sig=0;
#include <signal.h>

#ifndef TINY
#include "jacobi.c"
#endif

static int autogr=0;
void asksig(int signr) /* ========================================== asksig */
{
#if defined(DOS) || defined(EMX)
  if (autogr) closegraph();
  autogr=0;
#endif /*# defined(DOS) || defined(EMX) */
  if (sig++) exit(-1); /* ^C typed twice */

/* refresh the ^C handler - nothing is compatible */
#ifdef __EMX__
  (*signal)(SIGINT,SIG_ACK);
#else /*# __EMX__ */
  (*signal)(SIGINT,asksig);
#endif /*# __EMX__ */
  fprintf(stderr,
    "\nINTERRUPT(%d) ACCEPTED, wait to finish cycle or type ^C again to exit immediately\n",
    signr);
}

bond_t *b0;
struct angle_s *a0;
torsion_t *d0,*i0,*ar0,*cis0;
#ifndef TINY
fixdih_t *fd0;
#endif

/* WARNING unit conversion: e^2 / (4 PI eps0) = 332.063709 A kcal/mol */
/* based on CODATA 2006 */
#define obfuscate 332.063709

static double ULJ,Uel,signpairs;
static int verbose2,maxpairs,checkpairs;
static maxpair_t *maxpair;
static site_t *site;

#include "blendss.c"
#include "blenddm.c" /* dipolemoment() */

#ifdef GRAPHICS
static double zeye,eye;
static int cross=2;

static enum label_e {
  NONE, ID, TYPE, NUMBER, CHARGE,
  POLARIZABILITY, SATURATION,
  /* DIPOLE_eA, DIPOLE_D must be at the end */
  DIPOLE_eA, DIPOLE_D
} label;
#define NBONDCOL 4
static int bondcol=0,bondcols[NBONDCOL]={0,YELLOW,WHITE,DARKGRAY};
static int col,grmode=2;
struct {
  double q,dq; /* charge to find, to be in <q-dq,q+dq> */
  int i;       /* atom # to find or found */
  int x,y;     /* x,y to find (pick by mouse) */
  int l;       /* effective length of s[] */
  char s[16];  /* id or type to find */
  vector r;    /* (old) position of site[i].r */
} find = { 1e3,1e-5,-1,-1 };

static char pickinfo[160];
static unsigned char ch;
static int lasti=-1,moving,ns,redraw,center=-1,marklabeled;

static void draw3D(double x,double y,double z) /******************** draw3D */
{
  double q=zeye/(zeye-z);

  draw(x*q,y*q);
}

static int match3D(double x,double y,double z,int ix,int iy) /***** match3D */
{
  double q=zeye/(zeye-z);
  int rr=5;

  if (grmode==2) rr="\36\5\13"[cross];
  if (grmode==3) rr="\52\13\36"[cross];

  ix-=SX(x*q); iy-=SY(y*q);

#if defined(DOS) || defined(EMX)
  return abs(ix)+abs(iy)<=2;
#else /*# defined(DOS) || defined(EMX) */
  return ix*ix+iy*iy<=rr;
#endif /*# defined(DOS) || defined(EMX) */
}

/* stereo */
static void draw3Ds(double x,double y,double z,int stereo) /******* draw3Ds */
{
  double q=zeye/(zeye-z);

  draw(x*q+(1-q)*stereo*eye,y*q);
}

static void floodmark(int center) /****************************** floodmark */
{
  int n,i;

  site[center].count=1;

  loop (n,0,site[center].nnbr)
    if (!site[i=site[center].nbr[n]].count) floodmark(i);
}

static void makelabel(char *Us,double nr)
{
  char *c;

  sprintf(Us,outtext_f,nr);
  c=strend(Us);
  while (*(--c)=='0') *c=0;
  if (*c=='.') *c=0;
}

static void showlabel(int i) /*********************************** showlabel */
{
  site_t *s=site+i;

  if (label) {
#ifdef POLAR
    static char Us[32],*c;
#else /*# POLAR */
    static char Us[8],*c;
#endif /*# POLAR */

    settextstyle(0,0,1); /*?*/
    settextjustify(1,1);
    if (grmode==1) {
      setcolor((bondcols[bondcol]&7)==CYAN ? YELLOW : LIGHTCYAN); }
    else
      setcolor(col<=BLUE ? YELLOW : col==LIGHTGRAY ? CYAN : BLUE);
    up();
    draw3D(s->r[0],s->r[1],s->r[2]);
    switch (label) {
    case CHARGE:
      if (find.q<999 && fabs(find.q-s->charge)>find.dq) goto pickup;
      makelabel(Us,s->charge);
      outtext(Us);
      if (find.q<999) find.i=i;
      break;
#ifdef POLAR
    case DIPOLE_eA:
    case DIPOLE_D:
      if (s->polar) {
	static float dipoleunit[2]={1,4.8032068};
	makelabel(Us,s->mu*dipoleunit[label==DIPOLE_D]);
/*.....      strcat(Us,label==DIPOLE_D?"D":"eA");*/
	outtext(Us); }
      break;
    case SATURATION:
      if (s->polar && s->sat!=0) {
	makelabel(Us,s->sat);
/*.....      strcat(Us,label==DIPOLE_D?"D":"eA");*/
	outtext(Us); }
      break;
    case POLARIZABILITY:
      if (s->polar) {
        if (!s->pol) strcpy(Us,"-");
        else if (s->polar&POL_ISO)
	  makelabel(Us,((isotropicparm_t*)s->pol)->alpha);
	else {
	  makelabel(Us,((axial_t*)s->pol)->parm.alpha);
	  strcat(Us,",");
	  makelabel(strend(Us),((axial_t*)s->pol)->parm.alphazz); }
	outtext(Us); }
      break;
#endif /*# POLAR */
    case NUMBER:
      if (find.i>=0 && find.i!=i) goto pickup;
      sprintf(Us,"%d",i); outtext(Us);
      break;
    default:
      c = label==ID ? s->id : atom[s->type].name;
      if (find.s[0] && memcmp(find.s,c,find.l)) goto pickup;
      outtext(c);
      if (find.s[0]) find.i=i;  }
    if (marklabeled) site[i].count=1; }

 pickup:

  if (find.x>=0 && match3D(s->r[0],s->r[1],s->r[2],find.x,find.y)) {
    static vector last;
    static int llasti=-1,lllasti=-1;
    double rr;
    static vector dummyf;
    static angleparm_t ap; /* zeros OK */

    if (ch==RIGHTCLICK) {
      site[i].count=0;
      redraw=1; }
    else {
      double psi=-999;

      if (lasti<0) { VV(last,=s->r) lasti=i; }
      rr=SQRD(last,s->r);
      phi=-1;
      measure=2;
      if (i!=lasti && llasti!=lasti && lasti>=0 && llasti>=0) {
	anglepot(s->r,last,site[llasti].r,dummyf,dummyf,dummyf,&ap);
	if (lllasti>=0 && lllasti!=llasti && lllasti!=lasti && lllasti!=i)
	  psi=dihedral(s->r,last,site[llasti].r,site[lllasti].r); }
      sprintf(pickinfo,
	      infoxyz_f,
	      i,s->id,atom[s->type].name,s->charge,s->r[0],s->r[1],s->r[2]);
      if (rr>0) sprintf(strend(pickinfo),infor_f,lasti,sqrt(rr));
      if (phi>=0) sprintf(strend(pickinfo),infoa_f,llasti,phi*(180/PI));
      if (psi>-999) sprintf(strend(pickinfo),infod_f,lllasti,psi);

      if (strlen(pickinfo)>79) {
	char *b=strchr(pickinfo,'(');
	char *e=strchr(pickinfo,')');

	if (b && e && b<e) memmove(b+1,e,strlen(e)+1); }

      VV(last,=s->r)
      lllasti=llasti; llasti=lasti; lasti=find.i=i;
      if (ch==LEFTCLICK || center<0) center=i;
      switch (ch) {
        case MIDCLICK: floodmark(i);
        case LEFTCLICK: site[i].count=redraw=1; }
#ifdef X11
      fprintf(stderr,"%s\n",pickinfo);
#endif /*# X11 */
      }
    }
} /* showlabel */

static int darkcolor(int col) /********************************** darkcolor */
/* make dark color */
{
  if (col==LIGHTGRAY) col=DARKGRAY;
  else if (col==WHITE) col=LIGHTGRAY;
  else if ( (col-=8)<0 ) col=BLACK;

  return col;
}

#ifndef TINY
/* support for functions (w)rite and W=restore */
static struct remember_s {
  struct remember_s *next;
  int key;
  double x,y; } *remember0;

static void remember(int key,double x,double y)
{
  struct remember_s *rem;

  alloc(rem,sizeof(struct remember_s));
  rem->next=remember0;
  remember0=rem;
  rem->key=key;
  rem->x=x; rem->y=y;
}

static void forget(int back)
{
  struct remember_s *rem,*next;
  int i,ii,jj;
  double z;

  for (rem=remember0; rem; rem=next) {
    if (back) {
      if (rem->key<0) {
	loop (i,0,ns) {
	  site[i].r[0]-=rem->x;
	  site[i].r[1]-=rem->y; } }
      else {
	ii=rem->key; jj=(ii+1)%3;
	loop (i,0,ns) {
	  z=site[i].r[ii];
	  site[i].r[ii] =  rem->x*z - rem->y*site[i].r[jj];
	  site[i].r[jj] =  rem->y*z + rem->x*site[i].r[jj]; } } }

    next=rem->next;
    free(rem); }

  remember0=NULL;
}
#endif

void hotkeyE(species_t *spec,int E,double U0) /********************* hotkeyE */
{
  double tc,dipm;
  vector *r,*f;
#ifndef TINY
  vector Qdiag;
#endif
  FILE *geo;
  int n,i,j,k;
  char line[128];

  ns=spec->ns;
  strcpy(spec->ext,".geo");
  geo=fopen(spec->fn,"rt");

  alloc(r,ns*sizeof(r[0]));
  alloczero(f,ns*sizeof(f[0]));
  loop (i,0,ns) VV(r[i],=site[i].r)

  tc=totalcharge(spec,-1);
  dipm=dipolemoment(spec,tc
#ifndef TINY
		    ,Qdiag
#endif
		    );
  if (E)
    U0=Upot(r,f,spec,1);
  else {
    if (out!=stdout) fprintf(stderr,"U0 = %g\n",U0);
    prt("! U0=%.12g kcal/mol = %.12g J/mol = %.12g K kB",
	U0,U0*kcal,U0*(kcal/Eunit)); }

  prt("! charge=%.4f e  dipole moment = %.9f eA = %.9f D",
      tc,dipm,dipm*4.8032068);
#ifndef TINY
  prt("! quadrupole moment [e A^2] = diag(%.9f %.9f %.9f)",
      Qdiag[0],Qdiag[1],Qdiag[2]);
  prt("! quadrupole moment [C m^2] = diag(%g %g %g)",
      Qdiag[0]*1.60217733e-39,Qdiag[1]*1.60217733e-39,Qdiag[2]*1.60217733e-39);
  prt("! quadrupole moment [D A = Buckingham] = diag(%g %g %g)",
      Qdiag[0]*4.8032068,Qdiag[1]*4.8032068,Qdiag[2]*4.8032068);
#endif

#ifdef POLAR
  totalpolarizability(spec);
#endif

#ifdef CONTOURPLOT
  {
    int i,j;
    FILE *cd=fopen("contour.xyz","wb");

    loopto (i,-100,100) {
      r[0][0]=i*0.05;
      fprintf(stderr,"%5.2f\r",r[0][0]);
      loopto (j,-100,100) {
	float U0,UU;
	double q;

	r[0][1]=j*0.05;
	q=spec->site[0].charge;
	spec->site[0].charge=0;
	U0=Upot(r,f,spec,0);
	spec->site[0].charge=q;
	UU=Upot(r,f,spec,0)-U0;
	if (!(UU>-1e4 && UU<1e4)) UU=1e4;
	fwrite(&UU,1,4,cd); } }
    fclose(cd); }
#endif

  if (geo) {
    while (fgets(line,128,geo)) if (!strchr("!#",line[0])) {
      char *info=strchr(line,'!');
      if (!info) info=strchr(line,'#');
      if (!info) info="\n";

      n=sscanf(line,"%d%d%d",&i,&j,&k);
      if (i>=0 && i<ns && j>=0 && j<ns) {
	double ij=SQRD(r[i],r[j]);
	if (n==2) prt_("! dist %d-%d = %.11f %s",i,j,sqrt(ij),info);
	if (n==3 && k>=0 && k<ns) {
	  double jk=SQRD(r[j],r[k]);
	  prt_("! angle %d-%d-%d = %.9f %s",
	      i,j,k,
	      acos((ij+jk-SQRD(r[i],r[k]))/(2*sqrt(ij*jk)))*(180/PI),
	      info); } } }
    fclose(geo); }

  *spec->ext=0;
  free(f);
  free(r);
}

/* #include "dumpnextppm.c" */

static void show(double U,species_t *spec,int gr) /******************* show */
/***
    3D picture of the molecule.
    Algorithm originates in the HTH project, Pascal version (c) J.Kolafa 1990.

    gr=0: as in last step
    gr=1: show bonds only
    gr=2: balls-and-sticks model, draw bonds and atom-spheres
          ordered by z coordinate of their centres.
    gr=3: plus shading spheres
    gr=4: stereo (anaglyf)
    gr=-1: minimization by steepest descent, show bonds
    gr=-2: minimization by conjugate gradients, show bonds

***/
{
  bond_t *b;
  int i,j,ii=0,jj,sphere;
  static int oldstrlen=20;
  double
    minz,z,dx,dy;
  static species_t *oldspec;
  static double scalex,scaley;
  static float rscale; /* of RvdW = radii */
  static int erase=0, grid=0, init=1;
  static double zzeye=6; /* 3x screen height: static now */
  static double rpert=0.3;
  double
    dd=PI/32, /* move step */
    mouserot;
  vector rot;
  char *c,Us[24];

  ns=spec->ns;
  lasti=-1;
  if (gr>0) grmode=gr;

  if (init) {
    if (!b0 && grmode==1) grmode++;
    init=0; }

  /* ? is count used later ? */
  loop (i,0,ns) site[i].count=0;

  if (!autogr) {
    if (spec!=oldspec) {
      scaley=sqrt((double)ns)+0.5, /* empirical formula for automatic scale */
	oldspec=spec; /* do not change scaling unless other molecule */
      rscale=0.25;
      bondcol=0; }
#ifdef DOS
    pathtodriver=blendpath;
#endif /*# DOS */
    if (startgraph(-9)<0) exit(-1);
    doMouse(); }
  else switch (erase) {
  case 1: clearviewport(); break;
  case 0: {
    int i,mx=getmaxx(),my=getmaxy(),n=(long)mx*my/60L;

#if defined(DOS) || defined(EMX)
    static int x,y;
    /* DOS VGA 640x480: sophisticated patern shading */
    loop (i,0,n) {
      if ( (x+=107)>mx ) x-=mx;
      if ( (y+=173)>my ) y-=my;
#if 0
      /* white->gray */
      if ( (c=getpixel(x,y)) == WHITE ) putpixel(x,y,c-8);
      else
#endif /*# 0 */
      putpixel(x,y,BLACK); }
#else /*# defined(DOS) || defined(EMX) */
    /* no getpixel and mx,my may change => not so good random method */
    mx++; my++;
    loop (i,0,n) putpixel(irnd(mx),irnd(my),BLACK);
#endif /*# defined(DOS) || defined(EMX) */
  } }

  /*  dumpnextppm(); */

  mouserot=2*PI/(getmaxx()+getmaxy());

  setcolor(LIGHTCYAN);
  c="showing - press ? for help";
  if (gr<0) {
    setcolor(15+gr);
    c="minimizing - ESC to interrupt";
    autogr=1; }

  settextstyle(0,0,1);
  settextjustify(0,2);
  outtextxy(getmaxx()-236,2,c);

  settextstyle(0,0,2);
  sprintf(Us,"U=%.12g",U);
  if (erase!=1) {
    setfillstyle(0/*EMPTY_FILL*/,BLACK);
    bar(0,0,16+16*oldstrlen,16); }
  outtextxy(0,0,Us);
  oldstrlen=strlen(Us);

  for (;;) {
    scalex=scaley*getmaxx()/getmaxy();
    lwindow(0,0,0,0); /* within getmax */
    scale(-scalex,scalex,-scaley,scaley);
    zeye=zzeye*scaley; /* viewing point zeye*screen height from the screen */
    eye=scaley*eyedist;
    pickinfo[0]=0;

    if (grmode==1) {
      /* show bonds directly */
      setlinestyle(0,0,1);
      for (b=b0; b; b=b->next) {
	i=b->indx[0]; j=b->indx[1];
/*was:lline(site[i].r[0],site[i].r[1],site[j].r[0],site[j].r[1]); */
/*.....      setcolor(site[i].count && site[j].count ? LIGHTRED : bondcol);*/

	up();
	if (!bondcol) setcolor(atomcolor(site[i].type));
	else setcolor(bondcols[bondcol]);
	draw3D(site[i].r[0],site[i].r[1],site[i].r[2]);
	if (!bondcol) {
	  draw3D((site[j].r[0]+site[i].r[0])/2,(site[j].r[1]+site[i].r[1])/2,(site[j].r[2]+site[i].r[2])/2);
	  setcolor(atomcolor(site[j].type)); }
	draw3D(site[j].r[0],site[j].r[1],site[j].r[2]); }

      loop (i,0,ns) {
	showlabel(i);
	if (site[i].count || site[i].keep || site[i].nnbr==0) {
	  float qq=zeye/(zeye-site[i].r[2]);
	  int ix=SX(site[i].r[0]*qq),iy=SY(site[i].r[1]*qq);

	  setlinestyle(0,0,1);
	  if (site[i].nnbr==0) {
	    setcolor(WHITE);
	    circle(ix,iy,1); }
	  if (site[i].keep) {
	    setcolor(i==center?LIGHTGREEN:LIGHTRED);
	    circle(ix,iy,3); }
	  setlinestyle(0,0,3);
	  if (site[i].count) {
	    setcolor(i==center?LIGHTGREEN:LIGHTRED);
	    circle(ix,iy,3); } }
        setlinestyle(0,0,1); } }

    else if (grmode==4) {
      /* stereo (anaglyph): no circle over atom picked up */
      int stereo;

      setwritemode(1);
      setlinestyle(0,0,1);
      for (stereo=-1; stereo<2; stereo+=2) {
	setcolor(stereo<0 ? LIGHTRED : LIGHTCYAN);
	for (b=b0; b; b=b->next) {
	  i=b->indx[0]; j=b->indx[1];
	  up();
	  draw3Ds(site[i].r[0],site[i].r[1],site[i].r[2],stereo);
	  draw3Ds(site[j].r[0],site[j].r[1],site[j].r[2],stereo); } }
      setwritemode(0); }

    else { /* gr=2,3 */
#ifdef DOS
      char
#else /*# DOS */
      int
#endif /*# DOS */
	*bonddrawn,*sitedrawn;
      int ic,nc=0;
      bond_t *bdraw=NULL;

      double maxz=-9e9;

      settextstyle(0,0,1);
      settextjustify(1,1);

      for (b=b0; b; b=b->next) nc++;
      alloczero(bonddrawn,(nc+(nc==0))*sizeof(bonddrawn[0]));
      alloczero(sitedrawn,ns*sizeof(sitedrawn[0]));

      for (;;) {

	minz=9e9;
	sphere=-1;

	/* bonds */
	for (ic=0,b=b0; b; b=b->next,ic++) if (!bonddrawn[ic]) {
	  i=b->indx[0]; j=b->indx[1];
	  z=site[i].r[2]+site[j].r[2];
	  if (z<=minz) {
	    bdraw=b; ii=ic; minz=z;
	    sphere=0; } }

	/* atoms */
	loop (i,0,ns) if (!sitedrawn[i]) {
	  z=site[i].r[2]*2;
	  if (z<=minz) {
	    ii=i; minz=z;
	    sphere=1; } }

	/* Angstrom grid */
	if (grid) if (maxz*minz<0) {
	  int i,s;

	  s=(int)(scalex*grid);

	  setlinestyle(0,0,1);
	  loopto (i,-s,s) {
	    setcolor(i%10 ? i%5 ? DARKGRAY : LIGHTGRAY : WHITE);
	    up(); draw3D(-scalex,(double)i/grid,0); draw3D(scalex,(double)i/grid,0);
	    up(); draw3D((double)i/grid,-scalex,0); draw3D((double)i/grid,scalex,0); } }

	if (sphere<0) break; /* everything drawn */

	maxz=minz;
	if (maxz>=zeye*0.95) {
	  fputc('\a',stderr); zzeye=(zeye=maxz*1.1)/scaley; break; }

	if (sphere) {

	  float qq=zeye/(zeye-site[ii].r[2]);
          /* this is aproximate - should be ellipses ! */
	  float q=atom[site[ii].type].LJ[0].RvdW*rscale*qq;
	  unsigned rx=q*scaling.x,ry=q*scaling.y;
	  unsigned i,n;
	  int cx=SX(site[ii].r[0]*qq),cy=SY(site[ii].r[1]*qq); /* not exact! */
	  float hcx=cx+0.5,hcy=cy+0.5,co=-0.7373689,si=-0.6754903,x,y,xnew;
	  sitedrawn[ii]++;
	  setlinestyle(0,0,1);
	  col=atomcolor(site[ii].type);
	  setfillstyle(1 /*SOLID_FILL*/,col);
	  if (site[ii].count) {
	    setcolor(col==LIGHTRED?LIGHTGREEN:LIGHTRED);
	    if (ii==center) setlinestyle(0,0,3); }
	  else
	    setcolor(darkcolor(col));
	  fillellipse(cx,cy,rx,ry);

	  col=darkcolor(col);

	  if (label==NONE && cross) {
	    putpixel(cx,cy,col);
            if (cross==2 && rx>16) {
              putpixel(cx+2,cy,col);
              putpixel(cx,cy+2,col);
              putpixel(cx-2,cy,col);
              putpixel(cx,cy-2,col); }
            if (cross==2 || rx>16) {
	      putpixel(cx+1,cy,col);
	      putpixel(cx,cy+1,col);
	      putpixel(cx-1,cy,col);
	      putpixel(cx,cy-1,col); } }

	  if (grmode==3) {
	    /* shaded spheres */
	    n=rx*ry;
	    x=rx-0.5; y=0;
	    for (i=n; i>2; i--) {
	      xnew=x*co-y*si; y=y*co+x*si; x=xnew;
	      q=1-0.5/i/i;
	      si*=q; co*=q;
	      putpixel(hcx+x,hcy+y,col); } }

	  showlabel(ii); }

	else {
	  float ri[3],rj[3],dr[3],r=0,qi,qj;

	  bonddrawn[ii]++;

	  ii=bdraw->indx[0];
	  jj=bdraw->indx[1];

	  loop (j,0,3) {
	    dr[j] = (rj[j]=site[jj].r[j]) - (ri[j]=site[ii].r[j]) ;
	    r+=Sqr(dr[j]); }

	  r=rscale/sqrt(r);
	  qi=atom[site[ii].type].LJ[0].RvdW*r;
	  qj=atom[site[jj].type].LJ[0].RvdW*r;
	  if (qi+qj>1) {
	    fputc('\a',stderr); rscale/=qi+qj+0.02; }

	  setlinestyle(0,0,3);
	  setcolor(DARKGRAY);
	  up();
	  draw3D(ri[0]+=qi*dr[0],ri[1]+=qi*dr[1],ri[2]+=qi*dr[2]);
	  draw3D(rj[0]-=qj*dr[0],rj[1]-=qj*dr[1],rj[2]-=qj*dr[2]);

	  setlinestyle(0,0,1);
	  if (!bondcol) setcolor(atomcolor(site[ii].type));
	  else setcolor(bondcols[bondcol]);
/*.....        if (site[ii].count && site[jj].count) { setcolor(LIGHTRED); setlinestyle(0,0,3); }*/
	  up();
	  draw3D(ri[0],ri[1],ri[2]);
	  if (!bondcol) {
	    draw3D((rj[0]+ri[0])/2,(rj[1]+ri[1])/2,(rj[2]+ri[2])/2);
	    setcolor(atomcolor(site[jj].type)); }
	  draw3D(rj[0],rj[1],rj[2]); }

        } /* for(;;) */

      free(sitedrawn);
      free(bonddrawn); } /* gr>1 */

    if (pickinfo[0]) {
      setcolor(BLACK);
      setfillstyle(1,WHITE);
      bar(0,0,getmaxx(),8);
      settextstyle(0,0,1); /*?*/
      settextjustify(0,2);
      outtextxy(1,0,pickinfo); }

    if (autogr)
      if (!xkbhit(1)) goto ret;

  newkey:
    dx=dy=0;
    find.x=-1;

    if (moving) {
      char s[48];
      int sum=0;

      loop (i,0,ns) if (site[i].count) sum++;

      moving=1;
      if (center>=0) {
	sprintf(s,"moving %d atoms, center=%d",sum,center);
	if (sum>1) moving=2; }
      else
	sprintf(s,"moving %d atoms, no center",sum);

      setfillstyle(1,YELLOW);
      setwritemode(0);
      settextjustify(0,2);
      settextstyle(0,0,1);
      bar(0,getmaxy()-9,8*strlen(s)+16,getmaxy());
      setcolor(BLACK);
      outtextxy(4,getmaxy()-8,s); }

    if (redraw) {
      redraw=0;
      ch=140; }
    else {
      /*      dumpnextppm();*/
      showMouseCursor();
      ch=extch();
      hideMouseCursor(); }

    memset(rot,0,sizeof(rot));
    marklabeled=0;

    rpert=rpert*0.8+0.06;

    switch (ch) {
#ifdef X11
    case X_REDRAW_ME:
      lwindow(0,getmaxx(),0,getmaxy());
      if (xkbhit(1)) goto newkey;
      break;
#endif /*# X11 */
    case LEFTDRAG:
      if (mousedy) rot[0]=-mousedy*mouserot;
      if (mousedx) rot[1]=-mousedx*mouserot;
      break;

    case MIDDRAG:
      if (mousedx) dx=(double)mousedx*scalex*2/(getmaxx()+1);
      if (mousedy) dy=-(double)mousedy*scaley*2/(getmaxy()+1);
      break;

    case RIGHTDRAG:
      if (mousedy) scaley*=exp((double)mousedy/getmaxy());
      if (mousedx) rot[2]=-mousedx*mouserot;
      break;

    /* single click (no move) by any button */
    case LEFTCLICK:
    case RIGHTCLICK:
    case MIDCLICK:
      find.x=dosmouse.x; find.y=dosmouse.y;
    case 140: /* force redraw (redraw=1: see above) */
      break;

#ifdef DOS
    /* ignoring ... */
    case '@': PrtScr(); break;
#else /*# DOS */
    case '@':
      switch (extch()) {
        case 'k': writeMARKED(spec,1); break;
        case 'm': writeMARKED(spec,0); break;
        case 'K': readMARKED(spec,1); break;
        case 'M': readMARKED(spec,0); break; }
      break;
#endif /*# DOS */
    case 'O': case 'P':  case 'C': {
      static char fn[14]="blend000";
      char *ps=strchr("OC",ch);
      FILE *f;
      int x,mx=getmaxx()+1,y,my=getmaxy()+1,wbg,c;
      unsigned4 pix4;
      unsigned char *pix=(unsigned char *)&pix4;
      static unsigned char col[16][3]={
        {0,0,0}, {0,0,127}, {0,127,0}, {0,127,127},
        {127,0,0}, {127,0,127}, {127,127,0}, {153,153,153},
        {77,77,77}, {0,0,255}, {0,255,0}, {0,255,255},
        {255,0,0}, {255,0,255}, {255,255,0}, {255,255,255} };
#ifdef X11
      XImage *ximage=XGetImage(display,win,0,0,mx,my,AllPlanes,ZPixmap);
      unsigned long xcol;
      extern long unsigned *xcoltab;
#endif /*# X11 */

      strcpy(fn+8,ps?".ps":".ppm");
      fprintf(stderr,"writing %s\n",fn);
      f=fopen(fn,ps?"wt":"wb");
      if (ps) wbg=grmode==1 || grmode==4;
      else wbg=isupper(ch);
      if (wbg)
        col[0][0]=col[0][1]=col[0][2]=255,
        col[15][0]=col[15][1]=col[15][2]=240;

      if (ps) fprintf(f,"%%!PS-Adobe-2.0\n\
90 rotate 54 -550 translate %.2f 480 scale\n\
/DS %d string def\n\
%d %d 8 [ %d 0 0 %d 0 %d ]\n\
{ currentfile DS readhexstring pop } %simage",
            480.0f/my*mx, /* /ASPECT */
            ch=='C'?mx*3:mx,
            mx,my,
            mx,-my,my,
            ch=='C'?"false 3 color":"");
      else fprintf(f,"P6\n%d %d\n255\n",mx,my);
      loop (y,0,my) loop (x,0,mx) {
#ifdef X11
        xcol=XGetPixel(ximage,x,y);
        loop (c,0,16) if (xcoltab[c]==xcol) break;
        if (c>=16) c=0;
#else /*# X11 */
        c=getpixel(x,y);
#endif /*# X11 */

	copy(pix,col[c],3);
        /*	if (wbg) pix4=~pix4; no longer inversion */
        switch (ch) {
          case 'O':
	    fprintf(f,"%c%02x",x%20?' ':'\n',(30*pix[0]+59*pix[1]+11*pix[2])/100);
	    break;
          case 'C':
	    fprintf(f,"%c%02x%02x%02x",x%20?' ':'\n',pix[0],pix[1],pix[2]);
            break;
          default: /* PPM */
	    fwrite(pix,3,1,f); } }

      if (ps) fputs("\nshowpage\n",f);
      fclose(f);
#ifdef X11
      if (ximage) XDestroyImage(ximage);
#endif /*# X11 */
      sprintf(fn+5,"%03d",atoi(fn+5)+1);
      break; }
    case 'e': erase=(erase+1)%3; break;
    case 'M'|128: dx=dd*8; break;
    case 'K'|128: dx=-dd*7; break;
    case 'H'|128: dy=dd*8; break;
    case 'P'|128: dy=-dd*7; break;
#ifdef X11
    case 1: /* ctrl-A = Home */
#endif /*# X11 */
    case 'G'|128: zzeye/=exp(dd); break;
#ifdef X11
    case 5: /* ctrl-E = End */
#endif /*# X11 */
    case 'O'|128: zzeye*=exp(dd); break;
    case 'x': rot[0]=-dd; break;
    case 'X': rot[0]=dd; break;
    case 'y': rot[1]=-dd; break;
    case 'Y': rot[1]=dd; break;
    case 'z': rot[2]=-dd; break;
    case 'Z': rot[2]=dd; break;
    case '>': scaley/=2; break;
    case WHEELFORWARD:
    case '+': scaley*=0.86; break;
    case '<': scaley*=2; break;
    case WHEELBACKWARD:
    case '-': scaley*=1.1; break;
    case '*': if (dd<1) dd*=2; goto newkey;
    case '/': if (dd>1e-4) dd*=0.5; goto newkey;
    case 'b': bondcol+=2;
    case 'B': bondcol=(bondcol-1)%NBONDCOL; break;
    case '#':
    case '=': if ((grmode|1)==3)
                grid="\1\12\0\0\0\0\0\0\0\0"[grid]; /* 0->1->10->0 */
#ifdef X11
              else
                fprintf(stderr,"grid shown in graphic modes 2 and 3 only\n");
#endif               /*# X11 */
              break;
    case '\'': cross=(cross+1)%3; break;
    case 'r': rscale*=0.9; break;
    case 'R': rscale*=1.2; break;
    case 'g': grmode=(grmode&3)+1; break;
    case 'G': grmode=(grmode+3)&3; break;
    case 'c': if (colors[7]=='C') colors[7]=' ',colors[8]='C';
              else if (colors[8]=='C') colors[8]=' ',colors[11]='C';
              else colors[7]='C'; break;
    case 'm': moving=!moving; break;
      //    case 'M': if (center>=0) floodmark(center); break;
    case 'L': marklabeled=1; break;
    case ' ': label=NONE; break;
    case 'i': label=ID; break;
    case 't': label=TYPE; break;
    case 'n': label=NUMBER; break;
    case 'q': label=CHARGE; break;
#ifdef POLAR
    case 'p': label=POLARIZABILITY; break;
    case 'd': label=DIPOLE_eA; break;
    case 'D': label=DIPOLE_D; break;
    case 's': label=SATURATION; break;
#endif /*# POLAR */
    case 'F': find.s[0]=0; find.i=-1; find.q=1e3;
    case 'M':
    case 'u'&31:
    case 'u': loop (i,0,ns) site[i].count=0;
              center=-1;
              break;
    case 'I': loop (i,0,ns) site[i].count=!site[i].count;
              center=-1;
              break;
    case 'K': loop (i,0,ns) site[i].count=site[i].keep;
              break;
    case 'k': loop (i,0,ns) site[i].keep=(enum keep_e)(!!site[i].count);
#ifdef X11
              {
              int s=0;
              loop (i,0,ns) s+=site[i].keep;
              prt("\
%d marked atoms will be kept fixed and %d left free in minimization", s,ns-s);
              }
#endif /*# X11 */
              if (!spec->opt_k) {
#ifdef X11
                fprintf(stderr,"keep=1 set (the same as option -k1)\n");
#endif /*# X11 */
                spec->opt_k=1;
                }
              break;
    case 'l': label=(enum label_e)((label+1)%5); break;
#if defined(X11) && !defined(TINY)
    /* very ugly...*/
    case 'E': hotkeyE(spec,1,0); break;
#endif
    case 'f':
      closegraph();
      switch (label) {
        case ID:
          prts_("enter atom id");
          goto sss;
        case TYPE:
          prts_("enter atom type");
        sss:
          prts(" to find, may end by * :");
          scanf("%15s",find.s);
          find.l=strlen(find.s);
          if (find.s[find.l-1]=='*') find.l--; else find.l++;
          break;
        case NUMBER:
          prts("enter atom # to find :");
          scanf("%d",&find.i);
          break;
        case CHARGE:
          prts("enter charge to find and precision:");
          scanf("%lf%lf",&find.q,&find.dq);
          find.dq+=1e-14;
        default:; /* ... to make gcc happy */
        }
      if (startgraph(-9)<0) exit(-1); break;
#ifndef TINY
    case 'w': forget(0); break;
    case 'W': forget(1); break;
    case 'A': averagecharges(spec,&center); break;
#endif
    case '1': case '2': case '3': case '4': grmode=ch-'0'; break;
    case 3 :
      closegraph(); exit(0);
    case 'S'&31:
      if (spec->wrmol) writeMOL(spec);
      spec->wrmol=0;
      if (!spec->opt_p && !spec->opt_w) WARNING(("neither -p nor -w => nothing saved"))
      if (spec->opt_p) write3D(spec,3);
      if (spec->opt_w) write3D(spec,0);
      break;
    case ':': perturb(spec,NULL,0,rpert*=1.4); ch=',';
    case ';':
    case ',':
    case '.':
    case 27:
      if (moving) {
        moving=0;
        break; }
      if ((ch=='.'||ch==27) && !autogr) closegraph();
      sig=autogr;
      goto ret;
    default: {
#ifndef TINY
      static char *help[]={
        "mouse drag", "Left=rotate     Middle=move        Right=z-rot/resize",
        "mouse click","Left=mark atom  Middle=flood mark  Right=unmark atom",
        "", "",
        "cursors",    "move in x and y directions",
#ifdef X11
        "Home End",   "move eye (in z direction); also Ctrl-A Ctrl-E",
#else /*# X11 */
        "Home End",   "move eye (in z direction)",
#endif /*# X11 */
        "x X y Y z Z","rotate around axes",
        "w W",        "write/restore move/rot position",
        "+ -",        "resize molecule",
        "* /",        "double/halve step size for the above functions",
        "", "",
        "1 2 3 4",    "draw style, the same as gr=1,2,3,4",
        "g G"        ,"cycle draw style",
        "b B",        "cycle bond colors/style",
        "r R",        "resize spheres",
        "", "",
        "i t n q",    "label style: (i)d (t)ype (n)umber q=charge",
        "l"          ,"cycle label style",
        "f",          "find atom (labeling must be on)",
        "F",          "unfind (label all)",
#ifdef POLAR
        "d D p s",    "show: (d)ipoles[eA] (D)ipoles[D] (p)olarizability (s)aturation",
#endif /*# POLAR */
        "Space",      "redraw with no labels",
        "", "",
        "# =",        "toggle Angstrom grid (gr=2,3 only)",
        "'",          "cycle center of spheres style and mouse click radius",
        "c",          "gray/dark/cyan carbons",
        "e",          "erase/draw-over toggle; white bonds='ec'=erase gradually",
#ifdef X11
        "p P C O",      "dump of screen: (p)pm  (P)PM/white  (C)olor PS  O=B/W PS",
        "E",          "calculate energy",
#endif /*# X11 */
        "", "",
        "m",          "toggle move whole config/marked atoms",
        //        "M",          "flood mark (mark all bonded until marked atom)",
        "L",          "mark all labeled atoms",
        "I",          "invert marking",
        "k K",        "keep marked atoms fixed while minimizing/mark kept atoms",
        "A",          "average charges of marked atoms",
        "u ^u M",     "unmark all",
        "@m @k",      "write marked/kept atoms to a file (e.g. for mol2mol)",
        "@M @K",      "read marked/kept from file",
        "", "",
        "Esc",        "quit graphics/minimization, if not -o then get data",
        ", ; :",      "minimize (again or cg=1 `;'=w. sd `:'=w. ran)",
        ".",          "finish this species (no get data, as if `cg=-1;')",
        "Ctrl-C",     "kill"};

      clearviewport();
      settextstyle(0,0,1);
      settextjustify(0,2);

      j=2; jj=getmaxy()/(sizeof(help)/sizeof(help[0])/2);
      for (i=0; i<sizeof(help)/sizeof(char*); ) {
        setcolor(YELLOW);
        outtextxy(3,j,help[i]); i++;
        setcolor(LIGHTCYAN);
        outtextxy(110,j,help[i]); i++;
        j+=jj; }
#endif /*# TINY */
      goto newkey; } } /* switch (ch) */

    if (dx!=0 || dy!=0) {
      if (moving) {
	loop (i,0,ns)
         if (site[i].count) {
	   site[i].r[0]+=dx;
	   site[i].r[1]+=dy; } }
      else {
#ifndef TINY
	remember(-1,dx,dy);
#endif
      loop (i,0,ns) {
        site[i].r[0]+=dx;
        site[i].r[1]+=dy; }
      lasti=-1; } }

    if (moving && center>=0) {
      loop (j,0,3) if (rot[j]!=0) {
	dx=cos(rot[j]); dy=sin(rot[j]);
	ii=(j+1)%3; jj=(j+2)%3;
	loop (i,0,ns) if (site[i].count) {
	  double zz=site[i].r[jj]-site[center].r[jj];
	  z=site[i].r[ii]-site[center].r[ii];
	  site[i].r[ii] = site[center].r[ii] + dx*z + dy*zz;
	  site[i].r[jj] = site[center].r[jj] - dy*z + dx*zz; }
        } }
    else loop (j,0,3) if (rot[j]!=0) {
      dx=cos(rot[j]); dy=sin(rot[j]);
      ii=(j+1)%3; jj=(j+2)%3;
#ifndef TINY
      remember(ii,dx,dy);
#endif
      loop (i,0,ns) {
	z=site[i].r[ii];
	site[i].r[ii] =  dx*z + dy*site[i].r[jj];
	site[i].r[jj] = -dy*z + dx*site[i].r[jj]; }
      lasti=-1; }

    clearviewport(); }

 ret:
#ifndef TINY
  forget(0)
#endif
  ;
} /* show */
#endif /*# GRAPHICS */

static int nUpot;
static void times(int key)
{
  static time_t t0,t1;
  if (key) {
    time(&t1);
    prt("! duration=%us/%d=%gs",
        (unsigned int)(t1-t0), nUpot, (double)(t1-t0)/nUpot);
    t0=t1; }
  else
    time(&t0);
  nUpot=0;
}

#ifndef TINY
#include "blendpw.c"
#endif /*# TINY */

#ifdef POLAR
#define NRATES 29
static double polerr=0,polerr0,rate[NRATES];
static int polnit=0;
#endif /*# POLAR */

#ifndef TINY
static struct sym_s {
  struct sym_s *next;
  int axis;
  int i,j; } *sym0;

static void makesym(vector *r)
{
  struct sym_s *sym;

  for (sym=sym0; sym; sym=sym->next)
    if (sym->i==sym->j) r[sym->i][sym->axis]=0;
    else {
      double x=(r[sym->i][sym->axis]-r[sym->j][sym->axis])/2;
      VVVV(r[sym->i],=r[sym->j],=0.5*r[sym->i],+0.5*r[sym->j])
      r[sym->i][sym->axis]=x;
      r[sym->j][sym->axis]=-x; }
}
#endif

#include "blendu.c"

double minimize(species_t *spec,int print,enum keep_e mask) /***** minimize */
/***
    Random initial configuration (if option -r)
    Minimize energy for molecule spec.
    Greedy (steepest descent) or conjugate gradients method used
    if compiled with OMITKEPT:
    mask=FREE: all site-site interactions included
               (even if there are non-movable atoms INJAIL)
    mask=INJAIL: non-movable sites omitted
***/
{
  int ns=spec->ns,size=ns*sizeof(vector),i,by,wave,sd,cg,sdcg;
  unsigned it;
  int gr,help/*,keep*/;
  vector *r,*f,*f0,*r0,*G,*H;
  double h=1e-6,U,U0;
  double ran;

#ifdef GRAPHICS
  int g;

#ifdef EMX
  if (option('g')<0) {
    emxgrmode=5;
    option('g')=-option('g'); }
#endif /*# EMX */
#endif /*# GRAPHICS */

  if (print) times(0);

  cg = spec->opt_m *= ns>1;
  site=spec->site;
#ifdef OMITKEPT
  keepmask=mask;
#endif /*# OMITKEPT */

  /* enable ^C */
  sig=0;
#ifndef TINY
  if (spec->probe.ns==0)
#endif /*# TINY */
    {
      (*signal)(SIGINT,asksig);
      fprintf(stderr,"%s energy of %s: interrupt by ^C\n",
	      cg ? "minimizing" : "calculating",
	      spec->fn); }

  sd=(abs(cg)/4+2)*(cg<0); /* a few sd steps if opt_m<0 */
  cg=abs(cg);

  /* how often print U */
  if (option('g')) by=abs(option('g'));
  else by = cg>440 ? 20 : cg/22+1;

  ralloc(r0,size);
  ralloc(f0,size);
#ifndef TINY
  {
    FILE *f;
    char line[128];

    strcpy(spec->ext,".sym");
    f=fopen(spec->fn,"rt");

    if (f) {
      while (fgets(line,128,f)) if (!strchr("!#",line[0])) {
	unsigned char axis;
	int i,j;
	int n=sscanf(line,"%c%d%d",&axis,&i,&j);
	struct sym_s *sym;
	char *d=strchr(line,'-');

	if (axis=='.' || axis=='q') break;
	if (d) j=-atoi(d);
	if (n<2) ERROR(("%s: %s : bad line",spec->fn,line))
	if (n==2) j=i;
	if (axis>='x') axis-='x';
	else if (axis>='X') axis-='X';
	if (axis>2)
	  ERROR(("%s: %s : bad axis",spec->fn,line))
        if (i<0 || i>=ns || j<0 || j>=ns)
          ERROR(("%s: %s : bad site",spec->fn,line))
        do {
	  ralloc(sym,sizeof(struct sym_s));
	  sym->next=sym0;
	  sym0=sym;
	  sym->axis=axis;
	  sym->i=i;
	  if (d) sym->j=i;
	  else sym->j=j; }
	while (d && ++i<=j); }
      prt("! symmetry on");
      fclose(f); }
    *spec->ext=0;
  }
#endif

  /*** initial configuration from the molecule ***/
  loop (i,0,ns) VV(r0[i],=site[i].r)
#ifndef TINY
  makesym(r0);
#endif

#ifdef DEBUG
  prt("#### debug! site i<%d (-1=break) coordinate j={0,1,2}",spec->ns);
  for (;;) {
    static int i=0,j=0;
    static double h=1e-5;
    double e;

    for (;;) {
      getdata get(i) get(j) get(h) checkdata enddata
      if (i<0) goto debugdone;
      if (i<ns && j>=0 && j<3) break;
      prt("bad i or j"); }
    r0[i][j]-=h;
    U0=Upot(r0,f0,spec,0);
    r0[i][j]+=h+h;
    U=Upot(r0,f0,spec,0);
    r0[i][j]-=h;
    Upot(r0,f0,spec,i+1);
    e=f0[i][j]+(U-U0)/h/2;
    prt("i=%d j=%d h=%.1e f=%g -dU/dr=%g err=%g %g",
         i,   j,   h,    f0[i][j], (U0-U)/h/2, e, e/obfuscate);
  }
 debugdone:;
#endif /*# DEBUG */

#ifndef TINY
  if (spec->probe.ns==3) {
    vector HH[2];
    int i,j;

    /* special patch for probing by TIP3P water */
    i=spec->probe.i;
    U0=3e33;
    loop (j,0,12) {
      makeHH(&r0[i],j);
      U=Upot(r0,f0,spec,0);
      if (U<U0) {
	U0=U;
	copy(HH,&r0[i+1],sizeof(HH)); } }
    copy(&r0[i+1],HH,sizeof(HH));
    loop (i,0,ns) VV(site[i].r,=r0[i]) }
  else
#endif /*# TINY */

    U0=Upot(r0,f0,spec,print);

  if (U0>1e33) goto finish;
  if (print) times(1);

  if (out!=stdout) if (print) fprintf(stderr,"U = %.12g\n",U0);
  if (!spec->opt_m) {
    /* no minimization */
    sd=0; cg=-1;
#ifdef GRAPHICS
    if (option('s') || option('g')>0) {
      show(U0,spec,1);
      loop (i,0,ns) VV(r0[i],=site[i].r) }
#endif /*# GRAPHICS */
    goto finish; }

  ralloc(r,size); copy(r,r0,size);
  ralloc(f,size);
  ralloc(G,size); /* conjugate grad only */
  ralloc(H,size); /* conjugate grad only */

  do {
    sdcg=max(sd,cg);
    /* this happens if -m0 and show() ESCaped: no minimization but prompt */

    if (cg<=0) {
#ifdef GRAPHICS
      if (cg==0 && sd==0) show(U0,spec,1);
#endif
      goto cgdone; }

    h*=3;
    U0=Upot(r0,f0,spec,0); /* needed because show() may move atoms */

    /* re-consider step to prevent large moves when overlap etc. */
    U=0; loop (i,0,ns) Max(U,SQR(f0[i]))
    U=sqrt(U); /* =max force on 1 atom */
    if (U*h>0.5) h=0.5/U; /* max move in the 1st step = 0.5 A */
    if (U*h<0.001) h=0.001/U; /* min move in the 1st step = 0.001 A */

//  #include "blendmd.c" (not finished)

    if (sd==0 && cg && (option('a')==0 || option('b')==0) ) {
      prt("! NOTE: steepest descent will be used because of constraints (-a or -b)");
      sd=cg; cg=0; }

    if (sd) {
      /*** steepest descent method ***/
      if (print) prt("! %d steps of steepest descent:",sd);

      for (it=1; !sig; it++) {

        do {
          h*=0.55; /* magic value: works fine in a narrow valley */

	  /* converg. criterion changed from 1e-12 to to 1e-11 in V 2.0o */
          if (h<1e-11 || it>=sd) {
#ifdef GRAPHICS
            if (option('g')>0 && cg==0) show(U0,spec,-1);
#endif
            goto sddone; }

          loop (i,0,ns) VVV(r[i],=r0[i],+h*f0[i])
          U=Upot(r,f,spec,0);
        } while (U>U0);

        copy(r0,r,size);
        copy(f0,f,size);
        U0=U;

#ifdef GRAPHICS
        if (option('g')>0) {
          if (it && it%abs(option('g'))==0) {
            loop (i,0,ns) VV(site[i].r,=r0[i])
            show(U0,spec,-1);
            /* does NOT back copy rotated r here */ } }
        else
#endif /*# GRAPHICS */
          if (print) prtU(it,by,U,h);

        h*=2.5; /* magic: h should not shrink to zero */ } }

sddone:

    if (cg && option('b') && option('a')) {
      /* this is conjugate gradient, rewritten from Numerical Recipes */
      double dgg,gg,h0;
      /* brrrr! U0=fp r0=p  f0=xi */
      int j,maxj=24;

      if (!autogr && print) prt("! %d steps of conjugate gradients:",cg);
      /* from init, known r0 f0 U0 */
      h*=3;
      copy(G,f0,size);
      copy(H,f0,size);

      for (it=1; !sig && it<=cg; it++) {
        /***
          now we are looking for h>0 so that r=r0+h*f0 is minimum
          in most cases close to minimum we need only two
          evaluations of forces in one step
        ***/
        gg=0;
        loop (i,0,ns) gg+=SQR(f0[i]);
        if (gg==0) break;

        h0=h;
        for (j=0; ;j++) {
          loop (i,0,ns) VVV(r[i],=r0[i],+h*f0[i])
          U=Upot(r,f,spec,0);
          if (U<U0) break;
          TRACE('/')
          h*=0.5;
          if (j>=maxj) {
            /* unsuccessful - reset CG (should occur rarely!) */
            U0=Upot(r0,f0,spec,0);
            if (maxj==20) goto cgdone; /* unsuccessful again! */
            h=h0; maxj=20;
            copy(G,f0,size);
            copy(H,f0,size);
            TRACE('r')
            goto OK; } }

        maxj=5;

        dgg=0;
        loop (i,0,ns) dgg+=SCAL(f[i],f0[i]);
        dgg/=gg;
        if (dgg<0.8) {
          /* try one step of the secant method - efficient close to minimum */
          h*=1/(1-dgg);
          loop (i,0,ns) VV(r0[i],+=h*f0[i])
          U0=Upot(r0,f0,spec,0);
          if (U0<U) goto OK; }

        /***
  	  not better or too extrapolated - we must be happy with r,f,U
	  this occurs once a while
        ***/

        h*=4; maxj=8; /* might be caused by too low h */
        TRACE('-')
        U0=U;
        copy(f0,f,size);
        copy(r0,r,size);

OK:
#ifdef GRAPHICS
        if (option('g')>0) {
          if (it && (it-1)%abs(option('g'))==0) {
            loop (i,0,ns) VV(site[i].r,=r0[i])
            show(U0,spec,-2);
            /* does NOT back copy rotated r here */ } }
        else
#endif /*# GRAPHICS */
          if (print) prtU(it,by,U,h);

        gg=dgg=0;
        loop (j,0,ns) {
          gg+=SQR(G[j]);
          dgg+=(f0[j][0]-G[j][0])*f0[j][0]
              +(f0[j][1]-G[j][1])*f0[j][1]
              +(f0[j][2]-G[j][2])*f0[j][2]; }
        dgg/=gg;

        copy(G,f0,size);
        loop (j,0,ns) VVV(H[j],=G[j],+dgg*H[j])
        copy(f0,H,size); } }

cgdone:

    U0=Upot(r0,f0,spec,0);
    loop (i,0,ns) VV(site[i].r,=r0[i])
#ifdef GRAPHICS
    if (autogr) {
      closegraph(); autogr=0; show(U0,spec,0);
      loop (i,0,ns) VV(r0[i],=site[i].r) }
#endif /*# GRAPHICS */
    if (print) hotkeyE(spec,0,U0);
    sd=0;

    if (print) times(1);

    if (ch==',' || ch==';') {
      if (cg==0) cg=sdcg;
      if (ch==';') sd=cg; }
    else if (ch==27 && out==stdout
#ifndef TINY
                    && spec->probe.d==0
#endif /*# TINY */
                                     ) {
      /* read data */
      sig=0;
      keep=spec->opt_k;
#if defined(DOS) || defined(EMX)
      if (autogr) {
        closegraph();
        autogr=0; }
#endif /*# defined(DOS) || defined(EMX) */
#ifdef GRAPHICS
      g=option('g');
      if (g<0) g=0;
#endif /*# GRAPHICS */
      fprintf(stderr,"cg=%d keep=%d by=%d\n",cg,keep,by);
      fprintf(stderr,"enter data: `help=1;\' for help, ");
      spec->ext[0]=0;
      if (cg>=0) fprintf(stderr,"`cg=-1;\' to finish \"%s\"\n",spec->fn);
      else fprintf(stderr,"\n  `cg=VALUE;\' and/or `sd=VALUE;\' to continue minimizing %s\n",spec->fn);
      do {
        wave=gr=help=0;
        ran=0;

        getdata
          get(cg) get(sd) get(by) get(h) get(ran) get(wave)
#ifdef GRAPHICS
          get(gr) get(g)
#endif /*# GRAPHICS */
          get(keep) get(help)
          checkdata enddata
#ifdef GRAPHICS
        if (gr) {
          show(U0,spec,gr);
          if (ch==',' || ch==';') {
            if (sd==0 && cg==0) cg=1;
            if (ch==';') sd=cg;
            break; }
          loop (i,0,ns) VV(r0[i],=site[i].r) }
#endif /*# GRAPHICS */
        if (help) fprintf(stderr,"\n\
Enter list of assignments ended by \';\' :\n\
\n\
cg = # of conjugate gradient steps; cg=-1 to end minimization\n\
sd = # of steepest descent steps (set to zero after step)\n\
by = frequency of prints\n\
h = step for cg or sd (normally do not change)\n\
ran = random perturbation of free atoms by ran A\n\
wave = wave perturbation of free atoms (small integer)\n\
keep = keep/free kept atoms (overrides option -k)\n"
#ifdef GRAPHICS
"gr = show molecule now (gr=1: bonds, 2: balls-and-sticks, 3: shaded balls)\n\
g = frequency of show (with bonds) while minimizing (overrides by)\n\
gr, sd, wave, ran are set to zero after each step\n"
#endif /*# GRAPHICS */
"\n\
Example (double cg and print energy every second step):\n\
  cg*=2 by=2 ;\n\
\n\
More:\n\
  ? getdata help   $? scroll help\n");
        } while (gr | help);
#ifdef GRAPHICS
      option('g')=g;
#endif /*# GRAPHICS */
      spec->opt_k=keep;
      if (wave || ran) perturb(spec,r0,wave,ran);
      if (print) times(0); }
    else
      sd=0,cg=-1; /* none of [^S,.:;] nor stdout ==> force quit */

  } while (cg>=0);

  /* sometimes unnecessary, but it is difficult to trace all changes in r */
  U0=Upot(r0,f,spec,print);
  if (print) times(1);

finish:
  release(r0);
#ifndef TINY
  sym0=NULL;
/*.....prt("! symmetry off");*/
#endif
  return spec->Emin=U0;
}

#ifndef TINY

#include "blendar.c"
#include "blendnm.c"
#include "blendess.c"
#include "blendine.c"
#include "blendvir.c"

#endif /*# TINY */
