/* cc -o ev -O2 -Wall ev.c -I../gen -lm -lncurses

NEW 01/2012:
  - clipboard
NEW 11/2010:
  - trailing spaces silently removed on input
NEW 4/2010:
  - environment variable EV will replace 1st char to 2nd char
    (e.g., modulo is `, to use : for it, use export EV=":\`")
NEW 06/2009:
  - $COLUMNS instead of $SCROLL
NEW 11/2007:
  - support for a=b=c and a+=b
NEW 4/2007:
  - code cleaned
  - versions unified, DOS no longer supported
  - options changed
  - .evdata automatically read
*/

#include <stdlib.h>
#include <stdio.h>
#include <ctype.h>
#include <math.h>
#include <string.h>
#include <stdarg.h>

volatile int sig=0;
#include <signal.h>

#define CALC 4

int (*myprintf)(const char *format,...);
char *(*evgetline)(char *line);

#define Sqr(X) ((X)*(X))
#ifndef PI
#define PI 3.14159265358979323846
#endif

#define STDERR
#include "qmin.c"
#include "ms.h"

enum mode_e { ARGS, DUMB, PROMPT, FILTER, TEX } mode;

#include "prompt.c"

int cprintf(const char *format,...)
{
  int ret;
  char L[PROMPT_N*5],*tok;
  va_list args;

  va_start(args,format);
  ret=vsprintf(L,format,args);
  va_end(args);
  tok=strtok(L,"\n");
  while (tok) {
    _pr.point=_pr.pos=0;
    _pr.back=0; prompt_showline(tok); _pr.back=1;
    tok=strtok(NULL,"\n");
    addch('\n'); }
  refresh();

  return ret;
}

int texprintf(const char *format,...)
{
  int ret,com=0,res=0,nc=-2,inmac=0;
  char L[PROMPT_N*5],*c;
  va_list args;

  va_start(args,format);
  ret=vsprintf(L,format,args);
  va_end(args);
  for (c=L; *c; c++) {
    if (*c=='!') com++;
    if (!inmac) if (nc>44 && *c==' ' || nc>47 && !isalnum(*c) || nc>=50) {
      printf("\\\\\\hspace*{1em}");
      nc=0; }
    nc++;
    if (*c=='^') printf("$\\uparrow$");
    else if (*c=='_') printf("\\_");
    else if (*c==' ') printf("~");
    else if (*c=='{') printf("{}[");
    else if (*c=='}') printf("]");
    else if (*c=='[') {
      printf("{\\rm[\\Unit{"); inmac=1; }
    else if (*c==']') {
      printf("}]}"); inmac=0; }
    else if (*c=='') printf("$^\\circ$");
    else if (*c=='\\' && (c>=L && c[-1]!='\\') && c[1]!='\\') printf("$\\surd$");
    else if (*c=='@') { res=1; printf("\\tt\\mbox{}\\hfill$"); }
    else if (res && *c=='e') {
      res=3; printf("\\cdot10^{"); }
    else if (*c=='%') printf("\\%%");
    else if (*c=='~') printf("$\\sim$");
    else printf("%c",*c); }
  if (res) {
    if (res==3) printf("}");
    printf("$"); }
  printf("\\\\\n");

  return ret;
}

int dummyprintf(const char *format,...)
{
  return 0;
}

void asksig(int signr) /* ========================================== asksig */
{
  if (sig++) { /* ^C typed twice */
    if (mode==PROMPT) endwin();
    exit(-1); }
/* refresh the ^C handler - nothing is compatible */
#ifdef __EMX__
  (*signal)(SIGINT,SIG_ACK);
#else /*# __EMX__ */
  (*signal)(SIGINT,asksig);
#endif /*# __EMX__ */
  myprintf("\nINTERRUPT(%d) ACCEPTED (type ^C again to exit immediately)\n",
          signr);
}

static int err,comma;
static double res;
#define NFMT 30
static char fmt[NFMT+2];

typedef struct _Idlist_s {
  struct _Idlist_s *next;
  int isconst;
  double val;
  char id[1]; /* [var len] */ } _Idlist; /* cf. ground.h */
struct {
  struct _Idlist_s *head;
  int ignoreid; /* 0: unknown ID = error, 1: unknown ID = silent 0 */
  double nr;
  double degrad; } _Id = { NULL,0,0,1 };

static int isconst;
char assignop='=';

void assign(char *b)
{
  struct _Idlist_s *is;

  if (!strcmp(b,"pi") || !strcmp(b,"PI")) {
    myprintf("constant %s cannot be changed",b);
    return; }

  for (is=_Id.head; is; is=is->next)
    if (!strcmp(is->id,b)) {
      if (is->isconst) {
        myprintf("constant %s cannot be changed (~%s to remove)",b,b);
        return; }
      goto val; }
  is=malloc(sizeof(struct _Idlist_s)+strlen(b));
  if (!is) { myprintf("NO HEAP"); return; }
  is->next=_Id.head;
  _Id.head=is;
  strcpy(is->id,b);
  is->isconst=isconst;
 val:
  switch (assignop) {
    case ':':
    case '=': is->val=res; break;
    case '+': res=is->val+=res; break;
    case '-': res=is->val-=res; break;
    case '*': res=is->val*=res; break;
    case '/': res=is->val/=res;}
  assignop='=';
}

void deassign(char *b)
{
  struct _Idlist_s *is,*last=NULL;

  for (is=_Id.head; is; is=is->next) {
    if (!strcmp(is->id,b)) {
      if (last) last->next=is->next;
      else _Id.head=is->next;
/*.....  myprintf("variable %s freed\n",b);*/
      free(is); return; }
    last=is; }
  myprintf("variable %s not found\n",b);
}

int prompt=1;

#define CINT /* C-like integers */
#include "fortran.h"
#include "calc.c"

#include "powi.c"
#include "fortran.c"

static enum { DOUBLE,INT,UNSIGNED } cast;
char *toclip; /* getenv("TOCLIP") */

void clip(void)
{
  if (toclip) {
    char ll[1024],ff[128],*c;

    sprintf(ff,"%s %s",toclip,fmt);

    for (c=ff; c[1]; c++);
    if (*c=='\n') *c=0;
    sprintf(ll,ff+(ff[0]==' '),res);
    system(ll); }
}

void printres(void)
{
  switch (cast) {
    case DOUBLE: myprintf(fmt,res); break;
    case INT: myprintf(fmt,(int)res); break;
    case UNSIGNED: myprintf(fmt,(unsigned)res); break; }

  clip();

  _Id.nr=res;
}

void printres2(char *fmt2,double var2)
{
  char *f=malloc(strlen(fmt)+strlen(fmt2)+3);
  char *x;

  sprintf(f,"%s%s",fmt,fmt2);
  if ( (x=strchr(f,'\n')) ) *x=' ';
  strcat(f,"\n");
  switch (cast) {
    case DOUBLE: myprintf(f,res,var2); break;
    case INT: myprintf(f,(int)res,var2); break;
    case UNSIGNED: myprintf(f,(unsigned)res,var2); break; }

  clip();

  _Id.nr=res;
}

int nop,nparm;
double parm[4]; /* parm value ([4] is by 1 more) */
int isparm[4]; /* if parm assigned */
char *var,*eqn;
enum op_e { NONE,
            SOLVE,MIN,MAX,PLOT,
            INTEG,DERIV,SUM,PROD,ITER,REPEAT,CONTFRAC, /* see below */
            DEF,UNDEF,
            HELP /* just to mark that help has been printed */ };

struct mac_s {
  struct mac_s* next;
  char *X;
  char *E; } *mhead=NULL;

void defmacro(char *X,char *E)
{
  struct mac_s *m;

  for (m=mhead; m; m=m->next) if (!strcmp(m->X,X)) goto repl;
    m=malloc(sizeof(*m));
    m->E=NULL;
    m->X=NULL;
    m->next=mhead;
    mhead=m;

 repl:
  if (m->E) free(m->E);
  if (m->X) free(m->X);
  m->X=strdup(X);
  m->E=strdup(E);
}

void undefmacro(char *X)
{
  struct mac_s **mp,*m;

  for (mp=&mhead,m=*mp; *mp; mp=&((*mp)->next),m=*mp) if (!strcmp(m->X,X)) {
    if (m->E) free(m->E);
    if (m->X) free(m->X);
    *mp=m->next;
    free(m);
    return; }
  myprintf("%s not defined",X);
}

int verbose=3+8; /* 1: solve:f=err 2:macros 4:init.const 8:comments (FILTER only) */

void expmacro(char *expr)
{
  char *c;
  int nrepl=0;
  struct mac_s *m;
  int l,chg=0;

  if (!mhead) return;

  do {
    chg=0;
    for (c=expr; *c; c++) {
      if (isalpha(*c) || *c=='_') {
        char *b=c;
        while (*c && isalnum(*c) || *c=='_') c++;
        for (m=mhead; m; m=m->next) {
          if (c-b==strlen(m->X) && !memcmp(b,m->X,strlen(m->X))) {
            l=strlen(expr)+strlen(m->E)-strlen(m->X)+1;
            if (l>=PROMPT_N) {
              myprintf("too long expression after macro expansion");
              return; }
            if (nrepl++>10000) {
              myprintf("more than 10000 macro expansions (recursion?)");
              return; }
            memmove(b+strlen(m->E),b+strlen(m->X),strlen(b)-strlen(m->X)+1);
            memcpy(b,m->E,strlen(m->E));
            chg=1; } } } }
  } while (chg);

  if (nrepl && verbose&2) myprintf("[%s]",expr);

  return;
}


int oper(char *expr)
/*
  these operators are in the form OPER VAR=FROM[,TO,BY] EXPR
  exceptions:
    contfrac i=5,7 2++1//i ==> 2+1/(5+1/(6+1/(7)))
*/
{
  char *sp=strchr(expr,' '),*tok,*c,*eq;
  static struct op_s {
    char *name; /* keyword */
    int minparm; /* min # of parameters */
    int maxparm; /* max # of parameters */
    char *help; /* text to print if w/o parms */
    /* notes:
       * if less than maxparm parameters are give, undefined
         parameters adopt "reasonable" defaults, not previuos values
       * if minparm=0 then form "solve x x^2-2" is allowed, where x
         has been assigned before
    */
    } op[]= {
      /* index must correspond to enum op_e above */
      {"",0,0,""},
      {"solve",   0,3,"X[=INIT] EXPR (secants), X=FROM,TO[,BY] EXPR (search)"},
      {"min",     0,3,"X[=INIT] EXPR, X=FROM,TO[,BY] EXPR"},
      {"max",     0,3,"X[=INIT] EXPR, X=FROM,TO[,BY] EXPR"},
      {"plot",    2,3,"X=FROM,TO[,BY] EXPR"},
      {"integ",   2,2,"X=FROM,TO EXPR"},
      {"deriv",   0,1,"X[=VALUE] EXPR"},
      {"sum",     2,3,"N=FROM,TO[,BY] EXPR"},
      {"prod",    2,3,"N=FROM,TO[,BY] EXPR"},
      {"iter",    1,2,"X=FROM[,EPS] EXPR (repeat X=EXPR until |X-Xlast|<EPS)"},
      {"repeat",  2,2,"X=FROM,COUNT EXPR (COUNT* repeat X=EXPR)"},
      {"contfrac",2,3,"N=FROM,TO[,BY] [BEFORE++]NUM//DENOM"},
      {"def",     0,0,"X ANY_STRING, def X=A -> def X (A)"},
      {"undef",   0,0,"X"},
      {NULL,0,0,""} };
  int i,np;

  if (sp) *sp=0;

  for (nop=1; op[nop].name; nop++)
    if (!strcmp(op[nop].name,expr)) {
      if (!sp) {
        myprintf("%s %s\n",expr,op[nop].help);
        nop=HELP;
        return 1; }
      goto found; }

  if (!sp) return 0;
  *sp=' ';
  return 0;

 found:
  var=tok=strtok(sp+1," \t"); /* x=1,2,3 */

  if (!var) {
    myprintf("variable[=...] expected\n");
    return 0; }
  eq=strchr(tok,'=');
  if (eq) *eq++=0;
  res=Calc(var,&c);

  for (i=0; i<4; i++) { isparm[i]=0; parm[i]=0; }
  parm[0]=res;
  np=0;

  if (eq) {
    /* = */
    for (np=0; np<4; np++) {
      char *comma=strchr(eq,',');

      if (comma) *comma=0;
      if (np>=op[nop].maxparm) {
        myprintf("superfluous parameter %d",np+1);
        return 0; }
      if (strlen(eq)==0) {
        myprintf("expression expected after ,");
        return 0; }
      parm[np]=Calc(eq,&c);
      err=eq==c;
      if (err) {
        myprintf("parameter %d error",np+1);
        return 0; }
      isparm[np]=1;
      if (!comma) break;
      eq=comma+1; } }
  else {
    /* no = */
    if (op[nop].minparm>0) {
      myprintf("=from,to[,by] expected\n");
      return 0; } }
  np++;

  if (np<op[nop].minparm) {
    myprintf("at least %d parameters expected but %d given",
             op[nop].minparm,np);
    return 0; }

  if (var==c)
    /* undefined variable */
    if (np) {
      res=parm[0]; assign(var); }
    else {
      myprintf("undefined variable %s",var);
      return 0; }

  eqn=strtok(NULL,"");
  if (!eqn) {
    if (nop!=UNDEF) {
      myprintf("expression expected\n");
      return 0; } }

  return nop;
}

double setprec(double x)
{
  double d;
  char *c;
  double y,yy;
  int i;

  for (d=1e-33*(fabs(x)+1); d<1e99; d*=2) {
    int n=0;

    for (i=-5; i<=5; i+=2) {
      res=x+d*i;
      assign(var);
      err=eqn==c;
      if (err) { myprintf("bad equation\n"); return 0; }
      y=Calc(eqn,&c);
      if (i!=-5) n+=(yy!=y);
      yy=y; }
    if (n>2) return d; }

  return 0;
}

double integ(double d,double x0,double x1)
/* WARNING - setprec bad close to x^2-1 for x approx 0 and similar */
{
  int st=0,sg=1;
  double s1=0,s2=0;
  char *c;

/*.....  d=pow(d,.2);*/
  d=pow(d,.3);
/*.....  fprintf(stderr,"d0=%g\n",d);*/
  if (x0>x1) sg=-1,d=-d;
  while (!st) {
    if (sg*(x0+d-x1)>=0) st++,d=x1-x0;
    res=x0+0.21132486540519*d;
    assign(var);
    s1+=Calc(eqn,&c)*d;
    err=eqn==c;
    if (err) { myprintf("integ: bad expression\n"); return 0; }
    if (sig) { myprintf("integ: interrupted\n"); return 0; }

    res=x0+0.78867513459481*d;
    assign(var);
    s2+=Calc(eqn,&c)*d;
    x0+=d;
    d*=1.003; }
/*.....  fprintf(stderr,"d1=%g\n",d);*/
  return (s1+s2)/2;
}

int Solve(double eps)
/* eps<0 ==> print error messages */
{
  int prt=eps<0;
  char *c;

  eps=fabs(eps)*32; /*? - essentially 3e-8*/
  if (eps==0) eps=-1e-12;

  MS_BEGIN(res,eps)
    if (MS_it>1000) {
      if (prt) myprintf("solve: too many iterations\n");
      return 1; }
    assign(var);
    MS_f=Calc(eqn,&c);
    err=eqn==c;
    if (err) {
      if (prt) myprintf("solve: bad equation\n");
      return -1; }
    if (sig) {
      if (prt) myprintf("solve: interrupted\n");
      return -1; }
    MS_END(res,1)

  assign(var);

  return 0;
}

double MinMaxSg=1,MinMaxF;

int MinMax(double eps)
/* eps<0 ==> print error messages */
{
  int prt=eps<0;
  char *c;

  MinMaxF=0;

  qmin_prt=prt;
  eps=sqrt(fabs(eps));
  if (eps==0) eps=1e-5;

  QMIN_BEGIN(res,eps,5)
    if (QMIN_it>1000) {
      if (prt) myprintf("min/max: too many iterations\n");
      return 1; }
    assign(var);
    MinMaxF=QMIN_f=MinMaxSg*Calc(eqn,&c);
    err=eqn==c;
    if (err) {
      if (prt) myprintf("min/max: bad equation\n");
      return -1; }
    if (sig) {
      if (prt) myprintf("min/max: interrupted\n");
      return -1; }
    QMIN_END(res)

  assign(var);

  return 0;
}


double froot_ext;

int Checkres(double eps)
{
  char *c;

  assign(var);
  froot_ext=Calc(eqn,&c);

  return fabs(froot_ext)<fabs(300*eps); /* 3e-7 of typical y-value */
}

double getmass(char *c,char **ret)
{
  char elem[4];
  struct mend_s {
    char *name;
    double mass; } *m,mend[]={
      {"H",1.00794},
      {"D",2.014101778},
      {"T",3.016049268},
      {"He",4.002602},
      {"Li",6.941},
      {"Be",9.012182},
      {"B",10.811},
      {"C",12.0107},
      {"N",14.0067},
      {"O",15.9994},
      {"F",18.9984032},
      {"Ne",20.1797},
      {"Na",22.989770},
      {"Mg",24.3050},
      {"Al",26.981538},
      {"Si",28.0855},
      {"P",30.973761},
      {"S",32.065},
      {"Cl",35.453},
      {"Ar",39.948},
      {"K",39.0983},
      {"Ca",40.078},
      {"Sc",44.955910},
      {"Ti",47.867},
      {"V",50.9415},
      {"Cr",51.9961},
      {"Mn",54.938049},
      {"Fe",55.845},
      {"Co",58.933200},
      {"Ni",58.6934},
      {"Cu",63.546},
      {"Zn",65.39},
      {"Ga",69.723},
      {"Ge",72.64},
      {"As",74.92160},
      {"Se",78.96},
      {"Br",79.904},
      {"Kr",83.80},
      {"Rb",85.4678},
      {"Sr",87.62},
      {"Y",88.90585},
      {"Zr",91.224},
      {"Nb",92.90638},
      {"Mo",95.94},
      {"Tc",98},
      {"Ru",101.07},
      {"Rh",102.90550},
      {"Pd",106.42},
      {"Ag",107.8682},
      {"Cd",112.411},
      {"In",114.818},
      {"Sn",118.710},
      {"Sb",121.760},
      {"Te",127.60},
      {"I",126.90447},
      {"Xe",131.293},
      {"Cs",132.90545},
      {"Ba",137.327},
      {"La",138.9055},
      {"Ce",140.116},
      {"Pr",140.90765},
      {"Nd",144.24},
      {"Pm",145},
      {"Sm",150.36},
      {"Eu",151.964},
      {"Gd",157.25},
      {"Tb",158.92534},
      {"Dy",162.50},
      {"Ho",164.93032},
      {"Er",167.259},
      {"Tm",168.93421},
      {"Yb",173.04},
      {"Lu",174.967},
      {"Hf",178.49},
      {"Ta",180.9479},
      {"W",183.84},
      {"Re",186.207},
      {"Os",190.23},
      {"Ir",192.217},
      {"Pt",195.078},
      {"Au",196.96655},
      {"Hg",200.59},
      {"Tl",204.3833},
      {"Pb",207.2},
      {"Bi",208.98038},
      {"Po",209},
      {"At",210},
      {"Rn",222},
      {"Fr",223},
      {"Ra",226},
      {"Ac",227},
      {"Th",232.0381},
      {"Pa",231.03588},
      {"U",238.02891},
      {"Np",237},
      {"Pu",244},
      {"Am",243},
      {"Cm",247},
      {"Bk",247},
      {"Cf",251},
      {"Es",252},
      {"Fm",257},
      {"Md",258},
      {"No",259},
      {"Lr",262},
      {"Rf",261},
      {"Db",262},
      {"Sg",266},
      {"Bh",264},
      {"Hs",277},
      {"Mt",268},
      {"Uun",281},
      {"Uuu",272},
      {"Uub",285},
      {"Uuq",289},
      {NULL,0} };

  elem[0]=*c;
  elem[1]=elem[2]=elem[3]=0;
  *ret=c+1;
  if (islower(c[1])) {
    elem[1]=c[1];
    *ret=c+2;
    if (islower(c[2])) {
      elem[2]=c[2];
      *ret=c+3; } }

  for (m=mend; m->name; m++) if (!strcmp(m->name,elem)) return m->mass;

  myprintf("%s is unknown element");

  return 0;
}

void molarmass(char key,char *c)
/*
  in place replace M(FORMULA) by MMkey
  accepted expressions:
  C2H6
  CuSO4 H10O5 (chem. CuSO4.5H20)
  Cu
  Fe0.5
*/
{
  char *e=strchr(c,')');
  char *b=c;
  double m,x,s=0;

  c+=2;

  if (!e) {
    myprintf("M( : ) expected");
    return; }

  if (key>'Z') {
      myprintf("max 26 expressions M() allowed in 1 line");
      return; }

  for (;;) {
    while (*c==' ') c++;
    m=getmass(c,&c);
    if (*c==0)  {
      myprintf("M( : syntax or internal error");
      return; }
    x=strtod(c,&e);
    if (c==e) x=1;
    c=e;
    if (*c==0) {
      myprintf("M( : syntax or internal error");
      return; }
    s+=m*x;
    if (*c==')') {
      for (e=b; e<=c; e++) *e=' ';
      b[0]=b[1]='M';
      b[2]=key;
      b[3]=0;
      x=res;
      res=s;
      assign(b);
      res=x;
      b[3]=' ';
      return;
    }
  }
}

void calculate(char *expr)
{
  char *c;
  char *b,*e,key,extra;

  sig=0;

  if (strchr(expr,'%')) {
    /* format */
    if (strlen(expr)>NFMT) {
      myprintf("TOO LONG FORMAT (for modulo use `)\n");
      return; }
    cast=DOUBLE;
    if (!strcmp(expr,"%")) {
      strcpy(fmt," %.8g\n");
      printres();
      return; }
    if (expr[1]=='?') {
      myprintf("current format =%s",fmt);
      return; }
    if (strlen(expr)<NFMT) for (c="eEgGfiduxXoc"; *c; c++) if (strchr(expr,*c)) {
      strcpy(fmt,mode==TEX?"@":" ");
      strcat(fmt,expr);
      if (mode) strcat(fmt,"\n");
      if (strchr("idc",*c)) cast=INT;
      if (strchr("uxXo",*c)) cast=UNSIGNED;
      printres();
      return; }
    myprintf("BAD FORMAT (for modulo use `)\n");
    return; }

 {
   char x[8],*c;
   x[7]=0;
   memcpy(x,expr,7);
   for (c=x; *c; c++) if (!isalpha(*c)) *c=0;
   if (strcmp(x,"def") && strcmp(x,"undef")) expmacro(expr);
 }

  /* molar mass, e.g. M(H2O) */
  key='A'; /* A,B,..*/
  for (c=expr; *c; c++) if (*c=='M' && c[1]=='(') molarmass(key++,c);

  /* looking for id=expr */
  c=expr;

  while (*c==' ') c++;
  b=c;
  if (isalpha(*c) || *c=='_') while (isalnum(*c) || *c=='_') c++;
  e=c;
  while (*c==' ') c++;

  /* NEW := (the same as =) += -= *= /= (more later?) */
  extra='=';
  if (*c && strchr("+-*/:",*c) && c[1]=='=') {
    extra=*c; *c='='; c[1]=' '; }
  if (extra==':') extra='=';

  key=*c;

  if (key=='=' && c[1]!='=' || key=='~')
    if (b==e) {
      /* =id */
      /* NOTE: +=id should add res to id, but is caught as error earlier */
      c++;
      while (*c==' ') c++;
      b=c;
      if (isalpha(*c) || *c=='_') while (isalnum(*c) || *c=='_') c++;
      e=c;
      *e=0;
      if (b==e) {
        struct _Idlist_s *is,*isnext;

        /* remove all id's */
        if (key=='~') {
          for (is=_Id.head; is; is=isnext) {
            isnext=is->next;
            free(is); }
          _Id.head=NULL; }

        /* print all id's - reverse order */
        if (key=='=') {
          int n,i;
	  struct mac_s *m;

          for (n=0,is=_Id.head; is; is=is->next) n++;
          while (n) {
            n--;
            for (i=0,is=_Id.head; is; is=is->next) if (i++==n) break;
            myprintf(" %s = %.15g\n",is->id,is->val); }
          myprintf(" # = %.15g\n",res);
	  for (m=mhead; m; m=m->next)
            myprintf("def %s %s\n",m->X,m->E); } }
      else if (key=='~') deassign(b);
      else assign(b); /* assignop=extra (see above) */ }
    else {
      /* id=expr */
      calculate(c+1); /* called recursively: may handle cases as a=b=expr */
      if (!err) {
	int x=*e;
	*e=0;
	assignop=extra;
	assign(b);
	if (extra!='=') printres(); /* result of += -= *= /= */
	*e=x; } }
  else if (oper(expr)) switch (nop) {
    case SOLVE:
      if (isparm[1]) {
        int nr=0,n=500,i,j,flag;
#define NR 20
        double root[NR],froot[NR];
        double eps=1e-9*fabs(parm[1]-parm[0]);
        double from=parm[0],to=parm[1];
        double yrange=0;

        if (parm[1]<parm[0]) from=parm[1],to=parm[0];

        if (!isparm[2]) parm[2]=0;
        if (parm[2]) n=(int)(fabs((parm[1]-parm[0])/parm[2])+.5);
        if (n<2) n=2;
        if (n>100000) {
          myprintf("solve: too fine grid\n");
          return; }

        for (i=0; i<=n; i++) {
          char *c;

          res=parm[0]+(double)i/n*(parm[1]-parm[0]);
          if (sig) return;
          assign(var);
          res=Calc(eqn,&c);
          err=eqn==c;
          if (err) {
            myprintf("solve: bad equation\n");
            return; }
          yrange+=fabs(res); }
        yrange/=n;
        if (isnan(yrange)) yrange=9e99; /* ad hoc! */

        for (i=0; i<=n; i++) {
          res=parm[0]+(double)i/n*(parm[1]-parm[0]);
          if (!Solve(eps) && res>=from-eps && res<to+eps && Checkres(eps*yrange)) {
            flag=1;
            for (j=0; j<nr; j++) if (fabs(root[j]-res)<eps) {
              flag=0;
              break; }
            if (flag) {
              if (nr>=NR) {
                myprintf("solve: too many roots\n");
                return; }
              root[nr]=res;
              froot[nr]=froot_ext;
              nr++; } } }

        /* sort roots */
        do {
          flag=0;
          for (i=1; i<nr; i++)
            if ( (root[i-1]-root[i])*(parm[0]-parm[1])<0 ) {
              REAL aux;
              aux=root[i-1]; root[i-1]=root[i]; root[i]=aux;
              aux=froot[i-1]; froot[i-1]=froot[i]; froot[i]=aux;
              flag++; }
          } while (flag);

        for (i=0; i<nr; i++) {
          res=root[i];
          assign(var);
          if (verbose&1) printres2(" f=%.2g",froot[i]);
          else printres(); } }
      else {
        if (Solve(-1e-9*fabs(res=parm[0]))<0) return;
        printres(); }
      break;

    case MAX:
      MinMaxSg=-1;
      goto MINMAX;
    case MIN:
      MinMaxSg=1;

    MINMAX:
      if (isparm[1]) {
        /* in interval */
        int n=500,i,imin;
        double eps=1e-8*fabs(parm[1]-parm[0]);
        double from=parm[0],to=parm[1];
        double yrange=0;
        double absmin=9e99;

        if (parm[1]<parm[0]) from=parm[1],to=parm[0];

        if (!isparm[2]) parm[2]=0;
        if (parm[2]) n=(int)(fabs((parm[1]-parm[0])/parm[2])+.5);
        if (n<2) n=2;
        if (n>100000) {
          myprintf("min/max: too fine grid\n");
          return; }

        /* probably useless... */
        for (i=0; i<=n; i++) {
          char *c;
          res=parm[0]+(double)i/n*(parm[1]-parm[0]);
          if (sig) return;
          assign(var);
          res=Calc(eqn,&c);
          err=eqn==c;
          if (err) {
            myprintf("solve: bad equation\n");
            return; }
          yrange+=fabs(res); }
        yrange/=n;
        if (isnan(yrange)) yrange=9e99; /* ad hoc! */

        for (i=0; i<=n; i++) {
          res=parm[0]+(double)i/n*(parm[1]-parm[0]);
          if (sig) return;
          assign(var);
          res=MinMaxSg*Calc(eqn,&c);
          if (res<absmin) {
            imin=i;
            absmin=res; } }

        res=parm[0]+(double)imin/n*(parm[1]-parm[0]);

        if (imin>0 && imin<n) MinMax(eps);

        assign(var);
        printres2(MinMaxSg>0?" min=%.12g":" max=%.12g",MinMaxSg*MinMaxF); }
      else {
        if (MinMax(1e-8*fabs(res=parm[0]))<0) return;
        printres(); }
      break;

    case SUM: case PROD: {
      double s=nop==PROD;
      double x;

      if (!isparm[2]) parm[2]=0;
      if (parm[2]==0) parm[2]=parm[0]<=parm[1]?1:-1;
      if (parm[0]>parm[1]) x=parm[1],parm[1]=parm[0],parm[0]=x,parm[2]*=-1;
      for (res=parm[0]; res<=parm[1]; res+=parm[2]) {
        assign(var);
        x=Calc(eqn,&c);
        err=eqn==c;
        if (err) { myprintf("sum/prod: bad expression\n"); return; }
        if (sig) { myprintf("sum/prod: interrupted\n"); return; }
        if (nop==PROD) s*=x; else s+=x; }
      res=parm[0]; assign(var);
      res=s;
    enough:
      printres();
      break; }

    case REPEAT: case ITER: {
      int i,n=(int)parm[1],iserr=0;
      double last=parm[0];
      double err;

      res=parm[0];
      if (nop==ITER) n=100000;

      for (i=0; i<n; i++) {
        assign(var);
        res=Calc(eqn,&c);
        err=eqn==c;
        if (err) { myprintf("iter/repeat: bad expression\n"); return; }
        if (sig) { myprintf("iter/repeat: interrupted\n"); return; }
        if (nop==ITER) {
          iserr++;
          if (fabs(err=res-last)<=parm[1]) break;
          last=res; } }
      assign(var);
      if (iserr) printf("err=%g  ",err);
      printres();
      break; }

    case CONTFRAC: {
      /* format of command: contfrac VAR=FROM,TO[,BY] [BEFORE++]NUM//DENOM */
      double x=0,sg=1;
      double p=0,pp=1,q=1,qq=0,a,b,z;
      char *plus,*slash;

      if (!isparm[2]) parm[2]=0;
      if (parm[2]==0) parm[2]=parm[0]<=parm[1]?1:-1;
      if (parm[2]<0) sg=-1;
      plus=strstr(eqn,"++");
      if (plus) {
        *plus=0;
        x=Calc(eqn,&c);
        err=eqn==c;
        if (err) { myprintf("contfrac: bad expression before ++\n"); return; }
        eqn=plus+2; }

      slash=strstr(eqn,"//");
      if (!slash) {
        myprintf("contfrac: missing //\nSynopsis: contfrac N=FROM,TO[,BY] [BEFORE++]NUM//DENOM\n");
        return; }
      *slash=0;
      slash+=2;

      for (res=parm[0]; res*sg<=parm[1]*sg; res+=parm[2]) {
        assign(var);
        b=Calc(eqn,&c);
        err=eqn==c;
        if (err) { myprintf("contfrac: bad numerator\n"); return; }
        a=Calc(slash,&c);
        err=eqn==c;
        if (err) { myprintf("contfrac: bad denominator\n"); return; }
        if (sig) { myprintf("contfrac: interrupted\n"); return; }
        z=q*a+qq*b; qq=q; q=z;
        z=p*a+pp*b; pp=p; p=z;
/*..... printf("%g %g  %g+%g/%g\r\n",pp,qq,x,p,q); */
        }

      res=parm[0]; assign(var);
      res=x+p/q;
      goto enough; }

    case INTEG: {
      double x,d,s=0;

      /* WARNING bad algorithm for d !!!!*/
      d=setprec(parm[1]);
      x=setprec(parm[0]);
      if (x<d) d=x; /* ??? */
      if (d==0) d=1e-5; /* const */

      if (parm[0]*parm[1]<0) {
        x=setprec(0);
        if (x<d) d=x;
        d*=.1;
        s=integ(d,0,parm[1])-integ(d,0,parm[0]); }
      else if (fabs(parm[0])>fabs(parm[1]))
        s=-integ(d,parm[1],parm[0]);
      else
        s=integ(d,parm[0],parm[1]);
      res=s;
      goto enough; }

    case DERIV: {
      double d=setprec(parm[0]),y0[5],*y=y0+2;
      int i;

      d=pow(d,0.2);
      for (i=-2; i<=2; i++) if (i) {
        res=parm[0]+d*i;
        assign(var);
        y[i]=Calc(eqn,&c);
        err=eqn==c;
        if (err) {
          myprintf("deriv: bad expression\n");
          return; } }
      res=parm[0]; assign(var);
      res=(8*(y[1]-y[-1])-(y[2]-y[-2]))/(12*d);
      goto enough; }

    case PLOT: {
      FILE *q=fopen("/tmp/ev.tmp","wt");
      double y;
      int i,n=1000;

      if (!isparm[2]) parm[2]=0;
      if (parm[2]) n=(int)(fabs((parm[1]-parm[0])/parm[2])+.5);

      if (n<0 || n>1000001) {
        myprintf("plot: no or too many points\n");
        return; }

      for (i=0; i<=n; i++) {
        res=parm[0]+i*(parm[1]-parm[0])/n;
        assign(var);
        y=Calc(eqn,&c);
        err=eqn==c;
        if (sig) {
          myprintf("plot: interrupted\n");
          return; }
        if (err) {
          myprintf("plot: bad expression\n");
          return; }
        fprintf(q,"%.16g\t%.16g\n",res,y); }
      res=parm[0]; assign(var);
      fclose(q);
      if (system("export LAZYX11=-${LAZYX11}; plot /tmp/ev.tmp &"))
        myprintf("plot: system call failed\n"); }
      break;

    case DEF:
      defmacro(var,eqn);
      deassign(var);
      if (mode==FILTER) myprintf(" (defined)\n");
      break;
    case UNDEF:
      undefmacro(var);
      deassign(var);
      if (mode==FILTER) myprintf(" (undefined)\n");
      break;
      //    HELP:
      //      break;
    }
  else {
    res=Calc(expr,&c);
    err=c==expr;
    if (!err) {
      while (*c==' ') c++;
      if (*c) myprintf("ERROR at %s \n",c); }
    //    if (mode) myprintf("%s =",expr);
    if (err)
      myprintf(" ERROR\n");
    else
      printres(); }
}

int main(int narg,char **arg)
{
  char line0[PROMPT_N+1];
  char *line=line0+1;
  char *c,*semcol;
  char *evdata="/";
  FILE *f=NULL;
  int i,iarg,printed;
  int (*auxprintf)(const char *format,...);

  if (narg<2) {
    fprintf(stderr,"Command-prompt-oriented calculator, (c) J. Kolafa 1991-2012. Call by:\n\
  ev [OPTIONS] [EXPR [EXPR2 ..]]\n\
OPTIONS:\n\
  --         just a filter (the same as -d with no prompt)\n\
  --i[CONST] read file CONST with initial constants [default=$HOME/.evdata]\n\
  --i        do not read any initial constants (but pi=PI embedded in calc)\n\
  --v#       verbosity, sum of: 1=print f=(last value) after solve\n\
                                2=print expanded macros\n\
                                4=print initial constants\n\
                                8=print comments (FILTER only) [1+2+8]\n\
  --c[WIDTH] prompt-oriented calculator in current terminal (ncurses editing)\n\
             if WIDTH is missing, $COLUMNS is checked, otherwise 80 applies\n\
  --d        prompt-oriented calculator on a dumb terminal (no history/editing)\n\
  --fFMT     format (accepting double), LF appended; must be last option\n\
  --FFMT     as above, no LF appended; must be last option\n\
  --t        tex-formatted output\n\
  EXPR [EXPR2..] expression(s) (EXPR must not begin with \"--\")\n\
  \"?\"       print calculator help (also ? as a command)\n\
Order of options:\n\
  --i is applied first, then optional EXPR, then optional -c,-d,-f,-t\n\
Environment variables:\n\
  COLUMNS=# of columns in the terminal\n\
  HOME=home directory (used to locate .evdata)\n\
  TOCLIP=command to put string to the clipboard (e.g., xclip.sh)\n\
  EVFMT=format (the same as -f); for TOCLIP, ' ' and \\n are stripped off\n");

    exit(0); }

  myprintf=printf;
  if (getenv("EVFMT")) {
    strcpy(fmt,getenv("EVFMT"));
    strcat(fmt,"\n"); }
  else
    strcpy(fmt,"%.8g\n");

  //  toclip=getenv("TOCLIP");

  for (iarg=1; iarg<narg; iarg++)
    if (arg[iarg][0]=='-' && arg[iarg][1]=='-') switch(arg[iarg][2]) {
      case 'i':
	evdata=arg[iarg]+3;
	break;
      case 'v': verbose=atoi(arg[iarg]+2);
        break;
      case 'c': {
        mode=PROMPT;
        prompt=1;
        evgetline=mygetline;
        myprintf=cprintf;
	strcpy(fmt," %.8g\n");

        _pr.n=atoi(arg[iarg]+3);
        if (_pr.n==0) {
	  if (getenv("COLUMNS")) _pr.n=atoi(getenv("COLUMNS")); }
	if (_pr.n<8) _pr.n=80;
	_pr.n--;
        for (i=0; i<PROMPT_H; i++) strcpy(_pr.hist[i],"?>");

        initscr(); cbreak();
        noecho(); nl();
        scrollok(stdscr,TRUE);
        intrflush(stdscr, FALSE); }
        break;
      case 'd':
        mode=DUMB;
        prompt=1;
        strcpy(fmt," %.8g\n");
        myprintf=printf;
        evgetline=gets;
        break;
      case 0:
        mode=FILTER;
        prompt=0;
        myprintf=printf;
        evgetline=gets;
        break;
      case 'f':
        strcpy(fmt,arg[iarg]+3);
        strcat(fmt,"\n");
        break;
      case 'F':
        strcpy(fmt,arg[iarg]+3);
        break;
      case 't':
        mode=TEX;
        prompt=0;
	strcpy(fmt,"@%.8g\n");
        myprintf=texprintf;
        evgetline=gets;
        break;
      default: fprintf(stderr,"ev: wrong option (enclose expression starting with -- in ())\n");
        exit (0); }
    else break;

  signal(SIGINT,asksig);

  /* reading the initial constants */
  if (!strcmp(evdata,"/")) {
    char *h=getenv("HOME");
    if (h) {
      evdata=malloc(strlen(h)+9);
      strcpy(evdata,h);
      strcat(evdata,"/.evdata"); } }

  if (strlen(evdata)) f=fopen(evdata,"rt");

  auxprintf=myprintf;
  if (!(verbose&4)) myprintf=dummyprintf;

  isconst=1;
  if (f) while (fgets(line0,PROMPT_N+1,f)) {
    if ((unsigned char)line0[strlen(line0)-1]<' ') line0[strlen(line0)-1]=0;
    myprintf("%s\n",line0);
    if (strlen(line0)>2) calculate(line0); }
  isconst=0;

  myprintf=auxprintf;

  strcpy(line0,"#");

  if (mode==TEX)
    printf("\\documentclass[twocolumn]{article}\n\
\\usepackage{czech}\n\
\\usepackage{/home/jiri/tex/unitchem}\n\
\\parskip 0pt\\parindent 0pt\n\
\\topmargin -2cm\\oddsidemargin -1.6cm\n\
\\textwidth 19.12cm\\textheight 26cm\n\
\\columnwidth 9cm\\columnsep 0.62cm\n\
\\begin{document}\n\
\\pdfpagewidth=210mm \\pdfpageheight=297mm\n\
\\boldmath\n\
\\tt\n");

  for (;;) {
    if (prompt) {
      if (comma) strcpy(_pr.prompt," ,>");
      else strcpy(_pr.prompt," >");
      if (_Id.degrad==1) _pr.prompt[0]=toclip?'R':'r';
      else _pr.prompt[0]=toclip?'D':'d'; }
    else strcpy(_pr.prompt,"r>");

    /* ugly: for smart terminal, _pr.prompt used in mygetline (see prompt.c) */
    if (mode==DUMB) myprintf("%s",_pr.prompt);

    printed=0; /* echo only 1 comment/line: see --v8 and FILTER */

    if (iarg<narg) strcpy(line,arg[iarg++]);
    else if (!evgetline || !evgetline(line)) break;
    {
      char *c,*e=NULL;
      for (c=line; *c; c++) if ((unsigned char)*c<=' ') e=c; else e=NULL;
      if (e) *e=0;

      if (getenv("EV"))
        for (e=getenv("EV"); *e; e+=2)
          for (c=line; *c; c++) if (*c==e[0]) *c=e[1];
    }

    if (!strcmp(line,"exit") || !strcmp(line,"quit") || !strcmp(line,".")) break;

    if (mode==TEX) {
      if (!memcmp(line,"! #.",4)) {
        static int no=0;
        sprintf(line+1,"%2d",++no);
        line[3]='.'; }

      if (strlen(line)==0) strcpy(line,"!");
      texprintf("%s\n",line); }

    /* ! is comment */
    for (c=line; *c; c++) 
      if (*c=='!') {
        if (mode==FILTER && (verbose&8) && !printed) {
          printed++;
          puts(c); }
        *c=0; }

    if (comma) {
      for (c=line; *c; c++)
	if (*c==',')
	  if (c[1]==',') memmove(c+1,c+2,strlen(c+1));
	  else *c='.'; }

    do { /* loop over ; separating statements */
      semcol=strchr(line,';');
      if (semcol) *semcol=0;

      if (line[0]) {
	if (line[0]=='?') myprintf("\n\
3*sin(#+pi/6)     ! evaluate expression, #=last res\n\
solve x=1,2 x^2-2 ! solve equation (in range)\n\
S=sum x=1,9,3 x   ! calculate S=1+4+7 (not in expression)\n\
%%7.9f / %% / %%?    ! format: set / default / query\n\
deg / rad         ! set: degrees / radians\n\
-2*3    ! subtract 2*3 from last result\n\
 -2*3;S ! evaluate (-2*3) and S\n\
=id     ! assign last result to id\n\
id=2^3  ! evaluate and assign to id, also id1=id2=3; id+=1\n\
=       ! list all id's and macros (def X ANY_STRING)\n\
~id / ~ ! remove id / all id's\n\
time:   ! DD::hh:mm:ss (2:30=2.5, 13:06:36=13.11, 1::12=36)\n\
bin op: ! + - * / ^(power) `(modulo) & | $(xor) < > <= >= == <>\n\
func:   ! \\=sqrt abs sin cos tan asin.. exp ln log=log10 lb=log2\n\
        ! sinh.. asinh.. gamma lgamma erf erfc fact=fac(!)  pi=PI\n\
        ! int floor frac sign M(mol.mass)\n\
rnd(X): ! X<-1: seed, X=-1:[-1,1), 0:[0,1), 1:norm, N:{0..N-1}\n\
command:! solve min max integ deriv sum prod iter repeat contfrac\n\
        ! plot def undef (write command w/o arg to get help)\n\
sh cmd  ! shell command, status returned in #\n\
toggles:! comma(, instead of dec.pt. ,->,,)  > (prompt)  clip\n\
write FILE! write history to FILE [history.ev]\n\
. quit exit QUIT EXIT! quit\n\
");
	else if (!strcmp(line,"rad")) _Id.degrad=1;
	else if (!strcmp(line,"deg")) _Id.degrad=PI/180;
	else if (!strcmp(line,"comma")) comma=!comma;
	else if (!strcmp(line,"clip")) {
          if (toclip) toclip=NULL;
          else toclip=getenv("TOCLIP"); }
	else if (!memcmp(line,"sh ",3)) {
	  res=system(line+3);
	  printres(); }
	else if (!memcmp(line,"write ",6)) prompt_write(line+6);
	else if (!strcmp(line,"write")) prompt_write("history.ev");
	else if (!strcmp(line,">")) prompt^=1;
	else
	  if (strchr("+-*/^",line[0])) {
	    myprintf("# %c expr =",line[0]);
	    calculate(line0); }
	  else {
	    if (!memcmp(line,"def ",4)) {
	      /* DIRTY hack for equivalent syntax def A=x meaning def A (x) */
	      char *c=line+4;

	      while (*c==' ') c++;
	      for (; *c; c++) {
		if (*c=='=') {
		  memmove(c+1,c,strlen(c)+1);
		  c[0]=' ';
		  c[1]='(';
		  strcat(c,")");
		  break; }
		if (*c==' ') break; } }

	    calculate(line); }
        }
      if (semcol) memmove(line,semcol+1,strlen(semcol+1)+1);
    } while (semcol); }

  if (mode==PROMPT) endwin();
  if (mode==TEX) printf("\\end{document}\n");

  return -err;
}
