/* cc -o itexi -O2 itexi.c ; \mv itexi /home/jiri/bin
Math etc. TeXinfo preprocessor.  Should be called after ntexi.
Input lines starting with '/' are processed as in this example:
Input line
------------------------------------------------------------------
/ $a^2 \sum_{i=1}^{\infty} {\cos(i)\over i^2}$
------------------------------------------------------------------
gives on output
------------------------------------------------------------------
@tex
 $a^2 \sum_{i=1}^{\infty} {\cos(i)\over i^2}$
@end tex
@ifinfo
  a^2 SUM_i=1^INFTY cos(i) / i^2 
@end ifinfo
@ifhtml
 <i>a<sup>2</sup> SUM<sub>i=1</sub><sup>INFTY</sup> cos(i) / i<sup>2</sup></i>
@end ifhtml
------------------------------------------------------------------
Understood are: $$..$$, {\rm ..} etc.
BUG: bf,tt not translated to HTML
To create only specified mode, use
/i (only ifinfo)
/t (only tex)
/h (only html)
/it (ifinfo+tex)
etc.

//lines are copied with one / 

@comands in /lines are transformed only partly and similarly as
texi2html does, e.g., / @kbd{x} becomes <kbd>x</kbd>.  Nesting of
@commands is not allowed!

Only one level of indexes/exponents is treated

New paragraph (blank lines) are not correctly treated by texi2html.
Therefore any blank lines inside @table.. @end table and similar are
replaced by <p> and not ignored.
@table etc. MUST start from 1st column
*/

#include "../gen/include.h"

#define LINE 256
#define NB 100 /* buffer of remembered lines */

char line[LINE];
char block[NB][LINE];

int nb=0;
int type=-1;
int nlines;

void error(char *msg)
{
 int i;

 loop (i,0,nb) fprintf(stderr,"\"%s\"\n",block[i]);
 fprintf(stderr,"\"%s\"\n",line);
 Error(msg);
}

void tex()
{
 int i;
 
 printf("@tex\n");
 loop (i,0,nb) puts(block[i]);
 printf("@end tex\n");
}

char TOK[LINE],*left,*right;;

char *tok(char *l)
{
 static char *t;

 if (l) { t=l; return ""; }

 memset(TOK,0,LINE);

 if (*t=='\\') {
   int i=0;

   t++;
   while (isalpha(*t)) TOK[i++]=*t++;
/*.....   while (*t==' ') t++;*/
   if (!strcmp(TOK,"sum")
    || !strcmp(TOK,"prod")
    || !strcmp(TOK,"infty")) {
     char *c;
     for (c=TOK; *c; c++) *c=toupper(*c); }
   if (!strcmp(TOK,"over")) strcpy(TOK," /"); /* ratio */
   if (!strcmp(TOK,"left")) strcpy(TOK," ");
   if (!strcmp(TOK,"right")) strcpy(TOK," ");
   if (!strcmp(TOK,"langle")) strcpy(TOK,"<");
   if (!strcmp(TOK,"rangle")) strcpy(TOK,">");
   if (!strcmp(TOK,"AA")) strcpy(TOK,"A"); /* Angstrom */
   }

 else if (*t=='@') {
   /* @COMMAND or @COMMAND{..} */
   int nest=0,i=0;

   TOK[i++]=*t++;

   if (*t<' ') error("@ followed by EOL");

   if (!isalpha(*t)) {
     TOK[i++]=*t++;
     return TOK; }

   while (isalpha(*t)) TOK[i++]=*t++;

   left=right=NULL;
   if (*t!='{') return TOK;
   left=TOK+i;
   
   for (;; t++) {
     if (!*t) error("@COMMAND{..} not in one /-line");
     TOK[i++]=*t;
     if (*t=='{') nest++;
     if (*t=='}') {
       nest--;
       right=TOK+i-1; }
     if (!nest) return TOK; } }
     
 else {
   TOK[0]=*t++;
   if (TOK[0]=='$' && *t=='$') TOK[1]=*t++; }

return TOK;
} 

void info()
{
 int i;
 int nospace=0;
 
 printf("@ifinfo\n");
 loop (i,0,nb) {
   int lf;

   if (!memcmp(block[i],"@sp",3))
     loop (lf,0,atoi(block[i]+3)) putchar('\n');

   else {

     tok(block[i]);

     while (tok(NULL)[0])
       if (TOK[0]=='$') putchar(' ');
       else if (!strcmp(TOK," ")) { if (!nospace) putchar(' '); nospace=0; }
       else if (!strcmp(TOK,"~")) putchar(' ');
       else if (!strcmp(TOK,"rm")) nospace=1;
       else if (!strcmp(TOK,"bf")) nospace=1;
       else if (!strcmp(TOK,"tt")) nospace=1;
       else if (!strcmp(TOK,"{"));
       else if (!strcmp(TOK,"}"))/* putchar(' ')*/;
       else fputs(TOK,stdout);
     putchar('\n'); } }

 printf("@end ifinfo\n");
}

void html()
{
 int i;
 int dol=0,indx=0,nest=0,rm=0;
 int nospace=0,setsp=0;

 printf("@ifhtml\n");
 loop (i,0,nb) {
   int blank=1;

   tok(block[i]);

   while (tok(NULL)[0]) {

     if (setsp) {
       int i=atoi(TOK);
       while (i-- > 0) puts("<p>");
       setsp=0; }
     else if (!strcmp(TOK,"$")) {
       if (dol++&1) fputs("</i>",stdout);
       else fputs("<i>",stdout); }
     else if (!strcmp(TOK,"$$")) {
       if (dol++&1) fputs("<br></i>",stdout);
       else fputs("<i><br>",stdout); }
     else if (!strcmp(TOK," ")) { if (!nospace) putchar(' '); nospace=0; }
     else if (!strcmp(TOK,"~")) putchar(' ');
     else if (!strcmp(TOK,"<")) fputs("&lt;",stdout);
     else if (!strcmp(TOK,">")) fputs("&gt;",stdout);
     else if (!strcmp(TOK,"\"")) fputs("&quot;",stdout);
     else if (TOK[0]=='@') {
       if (isalpha(TOK[1])) {
	 if (left) {
	   char *cmd=TOK+1;
	   int ap=0; /* `..' */

	   *left=0;
	   if (right) *right=0;
	   else error("{} not matched");

	   /* most of @cmd unchanged ( @var{a} -> <var>a</var> )*/
	   if (!strcmp(cmd,"file")) cmd="tt",ap++;
	   else if (!strcmp(cmd,"key")) cmd="kbd";
	   else if (!strcmp(cmd,"samp")) ap++;
	   else if (!strcmp(cmd,"dfn")) cmd="strong";
	   else if (!strcmp(cmd,"url")) cmd="code"; /* ? */
	   else if (!strcmp(cmd,"email")) cmd="tt"; /* ? */
	   else if (!strcmp(cmd,"emph")) cmd="em";
	   if (ap) printf("<%s>`%s'</%s>",cmd,left+1,cmd);
	   else printf("<%s>%s</%s>",cmd,left+1,cmd); }
	 else if (TOK[1]=='s' && TOK[2]=='p') setsp++; }
       else if (TOK[1]==';') fputs("&nbsp;",stdout); }
     else if (!strcmp(TOK,"_")) fputs("<sub>",stdout),indx=1;
     else if (!strcmp(TOK,"^")) fputs("<sup>",stdout),indx=2;
     else if (!strcmp(TOK,"{")) { if (indx) nest=1; }
     else if (!strcmp(TOK,"rm")) fputs("</i>",stdout),rm=1,nospace=1;
     else if (!strcmp(TOK,"bf")) fputs("</i>",stdout),rm=1,nospace=1;
     else if (!strcmp(TOK,"tt")) fputs("</i>",stdout),rm=1,nospace=1;
     else if (!strcmp(TOK,"}")) {
       if (nest) {
	 nest=0;
	 if (indx&2) fputs("</sup>",stdout);
	 if (indx&1) fputs("</sub>",stdout);
	 if (rm) fputs("<i>",stdout);
	 indx=0; } }
     else {
       fputs(TOK,stdout);
       if (!nest) {
	 if (indx&2) fputs("</sup>",stdout);
	 if (indx&1) fputs("</sub>",stdout);
	 indx=0; } } 
     blank=0; }
   
   if (!blank) putchar('\n'); else puts("<br>"); }

 printf("@end ifhtml\n");
}


void flush(char *prt)
{
 char *cmd;

 if (type==0) tex(),info(),html();
 else if (type=='t') tex();
 else if (type=='i') info();
 else if (type=='h') html();
 else if (type=='h'+'i') info(),html();
 else if (type=='h'+'t') tex(),html();
 else if (type=='i'+'t') info(),tex();

 nb=0;
 if (prt) puts(prt);
 type=-1;
}

int main(int narg,char **arg)
{
 char atop[16][32];
 int atnest=0;

 fprintf(stderr,"TeXinfo math+html preprocessor: ntexi < FILE.ntex | itexi > FILE.texi\n");

 while (gets(line)) {
   char *l=line;
   
   while (*l && *l<=' ') l++;
   if (!*l) line[0]=0;

   if (line[0]=='@') {
     char tok[32],*t=tok,*l=line+1;
     
     while (isalpha(*l)) *t++=*l++;
     *t=0;
     if (!strcmp(tok,"end")) {
       char end[32],*e=end;
       while (*l==' ') l++;
       while (isalpha(*l)) *e++=*l++;
/*.....printf("@ifhtml\n<!-- END %d %s %s-->\n@end ifhtml\n",atnest,end,atop[atnest-1]);*/
       *e=0;
       if (!strcmp(end,atop[atnest-1]) || !strcmp(end,"group")) atnest--; }
     if (!strcmp(tok,"enumerate")
      || !strcmp(tok,"table")
      || !strcmp(tok,"itemize")
      || !strcmp(tok,"example")) {
/*.....printf("@ifhtml\n<!-- BEGIN %d %s -->\n@end ifhtml\n",atnest,tok);*/
       if (atnest>14) error("@nesting");
       strcpy(atop[atnest++],tok); } }
       

   /* ...line will also be copied to output */

   if (line[0]=='/') 
     if (line[1]=='/')
       flush(line+1);
     else {
       char *c=line+1;
       int newtype=0;

       for (;;) {
	 if (*c<=' ') break;
	 newtype+=*c;
	 c++; }

       while (*c==' ') c++;

       if (type<0) type=newtype;
       if (type==newtype) {
	 if (nb>=NB) error("block too long");
	 strcpy(block[nb++],c); }
       else {
	 flush(NULL);
	 type=newtype;
	 nb=0;
	 strcpy(block[nb++],c); }
       }
   else flush(line);

   if (line[0]==0) {
     if (atnest && atop[atnest-1][0]!='e') 
       /* make <p> within tables etc.; not within example */
       puts("@ifhtml\n<p>\n@end ifhtml"); }       
   nlines++; }

 printf("@comment filtered by itexi - math and html preprocessor\n");

 fprintf(stderr,"%d lines converted\n",nlines);
 return 0;
}
