/* cc -O2 -o lc lc.c
  rename file(s) to lowercase, (c) J.Kolafa 1996-2008
  2008: -8 changed to -7, -8 added (utf-8)
  2008: -f, -x, -L added
  2005: rename via other file (vfat) 
  2001: also Capitalize and UPPERCASE
  1999: check of file overwrite
*/
#include <stdlib.h>
#include <stdio.h>
#include <ctype.h>
#include <string.h>

enum {NONE,LOWER,CAPITALIZE,UPPER,EXT,UPPER2LOWER,PHOTO,REPRINT} style = LOWER;
enum {DIRECT,BIT7,CP1250,ISO88592,CP852,UTF8} code = DIRECT;

#define N 1024

void iso2asc(char *c)
{
  char asc[128]="~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~A~L$LS~~SSTZ~ZZ~a~l~ls~~sstz~zzRAAAALCCCEEEEIIDDNNOOOOxRUUUUYTSraaaalccceeeeiiddnnoooo-ruuuuyt";

  for (; *c; c++) *c = *c&128 ? asc[*c&127] : *c;
}

void win2asc(char *c)
{
  char asc[128]="~~~~~~~~~~S~STZZ~~~~~~--~~s~stzz~~~L~A~S~~S~~~~Z~~~l~~~~~as~L~l~RAAAALCCCEEEEIIDDNNOOOO~RUUUUYTSraaaalccceeeeiiddnnoooo~ruuuuyt";

  for (; *c; c++) *c = *c&128 ? asc[*c&127] : *c;
}

void dos2asc(char *c)
{
  char asc[128]="CueaauccleOoiZACELlooLlSsOUTtL caiouAaZzEe zCs~~~~~~~AAES~~~~Zz~~~~~~~Aa~~~~~~~ dDDEdNIIe~~~~TU~O~ONnnSsRUrUyYt      S~~~~~uRr~~";

  for (; *c; c++) *c = *c&128 ? asc[*c&127] : *c;
}

#include "4lc.c"

void utf82asc(char *s,char *end)
{
  char fn[N*2],*f=fn,*c; 
  struct utf8_s *u;
  
  memset(fn,0,N);

  for (c=s; *c; ) {
    for (u=utftab; u->utf; u++) 
      if (u->eq[0] && !memcmp(c,u->utf,strlen(u->utf))) {
        strcat(fn,u->eq);
        f+=strlen(u->eq);
        c+=strlen(u->utf); 
        goto cont; }
    *f++=*c++;
    cont:; }
     
  if (s+strlen(fn)>=end) {
    fprintf(stderr,"recoded name too long");
    exit(4); }

  strcpy(s,fn);      
}

int main(int narg,char **arg)
{
  int i,dry=0,vfat=0,haslc,repstat;
  char *reprint="--";
  char fn0[N],fn[N],*a,*f,*ar,xfrom=0,xto='-',yes=0;
  FILE *list,*over;
  char space=' ',*dot,*from;

  if (narg<2) {
    fprintf(stderr,"Rename files to lowercase and similar.  Call by:\n\
  lc [OPTION(s)] [@][PATH/]FILE [OPTION] [[@][PATH/FILE...]\n\
Recoding (performed first):\n\
  -7 = remove bit 8 (use ~ if not possible)\n\
  -8 = remove accents or try most similar from UTF-8 coding\n\
  -i = remove accents from ISO-8859-2 characters\n\
  -w = remove accents from CP1250 (windows) characters\n\
  -p = remove accents from CP852 (PC Latin II) characters\n\
Character change:\n\
  -xAB = change char A into B (cannot repeat)\n\
Case change (-nX etc.: change space into X, performed after -xAB)\n\
  -n = no case change\n\
  -l = lowercase (DEFAULT)\n\
  -L = Lowercase only names with all letters uppercase (A.SH->a.sh, Aa.c->Aa.c)\n\
  -c = Capitalize\n\
  -u = UPPERCASE\n\
  -e = lowercase extension (Aa.DOC->Aa.doc)\n\
  -f = lowercase first 6 chars and extension (IMG123-Az.JPG->img123-Az.jpg)\n\
  -rXAB = uppercase between chars AB (default=--), otherwise lowercase\n\
More options:\n\
  -d = dry run (show what would be renamed but do not actually rename)\n\
  -v = rename via ./vfat.aux; for case-insensitive filesystem (vfat)\n\
  -y = replace target without asking (default=asks for confirmation)\n\
File args:\n\
  PATH/ = path to file (not renamed)\n\
  FILE = file to rename\n\
  @FILE = list of [path/]files to rename\n\
Examples:\n\
Lowercase all files beginning with lowercase on a Windows-compatible disk:\n\
  lc -v -l_ /media/MyDisk/[a-z]*\n\
Remove accents from UTF8 file names, change spaces to _, do not change case:\n\
  lc -n_ -8 *\n\
See also: mvs ren\n\
");
    exit(0); }

  for (i=1; i<narg; i++) {

    if (arg[i][0]=='-') {
      switch (arg[i][1]) {
        case 'd': dry++; break;
        case 'n': style=NONE; break;
        case 'c': style=CAPITALIZE; break;
        case 'l': style=LOWER; break;
        case 'L': style=UPPER2LOWER; break;
        case 'e': style=EXT; break;
        case 'U':
        case 'u': style=UPPER; break;
        case 'f': style=PHOTO; break;
        case 'r': style=REPRINT; if (arg[i][3]) reprint=arg[i]+3; break;
        case '7': code=BIT7; break;
        case 'i': code=ISO88592; break;
        case 'w': code=CP1250; break;
        case 'p': code=CP852; break;
        case '8': code=UTF8; break;
        case 'x': xfrom=arg[i][2],xto=arg[i][3]; break;
        case 'v': vfat++; break;
        case 'y': yes++; break;
        default: fprintf(stderr,"%s is bad option\n",arg[i]); exit(0); }
      if (strchr("nclLuefr",arg[i][1]) && arg[i][2]) space=arg[i][2];
      continue; }

    else if (arg[i][0]=='@') {
      list=fopen(arg[i]+1,"rt");
      if (!list) {
        fprintf(stderr,"%s not found\n",arg[i]+1);
        exit(-1); } }
    else
      list=NULL;

    if (xfrom && !xto) {
      fprintf(stderr,"bad -x\n");
      exit(-1); }

    ar=arg[i];

  again:
    if (list) {
      if (fscanf(list,"%s",fn0)!=1) goto done;
      ar=fn0; }
   
    repstat=0;

    if (strlen(ar)>=N) {
      fprintf(stderr,"too long arg\n");
      exit(0); }
    strcpy(fn,ar);

    dot=NULL;
    from=fn;
    haslc=0;
    for (a=fn; *a; a++) {
      if (islower(*a)) haslc++;
      if (*a=='.') dot=a;
      if (*a=='/') from=a+1,dot=NULL,haslc=0; }

    /* coding change */
    switch (code) {
      case DIRECT: break;
      case BIT7:
        for (f=from; *f; f++) if (*(unsigned char*)f&128) {
          *(unsigned char*)f&=127;
          if (!isalnum(*f)) *f='~'; }
        break;
      case ISO88592: iso2asc(from); break;
      case CP1250:   win2asc(from); break;
      case CP852:    dos2asc(from); break;
      case UTF8:    utf82asc(from,fn+N); break; }

    /* lc, uc, capitalize */
    for (f=from; *f; f++) {
      switch (style) {
        case NONE: break;
        case UPPER: *f=toupper(*f); break;
        case LOWER: *f=tolower(*f); break;
        case REPRINT: 
          if (repstat==1) *f=toupper(*f);
          else *f=tolower(*f);
          if (*f==reprint[0] && !repstat) repstat=1;
          else if (*f==reprint[1] && repstat) repstat++;
          break;          
        case PHOTO: 
          if (f<from+6 || f>dot) *f=tolower(*f);
          break;
        case EXT: 
          if (f>dot) *f=tolower(*f);
          break;
        case UPPER2LOWER:
          if (!haslc) *f=tolower(*f);
          break;
        case CAPITALIZE:
          if (f==from) *f=toupper(*f); }
      if (xfrom) if (*f==xfrom) *f=xto;
      if (*f==' ') *f=space; }

    if (strcmp(ar,fn)) {
      printf("%s -> %s",ar,fn);
      if (!yes) {
        over=fopen(fn,"r");
        if (over) {
          char line[128];

          fclose(over);
          printf(" : target exists. Overwrite (y/n)? ");
          gets(line);
          if (!strchr("yY",line[0])) goto cont; } }

      if (dry)
        printf(" would be renamed\n");
      else {
        if (vfat) {
          if (rename(ar,"vfat.aux")) {
            printf(" : rename failed (no such file, access denied...)\n");
            exit(0); }
          if (rename("vfat.aux",fn)) {
            printf(" : rename failed (no such file, access denied...)\n");
            exit(0); } }
        else if (rename(ar,fn)) 
          printf(" : rename failed (no such file, access denied...)\n");
        else printf(" renamed\n"); }
    cont:; }
    if (list) goto again;
  done:; }

  return 0;
}
