/* cc -O2 -o makemake makemake.c
*/

#define METAMAKE "metamake" /* source metamake file */

/*** makemake.c *** (c) J.Kolafa 7/1993-10/2010 ***

Generate unix-style `makefile' or Turbo C project files from `metamake'.

Call by:
  makemake [unix|linux|emx|Prj|prj|tiny|small|compact|medium|large|huge] \
    [filelist] [DEF [...]]
The default type is linux

WARNING: switch makefile applies to unix/linux and emx only

Since NOTHING IS COMPATIBLE, I had to write this program which automatically
generates make files understandable by ALMOST ANY make utility.

History:
  10/10: linux mode made the default
  11/09: print PWD added
  10/09: getline renamed to getlongline (because of getline() in stdio)
   5/09: target-specific options OPT,LIBOPT added
   2/04: linking via CPLUS if at least 1 C++ module (because new gcc needs this)
         some string sizes increased
   7/01: !include statements added (because of MACSIMUS needs)
   9/00: bug fixed (conditionals with | &)
  11/99: small bug fixed
  11/97: conditionals (!if !else !endif ...) added,
         turboc.cfg automatically generated (because of NSK needs)
   9/96: $(OPT) now after -I (can use -I- in gcc)
   1/96: C++ support
   9/95: Turbo make format, all formats integrated
   6/95: aesthetical bug (too many empty lines printed) fixed
   1/95: bug in getline() fixed
  11/93: update
   7/93: 1st version

COMPILING `makemake':

UNIX, linux:
  cc -o makemake makemake.c
or
  gcc -O2 -o makemake makemake.c
or similar.

EMX port of gcc (OS/2 and DOS):
  gcc -O2 -o makemake.exe makemake.c

Turbo C 2.0
  Use IDE, recommended memory model=compact

Borland C++:
  Use IDE and turn off optimization, recommended memory model=compact


USING `makemake'
^^^^^^^^^^^^^^^^
Synopsis:
  makemake DEF [DEF ...]
where DEF is name (identifier) to define.
These DEFines may be used in the `metamake' file in conditional
processing (in the similar way as #ifdef NAME ... in the C preprocessor,
see below for details).

Example:
  makemake unix sgi

The following DEFines have special meaning and exactly one of them
must be specified:

  unix = generates UNIX-style `makefile'
  emx = for EMX port of gcc (DOS or OS/2), similar as `unix' but
        the executable name ends with .exe
  Prj = Projects for Turbo C 2.0 (IDE), generates *.prj files
      * use prjcnvt.exe to convert to Borland C++ (binary) *.prj files
  prj = as `Prj' + omits full path from files in the last directory of the
        !dir statement (assumed to be the working directory)
  tiny = for Turbo C 2.0 compiler (tcc): generates `makefile.dos', response
         `*.lnk' files (for tlink), and `turboc.cfg' (option file for tcc)
         for the Tiny memory model.
       * use `tcconfig.exe' to convert `turboc.cfg' <-->
         `tcconfig.tc' [options for IDE]
       * see also commands !cc !tclib !tcdir !tcopt below
  small = as above, Small memory model
  compact = as above, Compact memory model
  medium = as above, Medium memory model
  large = as above, Large memory model
  huge = as above, Huge memory model

In addition:
  linux = DEFines unix
  (linux is the default)
  dos = is automatically DEFined for Prj, prj, tiny, small, compact,
        medium, large, or huge
  filelist = will generate file "filelist" with a list of all files
             (relatively to !home, alphabetically)

Arrangement of work:
^^^^^^^^^^^^^^^^^^^^

* There are several directories containing source files (*.c), header
  files (*.h), and other files #included while compilation.  These
  directories are subdirectories (sub-subdirectories...) of one `home'
  directory (specified by statement `!home').  These directories
  (without the `home' part) are listed in statement `!dir'.  The order
  of directories corresponds to the order in which #include files are
  looked for.  For option `prj', the last directory in the list must
  be the actual (working or project) directory, for other options it
  does not matter.

* The `#include "***"' statements in the source files should contain
  only file names, not paths.  `makemake' finds the directory
  containing the file automatically but is confused if qualified names
  are given.

* Executables are obtained by linking from object modules (extensions
  given by statement `!objext') and libraries.  All executables and
  object modules are placed in the working directory, irrespective of
  the location of the corresponding sources.  (This arrangement allows
  to use the same source code with different options for one project
  and with another options for another project residing in another
  directory).


`metamake[.dos]'
^^^^^^^^^^^^^^^^
`makemake' reads file `metamake' (or `metamake.dos', see below) and
processes it, generating `makefile' (or `makefile.dos').  It
recognizes lines starting with `!' in the 1st column as metamake
statements.  All other statement are copied unchanged to the output .
(does not apply for prj and Prj).

For unix and emx: input file `metamake' should contain definitions of
the following make macros (variables):
  CC = <name of the C compiler>
  CPLUS = <name of the C++ compiler> // if needed
  OPT = <compile options (used for compiling *.c to object files)>
  LIBOPT = <libraries to link (e.g. `-lm')>

For dos, input file `metamake.dos' is tried first, if it does not
exist then `metamake'; however, it is recommended to have one file
`metamake' (instead of `metamake.dos') with conditional processing,
e.g.:
!if dos
myprog:myprog dosgraph
!else
myprog:myprog Xgraphics
!endif

For not dos (= unix or emx), `metamake' is read unconditionally


Metamake commands:
^^^^^^^^^^^^^^^^^^
* list means items separated by spaces
* lines may be continued in the makefile-like style by `\'

General commands:

!!       comment
!home = <fully qualified name of the home (root) directory, default = empty>
!dir = <list of directory names (max 9), must end with the working directory>
!include  include file, "" around name optional, use always `/' as dir sep
!objext = <ext. for obj-files (def. is `.o' for unix and emx, otherwise `.obj'>
!ignore = <list of (not qualified) files not to include into dependencies>
!treat = <list of files to undo 1 previous !ignore>
!<exe1> : [MOD:] [MOD:] <list of object file names without extension>
!<exe2> : [MOD:] [MOD:] <list of object file names without extension>
etc.

The modifier MOD serves to replace option $(OPT) and $(LIBOPT) and works
with Unix-style makefile only (no DOS):
MOD = OPT=<compile option to replace $(OPT) in compiling the target>
      LIBOPT=<linker option to replace $(LIBOPT)>
NOTE: `:' delimiting MODs must be in a single line, spaces around are allowed
          (list of names may then span more lines)
Example (test will be compiled with -O0 and linked with -lm;
         ground is compiled with the OPT set in the file):
  test : OPT=-O0 : LIBOPT=-lm : test ground

DOS commands:

!tcdir = <directory of the compiler (needed for tiny, small, ..., huge)>
!tclib = <list of extra libraries for linking, fully qualified>
!tcopt = <options for turboc.cfg>
!tcopt = +<options to be added to the default (see variable TCOPTDEF)>
!cc    = <DOS C-compiler, default=tcc>
Example: for Borland C++, use
!cc = bcc
NOTE: the DOS mode is tuned for Turbo C 2.0 and is probably usable
after eventual minor changes also for Borland C++.

Conditionals:

!if NAME     (max nesting level=8)
!ifdef NAME  (= !if NAME)
!else
!endif
!define NAME
!def NAME    (= !define NAME)
!undef NAME

These conditionals have similar meaning as in the C preprocessor.  The
differences are:
* NAMEs cannot have value, they are only defined or undefined
* Limited expressions only are allowed:
!if NAME1 | NAME2 | NAME3 ... (= !if NAME1 || NAME2 || NAME3 ... )
!if NAME1 & NAME2 & NAME3 ... (= !if NAME1 && NAME2 && NAME3 ... )
* `|' and `&' cannot be combined in one statement, parentheses not supported
* `!' in front of variable name means `not'
  there must not be a space after `!'

Example:
!if dos & !Windows

Another example:
!if dos
!objext=.obj
!else
!objext=.o
CC=gcc
!endif

Example of unsupported conditional:
!if unix & high | ! dos

Processing:
^^^^^^^^^^^
* The directory names (statement `!dir') are relative to `!home', the
  value of `home' (with `/' or `\' added) is prepended to the
  directory names to obtain fully qualified names.

* Dependencies and building (=linking) rules for exe's are created.
  The modules listed accept extension given by `!objext'

* Dependencies and building rules for all object files are created.
  Each object file encountered in the list is assumed to depend on the
  corresponding *.c file and consequently to all files that are
  included by `#include "***"' from it (system files included by
  `#include <***>' are not considered).  The `#include' statements may
  be nested.  All `#include "***"' statements (with # in the 1st
  column) are used though they need not be active because of certain
  values of #defines.  (The part of makemake calculating dependencies
  from sources is much simpler than the C preprocesor.  This is not
  disadvantage, however, because no dependency can be forgotten.  If
  #defines change in #included files, the makefile need not be
  calculated again; if they are set by compiler options -D, one must
  know the real dependencies of modules on these options anyway.)

* Output make-file has lines at most 80 characters long, longer are
  continued by `\`

Diagnostics:
^^^^^^^^^^^^
* If file (*.c or #included) is not found, $(D?) instead of the
  correct macro for the directory name is written and a message is
  printed

Bugs/caveats/features:
^^^^^^^^^^^^^^^^^^^^^^
* `#include "***"' within a comment in source files or within false
  conditionals is still treated as valid dependency (provided that #
  is in 1st column)

* The output listing with object files is in reverse order than one
  would expect.

* One file must not appear in more directories.  If it does, the first
  one only is considered.

***/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>

char *myfgets(char *s, int size, FILE *stream)
/* version of fgets with converting DOS -> UNIX */
{
  char *ret=fgets(s,size,stream),*c;

  if (ret)
    for (c=s; *c; c++)
      if (*c==13)
        if (c[1]==10)
          if (c[2]) *c=' ',c[1]=' '; /* (should not happen...) */
          else *c=10,c[1]=0; /* CR LF -> LF */
        else
          if (c[1]) *c=' '; /* (should not happen...) */
          else *c=0; /* CR -> LF */

  return ret;
}

#ifdef DOS
#define BUFLEN 1024
#define FNLEN 64
#define NDIRS (6+1) /* max 6 directories handled */
#define NINCL 4 /* # of nest levels for !include statements */
#else
#define BUFLEN 4096
#define FNLEN 256
#define NDIRS (9+1) /* max 9 directories handled */
#define NINCL 6 /* # of nest levels for !include statements */
#endif

#define LINELEN 77 /* for output */

char *dirsep;
char *targetname;
struct namelist_s {
  char name[FNLEN];
  struct namelist_s *next; };
struct project_s {
  char name[FNLEN];
  struct namelist_s *namelist;
  struct project_s *next; } *project0;

struct objlist_s {
  char name[FNLEN];
  struct objlist_s *next; } *objlist0;

FILE *in[NINCL],*out;
char *inname[NINCL];
char inpathname[BUFLEN];

char key,KEY;
char home[128];
int len=0;
int notfound=0;

int ndirs=1;
char dirs[NDIRS][FNLEN];
char objext[FNLEN];
char tcdir[FNLEN]="C:\\TC20\\";
char *tclib=""; /* extra lib */
char outsep='\\';

struct file_s {
  int type;
  int cplus; /* set if c++ */
  int todo;
  char name[FNLEN];
  char *xopt;
  struct list_s *dep;
  struct file_s *next; } *f0;

struct list_s {
  struct file_s *f;
  struct list_s *next; };

struct ignore_s {
  char name[FNLEN];
  struct ignore_s *next; } *ignore0;

int nottoignore(char *x) /************************************* nottoignore */
{
  struct ignore_s *ig;

  for (ig=ignore0; ig; ig=ig->next)
    if (!strcmp(ig->name,x)) {
      fprintf(stdout,"%s ignored\n",x);
      return 0; /* found in the list of ignored */ }

  return 1; /* not found */
}

void scopy_l(char *a,char *b,int l) /******************************** scopy */
{
  if (strlen(b)+5>=FNLEN) {
    fprintf(stderr,"too long name: \"%s\" in line %d\n",b,l);
    exit(4); }
  strcpy(a,b);
}

#define scopy(A,B) scopy_l(A,B,__LINE__)

void *myalloc(unsigned size) /************************************* myalloc */
{
  void *v=malloc(size);

  if (v!=NULL) return v;
  fprintf(stderr,"cannot malloc\n");
  exit(4);
}

char *mystrdup(const char *s) /************************************ mystrdup */
{
  char *n=strdup(s);

  if (n) return n;
  fprintf(stderr,"cannot malloc\n");
  exit(4);
}

char *scan(char **C,char sep[2]) /************************************ scan */
{
  char *c=*C,*t=c;

  key=' ';
  while (*c==' ' || *c=='\t') c++;
  if (*c==0) return NULL;
  t=c;
  while (*c!=' ' && *c!='\t' && *c!=sep[0] && *c!=sep[1] && *c!=0) c++;
  while (*c==' ' || *c=='\t') *c++=0;
  if (*c==sep[0] || *c==sep[1]) { key=*c; *c++=0; }
  *C=c;

  return t;
} /* scan */

void putobj(char *name) /******************************************* putobj */
/* prints name of 1 obj file */
{
  int l=strlen(name)+strlen(objext)+1;
  struct objlist_s *ol;

  if (KEY=='P') return;

  if ( (len+=l) > LINELEN ) {
    fprintf(out," \\\n "); len=l+1; }
  fprintf(out," %s%s",name,objext);

  if (KEY!='U' && KEY!='E') {
    ol=myalloc(sizeof(*ol));
    scopy(ol->name,name);
    ol->next=objlist0;
    objlist0=ol; }
}

int omitlastpath;
int depnames; /* to replace `targetname ()' in dependency by `targetname
                 (targetname.c)' because Turbo C does not like empty list */
char *depname=""; /* stored targetname */

#define MAXFILELIST 1000
char **filelist;
int nfilelist;

void includefilelist(char *di,char*name)
{
  char c[2*FNLEN];
  int i;

  sprintf(c,"%s/%s",di,name);
  for (i=0; i<nfilelist; i++) if (!strcmp(c,filelist[i])) return;

  if (nfilelist>=MAXFILELIST)
    fprintf(stderr,"too many files in filelist\n");
  filelist[nfilelist++]=mystrdup(c);
}

int cmpfilelist(const void *a,const void *b)
{
  return strcmp(*(char**)(a),*(char**)(b));
}

void putitem(char *name,int type) /******************************** putitem */
/* prints 1 item (typically file name) */
{
  int l,last=(type==ndirs-1)&omitlastpath;
  if (!type) return;

  if (filelist) includefilelist(dirs[type],name);

  if (KEY=='P') {
    l=strlen(name)+2;
    if (!last) l+=strlen(home)+strlen(dirs[type]);
    if (targetname) {
      if ( (len+=l) > LINELEN ) {  /*... this is stupid - must be 1st on line */
        fprintf(out," \n "); len=l+1; }
      if (type>0) {
        depnames=0;
        depname=myalloc(strlen(targetname)+strlen(home)+strlen(dirs[type])+4);
        if (last) {
          fprintf(out,"%s (",targetname);
          sprintf(depname,"%s.c",targetname); }
        else {
          fprintf(out,"%s%s\\%s (",home,dirs[type],targetname);
          sprintf(depname,"%s%s\\%s.c",home,dirs[type],targetname); } }
      else {
        fprintf(stderr,">>> ???\\%s\n",targetname);
        notfound=2;
        fprintf(out," ???\\%s (",targetname); } }
    else {
      if ( (len+=l) > LINELEN ) {
        fprintf(out," \n "); len=l+1; }
      if (type>0) {
        if (depnames++) fputc(' ',out);
        if (last) fputs(name,out);
        else fprintf(out,"%s%s\\%s",home,dirs[type],name); }
      else {
        fprintf(stderr,">>> ???\\%s\n",name);
        fprintf(out," ???\\%s",name); } }
    targetname=NULL; }
  else {
    l=strlen(name)+7;
    if ( (len+=l) > LINELEN ) {
      fprintf(out," \\\n "); len=l+1; }
    if (type>0)
      fprintf(out," $(D%d)%c%s",type,outsep,name);
    else {
      fprintf(stderr,">>> $(D?)/%s\n",name);
      fprintf(out," $(D?)/%s",name); } }
}

int containscplus=0;

struct file_s *dependency(char *name,int obj,char *xopt) /******* dependency */
{
  struct file_s *f;
  struct list_s *l;
  char fn[FNLEN];
  int i;

  for (f=f0; f; f=f->next)
    if (!strcmp(f->name,name)) return f; /* found in the list */

  /* not found */
  fprintf(stdout,"%s\n",name);
  f=myalloc(sizeof(struct file_s));
  f->type=-1;
  f->cplus=0;
  if (xopt) f->xopt=strdup(xopt);
  else f->xopt="$(OPT)";
  strcpy(f->name,name);
  f->next=f0; f0=f;
  f->dep=NULL;

  if (obj) {
    /* *.o file that has *.c direct dependency only (.o is not in the name) */
    scopy(fn,name); strcat(fn,".c");
    l=myalloc(sizeof(struct list_s));
    f->type=0;
    f->dep=l;
    l->next=NULL;
    l->f=dependency(fn,0,NULL);
    if (obj==2) {
      fprintf(stdout,"C++ ");
      containscplus++;
      l->f->cplus=2; }
    else if (l->f->type==-1) {
      strcat(fn,"c");
      fprintf(stdout,"C++ ");
      l->f=dependency(fn,0,NULL);
      containscplus++;
      l->f->cplus=1; } }

  else {
    /* search for file in all directories */
    for (i=1; i<ndirs; i++) {
      FILE *fi;
      char fullfn[FNLEN*2+128];
      char *r;
      char *x;

      r=myalloc(BUFLEN);
      strcpy(fullfn,home);
        strcat(fullfn,dirs[i]); strcat(fullfn,dirsep); strcat(fullfn,name);
      fi=fopen(fullfn,"rt");
      if (fi!=NULL) {
        f->type=i;
        while (myfgets(r,BUFLEN,fi))
          if (r[0]=='#') {
            char *c=r+1;
            if (!strcmp(scan(&c,"\"<"),"include")) if (key=='\"') {
              if ( (x=scan(&c,"\">")) )
                if (nottoignore(x)) {
                  strcpy(fn,x);
                  l=myalloc(sizeof(struct list_s));
                  l->next=f->dep;
                  f->dep=l;
                  l->f=dependency(fn,0,NULL); } } }
        fclose(fi);
        return f; }
      free(r); } }

  return f;
} /* dependency */

void alldep(struct file_s *f) /************************************** alldep */
{
  struct list_s *l;

  if (f) if (f->todo) {
    putitem(f->name,f->type);
    f->todo=0;
    for (l=f->dep; l; l=l->next) alldep(l->f); }
} /* alldep */

static char buf[BUFLEN]; /* for line incl. concatenation */

char *getlongline(void) /*************************************** getlongline */
/* reads one line incl. continuation lines ended by backslash */
{
  char *c=buf;

  if (!in[0]) return NULL;

  again:
    if (myfgets(c,(buf+BUFLEN)-c,in[0])==NULL) return NULL;
    while (*c!='\n' && *c!=0 && c<buf+BUFLEN) c++;
    if (c>=buf+BUFLEN) {
      fprintf(stderr,"concatenated line too long\n");
      exit(6); }
    if (c>buf) c--;
    while (*c==' ') {
      if (c==buf) { *c=0; goto retbuf; }
      c--; }
    if (*c=='\\') { *c++=' '; goto again; }

  *++c=0;

  retbuf:
  if (buf[0]=='\n' && buf[1]==0) buf[0]=0; /* oops ... empty line "\n" -> "" */

  return buf;
} /* getlongline */


struct def_s {
  struct def_s *next;
  char name[1]; } *def0,*def1;

char *name0="?",*name1="?";

int isdef(char *name)
{
  struct def_s *def;

  for (def=def0; def; def=def->next) if (!strcmp(def->name,name)) {
    name1=def->name;
    def1=def;
    return 1; }

  return 0;
}

int define(char *name)
{
  struct def_s *def;

  if (isdef(name)) return 1;

  def=myalloc(sizeof(struct def_s)+strlen(name));
  def->next=def0; def0=def;
  strcpy(def->name,name);

  return 0;
}

int undef(char *name)
{
  if (isdef(name)) {
    def1->name[0]=1; /* very cheap... */
    return 1; }

  return 0;
}

void unique(char newkey)
{
  if (KEY) {
    fprintf(stderr,"arguments \"%s\" and \"%s\" are exclusive\n",name1,name0);
    exit(16); }
  name0=name1;
  KEY=newkey;
}

char activestat[8]={1}; /* 1 if active (=defined) */
int ifnest;

int getactive(void)
{
  int i;
  for (i=0; i<=ifnest; i++) if (!activestat[i]) return 0;
  return 1;
}

char TCOPTDEF[]="-wamp -wcln -wuse -wrvl -wsig -wamb -wstv -wpro -wnod -wucp -1 -a -f87 -N -G -O -Z -d -DDOS";
char *tcoptdef=TCOPTDEF; /* must be array: strtok will be used */
char *tcoptplus="";
char *cc="tcc"; /* DOS C compiler, default for command !cc */
FILE *tcfg;

void puttcfg(char *str)
{
  char *tok=strtok(str," \n");

  while (tok) {
    fprintf(tcfg,"%s\n",tok);
    tok=strtok(NULL," \n"); }
}

void getxopt(char **opt,char *target,char *s)
{
  char *c;

  if (!memcmp(target,s,strlen(target))) {
    c=strchr(s,'=');
    if (!c) {
      fprintf(stderr,"%s not followed by =\n",target);
      exit(12); }
    c++;
    while (*c && *c<=' ') c++;
    *opt=c; }
}

void makepath(void) /********************************************* makepath */
{
  int i;
  char *x,*sl;

  inpathname[0]=0;
  for (i=NINCL-1; i>=0; i--) if (inname[i]) {
    if (inname[i][0]=='/')
      strcpy(inpathname,inname[i]);
    else {
      if (strlen(inpathname)+strlen(inname[i])>1023) {
        fprintf(stderr,"too long !include arguments\n");
        exit(4); }
      if (i==0 || strchr(inname[i],'/')) strcat(inpathname,inname[i]); }
    if (i && strchr(inname[i],'/')) {
      sl=inpathname;
      for (x=inpathname; *x; x++) if (*x=='/') sl=x;
      if (sl!=inpathname) sl[1]=0; } }

#ifdef DOS
  for (x=inpathname; *x; x++) if (x=='/') x='\\';
#endif
}

int main(int narg,char **arg) /**************************************** main */
{
  char *c,*token,*x,*makefile="";
  struct file_s *ff,*f;
  int i,nobj=0,active=1,ifop;
  /* ifnest= !if nesting level,  active shadows activestat[ifnest] */
  char ifstat[8]; /* 1=!if, 0=!else */
  struct def_s *def;

  struct project_s *project;
  struct namelist_s *namelist;

  fprintf(stderr,"\
*** makemake (part of MACSIMUS) *** (c) J. Kolafa 1993--2010 ***\n");

  if (narg<2) {
    fprintf(stderr,"\
Generate unix-style `makefile' from `metamake'.\n\
Call by:\n\
  makemake DEF [DEF..]\n\
where DEFs are keywords (conditionals) handled in `metamake'.\n\
Special DEFs:\n\
  filelist = generate `filelist' listing all needed files\n\
  linux = set makefile-style to linux (default; the same as unix)\n\
  makefile = all modules will depend also on the makefile\n\
             (= recompile all if the makefile changes)\n\
Legacy DEFs (see the comments in makemake.c):\n\
  unix emx Prj prj tiny small compact medium large huge\n\
Example:\n"
#ifdef DOS
"  makemake small"
#else
"  makemake lj polar colorgcc\n"
#endif
      );
    exit(4); }
  else
    for (i=1;i<narg;i++) define(arg[i]);

  fprintf(stderr,"makemake: PWD=%s\n",getenv("PWD"));

  if (isdef("linux")) define("unix");

  if (isdef("unix")) unique('U');
  if (isdef("emx")) unique('E');
  if (isdef("prj")) unique('p');
  if (isdef("Prj")) unique('P');
  if (isdef("tiny")) unique('T');
  if (isdef("small")) unique('S');
  if (isdef("medium")) unique('M');
  if (isdef("compact")) unique('C');
  if (isdef("large")) unique('L');
  if (isdef("huge")) unique('H');

  if (isdef("filelist")) filelist=malloc(MAXFILELIST*sizeof(filelist[0]));

  if (isdef("makefile")) makefile=" makefile";

  if (!KEY) {
    define("unix"); unique('U');
    fprintf(stderr,"type defaulted to unix/linux\n"); }

  omitlastpath=KEY>='a'; /* used only for p/P */
  KEY=toupper(KEY);

  if (strchr("PTSMCLH",KEY)) define("dos");
  else if (isdef("dos")) {
    fprintf(stderr,"cannot dos if unix or emx");
    exit(8); }

  if (KEY=='U' || KEY=='E')
    outsep='/';
  else {
    in[0]=fopen("metamake.dos","rt");
    if (in[0]!=NULL) fprintf(stderr,"WARNING: using \"metamake.dos\"\n"); }

  if (in[0]==NULL) in[0]=fopen(METAMAKE,"rt");
  if (in[0]==NULL) {
    fprintf(stderr,"file \""METAMAKE"\" not found\n");
    exit(4); }

  inname[0]=mystrdup("metamake"); /* irrelevant -- only (none) subdirectory needed */

  if (KEY=='P')
    out=NULL;
  else {
    out=fopen(KEY=='U' || KEY=='E' ? "makefile":"makefile.dos","wt");
    if (out==NULL) {
      fprintf(stderr,"cannot write out\n");
      exit(4); } }

  if (KEY=='E' || KEY=='U') {
    dirsep = "/"; strcpy(objext,".o"); }
  else {
    dirsep = "\\"; strcpy(objext,".obj"); }

  if (KEY!='P') {
    fputs("\
### this makefile was generated from metamake using the makemake utility ###\n\
# defined: ",out);
    for (def=def0; def; def=def->next) fprintf(out," %s",def->name);
    fprintf(out,"\n"); }

  for (;;) {
    while (!(c=getlongline())) {
      int i;

      if (in[0]==NULL) break;
      free(inname[0]);
      for (i=0; i<NINCL-1; i++) {
        in[i]=in[i+1];
        inname[i]=inname[i+1]; }
      in[NINCL-1]=NULL;
      inname[NINCL-1]=NULL; }

    if (!c) break; /* EOF on all !include files */

    if (*c!='!') {
      if (KEY!='P' && active) fprintf(out,"%s\n",c); }
    else {
      char *sem,*xopt,*xlibopt;

      c++;
/*.....    key=' ';*/
      if ( (token=scan(&c,":=")) ) switch(key) {

        case ' ':
          if (!strcmp(token,"!")) /* comment */;

          else if (!strcmp(token,"include")) {
            int i;

            if (in[NINCL-1]) {
              fprintf(stderr,"!include %s: too many nested levels\n",c);
              exit(3); }

            for (i=NINCL-1; i>0; i--) {
              in[i]=in[i-1];
              inname[i]=inname[i-1]; }
            x=scan(&c,"\"\"");
            if ( (x=strchr(c,'\"')) ) *x=0;
            inname[0]=mystrdup(c);
            makepath();
            in[0]=fopen(inpathname,"rt");
            if (in[0])
            fprintf(stdout,"including \"%s\"\n",inpathname);
          else {
              fprintf(stderr,"cannot open !include file \"%s\"\n",inpathname);
              exit(3); } }
          else if (!strcmp(token,"if") || !strcmp(token,"ifdef")) {
            if (ifnest>=7) {
              fprintf(stderr,"too many nested !if's\n");
              exit(12); }
            ifnest++;
            ifstat[ifnest]=1;
            ifop=0;
            while ( (x=scan(&c,"&|")) ) if (*x) {
              if (ifop && key!=' ' && key!=ifop) {
                fprintf(stderr,"!if: illegal combination of | and &\n");
                exit(12); }
              i=*x=='!';
              if (i) {
                x++;
                if (!*x) {
                  fprintf(stderr,"!if: '!' misplaced (or followed by space)\n");
                  exit(12); } }
              i^=isdef(x);
              if (!ifop) active=i;
              if (ifop=='&' && active) active=i;
              if (ifop=='|' && !active) active=i;
              ifop=key; }
            activestat[ifnest]=active;
            active=getactive(); }

          else if (!strcmp(token,"else")) {
            if (ifnest==0 || !ifstat[ifnest]) {
              fprintf(stderr,"misplaced !else\n");
              exit(12); }
            ifstat[ifnest]=0;
            activestat[ifnest]^=1;
            active=getactive(); }

          else if (!strcmp(token,"endif")) {
            if (ifnest<=0) {
              fprintf(stderr,"!endif without matching !if\n");
              exit(12); }
            ifnest--;
            active=getactive(); }

          else if (!strcmp(token,"def") || !strcmp(token,"define")) {
            if (active) if ((x=scan(&c,",,"))) if (*x) {
              define(x);
              if (KEY!='P') fprintf(out,"# %s defined\n",x); } }

          else if (!strcmp(token,"undef")) {
            if (active) if ((x=scan(&c,",,"))) if (*x) {
              undef(x);
              if (KEY!='P') fprintf(out,"# %s undefined\n",x); } }

          else if (!strcmp(token,"error")) {
            if (active) { fprintf(stderr,"!error %s\n",c);
            exit(12); } }

          break;

        case '=':
          if (!active) break;

          if (!strcmp(token,"!")) /* comment */;
          if (!strcmp(token,"home")) {
            if (home[0]) fprintf(stderr,"\aWARNING !home redefined\n");
            if (ndirs>1) {
              fprintf(stderr,"!home after !dir illegal\n");
              exit(4); }
            if ( (x=scan(&c,",,")) ) { strcpy(home,x); strcat(home,dirsep); } }

          if (!strcmp(token,"ignore"))
            while ( (x=scan(&c,",,")) ) {
              struct ignore_s *ig=myalloc(sizeof(struct ignore_s));
              if (strlen(x)>=FNLEN) {
                fprintf(stderr,"too long name in !ignore : %s\n",x);
                exit(4); }
              strcpy(ig->name,x);
              ig->next=ignore0; ignore0=ig; }

          if (!strcmp(token,"treat"))
            while ( (x=scan(&c,",,")) ) {
              struct ignore_s *ig;
              /* masked, not freed (memory leak) */

              for (ig=ignore0; ig; ig=ig->next)
                if (!strcmp(ig->name,x)) {
                  strcpy(ig->name,"-/void/");
                  break; } }

          if (!strcmp(token,"objext")) {
            if ( (x=scan(&c,",,")) ) {
              scopy(objext+(x[0]!='.'),x); } }

          if (!strcmp(token,"tcdir")) {
            if ( (x=scan(&c,",,")) ) {
              char *bsl;
              scopy(tcdir,x);
              bsl=tcdir+strlen(tcdir)-1;
              if (*bsl!='\\' && *bsl!='/') strcat(tcdir,dirsep); } }

          if (!strcmp(token,"tcopt")) {
            x=c;
            while (*x<=' ') x++;
            if (*x=='+') tcoptplus=mystrdup(x+1);
            else tcoptdef=mystrdup(x),tcoptplus=""; }

          if (!strcmp(token,"tclib")) {
            if (tclib[0]) fprintf(stderr,"!tclib redefined\n");
            tclib=mystrdup(c); }

          if (!strcmp(token,"cc")) {
            cc=mystrdup(c); }

          if (!strcmp(token,"dir")) {
            if (ndirs>1) fprintf(stderr,"more than one !dir statement (appended, %d dirs)\n",ndirs);
            if (home[0]==0) {
              fprintf(stderr,"!dir without !home\n");
              exit(4); }
            while ( (x=scan(&c,",,")) ) {
              scopy(dirs[ndirs],x);
              if (KEY!='P')
                fprintf(out,"D%d = %s%s\n",ndirs,home,x);
              ndirs++;
              if (ndirs>NDIRS) {
                fprintf(stderr,"too many dirs\n"); exit(4); } } }

          break;

        case ':':
          if (!strcmp(token,"!")) /* comment */break;
          if (!active) break;
          /* exe from obj command */
          containscplus=0;
          x=token;
          nobj++;

          /* list of exe's (used by P or make all,clean) */
          project=myalloc(sizeof(*project));
          scopy(project->name,x);
          project->namelist=NULL;
          project->next=project0;
          project0=project;

          if (KEY=='P') /* postponed */ ;
          else if (KEY=='U' || KEY=='E') {
            fprintf(out,"OBJ%d =",nobj); len=6+(nobj>9); }
          else {
            fprintf(out,"%s.exe : ",x); len=strlen(x)+7; }

          xlibopt="$(LIBOPT)";
          xopt=NULL; /* it means the default = $(OPT) */

          while ( (sem=strchr(c,':')) ) {
            while (*c<=' ') c++;
            getxopt(&xopt,"OPT",c);
            getxopt(&xlibopt,"LIBOPT",c);
            *sem=0; c=sem+1;
            while (sem[-1] && sem[-1]<=' ') { sem--; *sem=0; } }

          while ( (token=scan(&c,",,")) ) {
            char *plus=strchr(token,'+');

            if (plus) *plus=0;
            putobj(token);
            if (KEY=='P') {
              namelist=myalloc(sizeof(struct namelist_s));
              scopy(namelist->name,token);
              namelist->next=project->namelist;
              project->namelist=namelist; }
            dependency(token,1+(plus!=NULL),strcmp(x,token)?NULL:xopt); }

          /* print exe:obj dependencies and linking rule */
          if (KEY=='U' || KEY=='E') {
            fprintf(out,"\n%s%s :%s $(OBJ%d)\n",
                    x,KEY=='U'?"":".exe",makefile,nobj);
            fprintf(out,"\t$(%s) -o %s%s $(OBJ%d) %s\n\n",
                  /* linking requires CPLUS if at least one module is C++ */
                    containscplus?"CPLUS":"CC",
                    x,KEY=='U'?"":".exe",nobj,
                    xlibopt); }
          else if (KEY!='P') {
            char fn[FNLEN];
            FILE *lnk;
            struct objlist_s **op,**oo;

            strcpy(fn,x); strcat(fn,".lnk");
            lnk=fopen(fn,"wt");
            fprintf(lnk,"/c %sLIB\\C0%c+\n",tcdir,KEY);
            while (objlist0) {
              for (op=&objlist0; *op; op=&(*op)->next) oo=op;
              fprintf(lnk,"%s ",(*oo)->name);
              free(*oo); *oo=NULL; }
            fprintf(lnk,",+\n\
%s,+\n\
%s,+\n\
%sLIB\\MATH%c %sLIB\\GRAPHICS %sLIB\\FP87 %s %sLIB\\C%c\n",
                  x,x,tcdir,KEY,tcdir,tcdir,tclib,tcdir,KEY);
            fclose(lnk);
            fprintf(out,"\n\ttlink @%s\n",fn); }
          break;

        } /* switch */

    } }

  if (ifnest) {
    fprintf(stderr,"missing !endif\n");
    exit(12); }

  if (KEY=='P') {
    for (project=project0; project; project=project->next) {
      if (!out) {
        char fn[FNLEN];

        scopy(fn,project->name); strcat(fn,".prj");
        out=fopen(fn,"wt");
        if (!out) { fprintf(stderr,"cannot write to %s\n",fn); exit(4); } }

      /* print dependencies and compiling rules for obj files */
      for (f=f0; f; f=f->next) if (f->type==0) {
        for (namelist=project->namelist; namelist; namelist=namelist->next)
        if (!strcmp(namelist->name,f->name)) {
          targetname=f->name;
          len=2+strlen(f->name)+strlen(objext);
          for (ff=f0; ff; ff=ff->next) ff->todo=1;
          alldep(f);
          if (depnames!=0) depname="";
          fprintf(out,"%s)\n",depname); } }
      fclose(out);
      out=NULL; }
    }

  else {

    /* print dependencies and compiling rules for obj-files */
    for (f=f0; f; f=f->next) if (f->type==0) {
        fprintf(out,"%s%s :%s",f->name,objext,makefile);
      len=2+strlen(f->name)+strlen(objext);
      for (ff=f0; ff; ff=ff->next) ff->todo=1;
      alldep(f);
      if (KEY=='U' || KEY=='E') {
        fprintf(out,"\n\t$(%s)",f->dep->f->cplus?"CPLUS":"CC");
        fprintf(out," -o %s%s",f->name,objext);
        for (i=1; i<ndirs; i++) fprintf(out," -I$(D%d)",i);
        fprintf(out," %s",f->xopt); }
      else
        fprintf(out,"\n\t%s",cc);
      fprintf(out,
              " -c $(D%d)%c%s.%s\n\n",
              f->dep->f->type,outsep,f->name,f->dep->f->cplus&1?"cc":"c"); }

    {
      int l; 

      fprintf(out,"all :");
      len=6;
      for (project=project0; project; project=project->next) {
        l=strlen(project->name)+1;
        if ( (len+=l) > LINELEN ) {
          fprintf(out," \\\n "); len=l+1; }         
        fprintf(out," %s",project->name); }
      fprintf(out,"\n\n"); /* NOTHING after TAB ? */

      fprintf(out,"clean :\n\trm");
      len=10;
      for (project=project0; project; project=project->next) {
        l=strlen(project->name)+1;
        if ( (len+=l) > LINELEN ) {
          fprintf(out," \\\n "); len=l+1; }         
        fprintf(out," %s",project->name); }
      for (f=f0; f; f=f->next) if (f->type==0) {
        l=strlen(f->name)+strlen(objext)+1;
        if ( (len+=l) > LINELEN ) {
          fprintf(out," \\\n "); len=l+1; }         
        fprintf(out," %s%s",f->name,objext); }

      fprintf(out,"\n"); }

    fclose(out); }

  if (strchr("TSMCLH",KEY)) {
    tcfg=fopen("turboc.cfg","wt");
    fprintf(tcfg,"-m%c\n-I",tolower(KEY));
    for (i=1; i<ndirs; i++) fprintf(tcfg,"%s%s;",home,dirs[i]);
    fprintf(tcfg,"%sINCLUDE\n",tcdir);
    fprintf(tcfg,"-L%sLIB\n",tcdir);
    puttcfg(tcoptdef);
    puttcfg(tcoptplus);
    fclose(tcfg); }

  if (filelist) {
    qsort(filelist,nfilelist,sizeof(filelist[0]),cmpfilelist);
    out=fopen("filelist","wt");
    for (i=0; i<nfilelist; i++) fprintf(out,"%s\n",filelist[i]);
    fclose(out); }

  return notfound;
}
