/* make minmax
 : ground alloc
*/
#include "ground.h"
#include "alloc.h"
#include "matrix.c"
#include "ludecomp.c"
#include "ms.h"

int main(int narg,char **arg)
{
 char line[128];
 int ic,i,N;
 int iwid,nx,nit=6;
 double dx,maxx,wid,dwid=0;
 double *x,*y;

 initscroll(0);

 if (narg<5) {
   fprintf(stderr,"Find minima and maxima on a function (e.g., RDF). Call by:\n\
  minmax N DX MAXX WIDTH DWIDTH [NIT]< X-Y-FILE\n\
N = # of params (polynomial upto x^(N-1))\n\
DX = x grid\n\
MAXX = max x considered\n\
WIDTH = triangular window width (initial value)\n\
DWIDTH = next extrem uses WIDTH+=DWIDTH (default=0)\n\
NIT = # of iterations; nit<0: verbose (default=6)\n");
   exit(0); }

 N=atoi(arg[1]);
 dx=atof(arg[2]);
 maxx=atof(arg[3]);
 wid=atof(arg[4]);
 if (narg>5) dwid=atof(arg[5]);
 if (narg>6) nit=atoi(arg[6]);
 nx=maxx/dx;

 x=newvector(0,nx);
 y=newvector(0,nx);
 loop (i,0,nx) x[i]=y[i]=0;
 while (gets(line))
   if (line[0]!='#') {
     double xx,yy;
     
     if (2!=sscanf(line,"%lf%lf",&xx,&yy)) continue;
     i=xx/dx+0.1;
     if (i>=0 && i<nx) x[i]=xx,y[i]=yy; }

 iwid=wid/dx+1;
 ic=iwid;
 while (ic<nx-iwid) {
   int ic0=ic;

   if ((y[ic]-y[ic-iwid])*(y[ic]-y[ic+iwid])>0) {
     int k;
     double xc=x[ic],xx,xc0,*b0=newvector(0,N);
     int i0,i1,ia,ib,i;

     loop (k,0,abs(nit)) {
     
       double *sum=newvector(0,N*2);
       double **A=newmatrix(0,N,0,N),*b=newvector(0,N);
       int *indx=newintvector(0,N);

       i0=ic-iwid;
       i1=ic+iwid+1;
       xc0=xc;
       loop (ia,0,2*N) {
         sum[ia]=0;
         if (ia<N) b[ia]=0; }

       loop (i,i0,i1) {
         double Dx=x[i]-xc,q=wid-fabs(Dx);

         if (q<0)
	   if (fabs(q)>dx*2.0001) {
	     fprintf(stderr,
		     "%d %d %d %d %g %g too wide window or range problem\n",
		     i0,i,i1,ic,xc,q);
	     goto again; }
	   else q=0;
         
         if (i>i0 && fabs((x[i]-x[i-1])/dx-1)>1e-3)
           ERROR(("bad grid %d %g %g",i,x[i],x[i-1]))
         loop (ia,0,2*N) {
           sum[ia]+=q;
           if (ia<N) b[ia]+=q*y[i];
           q*=Dx; } 
         }

       loop (ia,0,N)
         loop (ib,0,N) A[ia][ib]=sum[ia+ib];
     
       LUdecomposition(0,N,indx,A);
       LUbacksubst(0,N,A,indx,b);
     
       loop (ia,0,N) b0[ia]=b[ia];
       loop (ia,0,N-1) b[ia]=(ia+1)*b[ia+1];

       xx=0;
       MS_BEGIN(xx,1e-7)
         if (MS_it>30) {
           fprintf(stderr,"%g: iterations failed\n",x[ic]);
           goto again; }
         MS_f=b[N-2];
         for (ia=N-3; ia>=0; ia--) MS_f=MS_f*xx+b[ia];
       MS_END(xx,0)

       xc+=xx;
       if (nit<0) fprintf(stderr,"%d %d %.9f\n",k,ic,xc);
       if (k!=abs(nit)-1) {
	 i=ic+xx/dx;
	 if (abs(i-ic)>iwid) {
	   prt("recenter too far");
	   goto again; }
	 if (x[i]-xc>dx/2) i--;
	 if (x[i]-xc<-dx/2) i++; 
	 ic=i; }

       freevector(0,N*2,sum);
       freevector(0,N,b);
       freeintvector(0,N,indx);
       freematrix(0,N,0,N,A); }

     {
       double s=b0[N-1];
       for (ia=N-2; ia>=0; ia--) s=s*xx+b0[ia]; 
       prt("%7.4f %7.4f %5.3f %.2g",xc,s,wid,xc-xc0);
     }
     if (nit<0)
       loop (i,i0,i1) {
         double Dx=x[i]-xc0,s=b0[N-1];
         for (ia=N-2; ia>=0; ia--) s=s*Dx+b0[ia]; 
         prt("%7.4f %7.4f",x[i],s); } 
     freevector(0,N,b0);

     ic+=iwid;
     wid+=dwid;
     iwid=wid/dx+1; 
     }

   again: ic++; 
   if (ic<=ic0) ic=ic0+1; }

return 0;
}
