/* cc -o ntexi -O2 ntexi.c ; \mv ntexi /home/jiri/bin
Node TeXinfo preprocessor.  Input lines starting with '/' are processed as
in this example:
Input line
------------------------------------------------------------------
/1 node1 // chaptername // menuitem
...
/m
/2 node2 // sectionname // menuitem
------------------------------------------------------------------
gives the output
------------------------------------------------------------------
@node node1, next, previous, Top
@chapter chaptername
...
@menu
* node2::        menuitem
@end menu
@node node2, next, previous, node1
@section sectionname
------------------------------------------------------------------
(with correct next/previous in the level /1 and /2 structure)

text after // in /1 nodes is translated via @chapter{}
text after // in /2 nodes is translated via @section{}
text after // in /3 nodes is translated via @subsection{}
text after // in /4 nodes is translated via @subsubsection{}
similarly:

/1u = @unnumbered
/2u = @unnumberedsec
/3u = @unnumberedsubsec, etc.
/1a = @appendix
/1a = @appendixsec, etc.

/m = generates @menu
/M generates @menu also for one level deep nodes

Other / lines are copied unchanged

Should be used BEFORE itexi preprocessor
*/

#include "../gen/include.h"

#define LINE 256
#define MENU 24 /* offset of menu description */

char *sec[6]={"","@chapter","@section","@subsection","@subsection","???"};
char *unn[6]={"","@unnumbered","@unnumberedsec","@unnumberedsubsec","@unnumberedsubsubsec","???"};
char *app[6]={"","@appendix","@appendixsec","@appendixsubsec","@appendixsubsubsec","???"};

struct line_s {
  struct line_s *next;
  char *text; } *line0;

struct node_s {
  struct node_s *next;
  struct node_s *prev;
  struct node_s *up;
  struct node_s *down;
  char key;
  char *name;
  char *menu;
  char *secname; 
  struct line_s *text; } *node0;

char line[LINE];

void endspace(char *c)
{
 char *e=NULL;

 for (;*c;c++) if (*c!=' ') e=c;
 if (e) e[1]=0;
}

int main(int narg,char **arg)
{
 int last=0,level;
 struct line_s **lapp,*l;
 struct node_s *n[6];

 fprintf(stderr,"TeXinfo node/section preprocessor: ntexi < FILE.ntex | itexi > FILE.texi\n");

 alloconezero(node0);
 n[0]=node0;
 lapp=&node0->text;

 while (gets(line)) {

   if (line[0]=='/') {
     level=line[1]-'0';

     if (level>=1 && level<=4) {
       struct node_s *nn;
       char *m,*sl=strstr(line,"//");
       char *c=line+2;
       char key;

       key=*c;
       if (key!=' ') c++;

       while (*c==' ') c++;
       if (!*c) Error("empty /level");

       if (sl) {
         *sl=0,sl+=2;
         while (*sl==' ') sl++;
         m=strstr(sl,"//");
         if (m) {
           *m++=0; m++;
           while (*m==' ') m++;
           endspace(m); }
         else m=sl; }

       else m=sl=c;


       endspace(c);
       endspace(sl);

       alloconezero(nn);

       nn->key=key;
       nn->name=strdup(line+3);
       nn->secname=strdup(sl);
       nn->menu=strdup(m);

       if (last>=level) {
         n[level]->next=nn;
         nn->prev=n[level];
         nn->up=n[level]->up;
         n[level]=nn; }
       else if (last<level) {
         if (level-last>1) {
           fprintf(stderr,"%s\n",line);
           fprintf(stderr,"last=%d level=%d\n",last,level);
           Error("missing level"); }
         n[level]=nn;
         n[last]->down=nn;
         nn->up=n[last]; }

       last=level;
       lapp=&n[level]->text; 
       goto done; }
     } 

   allocone(*lapp);
   (*lapp)->text=strdup(line);
   lapp=&(*lapp)->next;
   *lapp=NULL;

   done:; }

 level=0;
 if (n[0]!=node0) Error("");
 node0->name="Top";

 for (;;) {
   char *next="",*prev="",*up="Top";

   if (n[level]->next) next=n[level]->next->name;
   if (n[level]->prev) prev=n[level]->prev->name;
   if (n[level]->up) up=n[level]->up->name;

   if (level) {
     char key=n[level]->key;

     printf("@node %s,%s,%s,%s\n", n[level]->name,next,prev,up);
     printf("%s %s\n",key=='u'?unn[level]:key=='a'?app[level]:sec[level],
            n[level]->secname); }

   for (l=n[level]->text; l; l=l->next) {

     /* /m = auto generate menu */
     if (l->text[0]=='/' && tolower(l->text[1])=='m') {
       struct node_s *nn;

       printf("@menu\n");
       for (nn=n[level]->down; nn; nn=nn->next) {
         int i=MENU-printf("* %s::",nn->name);

         if (i<1) i=0;
         printf("%*s%s\n",i,"",nn->menu); } 

       if (l->text[1]=='M')
         /* /M = auto generate one-level nested menu */
         for (nn=n[level]->down; nn; nn=nn->next) if (nn->down) {
           struct node_s *nnn;
           int i=printf("\n%s\n",nn->secname)-2;

           while (i--) putchar('^');
           putchar('\n');

           for (nnn=nn->down; nnn; nnn=nnn->next) {
             i=MENU-printf("* %s::",nnn->name);

             if (i<1) i=0;
             printf("%*s%s\n",i,"",nnn->menu); }
         }

       printf("@end menu\n"); }

     else puts(l->text); }

   if (n[level]->down) {
     level++;
     n[level]=n[level-1]->down;
     continue; }

 again:

   if (n[level]->next) {
     n[level]=n[level]->next;
     continue; }

   level--;
   if (level<=0) break; else goto again;
   }

 printf("@comment filtered by ntexi - node/section preprocessor\n");
         
 return 0;
}
