/*  cc -O2 -Wall -I/home/jiri/gen -o polyfit polyfit.c  -lm
WARNING: use more general c/high/polyfit.c
(c) J. Kolafa
update 1995
update 2003
bug fixed 9/2004
*/

#include "include.h"
#include "matrix.c"
#include "ludecomp.c"
#include "rndgeni.c"

struct xy_s {
  struct xy_s *next;
  double x,y,std; } *head;

double **A,*b,*x,*s;
int *indx;
int nc;

void fitdata(double stdev)
/* fits the data
   stdev<0: Gaussian of stdev*xy->std added to data
   stdev>0: Gaussian of stdev added to data
*/
{
  struct xy_s *xy;
  int i,j,k;
  double w,y,g;

  loop (k,0,2*nc) s[k]=0;
  loop (k,0,nc) b[k]=0;

  looplist (xy,head) {
    w=1/Sqr(xy->std);
    y=xy->y;
    if (stdev) {
      g=rnd()-rnd()+rnd()-rnd()+rnd()-rnd()+rnd()-rnd()+rnd()-rnd()+rnd()-rnd();
      if (stdev<0) y+=g*stdev*xy->std;
      else y+=g*stdev; }
    loop (k,0,2*nc) {
      if (k<nc) b[k]+=w*y;
      s[k]+=w; w*=xy->x; } }

  loop (i,0,nc) loop (j,0,nc) A[i][j]=s[i+j];

  LUdecomposition(0,nc,indx,A);
  LUbacksubst(0,nc,A,indx,b);
}


int main(int narg,char **arg)
{
  int order=1,n=0;
  int i,j,k,nstd=0;
  char line[2048];
  struct xy_s *xy;
  double xx=0,yy,ss=1,sw,q,sum,sum1;
  double *b0,*bs,*bss,stdparm=1,fitstd;

  if (narg<2) {
    fprintf(stderr,"Polynomial fit of X Y STDEV data. Call by:\n\
  polyfit ORDER [[-]NSTD [RESCALE]] < DATA\n\
where\n\
  ORDER = polynomial order (# of constants = ORDER+1)\n\
  NSTD = # of random samples to calculate stderr of parameters,\n\
         stdev calculated from the fit is used for all points\n\
  -NSTD = as above, STDEV specified in data (3rd column) is used\n\
  RESCALE = STDEV given above is rescaled, resulting stderr is rescaled back\n\
  DATA = lines of form:\n\
    X Y STDEV : input X,Y,STDEV data\n\
    X Y       : input X,Y data (STDEV from previous data, or STDEV=1 if none)\n\
    Y         : input Y data (x=previous_x+1, x=0 if no previous_x)\n\
");
    exit(0); }

  order=atoi(arg[1]);
  if (order<1 || order>1024) Error("order");
  if (narg>2) nstd=atoi(arg[2]);
  if (narg>3) stdparm=fabs(atof(arg[3]));

  rndinit(0,0);

  while (gets(line)) {

    switch (sscanf(line,"%lf%lf%lf",&yy,&xx,&ss)) {
    case 0: 
      break;
    case 1: goto add;
    case 2: 
    case 3:
      sw=xx,xx=yy,yy=sw;
    add:
      n++;
      allocone(xy);
      xy->x=xx; xx+=1;
      xy->y=yy;
      xy->std=ss;
      xy->next=head;
      head=xy; } }

  nc=order+1;

  s=newvector(0,2*nc);
  b=newvector(0,nc);
  A=newmatrix(0,nc,0,nc);
  indx=newintvector(0,nc);

  fitdata(0);

  printf("\n# i coeff[i]\n");
  loop (i,0,nc) printf("# %d %.16g\n",i,b[i]);

  printf("\n# ");
  loop (i,0,nc-1) printf("%.16g+A*(",b[i]);
  printf("%.16g",b[nc-1]);
  loop (i,0,nc-1) printf(")");
  printf("\n");

  printf("\n#x y-poly(x) stdev\n");
  sum=sum1=0;
  looplist (xy,head) {
    q=1; yy=0;
    loop (k,0,nc) { yy+=q*b[k]; q*=xy->x; }
    printf("%.8g %.8g %.8g\n",xy->x,xy->y-yy,xy->std); 
    sum+=Sqr((xy->y-yy)/xy->std); 
    sum1+=Sqr(xy->y-yy); 
    xy->y=yy; }

  fitstd=sqrt(sum/(n-nc));
  printf("# stdev of fit=%g, n=%d, nc=%d, sumsq=%g\n",fitstd,n,nc,sum1);
  fprintf(stderr,"# stdev of fit=%g\n",fitstd);

  if (nstd) {
    if (nstd<0) {
      /* data-specified stdev */
      nstd=-nstd;
      stdparm=-stdparm; }
    else
      /* calculated stdev -- assumed the same for all data */
      stdparm*=sqrt(sum1/(n-nc)); 

    b0=newvector(0,nc);
    bs=newvector(0,nc);
    bss=newvector(0,nc);

    loop (i,0,nc) {
      b0[i]=b[i];
      bs[i]=bss[i]=0; }

    loop (j,0,nstd) {
      fitdata(stdparm);
      loop (i,0,nc) {
	bs[i]+=b[i];
	bss[i]+=Sqr(b[i]); } }

    loop (i,0,nc)
      printf("# %d %.16g %.6g\n",
	     i,b0[i],sqrt((bss[i]-Sqr(bs[i])/nstd)/(nstd-1))); 

    freevector(0,nc,bs);
    freevector(0,nc,bss); }

  freevector(0,2*nc,s);
  freeintvector(0,nc,indx);
  freematrix(0,nc,0,nc,A);
  freevector(0,nc,s);

  return 0;
}
