/* cc -O2 -o ren ren.c
 */
#include "../gen/include.h"

int main(int narg,char **arg)
{
  char *opt="",*op="mv";
  int from=3,i,ok=0,ko=0;
  int ask=1,dry=0,verb=1,all=0,cp=0;
  enum mode_e { ANYWHERE,BEGIN,END } mode = ANYWHERE;
  
  if (narg<4) {
    fprintf(stderr,"Rename files, replacement for the old good \"rename\". Call by:\n\
  ren OLD NEW [OPTION_STRING] FILE [FILE ...]\n\
Substrings OLD in file names are renamed to NEW.\n\
No regular expressions/wildcards are allowed in OLD and NEW!\n\
OPTION_STRING (merged as e.g. -fq):\n\
  --  empty option (good if FILE starts with '-')\n\
  -a  replace all occurrences (default=1st only)\n\
  -b  substring OLD can appear at the beginning only\n\
  -c  copy (cp) instead of rename (mv)\n\
  -d  dry run (only prints what would be renamed)\n\
  -e  substring OLD can appear at the end only\n\
  -f  force rename even if the target exists (default=ask for confirmation)\n\
  -p  flags-preserving copy (cp -p) instead of rename (mv)\n\
  -q  quiet (only final statistics printed)\n\
  -y  the same as -f\n\
Examples:\n\
  ren -2- -3- aaa-V-2-x.zip # the same as \"mv aaa2-V-2-x.zip aaa2-V-3-x.zip\"\n\
  ren .bak \"\" -e *.bak      # strip off all suffixes .bak\n\
  ren \"\" x- -yb sys-?.txt   # prepend x- before files sys-?.txt, overwrite\n\
                            # any existing file without asking\n\
See also: lc mvs\n\
");
    exit(0); }

  if (arg[3][0]=='-') {
    opt=arg[3],from=4;
    if (strchr(opt,'a')) all=1;
    if (strchr(opt,'c')) cp=1,op="cp";
    if (strchr(opt,'p')) cp=2,op="cp -p";
    if (strchr(opt,'q')) verb=0;
    if (strchr(opt,'b')) {
      if (mode==END) Error("options -eb are mutually exclusive");
      mode=BEGIN; } 
    if (strchr(opt,'d')) dry=1;
    if (strchr(opt,'e')) {
      if (mode==BEGIN) Error("options -eb are mutually exclusive");
      mode=END; } 
    if (strpbrk(opt,"fy")) ask=0; }    

  if (mode && all) Error("-a with -eb: only one replacement allowed at begin/end");

  loop (i,from,narg) if (strlen(arg[1])<=strlen(arg[i])) {
    char *match=NULL,*a,*b;
    FILE *f;
    char newname[1024];
    char line[256];

    switch (mode) {
      case ANYWHERE: match=strstr(arg[i],arg[1]);
                     break;
      case BEGIN: if (!memcmp(arg[i],arg[1],strlen(arg[1]))) match=arg[i];
                  break;
      case END: match=arg[i]+strlen(arg[i])-strlen(arg[1]);
                if (memcmp(match,arg[1],strlen(arg[1]))) match=NULL;
                break;
      default: Error("internal"); }
      
    if (strlen(arg[i])+strlen(arg[2])>512) Error("too long name");
    if (match) {
      int nrepl=0;
        
      strcpy(newname,arg[i]);
      strcpy(newname+(match-arg[i]),arg[2]);
      strcat(newname,match+strlen(arg[1])); 
        
      if (all) 
        while ( (match=strstr(newname,arg[1])) ) {
          memmove(match+strlen(arg[2]),match+strlen(arg[1]),strlen(match+strlen(arg[1]))+1);
          memcpy(match,arg[2],strlen(arg[2]));
          if (strlen(newname)>256) Error("too long name while replacement");
          if (nrepl++>256) Error("too many replacements (recursion?)"); }  
      
      if (!strcmp(arg[i],newname)) Error("the new name is identical to the old one");
      if (strlen(newname)>255) Error("the final name is longer than 255 B (ext2 limit)");
      
      if (ask) {
        f=fopen(newname,"r");
        if (f) {
          fclose(f);
          printf("%s exists. Overwrite (y/n)? ",newname);
          gets(line);
          if (line[0]!='y') goto ex; } }
          
      ko++;
      if (dry) {
        if (verb) printf("%s \"%s\" \"%s\" # DRY RUN\n",op,arg[i],newname); }
      else if (cp) {
        static char sys[10240];

        sprintf(sys,"%s \"%s\" \"%s\"",op,arg[i],newname);
        if (system(sys)) 
          fprintf(stderr,"%s %s %s failed\n",op,arg[i],newname);
        else {
          ok++;
          if (verb) 
            printf("\"%s\" copied to \"%s\"\n",arg[i],newname); } }
      else {
        if (rename(arg[i],newname)) 
          fprintf(stderr,"rename %s %s failed\n",arg[i],newname);
        else {
          ok++; 
          if (verb) 
            printf("\"%s\" renamed to \"%s\"\n",arg[i],newname); } }
          
      ex:; } }        

  ko-=ok;
  if (dry)
    printf("%d file%c to be renamed or copied\n",ko,"s"[ko==1]);
  else if (ko)
    printf("%d file%c %s and %d failed\n",ok,"s"[ok==1],cp?"copied":"renamed",ko);
  else
    printf("%d file%c %s\n",ok,"s"[ok==1],cp?"copied":"renamed");
      
  return 0;
}
