/* cc -O2 -o runint runint.c
 */
#include "../gen/include.h"

#define M 1024

double sum[M],subsum[M],x[M],last[M];
char line[M*32];
int n;

/* see Ralston, p.141, and http://mathworld.wolfram.com/Newton-CotesFormulas.html */
double coeff[8][9]={
  {0.5,0.5}, // trapezoid
  {1./3,4./3,1./3}, // Simpson
  {3./8,9./8,9./8,3./8}, // Simson 3/8
  {14./45,64./45,24./45,64./45,14./45}, // Boole
  {95./288,375./288,250./288,250./288,375./288,95./288},
  {41./140,216./140,27./140,272./140,27./140,216./140,41./140},
  {5257./17280,25039./17280,9261./17280,20923./17280,20923./17280,9261./17280,25039./17280,5257./17280},
  {3956./14175,23552./14175,-3712./14175,41984./14175,-18160./14175,41984./14175,-3712./14175,23552./14175,3956./14175} };

int nextline(void)
/* reads x[], returns number of columns (=n) on success */
{
  char *tok;
  static char lastn=0;
  int i;
  
  do {
    if (!gets(line)) return 0;
  } while (strchr("!#",line[0]));
  
  tok=strtok(line,"\t \r,");
     
  for (n=0;;n++) {
    if (n>=M) Error("runint: too many columns");
    if (!tok || sscanf(tok,"%lf",&x[n])!=1) break;
    tok=strtok(NULL,"\t \r,"); }
    
  if (n==0) return 0;
  
  if (lastn && n!=lastn) n=lastn;
  lastn=n;
    
  return n;
}

int main(int narg, char **arg)
{
  char *fmt="%.8g",*c;
  double h=0;
  int i,even=0,from=1;
  int method=1,phase;

  if (narg<2) {
    fprintf(stderr,"Newton-Cotes running integral from equidistant data. Call by:\n\
  runint OPTIONS < DATA\n\
DATA are normally lines of (cf. -h)\n\
  [x] y1 [y2 ..]\n\
OPTIONS:\n\
  -mMETHOD: number of subintervals, particularly:\n\
    -m1  2-point, 2nd order = trapezoid (default)\n\
    -m2  3-point, 4th order = Simpson\n\
    -m3  4-point, 4th order = Simpson 3/8, not recommended\n\
    -m4  5-point, 6th order = Boole\n\
    -m5  6-point, 6th order, not recommended\n\
    -m6  7-point, 8th order\n\
    -m7  8-point, 8th order, not recommended\n\
    -m8  9-point, 10th order, bad stability (negative coefficients)\n\
  -e the first interval is halved, METHOD must be even number\n\
         good for even functions (at x0, i.e., y(x-x0)=y(-x-x0)\n\
  -fFORMAT: C-style format (default = -f%s)\n\
  -hSTEP: there is no x column, x's are n*STEP (starts with n=0)\n\
#- and !-lines are omitted on input\n\
only data after the whole interval are printed\n\
Example:\n\
  runint -e -m2 -h.01 < spce256.Ptxz.cov | plot -\n\
See also:\n\
  runsum\n",fmt);
    exit(0); }
  
  loop (i,1,narg) switch (tolower(arg[i][1])) {
    case 'h': from=0; h=atof(arg[i]+2); break;
    case 'e': even++; break;
    case 'm': method=atoi(arg[i]+2); break;
    case 'f': fmt=arg[i]+2; 
      if (!strchr(fmt,'%')) fmt=string("%%%s",fmt);
      break; 
    default: Error("runint: bad option"); }

  if (method<1 || method>8) Error("runint: METHOD must be in interval [1,8]");

  if (even && method&1) Error("even function (-e) requires even method (-m)");

  if (even) even=method/2;

  if (!nextline()) Error("no data");
  
  for (;;) {
  
    loop (i,0,n) subsum[i]=0;
    if (from) h=x[0];

    loopto (phase,even,method) {
      if (phase!=even) if (!nextline()) return 0;
      //      printf("%g:%g\n",x[0],coeff[method-1][phase]);
      if (even && phase==even) 
        loop (i,from,n) subsum[i] += coeff[method-1][phase]*x[i]/2;
      else 
        loop (i,from,n) subsum[i] += coeff[method-1][phase]*x[i]; }
      
    if (from) {
      h=(x[0]-h)/method;
      if (even) h*=2;
      sum[0]=x[0]; }
    even=0;
    loop (i,from,n) sum[i] += h*subsum[i]; 
    loop (i,0,n) {
      printf(fmt,sum[i]);
      putchar(" \n"[i==n-1]); } }

}
