/* \make smoothg
*/
#include "ground.h"
#include "alloc.h"
#include "lineq.c"

typedef float vector[3];
#define NL 512
char line[NL];
int ncol;
double *data;
int K;
double FROMR,TOR,DR,d;
int DIM=3;

struct xy_s {
  struct xy_s *next;
  double x,y; } *xy0,*xy;

int nextline(void)
{
  char *c;
  int i;

  do
    if (fgets(line,NL,in)==NULL) return 0;
  while ( strchr("!#\n",line[0]) || !(c=strtok(line," \t\n")) );

  loop (i,0,ncol)
    if (c) {
      data[i]=atof(c);
      c=strtok(NULL," \t\n"); }
    else
      Error("not enough data in line");

  return 1;
}

double rectang=0,triang=0;
int colx=1,coly=2;

double window(double x)
{
  x=fabs(x);
  if (rectang<0) {
    x/=-rectang;
    if (x>9) return 0;
    else return exp(-x*x/2); }
		  
  else {
    if (x<rectang/2) return 1;
    if (x<rectang/2+triang) return (rectang/2+triang-x)/triang;
    return 0; }
}

double psi(int k,double r)
{
  return (powi(r+d,k+DIM)-powi(r-d,k+DIM))/(powi(r+d,DIM)-powi(r-d,DIM));
}

double regr(double x)
{
  double w;
  double **A,*B,*X,sum;
  void *mark;
  int j,k;

  rallocarrayzero(B,K);
  mark=B;
  rallocarrayzero(X,K);
  rallocarrayzero(A,K);
  loop (j,0,K) rallocarrayzero(A[j],K);

  for (xy=xy0; xy; xy=xy->next)
    if ( (w=window(xy->x-x)) ) {
      loop (k,0,K) {
	double psik=psi(k,xy->x);

	B[k]+=psik*w*xy->y;
	loop (j,0,K) {
	  double psij=psi(j,xy->x);
	  A[j][k]+=psik*w*psij; } } }

  GJlineq(0,K,A,B,X);

  sum=0;
  loop (j,0,K) sum+=X[j]*(j+DIM)/DIM*powi(x,j);

  release(mark);

  return sum;
}

int main(int narg, char **arg)
{
  char *fmt="%.12g %.12g";
  char *fn,*ch,*rg;
  int iline;

  initscroll(0);

  if (narg<6) {
    fprintf(stderr,"\
Smoothing of g(r) and rho(z) [obtained by histogram]. Call by:\n\
  smoothg WINDOW ORDER DR DATA[:COLR:COLG] FROMR[:TOR[:DR]] [DIM [FORMAT]]\n\
WINDOW    RECT[+TRIANG] rectangular width+triangular halfwidth\n\
          -WIDTH Gaussian halfwidth: exp[-(x/WIDTH)^2/2]\n\
ORDER     order of the polynomial (1=linear)\n\
DR        grid of R in DATAFILE (must match file!)\n\
DATA      file of r,g data\n\
COLR:COLG columns of r,g [default=1:2], COLR=0 means 0,1,..\n\
FROMR:TOR:DR output R\n\
DIM       dimension; 1=z-density profile, 2|3=g(r) in 2D|3D [default=3]\n\
FORMAT    output format for double x,y [default=\"%s\"]\n\
see also: smooth smooth3 smooth5 smoothlr\n",fmt);
   exit(0); }

  rectang=atof(arg[1]);
  ch=strchr(arg[1],'+');
  if (ch) triang=fabs(atof(ch+1));

  K=atoi(arg[2])+1;
  if (K<1 || K>10) Error("bad ORDER");

  d=atof(arg[3])/2;

  fn=strdup(arg[4]);
  ch=strtok(fn,":");
  if (ch) {
    ch=strtok(NULL,":");
    if (ch) colx=atoi(ch);
    else colx=1;
    ch=strtok(NULL,":");
    if (ch) coly=atoi(ch);
    else coly=2; }
  
  rg=strdup(arg[5]);
  FROMR=atof(rg);
  ch=strtok(rg,":");
  if (ch) {
    ch=strtok(NULL,":");
    if (ch) TOR=atof(ch);
    else TOR=10;
    ch=strtok(NULL,":");
    if (ch) DR=atof(ch);
    else DR=0.01; }

  if (narg>6) DIM=atoi(arg[6]);

  prt("# %s %s %s %s %s %d",arg[1],arg[2],arg[3],arg[4],arg[5],DIM);

  if (colx>=26 || colx<0) Error("arg 2 x column out of range");

  if (coly>=26 || coly<1) Error("arg 2 y column out of range");

  ncol=max(colx,coly);
  colx--,coly--;
  allocarray(data,ncol);
 
  if (strcmp("-",fn)) in=fopen(fn,"rt");

  if (!in) Error("arg 2 can't open file");

  iline=0;
  while (nextline()) {
    allocone(xy);
    if (colx<0) xy->x=iline++;
    else xy->x=data[colx];
    xy->y=data[coly];
    xy->next=xy0; xy0=xy; }
  fclose(in);

  if (!xy0->next) Error("not enough DATA");

  if (fabs((xy0->x-xy0->next->x)/d-2)>1e-6)
    fprintf(stderr,"WARNING: grid in file != DR, volume of slabs may be inaccurate");
 
  if (narg>7) fmt=arg[7];

  for ( ;FROMR<TOR+DR/2; FROMR+=DR) {
    printf(fmt,FROMR,regr(FROMR));
    printf("\n"); } 

  return 0;
}
