/* \make smoothlr
*/
#include "ground.h"
#include "alloc.h"
#include "linregr.h"

typedef float vector[3];
#define NL 512
char line[NL];
int ncol;
double *data;

struct xy_s {
  struct xy_s *next;
  double x,y; } *xy0,*xy;

int nextline(void)
{
  char *c;
  int i;

  do
    if (fgets(line,NL,in)==NULL) return 0;
    while ( strchr("!#\n",line[0]) || !(c=strtok(line," \t\n")) );

  loop (i,0,ncol)
    if (c) {
      data[i]=atof(c);
      c=strtok(NULL," \t\n"); }
    else Error("not enough data in line");

  return 1;
}

double rectang=0,triang=0;
int colx=1,coly=2;

double window(double x)
{
  x=fabs(x);
  if (x<rectang/2) return 1;
  if (x<rectang/2+triang) return (rectang/2+triang-x)/triang;
  
  return 0;
}

double regr(double x)
{
  double w;

  LRFree();
  for (xy=xy0; xy; xy=xy->next)
    if ( (w=window(xy->x-x)) )
      LRAdd("s",window(xy->x-x),xy->x-x,xy->y);

  if (!LRRes("s","?") || LRRes("s","N")<2) {
    fprintf(stderr,"too narrow window: 0 returned\n");
    return 0; }

  return LRRes("s","A");
}

double linint(double x)
{
  double x0,x1,y0,y1;
  double dx0=-3e33,dx1=3e33;

  for (xy=xy0; xy; xy=xy->next) {
    double dx=x-xy->x;
   
    if (dx>=0) {
      if (dx<dx1) dx1=dx,x1=xy->x,y1=xy->y; }
    else {
      if (dx>dx0) dx0=dx,x0=xy->x,y0=xy->y; } }

  if (dx0<-2.99e33) return y1;
  if (dx1>2.99e33) return y0;

  return (y0*(x1-x)+y1*(x-x0))/(x1-x0);
}

int main(int narg, char **arg)
{
  char *fmt="%.12g %.12g";
  char *fn,*ch;
  int iline;

  initscroll(0);

  if (narg<4) {
    fprintf(stderr,"\
Smoothing using rectangular+triangular window (sliding regression). Call by:\n\
  smoothlr RECTANG[+TRIANG] DATAFILE[:COLX:COLY] {INFILE|-}[:COLX] [FORMAT]\n\
RECTANG   rectangular window width, negative=linear interpolation only\n\
TRIANG    triangular window halfwidth (default=0)\n\
DATAFILE  file of xy data\n\
COLX:COLY columns of x,y (default=1,2), COLX=0 means 0,1,..\n\
INFILE    file of x data to calculate smoothed y; - denotes stdin\n\
FORMAT    output format for double x,y (default=\"%s\")\n\
see also: smoothg (with DIM=1)\n\
",fmt);
   exit(0); }

  rectang=atof(arg[1]) ;
  ch=strchr(arg[1],'+');
  if (ch) triang=fabs(atof(ch+1));

  fn=strdup(arg[2]);
  ch=strtok(fn,":");
  if (ch) {
    ch=strtok(NULL,":");
    if (ch) colx=atoi(ch);
    else colx=1;
    ch=strtok(NULL,":");
    if (ch) coly=atoi(ch);
    else coly=2; }

  if (colx>=26 || colx<0) Error("arg 2 x column out of range");
  if (coly>=26 || coly<1) Error("arg 2 y column out of range");

  ncol=max(colx,coly);
  colx--,coly--;
  allocarray(data,ncol);

  if (strcmp("-",fn)) in=fopen(fn,"rt");

  if (!in) Error("arg 2 can't open file");

  iline=0;
  while (nextline()) {
    allocone(xy);
    if (colx<0) xy->x=iline++;
    else xy->x=data[colx];
    xy->y=data[coly];
    xy->next=xy0; xy0=xy; }
  fclose(in);

  fn=strdup(arg[3]);
  ch=strtok(fn,":");
  if (ch) {
    ch=strtok(NULL,":");
    if (ch) colx=atoi(ch);
    else colx=1; }
  ncol=colx;
  free(data);
  allocarray(data,ncol);
  colx--;
 
  if (strcmp("-",fn)) in=fopen(fn,"rt");
  else in=stdin;
  if (!in) Error("arg 3 can't open file");
 
  if (narg>4) fmt=arg[4];
 
  iline=0;
  while (nextline()) {
    double x=iline++;
    if (colx>=0) x=data[colx];
    if (rectang<0) printf(fmt,x,linint(x));
    else printf(fmt,x,regr(x));
    printf("\n"); } 
  fclose(in);
 
  return 0;
}
