/*                                 cookdef.h
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
*/

#include "sitesite.h"

#define LINELEN 256 /* for line[LINELEN], not quite widely used... */

extern int nsites,nspec,nbfixes;
#ifdef WALL
extern struct wall_s {
  int 
    n, /* # of walls (1 or 2) */
    rep; /* 1: wall 1 repulsive, 2: wall 2 repulsive */
  double
    rho,  /* density of LJ atoms */
    minz,  /* where to shift the lowest atom (GOLD version) */
    g,   /* gravity */
    scalez; /* how to scale z while initializing */
  double 
    numden; /* number density */
} wall;
#endif /*# WALL */

/* site type group (support for distinguishing site types for rdf) */
typedef struct stitem_s {
  int sp; /* species */
  int i; /* site */
} stitem_t;

typedef struct stgroup_s {
  int n; /* # of atoms in the group */
  stitem_t stitem[1]; /* [n] */
} stgroup_t;

typedef struct {
  char name[8];    /* symbolic site name (CHARMM atom) */
  stgroup_t *g;    /* NULL if not a group */
  real M;          /* molar mass of the site in g/mol before equalization */
  siteparm_t LJ[2];/* LJ[0]: LJ parameters, LJ[1]: for 1-4 interactions */
#ifdef POLAR
  real alphapol;   /* polarisability in A^3 - to be used in self-field */
  real chargepol;  /* auxiliary charge in e - to be used in self-field */
#  if POLAR&1
  real kappa;      /* repulsive antipolarization (in 1/e) */
  int rep;         /* 2=2nd in the pair shell-core pair, 0=none */
#  endif /*# POLAR&1 */
#  if POLAR&2
  real Esat;       /* saturation energy for hyperpolarizability */
#  endif /*# POLAR&2 */
#endif /*# POLAR */
#ifndef FREEBC
  real sfweight; /* atom weight for structure factor, see also siteinfo_t */
#endif /*# FREEBC */
} sitedef_t;

extern nbfix_t *nbfix0;
extern sitedef_t *sitedef;

vector *initr(int sp);

#ifdef LINKCELL
extern real excrlimit; /* set by readblend */
#endif /*# LINKCELL */

int atomn(char *name,int sp,int i);
double readblend(void); /* returns minimum mass in g/mol */

void initss(double LJcutoff);

#ifdef WALL
void initwall(void);
#endif /*# WALL */
