/* this module depends on #defines WORM and CUT */
/* PARALLEL=3 support removed - see old+misc/simpot-par3.c, par3.c, parinter.c */

#include "ground.h"
#include "alloc.h"
#include "simglob.h"
#include "simdef.h"
#include "siminit.h"
#include "simpot.h"
#include "interpot.h"

#ifndef LINKCELL
#include "water.h"
#endif /*# LINKCELL */

#ifdef DUMP
/* WARNING: not correct for bonded */
#define DEBUGU(X) { \
  prt("%9s = %10.5f  total=%10.5f kcal/mol",\
  #X,(U+En.el-U0)*0.0019872,(U+En.el)*0.0019872); U0=U+En.el; }
double U0=0;
#else /*# 0 */
#define DEBUGU(X)
#endif /*# 0 */

#ifdef LINKCELL /* <<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<< LINKCELL */

double charmm1(vector f1[],molecule_t *m1,vector *rp) /************* charmm1 */
/* non-bonded forces: all pair forces solved by the link-cell method */
{
  int i,j,k,l;
  double U=spec[m1->sp]->zero_energy;
  vector *r1=rof(m1,rp);
  bond_t *b;
  angle_t *an;
  torsion_t *t;
  specinfo_t *sp;

#include "bonded.c"

  return U;
} /* charmm1 */

double charmm2(vector f1[],vector f2[],molecule_t *m1,molecule_t *m2,vector *rp)
/* LINKCELL: only intermolecular terms: used in configuration initializer */
{
  int i,j;
  double U=0,q,qq;
  vector *r1=rof(m1,rp), *r2=rof(m2,rp);
  real *r1i,*f1i;
  int ns1=m1->ns,ns2=m2->ns;
  siteinfo_t *si1=spec[m1->sp]->si;
  siteinfo_t *si2=spec[m2->sp]->si;
  sitesite_t *ss;

#define GLOB
#include "cookmeas.c"
#include "cookmm.c"
#include "cookundf.c"
#undef GLOB

  return U;
} /* charmm2 */

pot2_t *Potential(int i,int j) /********************************** Potential */
{
  i=spec[i]->pot; j=spec[j]->pot;
  if (i==5 && j==5) ERROR(("ST2 not supported for LINKCELL"))

  return charmm2;
}

#else /* ======================================================= no LINKCELL */

double charmm2(vector f1[],vector f2[],molecule_t *m1,molecule_t *m2,vector *rp)
/* not LINKCELL general function for inter and intra molecular forces */
{
  int i,j,k,l;

#ifndef POLAR
  double q,qq;
#endif /*# POLAR */

  vector *r1=rof(m1,rp);
  vector *r2=rof(m2,rp);

  real *r1i,*f1i;
  int ns1=m1->ns,ns2=m2->ns;
  bond_t *b;
  angle_t *an;
  torsion_t *t;
  specinfo_t *sp;
  sitesite_t *ss;
  siteinfo_t *si1=spec[m1->sp]->si;
  siteinfo_t *si2=spec[m2->sp]->si;
  double U=0;

  int j0,j1;
  exception_t *exc;

  if (m1!=m2) { /*** intermolecular interactions ***/

    if (measure) {

#define GLOB /* used to be active with PARALLEL=3 - kept for compatibility */

#include "cookmeas.c"
#include "cookmm.c"
#include "cookundf.c"

      En.LJmn=U; }
#undef GLOB
    else
#define GLOB
#include "cooknmea.c"
#include "cookmm.c"
#include "cookundf.c"
#undef GLOB
    }

  else { /*** intramolecular interactions ***/

    /* all site-site interactions */

#define GLOB
#define I0 0
#define I1 ns1

    if (measure) {
#include "cookmeas.c"
#include "cookm.c"
#include "cookundf.c"
      En.LJmn=U; }
    else
#include "cooknmea.c"
#include "cookm.c"
#include "cookundf.c"

    DEBUGU(qq+LJ)
    {
      /* bonded interactions: dirty trick!!! */
      double Ubonded=0;
#define U Ubonded
#include "bonded.c"
#undef U
      En.bonded+=Ubonded;
      if (No.first) if (m1-molec<No.first) En.bonded0+=Ubonded;
      U+=Ubonded; }

    U += spec[m1->sp]->zero_energy;

    } /* intramolecular */

#if 0
  prt("<0>: f1[0][0]=%g f1[%d][2]=%g U=%g En.el=%g En.vir=%g (simpot.c)",
      f1[0][0],ns1-1,f1[ns1-1][2],U,En.el,En.vir);
#endif /*# 0 */

  return U;
} /* charmm2 */

double atomatom2(vector f1[],vector f2[],molecule_t *m1,molecule_t *m2,vector *rp)
/* not LINKCELL optimized for monoatomic molecules */
{
#ifndef POLAR
  double q,qq;
#endif /*# POLAR */

  vector *r1=rof(m1,rp);
  vector *r2=rof(m2,rp);
  sitesite_t *ss;
  siteinfo_t *si1=spec[m1->sp]->si;
  siteinfo_t *si2=spec[m2->sp]->si;

  double U=0;

  if (m1!=m2) { /*** intermolecular interactions ***/

    if (measure) {
#define GLOB
#include "cookmeas.c"
#include "cookaa.c"
#include "cookundf.c"
      En.LJmn=U; }
#undef GLOB
    else {
#define GLOB
#include "cooknmea.c"
#include "cookaa.c"
#include "cookundf.c"
#undef GLOB
    } }

  else {
    En.LJmn=U;
    U += spec[m1->sp]->zero_energy; }

  return U;
} /* atomatom2 */


pot2_t *Potential(int i,int j) /********************************** Potential */
{
  if (spec[i]->ns==1 && spec[j]->ns==1) return atomatom2;

#ifndef POLAR
  i=spec[i]->pot; j=spec[j]->pot;
  if (i==j && option('x')<0) switch (i) {
    case 3: return TIP3P;
    case 4: return TIP4P;
    case 5: return ST2;
#ifdef WATERPLUS
#if WATERPLUS==2
    case 6: return CO2;
    case 7: return MeOH;
    case 8: return acetone;
    case 9: return dljd;
#endif
#if WATERPLUS==3
    case 10: return ionion; /* see waterion.c and watercut.c */
#endif
#endif /*# WATERPLUS */
    }

#if 0
  if (i==10 && j==3 || i==3 && j==10) return TIP3Pion; /* see waterion.c and watercut.c */
#endif

#endif /*! POLAR */

  return charmm2;
}

#endif /* ! LINKCELL */

#ifdef WALL
double mol_wall(vector f[],molecule_t *m,vector *rp) /************* mol_wall */
/*
  the molecule-wall potential
  should be optimized by making tables of LJA,LJAA,LJB,LJBB
  also gravity here
*/
{
  int ns=m->ns,i;
  vector *r=rof(m,rp);
  siteinfo_t *si=spec[m->sp]->si;
  double U=0;

  loop (i,0,ns) {
    double z=r[i][2],zz=z*z,z6=zz*zz*zz;
    static struct LJ_s {
      real A,AA,B,BB; } LJ1,LJ2; /* LJ2.A, LJ2.AA not used */
    sitesite_t *ss=sstab[nsites-1]+si[i].st;
    /* WARNING: LJ only */
    double sig6=Cub(ss->a.Sq);
    double eps4=wall.numden*ss->a.E4*sig6;
#ifdef GOLD
    double qq=Sqr(si[i].charge);

#ifdef FREEBC
    f[i][2] -= qq/(4*zz);
    if (measure) En.el -= qq/(4*z);
#else
#if COULOMB<0
#error COULOMB<0 not supported
COULOMB<0 not supported
#endif
    /* this is CUTELST version: cut-and-shift Coulomb needed */
    goldQQ(f[i],2*z,qq); /* updates En.el */
#endif /* FREEBC */
#endif /* GOLD */

    if (wall.g) {
      f[i][2]-=si[i].mass*wall.g;
      if (measure) U+=(si[i].mass*wall.g)*z; }

    if (ss->a.E4) {
      if (z<0) return 9e9; /* for the initializer only */
      LJ1.A=eps4*sig6*(PI/45);
      LJ1.AA=eps4*sig6*(PI/5);
      if (!(wall.rep & 1))
        LJ1.B=eps4*(PI/6),LJ1.BB=eps4*(PI/2);

      f[i][2]+=(LJ1.AA/z6-LJ1.BB)/(zz*zz);
      if (measure)
        U+=(LJ1.A/z6-LJ1.B)/(z*zz);

      if (wall.n==2) {
        if (!(wall.rep & 2))
          LJ2.B=eps4*(PI/6),LJ2.BB=eps4*(PI/2);
        z=box.L[2]-r[i][2],zz=z*z,z6=zz*zz*zz;
        if (z<0) return 9e9; /* for the initializer only */
        f[i][2]-=(LJ1.AA/z6-LJ2.BB)/(zz*zz);
        if (measure)
          U+=(LJ1.A/z6-LJ2.B)/(z*zz); } }
    }
  return U;
}
#endif
