/********** Calculator ********** (c) J.Kolafa 1991--NOW **********
NEW 2008: CALC=4 needed for some special functions (erfc, gamma, hyp.)
          new version CALC=4 is the same as old CALC=3
          (and CALC=4 becomes MACSIMUS default)
NEW 2007: _ allowed in identifiers
NEW 2006: binary operators ** (=^)  == >= <= <> added, floor added
NEW 2005: [day+]time format [DD::]hh:mm:ss (result in h)
          alternative modulo operator : changed to ` (% still OK)
          sign added
          null,nul added (null(0)=1,..; zero() interacts with tabproc x,y,z)
NEW 2004: operators < and > added, precedence between bitwise and +-
          (1<2 returns 1, 1>2 returns 0)
NEW 2002: hyperbolic functions added (sh=sinh, ash=asinh etc.)
          logarithms updated (ln,log,lb)
          bitwise functions added ( & | $=xor ), not(X), bit(X) (# of bits)
	  (all in CALC>=3)
NEW 2001: TABS allowed where ' '
NEW 1997: : is modulo (the same as %)
#define CINT : C-like integers (077 = 0o77 = octal, 0x2a = hex, 0b101 = bin)

Returns the value of expression made of numbers in usual format.
Binary operators: {+ - * / % ^} (% or ` is modulo, ^ is power)
Unary operators and functions: {+ - \ sin cos tan asin acos atan exp
  ln log lb sqrt abs int fac}
  ( \ is sqrt, fac or is factorial [fac 3=1*2*3, fac 2.5=0.5*1.5*2.5],
  ln=natural log, log=decadic log)
Parentheses: {( )}
Variables (#if CALC>=3): from a list (see _Id.head)
Last value (#if CALC>=2): # ( = _Id.nr )
PI or pi: 3.14159265358979323846
Time format: [DD::]hh[:mm[:ss]]
  (the result is on hours, e.g., 2::1=49, 11:45=11.75, 0:0:36=0.01)

End of expression is recognized by any character less than ' ' or a
character that would cause a syntax error (provided that the previous
part of expression is correct).  For CALC<3 the latter means any
character but binary operator at a position where this is expected.
E.g. given expression "(2+3)*1e-5A", parsing stops successfully at 'A'
while "2*3*A" is error because "2*3*" is error (does not apply for
CALC=3 version with variables).

Priorities:
  numbers and characters (incl. : or :: in time format)
  ()
  functions and unary \
  ^ ** (=power; associative from left as * or /)
  + - (unary)
  * / % `  (% and ` is modulo)
  + - (binary)
  < <= > >= <> == (binary relation operators)
  & $ (=xor)
  | (=or)
Examples:
  2e3       = 2000
  'a'       = 97 (ascii value)
  2^-1/2    = (2^(-1))/2
  -2^2      = -(2^2)
  3^3^3     = (3^3)^3
  cos pi^2  = (cos (pi))^2
  cos -pi^2 = cos (-(pi^2))
Function header is the same as for strtod:
  double Calc(char *buf,char **endscan)
Bugs:
  3^3^3  is not  3^(3^3)  but  (3^3)^3 (matter of taste, I guess)
  cos pi^2  is something else than  cos +pi^2
    I do not see any consistent way how to interpret such cases: use
    parentheses if you are in doubt
  In CALC<3 versions, only one or two letters are used to recognize function
    names so that e.g.:  sss=si=sin, sq=\=sqrt
**********************************************************************/

#ifdef DOS
#define opstacklen 48
#define nrstacklen 16
#else
#define opstacklen 96
#define nrstacklen 32
#endif

#include <time.h> /* for rnd seed */

static int precedence(char op)
{
 switch (op&127) {
   case 0: return 0; /* evaluation operator */
   case '=': return 1;
   case ')': return 2;
   case '(': return 3;
#if CALC>=3
   case '|': return 5;
   case '&': case '$': return 6;
   case '<': case '>': case 037: case 036: case 035: case 034: return 7;
#endif
   case '+': case '-': return op&128 ? 12 : 8; /* distinguish binary/unary */
   case '*': case '/': case '%': case '`': return 10;
   case '^': return 14;
   default: return 127; /* unary operator */ }
} /* precedence */


double Calc(char *buf,char **endscan)
{
  unsigned char *b,*end;
  unsigned char *op,Op[opstacklen]; /* operation stack, bit 7 set means unary operator */
  double *nr,Nr[nrstacklen]; /* number stack */

  b=(unsigned char*)buf;

  op=Op; *op='=';
  nr=Nr-1;

NRSCAN: /*** number, '(' or unary operator expected ***/

  if (op>Op+opstacklen-2) goto reterror;

  while (*b==' ' || *b=='\t') b++;

  if (*b==';' || *b<' ') goto reterror;

#if CALC>=2
  /* # = register (old value) */
  if (*b=='#') {
    nr++; if (nr>=Nr+nrstacklen) goto reterror;
    *nr=_Id.nr; b++;
    goto OPSCAN; }
#endif

#if CALC>2
  /* identifier */
  if (isalpha(*b) || *b=='_') {
    unsigned char *x=b;
    char id[64],*idp=id;
    _Idlist *l;
    char *funclist[]={"sin","cos","tan",
		      "asin","acos","atan",
#if CALC>3
		      "sinh","cosh","tanh","sh","ch","th",
		      "asinh","acosh","atanh","ash","ach","ath",
		      "erf","erfc","gamma","lgamma",
#endif
		      "not","bit",
		      "ln","log","lb",
		      "exp", "sqrt",
                      "fac","fact","rnd",
		      "pi","PI",
		      "abs","sign","int","float","floor","frac","null","nul",NULL};
    char **fptr;

    /* abc[12] allowed as identifier, abc[+2] or abc[-1] not! */
    while (isalnum(*x) || (*x && strchr("_.[]",*x))) *idp++=*x++;
    *idp=0;

    /* try to find the identifier in the list */
    for (l=_Id.head; l; l=l->next)
      if (!strcmp(l->id,id)) {
	nr++; if (nr>=Nr+nrstacklen) goto reterror;
        *nr=l->val; b=x;
        goto OPSCAN; }

    /* identifier not found ==> will be a function, CHECK now better than for
       CALC<3 (real scan will be done later) */
    for (fptr=funclist; *fptr; fptr++) if (!strcmp(*fptr,id)) goto fnfound;

    if (_Id.ignoreid) {
      /* identifier not found (not a function), :=0 */
      nr++; if (nr>=Nr+nrstacklen) goto reterror;
      *nr=0; b=x;
      goto OPSCAN; }

    goto reterror;

    fnfound:; }
#endif

  /* PI (in the list of valid keywords only pi and PI are allowed) */
  if (toupper(*b)=='P' && toupper(b[1])=='I'
   && !isdigit(b[2]) && !isalpha(b[2]) && b[2]!='_') {
    nr++; if (nr>=Nr+nrstacklen) goto reterror;
    *nr=PI; b+=2;
    goto OPSCAN; }

  /* number */
  if (isdigit(*b) || *b=='.') {
    nr++; if (nr>=Nr+nrstacklen) goto reterror;
#ifdef CINT
    /* numbers in the form 077=0o77 0xff 0b11 */
    if (*b=='0' && strchr("0123456789xbo",b[1])) {
      int base=8;

      end=b+1;
      *nr=0;
      switch (*end) {
        case 'x': base=16; goto plus;
        case 'b': base=2;
        case 'o':
plus:             end++; }

      for (;;end++) {
        int i=*end-'0';
        if (base==16)
          if (i>='A'-'0' && i<='F'-'0') i-='A'-'0'-10;
          else if (i>='a'-'0' && i<='f'-'0') i-='a'-'0'-10;
        if (i<0 || i>=base) break;
        *nr=*nr*base+i; }
      }
    else
#endif
      {
//	double x=strtod(b,&end);
//	double x=strtod((char*)b,(char**)&end);
	double x=fstrtod((char*)b,(char**)&end);
	double xx=0;
	double q=1;

	/* time format hh:mm:ss and day+time format DD::hh:mm:ss
	   the result is always in hours */
	if (*end==':' && b!=end && end[1]==':') {
	  b=end+2;
	  x=x*24+strtod((char*)b,(char**)&end); }

	while (*end==':' && b!=end) {
	  b=end+1;
	  q*=60;
	  xx+=strtod((char*)b,(char**)&end)/q; }
	*nr=x+xx;
      }
      //      *nr=strtod((char*)b,(char**)&end);


    if (b==end) goto reterror;
    b=end;
    goto OPSCAN; }

  /* character */
  if (*b=='\'') {
    nr++; if (nr>=Nr+nrstacklen) goto reterror;
    b++;
    if (*b=='\'') *nr=0;
    else *nr=*b++;
    if (*b++!='\'') goto reterror;
    goto OPSCAN; }

  /* left parenthesis */
  if (*b == '(') {
    *(++op) = *b++; goto NRSCAN; }

  /* function : op-code is the 1st letter or upcase 2nd letter for inv.func. */
  /* extended to hyp.func. */
  /* (this is cumbersome, but I'm too lazy to rewrite it...) */
  if (isalpha(*b)) {
    ++op;
    *op = *b=='a' ? toupper(b[1]) : *b ; /* sin,cos,tan + inv.; also abs */
    if (*b=='s' && b[1]=='q') *op='\\'; /* sqrt */
    if (*b=='s' && b[1]=='i' && b[2]=='g') *op='I'; /* sign */
    if (*b=='n' && b[1]=='u') *op='u'; /* nul,null */
    /* logarithms: */
    if (*b=='l') switch (b[1]) {
#if CALC>=3
#ifndef DOS
      case 'g': *op='G'; break; /* lgamma */
#endif
#endif
      case 'o': *op='d'; /* log=log_10, formerly dlog */
      case 'n': break;   /* ln */
      case 'b': *op='w'; /* lb=log_2 */ }
#if CALC>=3
    if (!memcmp(b,"float",5)) *op='A';
    if (!memcmp(b,"floor",5)) *op='F';
    if (!memcmp(b,"frac",4)) *op='Q';
#if CALC>3
    if (!memcmp(b,"erf",3)) *op=b[3]=='c'?'H':'h'; /* erf[c] */
    /* hyperbolic functions: (ANY function name with `h' in) */
    if (!memcmp(b,"sh",2) || !memcmp(b,"sinh",4)) *op='x';
    if (!memcmp(b,"ch",2) || !memcmp(b,"cosh",4)) *op='y';
    if (!memcmp(b,"th",2) || !memcmp(b,"tanh",4)) *op='z';
    if (!memcmp(b,"ash",3) || !memcmp(b,"asinh",5)) *op='X';
    if (!memcmp(b,"ach",3) || !memcmp(b,"acosh",5)) *op='Y';
    if (!memcmp(b,"ath",3) || !memcmp(b,"atanh",5)) *op='Z';
#endif
#endif
/*.....    fprintf(stderr,"<%s>%c\n",b,*op);*/
    *op |= 128;
    while (isalpha(*b)||*b=='_') b++; goto NRSCAN; }

  /* unary one-char operators like + - \ (\=sqrt) */
  *(++op) = 128 | *b++;

  goto NRSCAN;

OPSCAN: /*** binary operator or ')' or end expected ***/

  while (*b==' ' || *b=='\t') b++;
  *(++op) = *b++;

#if CALC>=3
  /* multicharacter operators: ** ==  >=  <=  <>
                               ^  31  30  29  28  (dec)
                                  037 036 035 034 (oct) */
  if (*op=='*' && *b=='*') *op='^',b++;
  if (*op=='=' && *b=='=') *op=037,b++;
  if (*op=='>' && *b=='=') *op=036,b++;
  if (*op=='<' && *b=='=') *op=035,b++;
  if (*op=='<' && *b=='>') *op=034,b++;

  if (!strchr(")+-*/%`^|$&<>\037\036\035\034",*op)) *op=0;
#else
  if (!strchr(")+-*/%`^",*op)) *op=0;
#endif

EVALSTACK: /*** stacked operations are tried to evaluate ***/

  if (op[-1]=='=' && *op==0) {
    *endscan=(char*)b-1;
    if (nr==Nr) return *nr;
    goto reterror; }

  if (*op==')' && op[-1]=='(') { op-=2; goto OPSCAN; }

  if (precedence(*op) > precedence(op[-1])) goto NRSCAN;

  op--;

  if (*op & 128) switch (*op & 127) {
    /* unary operators (functions) */
    case '+': break;
    case '-': *nr= -*nr; break;
    case '\\': *nr=sqrt(*nr); break;
#if CALC>=3
    case 's': *nr=sin(*nr*_Id.degrad); break;
    case 'c': *nr=cos(*nr*_Id.degrad); break;
    case 't': *nr=tan(*nr*_Id.degrad); break;
    case 'S': *nr=asin(*nr)/_Id.degrad; break;
    case 'C': *nr=acos(*nr)/_Id.degrad; break;
    case 'T': *nr=atan(*nr)/_Id.degrad; break;
    case 'n': *nr= ~((unsigned)*nr); break; /* logical not */
    case 'b': {
      /* bit count */
      unsigned i=0,u=(unsigned)*nr;
      while (u) { i+=u&1; u>>=1; } *nr=i; }
      break;
    case 'Q': *nr=*nr-(int)*nr; break;
//    case 'Q': *nr=frac(*nr); break;
#if 0 /* sh,th etc. : BAD QUALITY near zeros! */
    case 'x': *nr=(exp(*nr)-exp(-*nr))/2; break;
    case 'y': *nr=(exp(*nr)+exp(-*nr))/2; break;
    case 'z': *nr=(exp(*nr)-exp(-*nr))/(exp(*nr)+exp(-*nr)); break;
    case 'X': *nr=log(sqrt(*nr* *nr+1)+*nr); break;
    case 'Y': *nr=log(1/(*nr-sqrt(*nr* *nr-1))); break;
    case 'Z': *nr=log(sqrt((1+*nr)/(1-*nr))); break;
#else
#ifndef DOS
#if CALC>3
    case 'x': *nr=sinh(*nr); break;
    case 'y': *nr=cosh(*nr); break;
    case 'z': *nr=tanh(*nr); break;
    case 'X': *nr=asinh(*nr); break;
    case 'Y': *nr=acosh(*nr); break;
    case 'Z': *nr=atanh(*nr); break;
    case 'h': *nr=erf(*nr); break;
    case 'H': *nr=erfc(*nr); break;
#if 0
    /* WARNING: do not use gamma(*nr), it is sometimes gamma,
       sometimes ln(|gamma|) ... */
    /* signgam not supported on some systems */
    case 'g': { double x=lgamma(*nr); *nr=exp(x)*signgam; break; }

    /* this should work, but declaration is missing */
    extern double tgamma(double);
    case 'g': *nr=tgamma(*nr); break;

    /* more portable? */
    case 'g': { int sg; double x=lgamma_r(*nr,&sg); *nr=exp(x)*sg; break; }

    /* probably most fool-proof: */
#endif
    case 'g': {
      int sg=1;
      double x=lgamma(*nr);
      if (*nr<0) sg=(int)(*nr)&1?1:-1;
      *nr=exp(x)*sg;
      break; }
    case 'G': *nr=lgamma(*nr); break;
#endif
#endif
#endif
#else
    case 's': *nr=sin(*nr); break;
    case 'c': *nr=cos(*nr); break;
    case 't': *nr=tan(*nr); break;
    case 'S': *nr=asin(*nr); break;
    case 'C': *nr=acos(*nr); break;
    case 'T': *nr=atan(*nr); break;
#endif
    case 'l': *nr=log(*nr); break;
    case 'd': *nr=log10(*nr); break;
    case 'w': *nr=log(*nr)/log(2.); break; /* lb()=log_2 */
    case 'e': *nr=exp(*nr); break;
    case 'B': *nr=fabs(*nr); break; /* abs */
    case 'u': *nr=*nr==0; break; /* null, nul */
    case 'I': if (*nr>0) *nr=1; else if (*nr<0) *nr=-1; break; /* sign */
    case 'i': *nr=(int)(*nr); break;
    case 'F': *nr=floor(*nr); break;
    case 'A': /* float conversion: must cheat the optimization */
              { float F[2]; int i=1;
                F[1]=*nr;
                if (cos(2.3)>1) i=0;
                *nr=F[i]; }
               break;
    case 'r': {
      static int initialized=0;
      if (*nr<-1 || !initialized) {
        if (*nr<-1) {
          *nr=fabs(*nr);
          srand(-(int)*nr); }
        else {
          time_t init;
          time(&init);
          srand(init); }
        initialized=1; }

      if (*nr==-1)
        /* [-1,1) */
        *nr=rand()/(double)(RAND_MAX)*2.-1;
      else if (*nr==0)
        /* [0,1) */                                                            
        *nr=rand()/(double)(RAND_MAX);
      else if (*nr==1)
        /* normal Gaussian distribution */
	*nr=cos(rand()*(PI/RAND_MAX))*sqrt(-2*log(rand()/(double)RAND_MAX));
      else
        /* integer in [0,*nr-1] */
	*nr=(int)((double)rand()*(*nr/(double)RAND_MAX));
      } break;
    case 'f': if (*nr<0) goto reterror;
              if (*nr>1000) *nr=1000;
                /* *nr=1.0/0.0; else (not accepted by Microsoft C) */
              { double f=1; while (*nr>0) f*=*nr,*nr-=1e0; *nr=f; }
              break;
    default: goto reterror; }

  else { nr--; switch (*op) {
    /* binary operators */
#if CALC>=3
    case '&': *nr= (unsigned)*nr & (unsigned)nr[1]; break;
    case '|': *nr= (unsigned)*nr | (unsigned)nr[1]; break;
    case '$': *nr= (unsigned)*nr ^ (unsigned)nr[1]; break;
    case '<': *nr= *nr < nr[1]; break;
    case '>': *nr= *nr > nr[1]; break;
    case 037: *nr= *nr == nr[1]; break;
    case 036: *nr= *nr >= nr[1]; break;
    case 035: *nr= *nr <= nr[1]; break;
    case 034: *nr= *nr != nr[1]; break;
#endif
    case '+': *nr+=nr[1]; break;
    case '-': *nr-=nr[1]; break;
    case '*': if (*nr==0 || nr[1]==0) *nr=0; /* so that NaN*0=0 (2010-02-10) */
              else *nr*=nr[1]; break;
    case '/': *nr/=nr[1]; break;
    case '`': /* because of ev.c - % is in format */
    case '%': *nr= *nr-((int)(*nr/nr[1]))*nr[1]; break;
    case '^': *nr=pow(*nr,nr[1]); break;
    default: goto reterror; } }

  *op=op[1];

  goto EVALSTACK;

reterror: *endscan=buf; return 0.0;

}
