/*
  compatibility module from different graphics platforms
  and graphics primitives in x,y Cartesian coordinates

  #ifdef DOS : Turbo graphics + hotkeys, mouse support is in mouselib
  #ifdef EMX : some Turbo graphics + hotkeys functions emulated by EMX graphics,
               no mouse (yet)
  #ifdef X11 : some Turbo graphics functions + hotkeys emulated by XWindows
  none of these : plain version, all graphic + hotkey functions are void
*/

#include "prec.h"

#include "loop.h"

#if defined(DOS)

#include <graphics.h>
extern char *pathtodriver;
#define PALLEV 64 /* color levels in palette - VGA */

#elif defined(EMX)

#include <graph.h>
#define putpixel g_set
#define getpixel g_get
#define PALLEV 64 /* color levels in palette - VGA */

#elif defined(X11)

#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/Xos.h>
#include <X11/Xatom.h>
#include <X11/keysym.h>

#define PALLEV 256 /* color levels in palette - true colors */

#define X_REDRAW_ME   28
#define X_UNKNOWN_KEY 29

#include <time.h>

#ifdef GUI
extern struct gui_s {
  char *exe;
  char *name; 
} gui;
#endif

extern int extrabufptr;
void Imworking(int worktime);
/*
   worktime<0: resets timer, no change of parms
   worktime>0: sets timer to worktime seconds
               after this time, kbhit() stops and waits for an event
   worktime=0: kbhit() always stops and waits for an event
   NOTE: Turbo-like behavior of kbhit() is obtained by Imworking(large_number)
*/

extern struct xwindowhints_s {
  int icowidth,icoheight; /* X icon size */
/* unsigned ? */
  char *icobits; /* icon data */
  char *geometry;   /* standard x-window geometry (passed from -g option) */
  char *background; /* background color (passed from -bg option) */
  char *winname;    /* X window name */
  char *iconame;    /* X icon name */
  } xwindowhints;

extern int keybufptr; /* probably not needed extern */
extern char keybuf[128];
void putpixel(int x,int y,int c);
int hit(int wait);
void addkbdstring(char *s);

extern int x_xsize,x_ysize,x_color,x_fillcolor;
#define getmaxx() (x_xsize-1) 
#define getmaxy() (x_ysize-1)
#define getcolor() x_color
void line(int x0,int y0,int x1,int y1);
void bar(int x0,int y0,int x1,int y1);
void rectangle(int x0,int y0,int x1,int y1);

extern struct dosmouse_s {
  int x,y;
  int left,right,middle;
  int wheel; /* X11 only */
  int ex; /* patch to detect released button */
  } dosmouse;
extern Display *display;
extern Window win;

#define getpixel(X,Y) 0 /* NOT SUPPORTED !!! */

#else /* plain version */

#define getpixel(X,Y) 0 
#define getmaxx() 80
#define getmaxy() 24
#define getcolor() 0
void line(int x0,int y0,int x1,int y1);
void bar(int x0,int y0,int x1,int y1);
void rectangle(int x0,int y0,int x1,int y1);
void putpixel(int x,int y,int c);

#endif
#ifndef DOS

/* Turbo colors */
enum COLORS {
    BLACK,		    /* dark colors */
    BLUE,
    GREEN,
    CYAN,
    RED,
    MAGENTA,
    BROWN,
    LIGHTGRAY,
    DARKGRAY,		    /* light colors */
    LIGHTBLUE,
    LIGHTGREEN,
    LIGHTCYAN,
    LIGHTRED,
    LIGHTMAGENTA,
    YELLOW,
    WHITE
};

void closegraph(void);
void setcolor(int c);
void clearviewport(void);
int kbhit(void);
#undef getch
int getch(void);
void delay(unsigned msec);
void getimage(int x0,int y0,int x1,int y1,void *buf);
void putimage(int x,int y,void *buf,int mode);
void settextstyle(int a,int b,int size);
void settextjustify(int x,int y);
int outtextxy(int x,int y,char *s);
int outtext(char *s);
extern int font_width,font_height,font_top,font_base; /* DOS? */
void selectfont(int font);
void setlinestyle(int mode,int pattern,int thick);
void setfillstyle(int mode,int color);
void setwritemode(int mode);
void fillellipse(int x,int y,int rx,int ry);
void circle(int x,int y,int r);
void setpal(int col,int r,int g,int b,int maxbright); /* extend to DOS !*/

/* currentcolor: set by setcolor both in X11 and EMX */
#endif

#ifdef EMX
extern int gg_col,gg_fillcol;
#define getmaxx() (g_xsize-1) 
#define getmaxy() (g_ysize-1)
#define getcolor() gg_col
#define bar(X0,Y0,X1,Y1) g_box(X0,Y0,X1,Y1,gg_fillcol,G_FILL)
#define rectangle(X0,Y0,X1,Y1) g_box(X0,Y0,X1,Y1,gg_fillcol,G_OUTLINE)
#define line(X0,Y0,X1,Y1) g_line(X0,Y0,X1,Y1,gg_col)
#endif

/* common functions */

extern struct scaling_s { REAL fromx,x,toy,y; } scaling;
extern int pen;

int startgraph(int mode);
void up(void);
void lwindow(int x0,int x1,int y0,int y1);
void scale(REAL x0,REAL x1,REAL y0,REAL y1);
int SX(REAL x);
int SY(REAL y);
REAL iSX(int ix);
REAL iSY(int iy);
void draw(REAL x1,REAL y1);
void lline(REAL x1,REAL y1,REAL x2,REAL y2);
void dot(REAL x1,REAL y1);
void bigdot(REAL x1,REAL y1);
extern REAL pointsize;
extern int insidecolor;
void point(REAL x1,REAL y1,char type);
#define smartdot(X,Y) point(X,Y,'o'|128)
