#include "ground.h"
#include "alloc.h"
#include "mouselib.h"
#include "draw.h"
#include "drawplus.h"

#ifdef DOS
#include <conio.h> /* kbhit(), getch() */
#include <dos.h>   /* delay() */
#endif

struct mymouse_s New,Last;

char *mykeybuf;
unsigned mybufpos;

struct butt_s butt;

static int maxmybuflen,mouse;

void showMouse(void) { if (mouse) showMouseCursor(); }
void hideMouse(void) { if (mouse) hideMouseCursor(); }

struct buttons_s *buttons, *lastbutton;

void drawplus(int nbutt,int buflen,int mous)
{
butt.right=ESC;

if ( (mouse=mous) ) doMouse();
if (!buttons) {
  alloc(buttons,(butt.maxn=nbutt)*sizeof(buttons[0]));
  butt.n=0; }
if (!mykeybuf) {
  alloczero(mykeybuf,(maxmybuflen=buflen)*sizeof(mykeybuf[0]));
  mybufpos=0; }
}

void buttnl(void)
{
butt.aty+=butt.dy;
}

static void buttonborder(struct buttons_s *b,int col)
{
setcolor(col);
setwritemode(0);
rectangle(b->x0,b->y0,b->x1,b->y1);
rectangle(b->x0-1,b->y0-1,b->x1+1,b->y1+1);
}

void registerclickarea(int x0,int y0,          /********* registerclickarea */
                       int x1,int y1,
                       char ch)
{
if (butt.n<butt.maxn) {
  lastbutton=&buttons[butt.n];
  lastbutton->x0=x0;
  lastbutton->y0=y0;
  lastbutton->x1=x1;
  lastbutton->y1=y1;
  lastbutton->type=butt.type;
  lastbutton->ch=ch;
  butt.n++; }
else {
  ERROR(("button %c[%d] ofl",ch,ch))
  lastbutton=NULL; }
}


int extkey(void) /*************************************************** extkey */
/* getch with translated extended codes (as bit8) */
{
int ch=getch();
if (ch==ESC) mybufpos=0; /* ESC */
if (ch) return ch;
return getch()|128;
}

void addmykey(char ch)
{
if (mybufpos<maxmybuflen) mykeybuf[mybufpos++]=ch;
#ifdef X11
extrabufptr=mybufpos;
#endif
}

void nextevent(void) /******************************************** nextevent */
{
nextmouseevent();
while (kbhit() && mybufpos<maxmybuflen) {
  int e=extkey(); /* warning: extkey may change mybufpos */
  mykeybuf[mybufpos]=e;
  mybufpos++; 
#ifdef X11
  extrabufptr=mybufpos; 
#endif 
  }
}

int mykbhit(void) /************************************************ mykbhit */
{
nextevent();
#ifdef X11
extrabufptr=mybufpos;
#endif
return mybufpos ? mykeybuf[0] : 0;
}

int mykey(void) /**************************************************** mykey */
/* buffered getch */
{
again:
nextevent();
if (mybufpos) {
  /* get char from the buffer */
  unsigned char ch=mykeybuf[0];
  mybufpos--;
  memmove(mykeybuf,mykeybuf+1,mybufpos);
#ifdef X11
  extrabufptr=mybufpos;
#endif
  return ch; }
else
  /* buffer empty */
  if (mouse)
    /* waiting in a cycle for the next event */
    goto again;
  else
    /* waiting for char in standard DOS routine */
    return extkey();
}

void registerbutton(int fromx,int tox,char ch) /************* registerbutton */
/*
  registers the ch-button in area
    by 1 pixel wider than <fromx,tox> x <butt.aty,butt.aty+11>
*/
{
registerclickarea(fromx-1,butt.aty-1,tox+1,butt.aty+10,ch);
}

void createcolbutton(char *s,int dx,int col) /************** createcolbutton */
/*
  create button at (butt.atx+dx,butt.aty) incl. drawing s+1 
  and registering s[0]
  if dx<0 then dx=0 applies and butt.atx is advanced in x-direction by butt.dx
*/
{
int fromx;
char *ch;
int sg=dx<0;

if (sg) dx=0;
fromx=butt.atx+dx;

printinfo(fromx,butt.aty,s+1,col);
registerbutton(fromx,butt.lastx,s[0]);
buttonborder(lastbutton,butt.type==Wheel?LIGHTRED:LIGHTGRAY);

if (sg) butt.atx=butt.lastx+butt.dx;

/* highlighted hotkey, if present */
if ( (ch=strchr(s+1,s[0])) ) {
  static char ss[2];
  ss[0]=*ch;
  printinfo(-(fromx+8*(ch-(s+1))),butt.aty,ss,col==YELLOW?WHITE:YELLOW); }
}

void createbutton(char *s,int dx) /*************************** createbutton */
/*
  create button at (butt.atx+dx,butt.aty) incl. drawing s+1 and registering s[0]
*/
{
createcolbutton(s,dx,WHITE);
}

void createpmbutton(char *msg,char *smd,int change) /******* createpmbutton */
/*
  buttons with: smd[0]=info key, smd[1]=[*] key, smd[2]=[/] key
  if not change then [*] and [/] keys are not created (change not allowed)
  button padded by spaces to 4 chars 
*/
{
static char s[4];
static char ss[8];
int l=strlen(msg);

ss[0]=smd[0];
ss[1]=' ';

memcpy(ss+1+(l<3),msg,l<6 ? l+1 : 6);
if (strlen(ss)<5) strcat(ss," ");

createbutton(ss,0);

if (change) {
  s[0]=smd[1]; s[1]='*';
  createcolbutton(s,55,LIGHTCYAN);

  s[0]=smd[2]; s[1]='/';
  createcolbutton(s,74,LIGHTCYAN); }

buttnl();
}

#if 0 /* old version: scroll wheel under text bar */
void createwheel(char *msg,char ch) /************************* createwheel */
/*
  create dragging wheel (wheel-like scrollbar)
*/
{
static char wh[]="?=========";
butt.type=Wheel;
printinfo(butt.atx+2,butt.aty,msg,LIGHTGRAY);
butt.aty+=11;

wh[0]=ch;
createbutton(wh,-2);

buttnl();
}
#endif


void nextmouseevent(void) /********************************** nextmouseevent */
/*
  get next mouse event, put equivalent char to mykeybuf
  buttons[i] scanned for  butt.from<=i<butt.n
  butt.right is the code returned when the right mouse button is pressed
*/
{
static int iswheel;
static struct buttons_s *wheelp; /* dragging wheel pointer */
int i;

#ifdef X11
kbhit();
#endif

if (!mouse) return;

New.x=getMouseX();
New.y=getMouseY();
New.left=getButton(LEFTBUTTON);
New.middle=getButton(MIDDLEBUTTON);
New.right=getButton(RIGHTBUTTON);

if (memcmp(&Last,&New,sizeof(New))) {

  if (New.left==buttonDown) {
    if (iswheel) goto wheeldragged;

    loop (i,butt.from,butt.n) {
      struct buttons_s *b=&buttons[i];

      if (New.x>=b->x0 && New.x<=b->x1
       && New.y>=b->y0 && New.y<=b->y1) {

        if (b->type==Wheel) {
          /* no key returned! - first action when mouse moved */
          wheelp=b;
          iswheel=New.x;
          HIDEMOUSE
          buttonborder(b,RED);
          SHOWMOUSE
          delay(250); /*?*/

          break; }

        addmykey(b->ch);
        if (b->type==Drag) break;

        HIDEMOUSE
        buttonborder(b,BLACK);
        while (getButton(LEFTBUTTON)==buttonDown);
        buttonborder(b,LIGHTGRAY);
        New.left=buttonUp;
        SHOWMOUSE
        break; } } }
  /* left button up */
  else if (iswheel) {
    iswheel=0;
    goto wheeldragged; }

  if (New.middle==buttonDown) {
    while (getButton(MIDDLEBUTTON)==buttonDown);
    addmykey(ESC); }

  if (New.right==buttonDown) {
    while (getButton(RIGHTBUTTON)==buttonDown);
    addmykey(butt.right); }
  }
goto cont;

wheeldragged:

  HIDEMOUSE
  buttonborder(wheelp,iswheel?RED:LIGHTRED);
  SHOWMOUSE

  if (New.x!=Last.x) {

    if (!mybufpos || mykeybuf[mybufpos-1]!=wheelp->ch) {
      addmykey(wheelp->ch);
      butt.wheel=0; }
      /* else last event was Wheel and we must add it to the new one */

    butt.wheel+=(New.x-Last.x)/(REAL)(wheelp->x1-wheelp->x0); }

cont:
Last=New;
}

void panel(int xsize,int ysize,int col) /**************************** panel */
/*
  xsize is in characters with margins
  ysize is height in butt.dy, with black margin below
*/
{
int
  xx=butt.atx+xsize*8+2,
  yy=butt.aty+ysize*butt.dy;
setfillstyle(1,col);
bar(butt.atx-2,butt.aty,xx,yy-1);
setfillstyle(1,BLACK);
bar(butt.atx-2,yy,xx,yy+3);
butt.aty+=4;
}

void printinfo(int x,int y,char *msg,int bg) /******************* printinfo */
/*
  on background bg and with left margin 3 pixels, right 2 (=3 from char), 
    up and down 1 pixel
  thus, text starts at (x+3,y+1)
  butt.lastx is set to max x of the background rectangle
  PATCH: if x<0 then 1 pixel left and right margins are nor overwritten by
         background color and(x) is the x-coord
*/
{
int mar=2*(x<0);

x=abs(x);
butt.lastx=x+strlen(msg)*8+4;

HIDEMOUSE
setfillstyle(1,bg);
setwritemode(0);
setcolor(bg<7 || bg==LIGHTRED || bg==LIGHTGRAY || bg==LIGHTBLUE ? WHITE : BLACK);
bar(x+mar,y,butt.lastx-mar,y+9);
outtextxy(x+3,y+1,msg);
setcolor(BLACK);
SHOWMOUSE
}

REAL fint(REAL x) /**************************************************** fint */
/* conversion to integer coded as REAL: inefficient, but REAL-independent */
{
if (x<0)
  return -fint(-x);
else {
  /* m must be a power of 8 and m>=max integer coded in REAL possible */
#ifdef HIGHPREC
  REAL m=powi((REAL)2,32*PREC);
#else
  REAL m=Sqr(268435456e0);
#endif

  REAL ix=0,im;
  int i;

  if (x>m) WARNING(("fint(%g): fix ofl",Val(x)))

  while (m>=256) {
    m/=256;
    i=Int(x/m);
    im=i*m;
    x-=im; ix+=im; }
  return ix; }
}

#include "makegrid.c"
