/*
  #define DEBUG: detailed
  #define GUI : Gtk-based support for menu (in a separate window)

  environment variable LAZYX11
  positive value: waits given # of seconds (max 5) before redraw
                  to be used for lazy networks
  negative value: suppresses geometry message to stderr
  -0            : suppresses geometry message to stderr, not lazy
*/

int x_color=15,x_fillcolor=15; /* as set by setcolor, setfillstyle */
static int currentcolor=-1; /* as color of gc's */

#include "drawbit.c"

struct xwindowhints_s xwindowhints = {
  drawbit_width,drawbit_height,/* X icon size */
  drawbit_bits,   /* icon data */
  NULL,           /* standard x-window geometry (passed from -g option) */
  "black",        /* background color (passed from -bg option) */
  "JK X-draw",    /* X window name */
  "X-draw"        /* X icon name */
};

struct dosmouse_s dosmouse; /* to be used by mouselib to emulate DOS mouse */

Display *display;
int screen_num,display_width,display_height,no_colors;
Window win;
XSizeHints size_hints;
XEvent event;
XColor xcolor,hwcolor;
Colormap cmap;
GC gc;

#ifdef GUI
#include "config.h"
static Atom MyAtom;
struct gui_s gui;
#endif

long unsigned *xcoltab;

int x_xsize,x_ysize;

#define XERR(X,R) { fprintf(stderr,"draw/X11 error: %s\n",X); return R; }
#define XERE(X) { fprintf(stderr,"draw/X11 ERROR: %s\n",X); exit(-3); }

int keybufptr;
#define KEYBUFLEN 127
char keybuf[KEYBUFLEN+1];

void addkey(char ch) /*********************************************** addkey */
{
#ifdef DEBUG
  fprintf(stderr,"addkey: %c%c = %d\n", " ^"[ch<' '], ch>=' '?ch:ch+'@', ch);
#endif
  if (keybufptr>=KEYBUFLEN) {
    keybufptr--; XERR("keybuf ofl",) }
  keybuf[keybufptr++]=ch;
  keybuf[keybufptr]=0; /* for debugging only */
}

void addkbdstring(char *s) /*********************************** addkbdstring */
{
  while (*s) addkey(*s++);
}

int hit(int wait) /***************************************************** hit */
/*
  basic X event loop goes through here:
  if `wait' then waits until an event comes
  if not `wait' then cycles and stores keystrokes in buffer keybuf
    (other events are not bufered)
*/
{
  int pass=0;

#ifdef GUI
  XClientMessageEvent *evt;
  if (!MyAtom) MyAtom = XInternAtom(display, CTRL_ATOM_NAME, 0);
#endif

  if (!display) return 0;

  for (;;) {

    if (pass) usleep(100); // TUNE!

    if (wait) {
      if (pass) return pass; /* not to wait for the second event */
#ifdef DEBUG
      fprintf(stderr,"hit: waiting for the next event\n");
#endif
      XNextEvent(display,&event); }

    else {
#ifdef DEBUG
      //      fprintf(stderr,"chk-");
#endif
      if (!XCheckMaskEvent(display,0xffffffff,&event))
        if (!XCheckTypedEvent(display,ClientMessage,&event))
          return pass;
#ifdef DEBUG
      fprintf(stderr,"event came...");
#endif
    }

    pass++;

    event.type &= 0x7f; /* mask sent events */

    switch (event.type) {
#ifdef DEBUG
      /* ???: when a keyboard map is changed,
         keeps generating MappingNotify forever */
      case MappingNotify:
        fprintf(stderr,"MappingNotify\n");
        break;
#endif

      case Expose:
#ifdef DEBUG
        fprintf(stderr,"Expose\n");
#endif
        if (event.xexpose.count!=0) break;
        addkey(X_REDRAW_ME);
        break;

      case ConfigureNotify:
        x_xsize=event.xconfigure.width;
        x_ysize=event.xconfigure.height;
#ifdef DEBUG
        fprintf(stderr,"ConfigureNotify: x-size=%d y-size=%d\n",x_xsize,x_ysize);
#endif
        addkey(X_REDRAW_ME);
        break;

      case ButtonPress:
        dosmouse.x=event.xbutton.x;
        dosmouse.y=event.xbutton.y;
        if (event.xbutton.button==Button1) dosmouse.left=1;
        if (event.xbutton.button==Button2) dosmouse.middle=1;
        if (event.xbutton.button==Button3) dosmouse.right=1;
        if (event.xbutton.button==Button4) dosmouse.wheel++;
        if (event.xbutton.button==Button5) dosmouse.wheel--;
#ifdef DEBUG
        fprintf(stderr,"ButtonPress: x=%d y=%d buttons=%d%d%d w=%d\n",
                dosmouse.x,dosmouse.y,dosmouse.left,dosmouse.middle,dosmouse.right,dosmouse.wheel);
#endif
        break;

      case ButtonRelease:
        dosmouse.x=event.xbutton.x;
        dosmouse.y=event.xbutton.y;
        if (event.xbutton.button==Button1) dosmouse.left=0;
        if (event.xbutton.button==Button2) dosmouse.middle=0;
        if (event.xbutton.button==Button3) dosmouse.right=0;
        /* DIRTY - see mousekey.c */
        /*.....      if (event.xbutton.button==Button4) dosmouse.wheel=0;*/
        /*.....      if (event.xbutton.button==Button5) dosmouse.wheel=0;*/
#ifdef DEBUG
        fprintf(stderr,"ButtonRelease: x=%d y=%d buttons=%d%d%d\n",
                dosmouse.x,dosmouse.y,dosmouse.left,dosmouse.middle,dosmouse.right);
#endif
        break;

      case MotionNotify:
        dosmouse.x=event.xbutton.x;
        dosmouse.y=event.xbutton.y;
#ifdef DEBUG
        fprintf(stderr,"MotionNotify: x=%d y=%d buttons=%d%d%d\n",
                dosmouse.x,dosmouse.y,dosmouse.left,dosmouse.middle,dosmouse.right);
#endif
        break;

#ifdef GUI
      /* message received from xmenu */
      case ClientMessage:
        evt = (XClientMessageEvent *) &event;

        if (evt->message_type == MyAtom) {
#ifdef DEBUG
          time_t ttt;
          fprintf(stderr,"ClientMessage (from xmenu): type=%ld, %ld at %ld s\n",
                  (long int)evt->data.l[0],(long int)evt->data.l[1],(long int)time(&ttt));
#endif
          switch (evt->data.l[0]) {
            case ACTION_CONTROL_KEY:
              addkey(evt->data.l[1]);
              break;
            case ACTION_APP_KILL:
              exit(0); } }

        break;
#endif

      case KeyPress: {
        int n;
        char buf[128];
        KeySym key;
        XComposeStatus compose;

        n=XLookupString((XKeyEvent*)&event,buf,128,&key,&compose);

        switch (key) {

/* emulate DOS keyboard extended codes */
/* note that extkey() in drawplus.c converts these to 1-byte codes >=128 */
#define ADD(X) addkey(0); addkey(X); break;

          case XK_F1:       ADD('1')
          case XK_F2:       ADD('2')
          case XK_F3:       ADD('3')
          case XK_F4:       ADD('4')
          case XK_F5:       ADD('5')
          case XK_F6:       ADD('6')
          case XK_F7:       ADD('7')
          case XK_F8:       ADD('8')
          case XK_F9:       ADD('9')
          case XK_F10:      ADD('0')
          case XK_F11:      ADD(':')
          case XK_F12:      ADD(';')
          case XK_Right:    ADD('M')
          case XK_Left:     ADD('K')
          case XK_Up:       ADD('H')
          case XK_Down:     ADD('P')
          case XK_Insert:   ADD('R')
          case XK_Delete:   ADD('S')
          case XK_Begin:
          case XK_Home:     ADD('G')
          case XK_End:      ADD('O')
          case XK_Page_Up:  ADD('I')
          case XK_Page_Down:ADD('Q')

          /* only shift,ctrl,alt not generating any event */
          case XK_Shift_L: case XK_Shift_R:
          case XK_Control_L: case XK_Control_R:
          case XK_Alt_L: case XK_Alt_R:
            break;

          default:
            if (buf[0]) addkey(buf[0]); /* normal ASCII character */
            else addkey(X_UNKNOWN_KEY); /* any control/function key not in
                                           the above table */
        }
#ifdef DEBUG
        fprintf(stderr,"KeyPress: keysym=%d keybuf=/%c/%c/%c/%c/ bufptr=%d\n",
                (int)key,keybuf[0],keybuf[1],keybuf[2],keybuf[3],keybufptr);
#endif
      }
    }
  }
}

#if 0
/* simple */
int kbhit(void)
{
  hit(0);
  return keybufptr;
}
void Imworking(int worktime) /************************************ Imworking */
{
}
#else

int extrabufptr;
static int refreshtimer;
#ifdef HIGHPREC
static int worktime0=15;
#else
static int worktime0=3;
#endif

void Imworking(int worktime) /************************************ Imworking */
{
  if (worktime>=0) worktime0=worktime;
  refreshtimer=1;
}

int kbhit(void) /***************************************************** kbhit */
{
  static time_t ti,ti0;

  if (worktime0<=0)
    hit(keybufptr==0 && extrabufptr==0);
  else {

    if (ti0==0) ti0=time(&ti);

    if (ti-ti0>worktime0) {
      hit(keybufptr==0 && extrabufptr==0);
      refreshtimer++; }
    else time(&ti);

    if (hit(0) || refreshtimer) {
      ti0=time(&ti);
      refreshtimer=0; }
    }
  return keybufptr;
}
#endif

int getch(void) /***************************************************** getch */
{
  int ch;
  int i;

  while (!keybufptr) hit(1);

  ch=keybuf[0];

  keybufptr--;
  loop (i,0,keybufptr) keybuf[i]=keybuf[i+1];

#ifdef DEBUG
  fprintf(stderr,"getch: %c%c = %d\n", " ^"[ch<' '], ch>=' '?ch:ch+'@', ch);
#endif

  return ch;
}

#if 0
double fordelay=1;
void delay(unsigned msec)
/* anything to make the compiler happy... */
{
  int i;
  loop (i,0,msec*1000) fordelay=1.4-fordelay*fordelay;
}
#else
void delay(unsigned msec) /******************************************* delay */
/* not 100% portable ... */
{
  usleep(msec*1000); /* cf. also nanosleep () */
}
#endif

void clearviewport(void) /************************************ clearviewport */
{
  XClearWindow(display,win);
}

long unsigned winbackground;
static int xormode;

int startgraph(int mode) /*************************************** startgraph */
{
  int atx=1,aty=1,i;

  /* create X-window and set all necessary things */

  if (display) {
    /* already created - clear screen and raise instead */
    XRaiseWindow(display,win);
    XMapWindow(display,win);
    /*
       It looks like that XRaiseWindow is not sufficient, but why
       XMapWindow works, is a puzzle
    */
    clearviewport();

    {
      static XEvent dummy;
      dummy.xexpose.type=Expose;
      dummy.xexpose.display=display;
      dummy.xexpose.window=win;
      XSendEvent(display,win,False,ExposureMask,&dummy);
    }

    /*
       to eat all window exposure events (?!- at least one must come)
       (in fact, I don't know how to pop a window WITHOUT any event
    */
    i=getch();
    if (i!=X_REDRAW_ME) fprintf(stderr,"unexpected event/key %c%c = %d\n",
                                " ^"[i<' '], i>=' '?i:i+'@', i);
    while (kbhit() && keybuf[0]==X_REDRAW_ME) (void)getch();

    return abs(mode); }

  if (mode>16) no_colors=mode;
  else if (mode>0 && (mode&1)==0) no_colors=256;
  else no_colors=16;
  xcoltab=(long unsigned*)malloc(no_colors*sizeof(xcoltab[0]));
  if (xcoltab==NULL) XERE("no heap")
  memset(xcoltab,0,no_colors*sizeof(xcoltab[0]));

  display=XOpenDisplay(getenv("DISPLAY"));
  if (!display) XERR("XOpenDisplay",-1)
  screen_num=DefaultScreen(display);
  display_width=DisplayWidth(display,screen_num);
  display_height=DisplayHeight(display,screen_num);

#ifndef min
#define min(A,B) ((A)<(B)?(A):(B))
#endif
  x_xsize=min(640,display_width*2/3);
  x_ysize=min(480,display_height*2/3);

  if (xwindowhints.geometry) {
    char *ch;
    /* -g option */
    x_xsize=atoi(xwindowhints.geometry);
    if ( (ch=strchr(xwindowhints.geometry,'x')) ) {
      x_ysize=atoi(ch+1);
      if ( (ch=strpbrk(ch,"+-")) ) {
	atx=atoi(ch);
	if ( (ch=strpbrk(ch+1,"+-")) ) aty=atoi(ch); } } }

  if (x_xsize<8) x_xsize=100;
  if (x_xsize>display_width) x_xsize=display_width;

  if (x_ysize<8) x_ysize=100;
  if (x_ysize>display_height) x_ysize=display_height;

  { char *c=getenv("LAZYX11");
    if (!c || *c!='-')
      fprintf(stderr,"window geometry=%dx%d%+d%+d\n",x_xsize,x_ysize,atx,aty);
    }

  if (atx<0) atx+=display_width-x_xsize;
  if (aty<0) aty+=display_height-x_ysize;

  cmap=DefaultColormap(display,screen_num);

  /*.....winbackground=BlackPixel(display,screen_num);*/
  /* does not work generally for xor */

  if (!xwindowhints.background) xwindowhints.background="black";

  if (XAllocNamedColor(display,cmap,xwindowhints.background,&xcolor,&hwcolor))
    winbackground=xcolor.pixel;

  /*.....fprintf(stderr,"background=%ld\n",xcolor.pixel);*/

  win=XCreateSimpleWindow(display,RootWindow(display,screen_num),
			  atx,aty,
			  x_xsize,x_ysize,
			  0,
			  WhitePixel(display,screen_num),
			  winbackground);

#ifdef DEBUG
  fprintf(stderr,"Window ID: %d\n", (int)win);
#endif
#ifdef GUI
  {
    char *path=getenv("GUIPATH");

    if (path && gui.exe && gui.name) {
      int l=strlen(path)*3+strlen(gui.exe)+strlen(gui.name)*2+128;
      char *s=malloc(l);

      sprintf(s,"%s/%s -wid %ld -display %s -%s %s/xml/%s.xml -path %s &",
              path,gui.exe,
              (long int)win,
              DisplayString(display),
              gui.name,path,gui.name,
              path);
      fprintf(stderr,"%s\n",s);
      system(s);
      free(s); }
  }
#endif

  if (x_xsize>=DisplayHeight(display,screen_num) ||
      x_ysize>=DisplayHeight(display,screen_num)) {
    size_hints.flags=PPosition | PSize | PMinSize | PMaxSize | PWinGravity;
    size_hints.x=0;
    size_hints.y=0;
    size_hints.min_width=x_xsize;
    size_hints.max_width=x_xsize;
    size_hints.width=x_xsize;
    size_hints.min_height=x_ysize;
    size_hints.max_height=x_ysize;
    size_hints.height=x_ysize;
    size_hints.win_gravity=CenterGravity; }
  else {
    size_hints.flags=PPosition | PSize | PMinSize;
    size_hints.min_width=x_xsize;
    size_hints.min_height=x_ysize; }

  {
    static XWMHints wm_hints;
    static XClassHint class_hint;
    static int xnarg; /*?*/
    static char **xarg; /*?*/

    XTextProperty winname,iconame;

    Pixmap icon_pixmap=XCreateBitmapFromData(display,win,
      xwindowhints.icobits,xwindowhints.icowidth,xwindowhints.icoheight);

    XStringListToTextProperty(&xwindowhints.winname,1,&winname);
    XStringListToTextProperty(&xwindowhints.iconame,1,&iconame);

    wm_hints.initial_state=NormalState;
    wm_hints.input=True;
    wm_hints.flags=StateHint | InputHint | IconPixmapHint;

    wm_hints.icon_pixmap=icon_pixmap;

    XSetWMProperties(display,win,&winname,&iconame,xarg,xnarg,&size_hints,&wm_hints,&class_hint);
}

  XSelectInput(display,win,
	       ExposureMask | StructureNotifyMask
	       | KeyPressMask
	       | ButtonPressMask | ButtonReleaseMask | PointerMotionMask );

  {
    XGCValues gcvalues;

#if 1
    static char *turbo[16]={
      "#000000",
      "#000080",
      "#008000",
      "#008080",
      "#800000",
      "#800080",
      "#804000",
      "#c0c0c0",
      "#707070",
      "#0000ff",
      "#00ff00",
      "#00ffff",
      "#ff0000",
      "#ff00ff",
      "#ffff00",
      "#ffffff"};
#else
    static char *turbo[16]={
      "black",
      "darkblue",
      "darkgreen",
      "darkcyan",
      "darkred",
      "darkmagenta",
      "chocolate",
      "gray70",
      "gray40",
      "blue",
      "green",
      "cyan",
      "red",
      "magenta",
      "yellow",
      "white"};
#endif
    int i;

    gcvalues.function=GXcopy;
    gc=XCreateGC(display,win,GCFunction,&gcvalues);

    if (no_colors==16) {
      loop (i,1,16)
	if (!XAllocNamedColor(display,cmap,turbo[i],&xcolor,&hwcolor))
	  fprintf(stderr,"draw/X11: %s is unknown color\n",turbo[i]);
	else xcoltab[i]=xcolor.pixel;
      xcoltab[0]=winbackground;
      setcolor(WHITE); }
    }

  XMapWindow(display,win);

  /* wait until window is actually created, i.e. the first Expose */
  do {
    XNextEvent(display,&event);
#ifdef DEBUG
    fprintf(stderr,"startgraph: %s\n",
	    event.type==Expose?"Expose":"Other event");
#endif
    } while (event.type!=Expose);

  { char *c=getenv("LAZYX11");
    int i;

    if (c) {
      i=abs(atoi(c));
      if (i>5) i=5;
      if (i) sleep(i); } }

  return abs(mode);
}

void closegraph(void) /****************************************** closegraph */
/*
   graphics shut down ignored
   window remains created but inactive for future
*/
{
  if (!display) return;
  setwritemode(0);
  setfillstyle(1,LIGHTRED);
  setcolor(YELLOW);
  bar(0,0,90,40);
  settextstyle(0,0,1);
  settextjustify(0,2);
  outtextxy(12,8," WINDOW");
  outtextxy(12,22,"INACTIVE");
  setcolor(LIGHTRED);
  rectangle(0,0,getmaxx(),getmaxy());
  rectangle(1,1,getmaxx()-1,getmaxy()-1);
  hit(0);
  keybufptr=0; /* clear the buffer of keys/events */
#if 0
  XFreeGC(display,gc);
  XCloseDisplay(display);
  display=NULL;
#endif
}

void restorecrtmode(void) {}

void setpal(int col,int r,int g,int b,int maxbright) /*************** setpal */
/* maxbright<0 requests freeing colors-??? */
{
  int ret;
  if (col<0 || col>no_colors)
    XERR("color number out of range",)
  else {
    if (maxbright<0) {
      if (xcoltab[col]) XFreeColors(display,cmap,xcoltab+col,1,0);
      maxbright=-maxbright; }
    if (r+g+b==0) xcoltab[col]=winbackground;
    else {
      xcolor.red=r*0x10000/maxbright;
      xcolor.green=g*0x10000/maxbright;
      xcolor.blue=b*0x10000/maxbright;
      ret=XAllocColor(display,cmap,&xcolor);
      xcoltab[col]=xcolor.pixel;
      if (ret==0) XERR("cannot allocate color",) } }
}

void setcolor(int color) /***************************************** setcolor */
{
  static int first=1;
  if (first) {
    XSetForeground(display,gc,xcoltab[currentcolor=WHITE]);
    XSetBackground(display,gc,BLACK);
    first=0; }
  x_color=color;
}

void setgccolor(int color) /************************************* setgccolor */
{
  if (color<0 || color>=no_colors)
    XERR("color",)
      if (currentcolor!=color) {
    if (xormode)
      XSetForeground(display,gc,xcoltab[currentcolor=color]^winbackground);
    else
      XSetForeground(display,gc,xcoltab[currentcolor=color]); }
}

void setfillstyle(int mode,int color) /************************ setfillstyle */
{
  /* only mode=1 (normal fill) supported */
  x_fillcolor=color;
}

void setlinestyle(int mode,int pattern,int thick)  /*********** setlinestyle */
/* only 1 dotted line supported */
/* thick more general than for DOS */
{
  XGCValues gcvalues;
  unsigned long mask=GCLineStyle|GCLineWidth;

  if (mode==0)
    gcvalues.line_style=LineSolid;
  else {
    mask|=GCDashList;
    gcvalues.line_style=LineOnOffDash;
    gcvalues.dashes=2;
    XSetDashes(display,gc,2,"\1\3",2);
  /* "\1\3" is dot-black-black-black */ }

  gcvalues.line_width=thick;
  XChangeGC(display,gc,mask,&gcvalues);
}

void setwritemode(int mode) /********************************** setwritemode */
{
  XGCValues gcvalues;
  if (mode) gcvalues.function=GXxor;
  else gcvalues.function=GXcopy;
  XChangeGC(display,gc,GCFunction,&gcvalues);
  xormode=mode;
  currentcolor=-1;
}

static int ggx,ggy;

void line(int x,int y,int xx,int yy)  /******************************** line */
{
  setgccolor(x_color);
  XDrawLine(display,win,gc,x,y,ggx=xx,ggy=yy);
}

void draw(REAL x1,REAL y1) /******************************************* draw */
{
  int gx=SX(x1), gy=SY(y1);
  setgccolor(x_color);
  if (pen) XDrawLine(display,win,gc,ggx,ggy,gx,gy);
  ggx=gx;
  ggy=gy;
  pen=1;
}

void bar(int x0,int y0,int x1,int y1) /********************************* bar */
{
  /* warning: good only if x1>x0, y1>y0 !!! */
  setgccolor(x_fillcolor);
  XFillRectangle(display,win,gc,x0,y0,x1-x0+1,y1-y0+1);
}

void rectangle(int x0,int y0,int x1,int y1) /********************* rectangle */
{
  setgccolor(x_color);
  XDrawRectangle(display,win,gc,min(x0,x1),min(y0,y1),abs(x1-x0),abs(y1-y0));
}

void putpixel(int x,int y,int color) /***************************** putpixel */
{
  setgccolor(color);
  XDrawPoint(display,win,gc,x,y);
}

void fillellipse(int x,int y,int rx,int ry) /******************* fillellipse */
{
  setgccolor(x_fillcolor);
  XFillArc(display,win,gc,x-rx,y-rx,rx*2,ry*2,0,360*64);
  setgccolor(x_color);
  XDrawArc(display,win,gc,x-rx,y-rx,rx*2,ry*2,0,360*64);
}

void circle(int x,int y,int r) /************************************* circle */
{
  setgccolor(x_color);
  XDrawArc(display,win,gc,x-r,y-r,r*2,r*2,0,360*64);
}
