/* FOURIER.C - translated from FOURIER.PAS */

/*================== FOURIER.PAS ==== (c) J.Kolafa 1990 ==================

     Generalized fast Fourier transform
     To be #included directly (deprecated version), cf. fourier.h
     
     See fft.c,fft.h which are in the form of a module


Declarations required:
^^^^^^^^^^^^^^^^^^^^^^
  Type  real == <real | single | double | extended>;
  Const Max3Dh == <max grid for (3D>; ($IfDef ER3DL || ER3D only)
  void ScanKey: to allow scan of pressed key (may be empty)

Procedures available:
^^^^^^^^^^^^^^^^^^^^^
  Fourier(var a:CVectorPtr; n:integer);
  FourierER(a:RVectorPtr; nh,by:integer);        ($IfDef ER )
  FourierER4(a,b,c,d:RVectorPtr; nh,by:integer); ($IfDef ER )
  FourierER3D(var a:VectorER3D; nh:integer);     ($IfDef ER && ER3D)
  FourierER3DL(var a:Large3D; nh:integer);       ($IfDef ER && ER3DL)

============================ FOURIER.PAS ==============================*/

#define varlen 1

typedef struct { real re,im; } complex;
typedef struct Qstruct {
  struct Qstruct *next;
  int NofQ;
  complex Q[varlen]; } Qtype,*QPtr;

QPtr QHead=NULL;


QPtr InitFour(int n) /*======================== InitFour */
/* returns pointer to the table of Q[j]==exp(i*j/n), j==0..n-1,
  creates new table if (necessary */
{
real x;
int i;
QPtr Qn;

Qn=QHead; /* searching for (Q-table */
while (Qn!=NULL)
  if (Qn->NofQ==n) return Qn;
  else Qn=Qn->next;

/* creating new Q-table */
alloc(Qn,sizeof(Qtype)+(abs(n)-1)*sizeof(complex));
Qn->next=QHead; QHead=Qn;
Qn->NofQ=n;
for (i=0; i<abs(n); i++) {
  x=2*PI/n; Qn->Q[i].re=cos(x*i); Qn->Q[i].im=sin(x*i); }
return Qn;
} /* InitFour */


void Fourier(complex *a, int n) /*============== Fourier

complex array a[0..m-1], m==abs(n), is replaced by the Fourier transform:

           m-1
    a[k] = SUM (a[j]*exp[i*j*k/n]), k==0..m-1
           j=1

NOTES:
* n<0 denotes back transform.
* Efficiency decreases if  n  contains large prime numbers,
  recommended values are  n == 2**k1 * 3**k2, where k2 is not too big. */

{
int iaux,iw,ia,iq,i,j,k,l,m,d,blk,NewBlk,NdD,jBlk,jNewBlk,kNdD;
unsigned int bLen;
complex aux,w,ww,w1,w2,ww1,ww2,r3;
complex *aparm,*b,*swap,*Q;
QPtr Qn;

r3.re=-0.5; r3.im=sqrt(0.75);
if (n<0) r3.im=-r3.im;
Qn=InitFour(n);
n=abs(n);
blk=n; d=2;
bLen=n*sizeof(complex); alloc(b,bLen);

/* orig:
REPEAT
  a2b(a^,b^);
  if (blk>1) a2b(b^,a^) else move(b^,a^,bLen);
  UNTIL blk<==1;
*/

aparm=a;

do { /* one step */

while (blk % d != 0) d++;

m=n/blk; NewBlk=blk/d; NdD=n/d;

jBlk=0; jNewBlk=0;
for (j=0; j<m; j++) {
  for (i=0; i<NewBlk; i++)
    switch (d) {

case 2: /* optimized code */
      ia=jBlk+i;
      aux=a[ia+NewBlk];
      Q = Qn->Q + jNewBlk;
      w.re = Q->re*aux.re - Q->im*aux.im;
      w.im = Q->im*aux.re + Q->re*aux.im;
      iw=jNewBlk+i;
      aux=a[ia];
      b[iw].re=aux.re+w.re; b[iw].im=aux.im+w.im;
      b[iw+NdD].re=aux.re-w.re; b[iw+NdD].im=aux.im-w.im;
      break; /*2*/

case 3: /* optimized code */
      ia=jBlk+i; iq=jNewBlk; iaux=ia+NewBlk;
      aux=a[iaux];
      Q = Qn->Q + iq;
      w.re = Q->re*aux.re - Q->im*aux.im;
      w.im = Q->im*aux.re + Q->re*aux.im;
      aux=a[iaux+NewBlk];
      Q = Qn->Q + iq*2 % n;
      ww.re = Q->re*aux.re - Q->im*aux.im;
      ww.im = Q->im*aux.re + Q->re*aux.im;
      iw=jNewBlk+i;
      aux=a[ia];
      b[iw].re=aux.re+w.re+ww.re; b[iw].im=aux.im+w.im+ww.im;
      w1.re=w.re*r3.re; w2.re=w.im*r3.im;
      w2.im=w.re*r3.im; w1.im=w.im*r3.re;
      ww1.re=ww.re*r3.re; ww2.re=ww.im*r3.im;
      ww2.im=ww.re*r3.im; ww1.im=ww.im*r3.re;
      iaux=iw+NdD;
      b[iaux].re=aux.re+w1.re-w2.re+ww1.re+ww2.re;
      b[iaux].im=aux.im+w1.im+w2.im+ww1.im-ww2.im;
      iaux += NdD;
      b[iaux].re=aux.re+w1.re+w2.re+ww1.re-ww2.re;
      b[iaux].im=aux.im+w1.im-w2.im+ww1.im+ww2.im;
      break; /*3*/

default: /* d>3: not optimized */
      kNdD=0; iw=jBlk+i;
      for (k=0; k<d; k++) {
        ia=iw;
        w=a[ia];
        iaux=jNewBlk+kNdD; iq=iaux;
        for (l=1; l<d; l++) {
          Q = Qn->Q + iq;
          ia += NewBlk; iq=(iq+iaux) % n;
          aux=a[ia];
          w.re += aux.re*Q->re - aux.im*Q->im;
          w.im += aux.re*Q->im + aux.im*Q->re; }
        b[jNewBlk+i+kNdD]=w;
        kNdD += NdD; } /*k*/

    } /*CASE, i*/
  jBlk += blk; jNewBlk += NewBlk;
  } /*j*/

blk=NewBlk;

swap=a; a=b; b=swap;


} while (blk>1);

/* to preserve the stack-like structure of the heap */
if (a!=aparm) {
  swap=a; a=b; b=swap;
  copy(a,b,bLen); }

free(b);
} /* Fourier */
#ifdef ER

void FourierER(real *a, int nh, int by) /* ========== FourierER

Fourier transform of 1 Even Real function a^ [0..nh] - not optimized !
Period of the function is n==2*nh.
'a^' may be a non-contiguous vector, increment is 'by' in units of
real size (by==1 corresponds to a usual contiguous vector). */
{
complex *f;
unsigned int fLen;
int i,n;

InitFour(nh*2);
fLen=sizeof(complex)*nh*2; alloc(f,fLen);
nh=abs(nh); n=2*nh;

for (i=0; i<=nh; i++) { f[i].re=a[i*by]; f[i].im=0; }
for (i=1; i<=nh; i++) f[n-i]=f[i];

Fourier(f,n);

for (i=0; i<=nh; i++) a[i*by]=f[i].re;
free(f);
} /* FourierER */

void FourierER4(real *a, real *b, real *c, real *d, int nh, int by) /* FourierER4 */
/* Fourier transform of 4 Even Real functions a^,b^,c^,d^ [0..nh].
  Optimized code, see comments of FourierER. */
{
int iby,i,j,n;
unsigned int fLen;
real c0,c1,d0,d1;
complex fi,fni,*f,*Q;
QPtr Qn;

nh=abs(nh); n=nh*2;
Qn=InitFour(n);
fLen=sizeof(complex)*n; alloc(f,fLen);

for (i=0; i<n; i++) {
  Q = Qn->Q + i;
  j=n-i; if (j>i) j=i; j=j*by;
  f[i].re = a[j] + Q->re*c[j] - Q->im*d[j];
  f[i].im = b[j] + Q->im*c[j] + Q->re*d[j]; }

Fourier(f,n);

c0=(c[0]+c[nh*by])/2; c1=(c[0]-c[nh*by])/2;
d0=(d[0]+d[nh*by])/2; d1=(d[0]-d[nh*by])/2;

for (i=1; i<nh; i++) {
  iby=i*by;
  Q = Qn->Q + i;
  c0=c0+c[iby]; c1=c1+c[iby]*Q->re;
  d0=d0+d[iby]; d1=d1+d[iby]*Q->re; }

c[0]=2*c0; c[by]=2*c1; d[0]=2*d0; d[by]=2*d1;

for (i=1; i<nh; i++) {
  iby=i*by;
  fni=f[n-i]; fi=f[i];
  c[iby+by]=fi.re-fni.re+c[iby-by];
  a[iby]=(fi.re+fni.re-c[iby-by]-c[iby+by])/2;
  d[iby+by]=fi.im-fni.im+d[iby-by];
  b[iby]=(fi.im+fni.im-d[iby-by]-d[iby+by])/2; }

a[0]=f[0].re-c[by]; b[0]=f[0].im-d[by];
a[nh*by]=f[nh].re-c[nh*by-by]; b[nh*by]=f[nh].im-d[nh*by-by];

free(f);
} /* FourierER4 */
#endif




#ifdef ER3DL
/* NOTE: this is old DOS version, originally from Pascal, for >=32 bit
compilers use OZ3D */

void FourierER3DL(real **xPtr, int nh) /* FourierER3DL */
/* as FourierER3D but xPtr[x]^[y*span+z] replaces a[x,y,z] */
{
int span,x,y,z,yz,i;
unsigned int AuxLen;
real *p[4];
int yy[4],zz[4];
real *aux,*off;

span=nh+1;
InitFour(2*nh); /* to allocate Q before Aux */
AuxLen=sizeof(real)*4*span; alloc(aux,AuxLen);

for (i=0; i<4; i++) p[i]=aux+i*span;

i=0;
for (y=0; y<=nh; y++) for (z=0; z<=nh; z++) {
  off=p[i]; yz=y*span+z;
  for (x=0; x<=nh; x++) off[x]=xPtr[x][yz];
  yy[i]=y; zz[i]=z; i++;
  if (i==4) {
    Ff4(p,nh,1);
    for (i=0; i<4; i++) {
      off=p[i]; yz=yy[i]*span+zz[i];
      for (x=0; x<=nh; x++) xPtr[x][yz]=off[x]; }
    i=0; } }
if (i==1) {
  Fflush(p,nh,1,i);
  yz=yy[0]*span+zz[0]; /*off==p[0]*/
  for (x=0; x<=nh; x++) xPtr[x][yz]=off[x]; }

i=0;
for (x=0; x<=nh; x++) for (z=0; z<=nh; z++) {
  p[i]=xPtr[x]+z; i++;
  if (i==4) { Ff4(p,nh,span); i=0; } }
Fflush(p,nh,span,i);

i=0;
for (x=0; x<=nh; x++) for (y=0; y<=nh; y++) {
  p[i]=xPtr[x]+y*span; i++;
  if (i==4) { Ff4(p,nh,1); i=0; } }
Fflush(p,nh,1,i);

free(aux);

} /*FourierER3DL*/
#endif

#ifdef ER3DSYM
/*** version of ER3D optimized for x<-->y<-->z symmetric functions ***/
void Ff4(real **p, int nh, int span)
{
FourierER4(p[0],p[1],p[2],p[3],nh,span);
}

void Fflush(real **p, int nh, int span, int i)
{
switch (i) {
  case 0: return;
  case 1: FourierER(p[0],nh,span); break;
  case 2: {
    real *aux;
    alloczero(aux,(span*(nh+1))*2*sizeof(real));
    FourierER4(p[0],p[1],aux,aux+span*(nh+1),nh,span);
    free(aux); } break;
  case 3: {
    real *aux;
    alloczero(aux,span*(nh+1)*sizeof(real));
    FourierER4(p[0],p[1],p[2],aux,nh,span);
    free(aux); } }
}

#ifndef PARALLEL

void FourierER3D(real *a,int nh) /* ====== FourierER3D/serial

3D fourier transform of Even Real 3D function a[0..nh,0..nh,0..nh].
Period is 2*nh in each coordinate.
For a[x,y,z]=a[y,x,z]=a[y,z,x]...

NOTE map function: a[x,y,z] = x*span^2 + y*span + z, where span=nh+1
NOTE a is one large array, not good for 16bit DOS...
*/
{
int span=nh+1;
int x,y,z,i;
real *p[4];
InitFour(2*nh); /* to allocate Q before Aux */

#define A(X,Y,Z) a[X*(span*span)+Y*span+Z]

i=0;
for (y=0; y<=nh; y++) for (z=0; z<=y; z++) {
  p[i]=a+y*span+z; i++;
  if (i==4) { Ff4(p,nh,Sqr(span)); i=0; } }
Fflush(p,nh,Sqr(span),i);
for (y=0; y<=nh; y++) for (z=0; z<y; z++)
  for (x=0; x<=nh; x++) A(x,z,y)=A(x,y,z);

i=0;
for (x=0; x<=nh; x++) for (z=0; z<=nh; z++) {
  p[i]=a+x*Sqr(span)+z; i++;
  if (i==4) { Ff4(p,nh,span); i=0; } }
Fflush(p,nh,span,i);

i=0;
for (x=0; x<=nh; x++) for (y=0; y<=x; y++) {
  p[i]=a+x*Sqr(span)+y*span; i++;
  if (i==4) { Ff4(p,nh,1); i=0; } }
Fflush(p,nh,1,i);
for (x=0; x<=nh; x++) for (y=0; y<x; y++)
  for (z=0; z<=nh; z++) A(y,x,z)=A(x,y,z);

#undef A

} /*FourierER3D*/

#else /* PARALLEL */

static real *aa;
static int nhnh,sspan;
static int nthreads;
static int indx1[25],indx2[25];

void par_1(int ithread)
{
int i=0,j=0,y,z;
int from=indx1[ithread],to=indx1[ithread+1];
real *p[4];

for (y=0; y<sspan; y++) for (z=0; z<=y; z++) {
  p[i]=aa+y*sspan+z; i++;
  if (i==4) {
    if (j>=from && j<to) Ff4(p,nhnh,Sqr(sspan));
    j++; i=0; } }

if (ithread==0) Fflush(p,nhnh,Sqr(sspan),i);
}

void par_2(int ithread)
{
int i=0,j=0,x,z;
int from=indx2[ithread],to=indx2[ithread+1];
real *p[4];

for (x=0; x<sspan; x++) for (z=0; z<sspan; z++) {
  p[i]=aa+x*Sqr(sspan)+z; i++;
  if (i==4) {
    if (j>=from && j<to) Ff4(p,nhnh,sspan);
    i=0; j++; } }
if (ithread==0) Fflush(p,nhnh,sspan,i);
}

void par_3(int ithread)
{
int i=0,j=0,x,y;
int from=indx1[ithread],to=indx1[ithread+1];
real *p[4];

for (x=0; x<sspan; x++) for (y=0; y<=x; y++) {
  p[i]=aa+x*Sqr(sspan)+y*sspan; i++;
  if (i==4) {
    if (j>=from && j<to) Ff4(p,nhnh,1);
    i=0; j++; } }

if (ithread==0) Fflush(p,nhnh,1,i);
}

void FourierER3D(real *a,int nh) /* ====== FourierER3D/parallel

3D fourier transform of Even Real 3D function a[0..nh,0..nh,0..nh].
Period is 2*nh in each coordinate.
For a[x,y,z]=a[y,x,z]=a[y,z,x]...

NOTE map function: a[x,y,z] = x*span^2 + y*span + z, where span=nh+1
NOTE a is one large array, not good for 16bit DOS...
*/
{
int x,y,z,i,iii;

nhnh=nh;
sspan=nh+1;
aa=a;

if (nthreads==0) {
  char *e=getenv("MPC_NUM_THREADS");
  if (!e) ERROR(("no env MPC_NUM_THREADS"))
  nthreads=atoi(e);
  if (nthreads<1 || nthreads>24) ERROR(("bad MPC_NUM_THREADS")); }

InitFour(2*nh); /* to allocate Q before Aux */

#define A(X,Y,Z) a[X*(sspan*sspan)+Y*sspan+Z]

/* decomposition for triangular coordinates (both par_1 and par_3) */
z=sspan*(sspan+1)/2;
y=z/4;
loopto (i,0,nthreads) indx1[i]=y*i/nthreads;

/* decomposition for square coordinate y (par_2) */
z=sspan*sspan;
y=z/4;
loopto (i,0,nthreads) indx2[i]=y*i/nthreads;

#pragma parallel
#pragma local(iii) shared(nthreads) 
  {
#pragma pfor iterate(iii=0;nthreads;1) schedtype(simple)
  loop (iii,0,nthreads) par_1(iii);
  }
for (y=0; y<=nh; y++) for (z=0; z<y; z++)
  for (x=0; x<=nh; x++) A(x,z,y)=A(x,y,z);

#pragma parallel
#pragma local(iii) shared(nthreads) 
  {
#pragma pfor iterate(iii=0;nthreads;1) schedtype(simple)
  loop (iii,0,nthreads) par_2(iii);
  }

#pragma parallel
#pragma local(iii) shared(nthreads) 
  {
#pragma pfor iterate(iii=0;nthreads;1) schedtype(simple)
  loop (iii,0,nthreads) par_3(iii);
  }
for (x=0; x<=nh; x++) for (y=0; y<x; y++)
  for (z=0; z<=nh; z++) A(y,x,z)=A(x,y,z);

#undef A

} /*FourierER3D*/

#endif /* PARALLEL */

#endif /* ER3DSYM */
