/*
  WARNING: because of some agressive #defines, must be #included after all
  system #include's !!! 
*/

#ifdef __cplusplus
extern "C" {
/*
  it is assumed that a conversion function/macro Val(X) returning double 
  exists for each class X that is to be used in macros put, put2, put3 
  and similarly for Int (not used here)
*/
#else
#define Val(X) (double)(X)
#define Int(X) (int)(X)
#endif

#ifndef CALC
#define CALC 4
#endif

#include <stdlib.h>
#include <stdio.h>
#include <ctype.h>
#include <limits.h>
#include <stdarg.h> /* added 8/2009 - removed from ground.c, alloc.c */

#include <string.h>
#include <math.h>   /* on stellar: #include <fastmath.h> */
#ifndef FP_NAN
#define FP_NAN (1./(exp(0)-1)*(cos(0)-1))
#endif
#ifndef isfinite
#define isfinite(X) (X>-1.797693e+308 && X<1.797693e+308)
#endif
#if CALC
double Calc(char *buf,char **endscan);
#endif

extern FILE *in,*out;

#include "loop.h"
#include "int4.h"

#define PI_DOUBLE 3.14159265358979323846
#ifndef PI
#define PI PI_DOUBLE
#endif

#ifndef Max
/* A=max(A,B) */
#define Max(A,B) { if ((B)>(A)) (A)=(B); }
/* A=min(A,B) */
#define Min(A,B) { if ((B)<(A)) (A)=(B); }
#endif

#ifndef max
/* see below for fmax and fmin */
#define max(A,B) ((A)<(B)?(B):(A))
#define min(A,B) ((A)<(B)?(A):(B))
#endif

#ifndef copy
#ifdef __TRANSPUTER__
#define copy block_move
void block_move(void *, const void *, int);
#else
#define copy memcpy
#endif
#endif

#define Sqr(X) ((X)*(X))
#define Cub(X) ((X)*(X)*(X))
#define Pow4(X) Sqr(Sqr(X))
#define Pow5(X) ((X)*Sqr(Sqr(X)))
#define Pow6(X) Cub(Sqr(X))

// cbrt(): C99, POSIX.1-2001
#define cubrt(X) cbrt(X)
//#define cubrt(X) pow((X),1e0/3)

#define REAL double
#define FN(X) X
#include "func.h"
#undef REAL
#undef FN

/*
  PRECISION==0 (float, not tested) and PRECISION==2 (long double) C versions.
  The get data module works in double.  Thus, there are two versions (double 
  and given by PRECISION) of all these functions.
  If "prec.h" is used, double versions are #defined identical to 
  float/long double; also REAL is #defined in prec.h and not here.
  PRECISION>2 (emulated) versions are in C++ and use overloading, so that no 
  such naming conventions are necessary.
*/
#if PRECISION==2 || PRECISION==0
#undef REAL
#undef FN
#define CAT(X,Y) X##Y
#if PRECISION==2
#define REAL long double
#define FN(X) CAT(X,l)
#else 
#define REAL float
#define FN(X) CAT(X,f)
#endif
#include "func.h"
#undef FN
#undef CAT
#undef REAL
#endif

#define Sign(X) ((X)<0?-1:(X)>0?1:0)

/****** prt and prtc ******/

int prtc(char c);
int prt(const char *format, ...); 
int prt_(const char *format, ...);
int prts(const char *c);
int prts_(const char *c);

/****** scroll ******/

void setscroll(int lines,int columns,int clear,int closegraph);
int initscroll(int buflen);
void scroll(void);

void extError(const char *msg, const char *f, int l);
#define Error(X) extError(X,__FILE__,__LINE__)

/*
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
     new more sophisticated version of error/warning handling (not PAR)
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

examples of usage (note double parentheses!):

if (length_of_rope<0) DISASTER(("rope length is negative"))
if (length_of_rope<1) ERROR(("too short rope: %f",length_of_rope))
if (length_of_rope>5) {
  WARNING(("too long rope: %f (cut to 5) ",length_of_rope))
  length_of_rope=5; }
*/

extern struct myErrorInfo_s {
  int Line;       /* to store value of __LINE__ macro */
  const char *File;     /* to point to __FILE__ macro string */
  int Level;      /* set by WARNING/ERROR/DISASTER : */
                       /* value status    batch mode  interactive mode */
                       /* 0     WARNING   continue    asks for cont. */
                       /* 1     ERROR     stop        asks for cont. */
                       /* 2     DISASTER  stop        stop */
} myErrorInfo;

void myError(const char *format, ...);

#define WARNING(_X) { myErrorInfo.Level=0; myErrorInfo.Line=__LINE__; myErrorInfo.File=__FILE__; myError _X; }
#define ERROR(_X) { myErrorInfo.Level=1; myErrorInfo.Line=__LINE__; myErrorInfo.File=__FILE__; myError _X; }
#define DISASTER(_X) { myErrorInfo.Level=2; myErrorInfo.Line=__LINE__; myErrorInfo.File=__FILE__; myError _X; }


/****** PUT DATA  ******/

extern const char *_putformat;
#define put(_X) prt(_putformat,#_X,Val(_X));
#define put_(_X) prt_(_putformat,#_X,Val(_X));
#define putkey(_X,_Y) { _PrtKey_(#_X,Int(_X),_Y); _n }
#define putkey_(_X,_Y) _PrtKey_(#_X,Int(_X),_Y);
#define put2(_X,_Y) { put_(_X) put(_Y) }
#define put3(_X,_Y,_Z) { put_(_X) put_(_Y) put(_Z) }
#define _n prtc('\n');
#define putv(_X) prt("%11s=[%13.6g %13.6g %13.6g ]", \
  #_X,Val(_X[0]),Val(_X[1]),Val(_X[2]));


/****** GET DATA ******/

struct keys_s {
  char *key; 
  int val; 
};

#ifndef _GetBufLen
#ifdef DOS
#define _GetBufLen 128
#else
#define _GetBufLen 256
#endif
#endif

typedef struct _Idlist_s {
  struct _Idlist_s *next;
  double val;
  char id[1]; /* [var len] */ 
} _Idlist;

extern struct _Id_s {
  char fmt[16];         /* output format for nr */
  char buf[_GetBufLen]; /*                AB[-123]XY = 2*(3+4) next... */
#if CALC>=3
  double degrad;/* 1=radians, pi/180=degrees, ... */
  int list;     /* switch for creating the list of identifiers */
  int ignoreid; /* ignore unknown ID at rhs -- use zero value */
  _Idlist *head;/* list of identifiers */
#endif
  double nr;  /* numeric value            ^ ^     ^            ^       */
  char *b;    /* running pointer to buf   ^ ^     ^            *       */
  char *id;   /* pointer to identifier    * ^     ^                    */
  char *br;   /* pointer to '[' (if any)    *     ^                    */
  char *suff; /* pointer to suffix (after ']')    *                    */
  char *str;  /* string given by ID="..." */
  int indx;   /* value of subscript */
  int prt;    /* id=result of expr printed; prt=2: command `??' */
  int asg;    /* assign val to register (#); op for += _= ... */
  int echo;   /* toggle by ?=:
		 0:no echo, 1:copy input, 2:echo (print results) 
                 -1: ignore some errors */
  int end;    /* set if ';' */
  int empty;  /* set if empty line or ";" only */
  int used;   /* set if match */
  int col;    /* packing identifiers to a line for the ?? function */
} _Id;


int _GetId(void);
#if CALC<2
int _GetKey(char *id,struct keys_s *key);
int _GetCmp(char *id);
int _GetVecCmp(char *id,char *suff,int maxindex);
#else /* CALC>=2 */
int _GetKey(char *id,double val,struct keys_s *key);
int _GetCmp(char *id,double val);
int _GetVecCmp(char *id,char *suff,int maxindex,double val);
#endif
void _CheckData(void);
int _GetEnd(void);

#if CALC<3
#define getdata { *(_Id.b=_Id.buf)=0; do { if (_GetId()) {
#else
#define getdata { *(_Id.b=_Id.buf)=0; _Id.list=1; do { if (_GetId()) {
#endif

#if CALC<2
#define get(_X) if (_GetCmp(#_X)) _X=_Id.nr;
#define getkey(_X,_Y) if (_GetKey(#_X,_Y)) _X=_Id.nr;
#define getmark(_X,_M) if (_GetCmp(#_X)) _M++,_X=_Id.nr;
#define getvec(_X,_S,_MI) if (_GetVecCmp(#_X,#_S,_MI)) _X[_Id.indx]_S=_Id.nr;
#define getvecmark(_X,_S,_MI,_M) if (_GetVecCmp(#_X,#_S,_MI)) _M++,_X[_Id.indx]_S=_Id.nr;
#else /* CALC>=2 */
#define get(_X) if (_GetCmp(#_X,Val(_X))) _X=_Id.nr;
#define getkey(_X,_Y) if (_GetKey(#_X,Val(_X),_Y)) _X=_Id.nr;
#define getmark(_X,_M) if (_GetCmp(#_X,Val(_X))) _M++,_X=_Id.nr;
#define getvec(_X,_S,_MI) if (_GetVecCmp(#_X,#_S,_MI,Val(_X[_Id.indx]_S))) _X[_Id.indx]_S=_Id.nr;
#define getvecmark(_X,_S,_MI,_M) if (_GetVecCmp(#_X,#_S,_MI,Val(_X[_Id.indx]_S))) _M++,_X[_Id.indx]_S=_Id.nr;
#endif

#define checkdata _CheckData();
#define enddata  } } while (!_GetEnd());  }

int prtkey_(int val,struct keys_s *key,int how);
void _PrtKey_(char *id,int val,struct keys_s *key);

/****** headers of tables ******/

void putline(char c,int l);
void header(char *h);
void underline(char *t);

void graph(double x,int linelen);
/*
Pseudo-graphics with a char/4 resolution.
For 0<=x<=1, a multi-character is printed at the position scaled by linelen.
*/

#ifdef __cplusplus
}
#endif

#ifdef MYSQRT
/*
  this is optional sqrt function replacement
  implementation requirements:
    sizeof(unsigned)=4
    sizeof(long unsigned)=8,
    sizeof(double)=8, IEEE format
  significant improvement detected on DEC Alpha processors
  no improvement found on PC (gcc/emx) nor R4400 SGI
  options:
    #define MYSQRT 1 : sqrt is a function
    #define MYSQRT 2 : sqrt is a macro (a bit faster)
*/
#ifdef PRECISION
#if PRECISION!=1
#error MYSQRT requires PRECISION=1
#endif
#endif

#if MYSQRT==1
/* function */
double mySqrt(double);
#define sqrt mySqrt

#else
/* macro (inline), see ground.h */
extern unsigned sqrt_tab[64];
extern long unsigned sqrt_u;
extern double sqrt_t, sqrt_x;
extern union sqrt_un { long unsigned u; double d; } sqrt_ud;

#define sqrt(_X)  (sqrt_ud.d = _X,\
  sqrt_u = 0x5fe80000UL-(sqrt_ud.u>>33),\
  sqrt_ud.u = sqrt_u-(long unsigned)sqrt_tab[(sqrt_u>>14) & 63UL]<<32,\
  sqrt_x = sqrt_ud.d,\
  sqrt_x *= 3.0-_X*sqrt_x*sqrt_x, \
  sqrt_x *= 12.0-_X*sqrt_x*sqrt_x, \
  sqrt_x *= 0.0625, \
  sqrt_t = _X*sqrt_x,\
  sqrt_t+sqrt_x*0.5*(_X-sqrt_t*sqrt_t))
#endif
#endif

#include "fortran.h"
