/*
  2009: x error bars added (PLOTERRBAR)
  2006: bug fixed (asymmetric error bars)
  This module draws a file, using formula, 2004.
  Error bars supported.
  Used by plot.
*/
#include "ground.h" /* #define CALC 3 required */
#include "alloc.h"
#include "draw.h"
#include "ploterr.h"
#include "mydraw.h"

#ifdef DOS
#define LINELEN 1024
#else
#define LINELEN 16384
#endif

/* separator characters; must not include: ! # (and any number components) */
#define SEP " \t\n\r,;"

double *count;
struct _Idlist_s *id;

int drawerrbar(char *fn,
               char *colx,char *coly,
               char *coldy1,char *coldy2,
               int maxcol,int color,int style,int err)
{
  FILE *f;
  char line[LINELEN];
  int i,OK;
  double xy[4];
  char *c=getenv("PLOTERRBAR");
  enum errmode_e {YY,XX,XY} errmode=YY;

  if (c && tolower(c[0])=='x') {
    errmode++;
    if (tolower(c[1])=='y') errmode++; }

  if (!fn || fn[0]==0) f=stdin;
  else f=fopen(fn,"rt");

  if (!f) {
    fprintf(stderr,"\"%s\" does not exist\n",fn);
    return -1; }

  mysetcolor(color);

  if (err) {
    mysetlinestyle(0,0,1);
    if (style<=0) pointsize=(-1-style)*0.798; }
  else {
    if (style>1) mysetlinestyle(0,0,2*style+1);
    else if (style>0) mysetlinestyle(style,0,style);
    else my_style.pointsize=pointsize=(-1-style)*0.798; }
  my_style.pointsize=pointsize; /* ?? */
//fprintf(stderr,"%g\n",(double)my_style.pointsize);
  myup();
  *count=0;

  while (fgets(line,LINELEN,f)) {
    char *e,*tok;
    int n;

    tok=strtok(line,SEP);

    if (!tok || strchr("!#",tok[0])) {
      myup();
      if (!tok) *count=0; }
    else {

      /* _Id.head->val = count, _Id.head->next->val = 1st number, etc. */
      n=1;
      id=_Id.head->next;

      do {
        id->val=strtod(tok,&e);
        if (e==tok) id->val=FP_NAN;
        id=id->next;
        n++;
        tok=strtok(NULL,SEP);
      } while (tok && id);

      OK=1;
      loop (i,0,2+err) {
        double z;
        char *expr =
          i==0 ? colx :
          i==1 ? coly :
          i==2 ? coldy1 : coldy2;
        e=expr; /* err */
        if (maxcol<n) z=Calc(expr,&e); else OK=0;
        if (e==expr) OK=0;
        if (OK) xy[i]=z; }

      *count+=1.0;

      if (isfinite(xy[0]) && isfinite(xy[1])) {
        if (err)
          switch (errmode) {
            case YY: myerrybar(xy[0],xy[1]-xy[2],xy[1]+xy[3]); break;
            case XX: myerrxbar(xy[0]-xy[2],xy[0]+xy[3],xy[1]); break;
            case XY: myerrxbar(xy[0]-xy[2],xy[0]+xy[2],xy[1]);
                     myerrybar(xy[0],xy[1]-xy[3],xy[1]+xy[3]); }
        else
          if (OK) {
            if (style>=0) mydraw(xy[0],xy[1]);
            else mysmartdot(xy[0],xy[1]); }
          else
            myup(); } } }

  fclose(f);

  return 0;
}

int drawfile(char *fn,char *colx,char *coly,int maxcol,int color,int style)
{
  return drawerrbar(fn,colx,coly,NULL,NULL,maxcol,color,style,0);
}

int minmaxfile(char *fn,int cont,char *colx,char *coly,int maxcol,
               double *X0,double *X1,double *Y0,double *Y1)
{
  FILE *f;
  char line[LINELEN];
  int i,OK;
  double xy[2];

  *count=0;

  if (cont==0) {
    *X0=*Y0=3e33;
    *X1=*Y1=-3e33; }

  if (fn==NULL || (f=fopen(fn,"rt"))==NULL) {
    fprintf(stderr,"\"%s\" does not exist\n",fn);
    return -1; }

  while (fgets(line,LINELEN,f)) {
    char *e,*tok;
    int n;

    tok=strtok(line,SEP);

    if (!tok || strchr("!#",tok[0])) {
      myup();
      if (!tok) *count=0; }
    else {

      /* _Id.head->val = count, _Id.head->next->val = 1st number, etc. */
      n=1;
      id=_Id.head->next;

      do {
        id->val=strtod(tok,&e);
        if (e==tok) id->val=FP_NAN;
        id=id->next;
        n++;
        tok=strtok(NULL,SEP);
      } while (tok && id);


      OK=1;
      loop (i,0,2) {
        double z;
        char *expr=i?coly:colx;

        e=expr; /* err */
        if (maxcol<n) z=Calc(expr,&e); else OK=0;
        if (e==expr) OK=0;
        if (OK) xy[i]=z; }

      *count+=1.0;

      if (OK) {
        if (isfinite(xy[0])) { Min(*X0,xy[0]) Max(*X1,xy[0]) }
        if (isfinite(xy[1])) { Min(*Y0,xy[1]) Max(*Y1,xy[1]) } } } }

  fclose(f);

  return 0;
}
