#include "gui_blend.h"

namespace gctrl {

BlendApp::BlendApp ( long wid, char *xml_file )
  : GUIApp( wid, xml_file, "Blend menu" )
{
  b_ballsize_inc = b_ballsize_dec = 0;
  b_id = b_type = b_number = b_charge = 0;
  b_cancel_description = 0; 
  b_energy =  b_minimize = b_random = 0;
  b_move = b_unmark = 0;
  b_grid = b_projection_home = b_projection_end = 0;
  b_wires = b_sticks = b_spheres = b_anaglyph = 0;
  b_go = 0; 

  builder->get_widget( "b_ballsize_dec",  b_ballsize_dec );
  builder->get_widget( "b_ballsize_inc",  b_ballsize_inc );
  builder->get_widget( "b_id",            b_id );
  builder->get_widget( "b_number",        b_number );
  builder->get_widget( "b_type",          b_type );
  builder->get_widget( "b_charge",        b_charge );
  builder->get_widget( "b_cancel_description", b_cancel_description );
  builder->get_widget( "b_energy",        b_energy );
  builder->get_widget( "b_minimize",      b_minimize );
  builder->get_widget( "b_random",        b_random );
  builder->get_widget( "b_move",          b_move );
  builder->get_widget( "b_unmark",        b_unmark );
  builder->get_widget( "b_grid",          b_grid );
  builder->get_widget( "b_projection_home",    b_projection_home ); // projection
  builder->get_widget( "b_projection_end",     b_projection_end ); // projection
  builder->get_widget( "b_wires",         b_wires );
  builder->get_widget( "b_sticks",        b_sticks );
  builder->get_widget( "b_spheres",       b_spheres );
  builder->get_widget( "b_anaglyph",      b_anaglyph );
  builder->get_widget( "b_go",            b_go );


  // GUIApp's widgets
  if ( b_save ){
    b_save->show();
    connect_action_clicked(b_save, 's'&31); // ctrl-s
  }  


  // BlendApp's widgets

  if ( b_go )
    b_go->signal_clicked().connect( 
      sigc::mem_fun(*this, &BlendApp::on_go) );

  connect_action_clicked( b_ballsize_inc, 82, "list-add.png");      // R
  connect_action_clicked( b_ballsize_dec, 114,"list-remove.png");   // r
  connect_action_clicked( b_id,           105); // i
  connect_action_clicked( b_number,       110); // n
  connect_action_clicked( b_type,         116); // t
  connect_action_clicked( b_charge,       113); // q
  connect_action_clicked( b_cancel_description, 32); // space
  connect_action_clicked( b_energy,       69);  // E
  connect_action_clicked( b_minimize,     44);  // ,
  connect_action_clicked( b_random,       58);  // :
  connect_action_clicked( b_move,         109); // m
  connect_action_clicked( b_unmark,       21 ); // ctrl - u
  connect_action_clicked( b_grid,         61,   "grid.png");          // =
  connect_action_clicked( b_wires,        49,   "wire.png");          // 1
  connect_action_clicked( b_sticks,       50,   "bsticks_col.png");   // 2
  connect_action_clicked( b_spheres,      51,   "spheres.png");       // 3
  connect_action_clicked( b_anaglyph,     52,   "anaglyph.png");      // 4
  connect_action_clicked( b_projection_home,   1, "projection_near.png"); // home
  connect_action_clicked( b_projection_end,    5, "projection_far.png");  // end
}

BlendApp::~BlendApp ()
{
  delete_widget( b_ballsize_inc );
  delete_widget( b_ballsize_dec );
  delete_widget( b_id );
  delete_widget( b_number );
  delete_widget( b_type );
  delete_widget( b_charge );
  delete_widget( b_cancel_description );
  delete_widget( b_energy );
  delete_widget( b_minimize );
  delete_widget( b_random );
  delete_widget( b_move );
  delete_widget( b_unmark );
  delete_widget( b_grid );
  delete_widget( b_projection_home );
  delete_widget( b_projection_end );
  delete_widget( b_wires );
  delete_widget( b_sticks );
  delete_widget( b_spheres );
  delete_widget( b_anaglyph );
}

void
BlendApp::on_go ()
{
  // send '.' (ASCII 46)
  x11::send_event(tgt_win_id_, g_display_name, CTRL_ATOM_NAME, ACTION_CONTROL_KEY, 46); 
  Gtk::Main::quit();
}

}; // namespace gctrl
