#ifndef _GUIAPP_H  
#define _GUIAPP_H

#include <gtkmm.h>
#include <glibmm.h>
#include <iostream>

#include "config.h"
#include "x11.h"

// declared in main()
extern char *g_display_name;
extern Glib::ustring g_base_path; 

namespace gctrl {

/////////////////////////////////////////////////////////////////////

inline void delete_widget ( Gtk::Widget *w )
{
  delete w;
  w = 0;
};

GdkFilterReturn client_message_filter ( GdkXEvent *xev, GdkEvent *ev, gpointer data);

/////////////////////////////////////////////////////////////////////
// HelpDialog - helper class

class HelpDialog : public Gtk::Window
{
  private:
    Gtk::Label  *lbl;
    Gtk::VBox   *vbox;
    Gtk::Button *btn;

    void on_quit ()
    {
      this->hide();
    };

  public:
    HelpDialog ( const Glib::ustring &str )
    {
      this->set_title("Help");
      this->resize(320,240);
      this->set_resizable(false);

      lbl  = new Gtk::Label(str);
      btn  = new Gtk::Button(Gtk::Stock::CLOSE);
      vbox = new Gtk::VBox();

      if ( !lbl || !btn || !vbox ){
        std::cerr << "Error: HelpDialog::HelpDialog(): can't initialize widgets.\n";
        return;
      } 

      vbox->pack_start(*lbl);
      vbox->pack_start(*btn);
      vbox->set_border_width(10);
      vbox->set_spacing(10);
      this->add(*vbox);

      btn->signal_clicked().connect( sigc::mem_fun(*this, &HelpDialog::on_quit ) );

      btn->show();
      lbl->show();
      vbox->show();
    };

    ~HelpDialog ()
    {
      delete_widget(lbl);
      delete_widget(btn);
      delete_widget(vbox);
    };

    void set_text ( const Glib::ustring &str )
    {
      lbl->set_markup(str);
    };
};


/////////////////////////////////////////////////////////////////////
// GUIAPP - base class

class GUIApp : public Gtk::Window
{
  private:
    Gtk::Label label;
    bool on_key_press_event ( GdkEventKey *event );

  protected:
    long tgt_win_id_;
    long this_win_id_;
    Glib::ustring window_title;
    Glib::RefPtr<Gtk::Builder> builder;

    Glib::ustring str_help;

    Gtk::VBox *vbox_main;
    Gtk::HButtonBox button_box;
    Gtk::HSeparator *separator;

    Gtk::Button *b_help, *b_quit, *b_save;
    HelpDialog *help_dialog;

    void connect_action_clicked ( Gtk::Button *b, long data, 
                                  long action = ACTION_CONTROL_KEY );
    void connect_action_clicked ( Gtk::ToolButton *b, long data, 
                                  Glib::ustring img_name = "", 
                                  long action = ACTION_CONTROL_KEY );

    bool on_delete_event ( GdkEventAny *event );


  public:
    GUIApp (  long tgt_win_id, 
              char *xml_file = 0, 
              Glib::ustring window_title = "GUIApp - base clas for menus" );
    ~GUIApp ();

    void set_position ();
    void send_event ( long action, long data = 0 );

    void set_help ( Glib::ustring str );
    void show_help ();
    void on_quit ();
};

}; // namespace gctrl
#endif //_GUIAPP_H
