#include "x11.h"

namespace x11 {

bool check_winid ( Display *display, Window window, unsigned long id )
{
  Window  parent, 
          *children;
  unsigned int  n_children,
                i;

  if ( XQueryTree (display, window, &window, &parent, &children, &n_children) != 0 ){
    if (n_children > 0){
      for (i=0; i < n_children; ++i) {
        if (id == children[i] || check_winid(display, children[i], id) > 0) 
          return true;
      } // for n_children
    } // if n_children > 0
  }
  return false;
}


bool send_event ( long win_id, 
                  const char *display_name,
                  const char *atom_name, 
                  long action,   
                  long data )
{
  bool ret = true;
  Display *display = XOpenDisplay(display_name);

  // check whether 'show' window exists
  if ( check_winid( display, RootWindow(display, DefaultScreen(display)), win_id ) ) {
    XClientMessageEvent xevent;
    xevent.type         = ClientMessage;
    xevent.message_type = XInternAtom(display, atom_name , 0);
    xevent.format       = 32;
    xevent.data.l[0]    = action;
    xevent.data.l[1]    = data;
    xevent.data.l[2]    = 0;
    xevent.data.l[3]    = 0;
 
    XSendEvent ( display, win_id, 0, 0, (XEvent *) &xevent );
  }
  else {
    std::stringstream ss;
    ss << "Window " << win_id << " not found.";
    std::cerr << "Error: GUIApp::send_event(): " << ss.str() << std::endl;
    ret = false;
  }

  XCloseDisplay(display);
  return ret;
}


bool get_xwindow_attributes (  unsigned long win_id, 
                               const char *display_name, 
                               xwindow_attributes &attrs )
{
  bool ret = false;
  Display *display = XOpenDisplay(display_name);
  if (x11::check_winid(display,RootWindow(display, DefaultScreen(display)), win_id)){
    if ( XGetWindowAttributes(display, win_id, &attrs) )
      ret = true;
  }
  XCloseDisplay(display);
  return ret;
}

}; // namespace x11
