/* make plot

  number color point       line
  1      W / K full circ   ----
  2      Y   B open circ   -- --
  3      C   R full sq     --.--.
  4      M   G open sq     ......
  5      G   M full diam   - - - -
  6      R   C open diam   -..-..
  7      B   Y full ^      -...-...
  8      brown open ^      .. .. ..
  9      cyan  full v      ... ...
  10     mag   open v      -- - -- -
  11     gray  +           . . . . .
  12     green x           . ... . ...
  13     red   *           . .. . ..
  14     blue  odot        -- - -- -
  15  =1, etc

  7/2009 interface to fit.c extended, some hot keys changes

  1/2007 interface to fit.c - hot key t,T,ctrl-t
    must be called as (example):
      plot 2fit.dat:A:B:o:C :"a+b*A+c*A^2"
    file 2fit.dat must have x in 1st column, y in 2nd column, and
      optional std in 3rd column
    max PARM parms, far from fool-proof
    uses external routine exp2pow to convert ^ to pow or multiplication
    hot key '=' can serve to change parameters sent to minimize
                (see FIT4PLOT/fit.c and gen/minimize.h)

  6/2004 double leftclick emits coordinates and if env TOCLIP is set
    then puts it to the clipboard
  TOCLIP is the command to paste the text (appended 1 arg) to the clipboard
  e.g.:
    "dcop --user jiri klipper klipper setClipboardContents" (for KDE)
    "xclip.sh", where xclip.sh = shell script "echo $* | xclip"
  The result will be called as by system()

  4/2004 environment variables a b c d e f (see PARM) now provide default
           values for parameters
  2/2004 mouse interface changed from drawplus to mousekey

  7/2000 update: plot and plotc unified
  for the old version ("LEVEL=1") see plot0.c
  10/2000: Any redraw-causing event that comes during drawing is ignored.
           This may cause that the graph is not redrawn when a window
           is minimized or eclipsed during long drawing, but
           suppresses unnecessary redrawing on systems that
           occasionally generate several Expose events on start,
           maximize or so.
           drawx11: environment variable LAZYX11 causes waiting
           LAZYX11 seconds after creating a window (otherwise more
           Expose events can come later causing redraws)
environment variable PLOTINIT can contain the inital string to execute
e.g., to make PostScript and quit, use (csh/tcsh)
  setenv PLOTINIT "#q"
or (sh/bash)
  export PLOTINIT="#q"
*/

/* allows changing param a by hotkeys a A etc. */
#include "parm.h"
/* maximum 10 (params=a..j): code has to be changed at several places */

#ifndef PARM
#define PARM 0
#endif

/* PostScript definitions to be placed in file "ps.def"
   all lengths are in pt if not stated otherwise
   values in [] are optional or, if inside <>, defaults
  ! <comment>
  # <comment>
  b B/W mode: replaces color lines/points by different style black
  s <font size [14]> [<default font [Helvetica]>]
  m <mode: Landscape Portrait Encapsulated [L]>
    !!! Portrait swaps xsize <--> ysize if used after command w
  x <left margin [100]> <right margin [5]> <xaxis text []>
  y <bottom margin [100]> <top margin [5]> <yaxis text []>
  w <x window size [576=8*72]> <y window size [432=6*72]>
    (negative = in cm)
    !! command m p (used after w) swaps xsize <--> ysize
  t <line thickness [1]> <frame thickness [0.5]>
  f <approx. # of ticks in x [5]> <# of ticks in y [5]> <tick length [5]>
    negative: abs. tick span
    this option may repeat, but labeling applies to the first f
  # R G B[,<dash>,<linewidth>[,<pointsize>,<point>]
    (set PS color and style, #=1,2,...;
    <dash>=line pause [line pause ...]
    <point>={circle,fullcircle...}
    example: 1 1 0 0,5 3 1 3,0.5,2,opensquare )
    BUG: numbering is wrong (cf. line[16] below), b recommended instead
  r <rotation angle for l and y (y needs EXTPSSTRING)
  l x y STRING : string to write; can use _Index ^Exponent \Symbol
  l x y $#-STRING as above, prepend by a line of style # (=1,2,3..)
  l x y $#.STRING as above, prepend by a symbol of style # (=1,2,3..)
  l x y $#+STRING as above, both keys (line+symbol)
  l x y $#=STRING as above, errbar
        WARNING: in scripts (incl. cat > ps.def <<EOF), protect $ by \$
  l x +[dy] STRING : (next y by dy below)
*/

#include "ground.h"
#include "alloc.h"
#include "mouselib.h"
#include "draw.h"
#include "ploterr.h"
#include "tabinc.c"
#include "mydraw.h"
#include "mousekey.c"

#ifdef X11
#define HIDEMOUSE
#define SHOWMOUSE
#else
#define HIDEMOUSE hideMouse();
#define SHOWMOUSE showMouse();
#endif
#define fint(R) (REAL)(int)(R)
#include "makegrid.c"

#ifdef DOS
#include <graphics.h>
#include "prtscr.c"
#endif

static int color[14]={
  WHITE,YELLOW,LIGHTCYAN,LIGHTMAGENTA,LIGHTGREEN,LIGHTRED,LIGHTBLUE,
  BROWN,CYAN,MAGENTA,LIGHTGRAY,GREEN,RED,BLUE};

#ifdef X11
#include "plotbit.c"
#include "dumpscr.c"
#endif

void roundrg(double *x,double *X,int enlarge) /********************* roundrg */
{
  double dx=*X-*x,q=1;

  if (dx<=0) return;
  while (dx>50) dx/=10,q*=10;
  while (dx<5) dx*=10,q/=10;

  *x=((int)(*x/q+0.5001*(Sign(*x)-enlarge)))*q;
  *X=((int)(*X/q+0.5001*(Sign(*X)+enlarge)))*q;
}

#ifdef DOS
#define ARGLEN 128
char ext[6];

int hasext(char *a) /******************************************* DOS: hasext */
{
  char *c,*e=NULL;

  for (c=a; *c && *c!=':'; c++)
    if (*c=='\\') e=NULL;
    else if (*c=='.') e=c;

  if (e) copy(ext,e,5);
  if ( (c=strchr(ext,':')) ) *c=0;

  return e!=NULL;
}

void inclext(char *fn) /*************************************** DOS: inclext */
{
  char *col=strchr(fn,':');

  if (col) {
    memmove(col+strlen(ext),col,strlen(col)+1);
    copy(col,ext,strlen(ext)); }
  else strcat(fn,ext);
}
#else
#define ARGLEN 102406
#endif

char *strend1(const char *str)
/* returns pointer to the last character (before 0) */
{
  char *c;

  for (c=(char*)str; *c; c++);

  return c-1;
}

char *dupstr(char *tok) /******************************************** dupstr */
{
  char *c=strdup(tok),*e;

  if (!c) {
    fprintf(stderr,"no heap\n");
    exit(0); }

  e=strend1(c);
  if (*e=='\n') *e=0;

  return c;
}

int smartcol(char **pT) /******************************************* smartcol */
/*
  changes column numbers 0,1,...,26 into @,A,...Z
  new: also 27 into c27 (then, allocates a new copy)
  01 changed into A, 1. or +1 etc. not changed (=expression)
  returns whether T started with 0 (turns off hotkeys 1..9)
*/
{
  char *T=*pT;
  int i=atoi(T),r=T[0]=='0';
  char *c;

  for (c=T; *c; c++) if (!isdigit(*c)) return r; /* no digit ==> no change */
  if (i<=26) T[0]=i+'@',T[1]=0;
  else if (r) T[0]='c';
  else {
    alloc(T,strlen(T)+2);
    strcpy(T+1,*pT);
    T[0]='c';
    *pT=T; }

  return r;
}

char *coltok(char *s) /********************************************** coltok */
{
  static char *S;
  char *o;

  if (s) S=s;
  if (!S) return NULL;

  o=S;
  S=strchr(S,':');
  if (S) *S++=0;

  return o;
}

struct lett_s {
  struct lett_s *next;
  float x,y,r;
  char *s; } *lett;

#if PARM
struct parm_s {
  double *parm;
  double step;
  int on; } parm[PARM];
int lastparm=-1;

/* see FIT4PLOT/fit.c for info on parms */
char *FIT4PLOT; /* where the code for fitting is located */
double D=1e-5,eps=1e-8,err=-2;
int method=-2,maxit=10000,nerr=0;
#endif

int percnr=0;
char *perc=NULL;
char *numberedfile(char *template)
{
  if (perc)
    return string(template,percnr);
  else
    return template;
}

int main(int narg,char **arg) /**************************************** main */
{
  int iarg,I,I0=1,advI=1,col,grmode;
  int fixcoly=0; /* see struct f_s */
  int difcol;
  static char colA[]="A",colB[]="B";
  char *colx=colA,*coly=colB;
  static struct rgfn_s {
    struct rgfn_s *next;
    char name[L_tmpnam]; } *rgfn0,*rgfn;
  double x;
  static struct rg_s
    rginit={3e33,-3e33,3e33,-3e33},
    rg    ={3e33,-3e33,3e33,-3e33},
    rgset ={3e33,-3e33,3e33,-3e33},
    rgundo={3e33,-3e33,3e33,-3e33};
  int ix0,ix1,iy0,iy1,lbl=1,lblx=0,lbly=0,lbldy=0,psty=0,lsty=0,fix=0,i;
  struct f_s {
    struct f_s *next;
    char *colx,*coly; /* expressions for x,y columns */
    char *coldy1,*coldy2; /* for error bars */
    int errbar;
    int maxcol;
    int fixcoly;   /* column with leading 0 ==> not changed for hotkeys 1..0 */
    int col;       /* color */
    int lsty,psty,fix; /* linestyle, pointstyle, fixed style=SPACE inactive */
    char *colon;   /* pos of : in fn (if any) */
    char fn[1];    /* filename [var.len] */
    } *head=NULL,*f,*last=NULL;
  char *tok1,*tok2,*tok3,*ch;
  char *tok4,*tok5;
  char fn[ARGLEN], A[ARGLEN];
  int key;
  static FILE *resp;
  static FILE *fig_def;
  int isstdin=0;

  initscroll(0);

  if (narg<2) {
    fprintf(stderr,"\
MACSIMUS command-line formula and file plot, (c) J.Kolafa 96-10.  Call by:\n\
  plot [RANGE] [FILE][COLUMNS] [[FILE][COLUMNS] ..]\n\
where\n\
  RANGE = [FROMX:TOX,FROMY:TOY] or [FROMX:TOX][FROMY:TOY] (incl. [])\n\
    missing bounds (e.g., [0:][0:]) determined from FILEs\n\
  FILE =\n\
    FILENAME : plot file\n\
    -FILENAME : do not plot (right now), but advance colors/plotstyle\n\
    [INT] : (incl. []) as file of INT+1 equidistant points in FROMX:TOX\n\
    [INT:FROMX:TOX] : as above, override range\n\
    - : stdin\n\
    @ : dummy (useful to set X,Y,STYLE)\n\
    @RESPONSEFILE : read arguments from file\n\
  COLUMNS = { :X:Y | :Y | :X:Y:STYLE:DY | :X:Y:STYLE:DYLOW:DYHIGH |\n\
    :X:Y:STYLE:DXLOW:DXHIGH | :X:Y:STYLE:DX:DY (see PLOTERRBAR)}\n\
  X,Y = x and y coordinates to plot: column # or expression\n\
    column 1 of file = 1 = #1 = A = x  (01 = 1 but hotkeys 0..9 not active)\n\
    line # (counts from 0, reset by blank line) = 0 = #0 = @ = n\n\
  STYLE = string of chars:\n\
    LINES -:solid =:thick d:dotted\n\
    DOTS  .:pixel p:small P:medium o:large O:huge\n\
    COLORS advanced in order: W Y M G R B Brown darkC darkM Gray darkG darkB\n\
      C:advance color on/off, c:set W, cc:set Y, ...; 1=set W+advance off...\n\
    s:hotkey SPACE enabled again\n\
Missing FILE,X,Y,STYLE kept from previous arg, default is @:1:2:-\n\
Environment:\n"
#ifdef X11
"  PLOTGEOMETRY = X geometry (default=640x480+1+1)\n"
#endif
"\
  PLOTINIT = initial sequence of keys (use \"#Q\" to generate [E]PS)\n\
  TOCLIP = commant to put data into clipboard (after left click)\n\
  LAZYX11 = waiting time if slow link\n\
  PLOTNAME = window and icon name\n\
  PLOTERRBAR = { yy | xx | xy } error bar style\n\
  PRTSCR = print screen file name (hot keys @M @O etc.)\n\
  FIT4PLOT = path to fitting support\n\
Fitting (data must be in columns 1,2[,3]) example:\n\
  plot FILE :A:B:o:C :\"(a+b*A)/(1+c*A)\"\n\
  then pre-adjust parameters a,b,c by hotkeys aAbBcB/* and fit by t\n\
  highprec fit: T (slow, good for linear problem or increase precision)\n\
  calculate standard errors of parms (w. optional column 3=data stdev):\n\
    set the number of sampling points by Nn, then use t\n\
  max. parm=%c, enter parms/options directly: =\n"
#ifdef DOS
"FILE extension remembered, e.g. \"plot a.dat b c.\" = \"plot a.dat b.dat c.\""
#endif
,PARM+'a' );
  exit(0); }

#ifdef X11
{
  char *name,iconame[24];

  if (!(name=getenv("PLOTNAME"))) {
    int i,l=5;

    loop (i,1,narg) l+=1+strlen(arg[i]);
    alloc(name,l);
    strcpy(name,"plot");
    loop (i,1,narg) {
      strcat(name," ");
      strcat(name,arg[i]); } }

  xwindowhints.winname=name;
  name+=5;
  if (strlen(name)>23) {
    copy(iconame,name,24); iconame[23]=0; }
  else
    strcpy(iconame,name);
  xwindowhints.iconame=iconame;
  xwindowhints.icowidth=plotbit_width;
  xwindowhints.icoheight=plotbit_height;
  xwindowhints.icobits=(char*)plotbit_bits;
  if ((name=getenv("PLOTGEOMETRY"))) xwindowhints.geometry=name;
  Imworking(1);
  if ((name=getenv("PLOTINIT"))) addkbdstring(name);
}
#endif

/* making a list of id's for Calc() */
#if PARM
  FIT4PLOT=getenv("FIT4PLOT");
  if (!FIT4PLOT) {
    char *USER=getenv("USER");
    if (!USER) USER=getenv("USERNAME");
    if (!USER) USER="plot";
    FIT4PLOT=strdup(string("/home/%s/macsimus/c/fit4plot",USER)); }

  loop (i,0,PARM) {
    char *ge;

    alloczero(id,sizeof(struct _Idlist_s)+1);
    id->next=_Id.head;
    _Id.head=id;
    parm[i].parm=&id->val;
    parm[i].step=1;
    parm[i].on=0;
    id->id[0]='a'+i;
    id->val=0;
    if ( (ge=getenv(id->id)) ) id->val=atof(ge); }
#endif

   columnidlist();

  for (iarg=I=0;;) {

    if (resp) do {
      if (!fgets(A,ARGLEN,resp)) {
        fclose(resp);
        resp=NULL;
        goto nextarg; }
      } while (A[0]=='!');
    else {
     nextarg:
      if (++iarg>=narg) break;
      if (arg[iarg][0]=='@' && arg[iarg][1]!=':') {
        resp=fopen(arg[iarg]+1,"rt");
        if (!resp) fprintf(stderr,"no file %s\n", arg[iarg]+1);
        continue; }
      else
        strcpy(A,arg[iarg]); }

    while ( *(ch=strend1(A))<=' ' && ch>=A) *ch=0;

    I+=advI;

    /* optional range: must be 1st arg and must be of form [...] */
    if (iarg==1 && A[0]=='[' && *strend1(A)==']') {
      I0=2;
      ch=A+1;
      if (*ch!=':') rgset.x=atof(ch);
      if ( (ch=strchr(ch,':')) ) {
        ch++;
        if (!strchr("[],",*ch)) rgset.X=atof(ch);
        ch=strchr(tok1=ch,'[');
        if (!ch) ch=strchr(tok1,',');
        if (ch) {
          ch++;
          if (*ch!=':') rgset.y=atof(ch);
          if ( (ch=strchr(ch,':')) ) {
            ch++;
            if (*ch!=']' && *ch) rgset.Y=atof(ch); } } } }

    else {
      /* filename */
      if (A[0]==':' && last!=NULL) {
        strcpy(fn,last->fn);
        if ( (tok1=strchr(fn,':')) ) *tok1=0;
        strcat(fn,A); }
      else if (A[0]=='[') {
        /* # of intervals */
        static int n=1000;
        static double r=3e33,R=-3e33;
        FILE *rgf;

        ch=A+1;
        if ( !(tok1=strchr(ch,']')) ) {
          fprintf(stderr,"no matching ]\n");
          goto shut; }
        if (!strchr(":],",*ch)) n=atoi(ch);
        if ( (ch=strchr(ch,':')) && ch<tok1) {
          ch++;
          if (!strchr(":],",*ch)) r=atof(ch);
          if ( (ch=strchr(ch,':')) && ch<tok1) {
            ch++;
            if (!strchr(":],",*ch)) R=atof(ch); } }

        if (r>2e33) r=rgset.x;
        if (R<-2e33) R=rgset.X;

        if (r>2e33) r=rg.x;
        if (R<-2e33) R=rg.X;

        if (r>2e33 || R<-2e33 || n<=0 || n>1000000) {
          fprintf(stderr,"no range or bad # of points\n");
          goto shut; }

        alloc(rgfn,sizeof(*rgfn));
        tmpnam(rgfn->name);
        rgfn->next=rgfn0;
        rgfn0=rgfn;
        rgf=fopen(rgfn->name,"wt");
        loopto (i,0,n)
          fprintf(rgf,"%.20g\n", i*(R-r)/n+r);
        fclose(rgf);
        strcpy(fn,rgfn->name);
        tok1++;
        if (tok1[0]!=':') strcat(fn,":");
        strcat(fn,tok1); }
      else
#ifdef DOS
        if (!hasext(A)) {
          strcpy(fn,A);
          inclext(fn); }
        else
#endif
          strcpy(fn,A);

      if (last || !strchr(":@",fn[0]))
        alloczero(f,sizeof(struct f_s)+strlen(fn));
      else
        f=NULL,I--;
      if (f) strcpy(f->fn,fn);

      if (fn[0]=='-' && (fn[1]==0 || fn[1]==':') && !isstdin) {
        /* make copy of stdin: not good for DOS */
        FILE *ff;
        int ch;

        isstdin++;
        ff=fopen("-","wt");

        while ( (ch=getchar())>=0 ) fputc(ch,ff);
        fclose(ff); }

      if ( (perc=strchr(fn,'%')) && strchr("0123456789-+",perc[1]))
        percnr=1;

      tok1=coltok(fn);
      tok2=tok3=tok4=tok5=NULL;
      tok1=coltok(NULL);
      if (tok1) {
        tok2=coltok(NULL);
        if (tok2) {
          tok3=coltok(NULL);
          if (tok3) {
            tok4=coltok(NULL);
            if (tok4)
              tok5=coltok(NULL); } }
        else
          tok2=tok1,tok1=""; /* :XXX denotes y-axis */ }

      if (tok1 && *tok1) {
        smartcol(&tok1);
        to_colon(tok1);
        colx=dupstr(tok1); }

      if (tok2 && *tok2) {
        fixcoly=smartcol(&tok2);
        to_colon(tok2);
        coly=dupstr(tok2); }

      if (tok3 && *tok3) {
        int II=I0;

        psty=0; lsty=-1; fix=1;
        for (ch=tok3; *ch; ch++) switch (*ch) {
          case 's': fix=0; break;
          case 'c': I=II++; break;
          case 'C': advI^=1; break;
          case '.': psty=-1; break;
          case 'p': psty=-2; break;
          case 'P': psty=-3; break;
          case 'o': psty=-4; break;
          case 'O': psty=-6; break;
          case '-': lsty=0; break;
          case '=': lsty=2; break;
          case 'd': lsty=1; break;
          default:
            if (isdigit(*ch)) {
              advI=0;
              II=atoi(ch);
              if (II>9) ch++;
              I=I0-1+II++; }
            else fprintf(stderr,"%c bad key\n",*ch); } }

      if (tok4 && *tok4) {
        smartcol(&tok4);
        to_colon(tok4); }

      if (tok5 && *tok5) {
        smartcol(&tok5);
        to_colon(tok5); }

      if (f) {
        f->colx=colx;
        f->coly=coly;
        f->errbar=0;
        if (tok4) {
          f->errbar++;
          /* file:A:B:-: is the same as file:A:B:-:C; only until Y */
          if (strlen(tok4)==0 && tok2[0]>='@' && tok2[0]<'Z') {
            f->coldy1=strdup(tok2);
            f->coldy1[0]++; }
          else
            f->coldy1=strdup(tok4);
          if (tok5) f->coldy2=strdup(tok5);
          else f->coldy2=f->coldy1; }
        f->fixcoly=fixcoly;
        f->psty=psty;
        f->lsty=lsty;
        f->fix=fix;
        f->col=color[(I-I0)%14];

        f->maxcol=max_column;
 //       f->maxcol=0;
 //       loopto (cc,'@','Z')
 //         if (strchr(f->colx,cc) || strchr(f->coly,cc)) f->maxcol=cc-'@';

        if (!minmaxfile(numberedfile(fn),1,colx,coly,f->maxcol, &rg.x,&rg.X,&rg.y,&rg.Y)) {
          if (head) last->next=f,last=f;
          else head=last=f;
          f->next=NULL;
          if (rgset.x< 2e33) rg.x=rgset.x;
          if (rgset.y< 2e33) rg.y=rgset.y;
          if (rgset.X>-2e33) rg.X=rgset.X;
          if (rgset.Y>-2e33) rg.Y=rgset.Y; }
        rginit=rg;

        f->colon=strchr(f->fn,':');
        if (f->colon) f->colon[0]=0; }
      }
    } /* iarg & @resp-file */

#ifdef DEBUG
  fprintf(stderr,"rgx=%.9g %.9g  rgy=%.9g %.9g\n",rg.x,rg.X,rg.y,rg.Y);
#endif

#if 1
  if (rg.x>=rg.X || rg.y>=rg.Y) {
    fprintf(stderr,"rgx=%.9g %.9g  rgy=%.9g %.9g\n",rg.x,rg.X,rg.y,rg.Y);
    fprintf(stderr,"not enough points (in given ranges)\n");
    goto shut; }
#endif

  psty=lsty=0;

  grmode=startgraph(isstdin?9:-9);
  if (grmode<0) {
    fprintf(stderr,"%d\n",grmode);
    exit(-1); }

 notify:

  lwindow(lbl?48:4,getmaxx()-4,4,getmaxy()-(lbl?11:4));

  for (;;) { /* main loop */

#ifdef DEBUG
    fprintf(stderr,"drawing...\n");
#endif

    if (rg.x>=rg.X) rg.x-=1,rg.X+=1;
    if (rg.y>=rg.Y) rg.y-=1,rg.Y+=1;

    scale(rg.x,rg.X,rg.y,rg.Y);
    /* bad implementation of identical x,y scaling: */
    if (getenv("PLOTSCALE")) scaling.x=scaling.y=fmin(scaling.x,scaling.y);
    clearviewport();
    setwritemode(0);
    setfillstyle(1,BLACK);
    makegrid(lbl,grmode==7?2.4:4, 1,rg.x,rg.X,rg.y,rg.Y,getmaxy()-8);
    makegrid(lbl,grmode==7?2.4:4, 0,rg.y,rg.Y,rg.x,rg.X,0);

    if (my_style.ps && lbl) myclip(); /* contains gsave ! */

    for (f=head; f; f=f->next) {
      if (grmode==7) f->col=WHITE;

      if (f->errbar)
        drawerrbar(numberedfile(f->fn),
                   f->colx,f->coly,f->coldy1,f->coldy2,f->maxcol,
                   f->col,f->psty<0?f->psty:psty<0?psty:-6,2);

      if (f->fix) {
        if (f->lsty>=0)
          drawfile(numberedfile(f->fn),
            f->colx,f->coly,f->maxcol, f->col,f->lsty);
        if (f->psty<0)
          drawfile(numberedfile(f->fn),
            f->colx,f->coly,f->maxcol, f->col,f->psty); }
      else {
        if (psty<0)
          drawfile(numberedfile(f->fn),f->colx,f->coly,f->maxcol, f->col,psty);
        if (lsty>=0)
          drawfile(numberedfile(f->fn),f->colx,f->coly,f->maxcol, f->col,lsty);
        if (lsty<0 && psty>=0)
          drawfile(numberedfile(f->fn), f->colx,f->coly,f->maxcol, f->col,0); }
      setlinestyle(0,0,1); /*???*/ }

    if (lbldy) {
#if PARM
      char as[32],*env;
      static char name[8]="aname";
      int ii;
      static int maxparm;

      Max(maxparm,lastparm)

      if (lastparm>=0 && lastparm<PARM) loopto (ii,0,maxparm) {
        name[0]='a'+ii;
        if ( (env=getenv(name)) )
          sprintf(as,"%c: %s = %g",'a'+ii,env,*parm[ii].parm);
        else
          sprintf(as,"%c=%.12g",'a'+ii,*parm[ii].parm);
        mysetcolor(3);
        bar(lblx-2,lbly-lbldy*(maxparm+1-ii)-2,lblx+128,lbly-lbldy*(maxparm+1-ii)+9);
        mysetcolor(15);
        outtextxy(lblx,lbly-lbldy*(maxparm+1-ii),as); }
#endif

      for (col=0,f=head; f; col++,f=f->next) {
        mysetcolor(f->col);
        if (f->colon) f->colon[0]=':';
        bar(lblx-2,lbly+col*lbldy-2,lblx+8*strlen(numberedfile(f->fn)),lbly+col*lbldy+9);
        outtextxy(lblx,lbly+col*lbldy,numberedfile(f->fn));
        if (my_style.ps)
          fprintf(my_style.ps,"newpath %.2f %.2f moveto (%s) show\n",
                  lblx/(double)getmaxx()*my_style.xsize,
                  (1-lbly/(double)getmaxy())*my_style.ysize-(col+1)*my_style.fontsize*1.15,
                  numberedfile(f->fn));
        if (f->colon) f->colon[0]=0; } }

    SHOWMOUSE

    if (fig_dat) {
      if (fig_def) {
        char line[ARGLEN];
        while (fgets(line,ARGLEN,fig_def)) fputs(line,fig_dat);
        fclose(fig_def); fig_def=NULL; }
      fclose(fig_dat);
      fig_dat=NULL; }

    if (my_style.ps) {
      struct lett_s *l;
      int it;

      myup();
      if (lbl) fprintf(my_style.ps,"grestore\n");
      loop (it,0,my_style.nticks) {
        myaxis(it,1,1);
        myaxis(it,1,-1);
        myaxis(it,-1,1);
        myaxis(it,-1,-1); }

      for (l=lett; l; l=l->next) {
        fprintf(my_style.ps,"\n%%----- %s -----\n",l->s);
        if (l->r)
          fprintf(my_style.ps,
                  "gsave\n%.2f %.2f translate %.2f rotate\n0 0 moveto\n",
                  l->x,l->y,l->r);
        else
          fprintf(my_style.ps,"%.2f %.2f moveto\n",l->x,l->y);
        mypssetxy(l->x,l->y,color[(atoi(l->s+1)-1)%14]);
        psstring(l->s,-1);
        if (l->r) fprintf(my_style.ps,"grestore\n"); }

      fprintf(my_style.ps,"grestore\n");
      mycloseplotps();
      my_style.ps=NULL; }

#ifdef DEBUG
    fprintf(stderr,"plot: drawn\n");
#endif

/*.....   do { key=extch(); } while (key==X_REDRAW_ME);*/

    /* all extra X_REDRAW_ME (typically Expose) coming just now are ignored */
    /* This is not the best solution :-( */
    for (;;)
      if (xkbhit(1)) {
        if ( (key=extch())==X_REDRAW_ME) {
          continue; }
        else
          break; }
      else {
        key=extch();
        break; }

#ifdef DEBUG
    fprintf(stderr,"plot: switch key=%d\n",key);
#endif

    if (key=='u')
      rg=rgundo;
    else {
      rgundo=rg;

      switch (key) {
        case '1'|128: /* F1 */
        case '?': {
          static char *help[]={
"=== HOT KEYS ===",
"+ -       zoom in, zoom out",
"u ^z  U k undo last zoom/all (also k)",
"x X y Y   zoom x or y only",
"^x ^y     make 0 center in x or y",
"r R ^r    redraw, Round up, round",
"cursors   move viewpoint",
"PgUp PgDn place file info",
"Del       no file info",
": l L | \\ line style",
". p P     point size",
"z Z       grid+label style (also o O)",
"SPACE     style toggle (if not spec.)",
"!         make fig.dat (use fig.def)",
"@o @O     prtscr: postscript, On white bg",
#ifdef DOS
"@l @e @p  prtscr: Laser Epson9 ePson24",
#endif
"@m @M       ppM, M=on white background",
"#         make plot.[e]ps (use ps.def)",
"1 2.. [ ] change column Y, also 010,011,..",
#if PARM
"=         set/calc parms from prompt",
"a A b B.. change parms a..j; ^a ^b to zero",
"/ *       change step for changing parms",
"t T ^t    fit long/high prec (slow)/double",
"n N       # of sampling pts for fit errors",
#endif
"{ }       cycle files (if format in name)",
"q Esc K   quit, quit, Kill all plots",
"=== MOUSE ===",
"click left button: print x y dx dy (digitize)",
"click right button: place file info",
"click mid button: quit",
"drag left button: rectangle to rescale",
"drag mid/right button: move/scale",
"Press any key...",
NULL};
          int i;
          setfillstyle(1,grmode==7?BLACK:DARKGRAY);
          bar(0,0,340,sizeof(help)/4*11-4);
          setcolor(WHITE);
          for (i=0; help[i]; i++) outtextxy(2,2+i*(11-(grmode==7)*3),help[i]);
          if (extch()!=27) break; }

#ifdef DOS
        case '@': HIDEMOUSE PrtScr(extch()); SHOWMOUSE break;
#endif

#ifdef X11
        case '@': DumpScr(extch(),""); break;
        case X_REDRAW_ME: /* event forcing redraw */
#ifdef DEBUG
          fprintf(stderr,"plot: redraw requested\n");
#endif
/*.....         delay(50);*/
/*.....         while (mykbhit()==X_REDRAW_ME) (void)extch();*/
          goto notify;
        case X_UNKNOWN_KEY: break; /* some more non-alpha keys */
        case 'K': {
/*.....         system("Kill plot");*/
          system("killall plot");
         /* .. this is dirty, because the caller is supposed to be killed */
          }
#endif

        /*** MOUSE ***/
        case LEFTCLICK:
#if X11
          { /* print x,y [dx,dy] */
            char clipit[128],*name;
            double x=iSX(dosmouse.x),y=iSY(dosmouse.y);
            static double lastx,lasty,col=LIGHTCYAN;

            printf("%14.10g %14.10g %14.10g %14.10g\n",x,y,x-lastx,y-lasty);
            lastx=x; lasty=y;
            if (name=getenv("TOCLIP")) {
 	      sprintf(clipit,"%s \'%.9g %.9g\'",name,x,y);
 	      system(clipit); }
            sprintf(clipit,"%.9g %.9g",x,y);    
            if (col==LIGHTCYAN) col=YELLOW; else col=LIGHTCYAN;
            setcolor(col);
            line(dosmouse.x-8,dosmouse.y,dosmouse.x+8,dosmouse.y);
            line(dosmouse.x,dosmouse.y-8,dosmouse.x,dosmouse.y+8);
            XFlush(display);
            usleep(200000);
            break;
          }
#endif
        case MIDCLICK:
        case 'q': case 'Q':
          //        case 3:
        case 27: goto end;

        case RIGHTCLICK: /* show info */
          lbly=dosmouse.y;
          lblx=dosmouse.x;
          lbldy=12;
          break;

        case MIDDRAG: {
          double q=mousedx/scaling.x;

          rg.x-=q; rg.X-=q;
          q=mousedy/scaling.y;
          rg.y+=q; rg.Y+=q; }
          break;

        case LEFTDRAG:
          setcolor(WHITE);
          setwritemode(1);
          ix0=ix1=dosmouse.x;
          iy0=iy1=dosmouse.y;
          while (xkbhit(1)) {
            extch();
            HIDEMOUSE
            rectangle(ix0,iy0,ix1,iy1);
            rectangle(ix0,iy0,ix1=dosmouse.x,iy1=dosmouse.y);
            SHOWMOUSE }

          if (ix0==ix1 && iy0==iy1) {
            lblx=ix0; lbly=iy0; lbldy=12; }
          else if (ix0!=ix1 && iy0!=iy1) {
            rg.x=iSX(ix0); rg.X=iSX(ix1);
            if (rg.x>rg.X) x=rg.x, rg.x=rg.X, rg.X=x;

            rg.y=iSY(iy0); rg.Y=iSY(iy1);
            if (rg.y>rg.Y) x=rg.y, rg.y=rg.Y, rg.Y=x; }
          break;

        case RIGHTDRAG: {
          double x,y;

          x=(rg.X-rg.x)*mousedx*(2./getmaxx()); rg.X-=x; rg.x+=x;
          y=(rg.Y-rg.y)*mousedy*(2./getmaxy()); rg.Y+=y; rg.y-=y; }
          break;

        case 'k': case 'Z'&31:
        case 'U': rg=rginit; break;
        case 'u': rg=rgundo; break;
        case '.': psty=lsty=-1; break;

#ifdef PARM
        case 'N': if (nerr==0) nerr=50; else nerr+=12+nerr*2;
        case 'n': if (nerr>5) nerr/=2; else nerr=0;
          prt("nerr=%d (# of samples to calc. stderr of fit) will be used by hot keys t crtl-t T",nerr);
 	  break;
        case 'T'&31:
        case 't':
        case 'T': {
	  FILE *sh=fopen("plotfit.sh","wt");
	  int i,np=0,prec=(key=='t')+2*(key=='T');
	  char *c,*precname[3]={"double","long","high"};
	  static int lastnp[3]={-1,-1,-1};

	  if (!head->next) break;

	  fprintf(sh,"#!/bin/bash\n");
	  fprintf(sh,"echo method=%d par=1 D=%g eps=%g maxit=%d nerr=%d err=%g > fit.get\n",
		 method==-2?(key==2?2:0):method,D,eps,
		 method==-2?(key==2?10:5000):maxit,nerr,err);

	  /* determine np */
	  for (c=head->next->coly; *c; c++)
	    if (*c>='a' && *c<='`'+PARM) {
	      /* to extend if _ used */
	      if (isalnum(c[1])) continue; /* part of longer id */
	      if (c>head->next->coly) if (isalnum(c[-1]) || c[-1]=='.') continue; /* part of exp number */
	      Max(np,*c-'`') }

 	  loop (i,0,np)
	    fprintf(sh,"echo \"A[%d]=%.15g\" >> fit.get\n",i,*parm[i].parm);
	  fprintf(sh,"echo \"; method=9;\" >> fit.get\n");
	  fprintf(sh,"cd %s\n",FIT4PLOT);
	  if (np!=lastnp[prec]) {
	    fprintf(sh,"echo \"#define N %d\" > n.h\n",np);
	    fprintf(sh,"rm f.c\ntouch f.c\n");
	    loop (i,0,np) fprintf(sh,"echo \"#define %c (P[%d])\" >> f.c\n",i+'a',i);
	    fprintf(sh,"exp2pow \"%s\" >> f.c\n",head->next->coly);
	    loop (i,0,np) fprintf(sh,"echo \"#undef %c\" >> f.c\n",i+'a');
	    fprintf(sh,"cp makefile.%s makefile\nmake fit.%s\n",
	      precname[prec],precname[prec]);
	    lastnp[prec]=np; }
	  fprintf(sh,"cd -\n");
	  fprintf(sh,"%s/fit.%s %s < fit.get\n",FIT4PLOT, precname[prec], head->fn);
 	  fclose(sh);
	  system("bash plotfit.sh");
	  sh=fopen("fit.fit","rt");
	  if (sh) {
	    char line[32];

	    loop (i,0,np) {
	      fgets(line,32,sh);
	      sscanf(line+4,"%lf",parm[i].parm); }
	    fclose(sh); }
	  break; }
#endif

        /* POSTSCRIPT */
        case 'P'&31:
        case '#': my_rg=rg; {
          FILE *psdat=fopen("ps.def","rt");
          char line[ARGLEN];
          float rot=0;
          int istick=0;

          if (psdat) while (fgets(line,ARGLEN,psdat)) {
            double x=atof(line+2),y=0;
            char *c=strchr(line+3,' '),*cc=NULL,*parm2=NULL;
            int relx=line[2]=='+',rely=0;

            if (c) {
              c++;
              if (*c=='+') rely=1,c++;
              y=atof(parm2=c);
              cc=strchr(c,' ');
              if (cc) cc++; }

            if (isdigit(line[0])) {
              int col=color[atoi(line)%14];

              strcpy(strend1(line),",,,,,");

              cc=strchr(c=line+2,','); *cc++=0;
              my_style.line[col].rgb=dupstr(*c?c:"0 0 0");
              c=strchr(cc,','); *c++=0;
              my_style.line[col].dash=dupstr(cc);
              cc=strchr(c,','); *cc++=0;
              if (*c) my_style.line[col].thick=atof(c);
              c=strchr(cc,','); *c++=0;
              if (*c) my_style.line[col].size=atof(cc);
              cc=strchr(c,','); *cc++=0;
              my_style.line[col].symbol=dupstr(*c?c:"fullcircle"); }

            else switch (toupper(line[0])) {
              case 'B': {
                static  struct my_style_line_s line[16] = {
     /*-*/    { "1 1 1", "", -1,"fullcircle",-1},
     /*14*/   { "0 0 0", "3 3 3 6", -1,"dotcircle",-1},
     /*12*/   { "0 0 0", "1 2 1 4 1 4 1 2", -1,"cross",-1},
     /*9*/    { "0 0 0", "1 2 1 2 1 4", -1,"fulldown",-1},
     /*13*/   { "0 0 0", "1 2 1 4 1 4", -1,"star",-1},
     /*10*/   { "0 0 0", "6 4 3 4", -1,"opendown",-1},
     /*8*/    { "0 0 0", "1 2 1 4", -1,"openup",-1},
     /*11*/   { "0 0 0", "1 4", -1,"plus",-1},
     /*-*/    { "0 0 0", "", -1,"fullcircle",-1},
     /*7*/    { "0 0 0", "6 3 1 3 1 3 1 3", -1,"fullup",-1},
     /*5*/    { "0 0 0", "3 4", -1,"fulldiamond",-1},
     /*3*/    { "0 0 0", "6 3 1 3", -1,"fullsquare",-1},
     /*6*/    { "0 0 0", "6 3 1 3 1 3", -1,"opendiamond",-1},
     /*4*/    { "0 0 0", "1 2", -1,"opensquare",-1},
     /*2*/    { "0 0 0", "6 4", -1,"opencircle",-1},
     /*1*/    { "0 0 0", "", -1,"fullcircle",-1} };
                copy(my_style.line,line,sizeof(line)); }
                break;
              case 'R': rot=x; break;
              case 'S':
                my_style.fontsize=x;
                if (parm2 && *parm2) {
                  if (cc) { cc[-1]=0; my_style.enc=atoi(cc); }
                  my_style.font=strdup(parm2); }
                break;
              case 'L': {
                struct lett_s *l;
                static double lastx,lasty;

                allocone(l);
                l->next=lett;
                lett=l;
                l->x=x;
                if (relx) l->x=lastx+=x;
                else lastx=l->x=x;
                if (rely) {
                  if (!y) y=my_style.fontsize*1.5;
                  l->y=lasty-=y; }
                else
                  lasty=l->y=y;
                l->r=rot;
                l->s=dupstr(cc); }
                break;
              case 'X': my_style.bx=x;
                        if (c) my_style.bX=y;
                        if (cc) my_style.xlabel=dupstr(cc);
                        break;
              case 'Y': my_style.by=x;
                        if (c) my_style.bY=y;
                        my_style.yrot=rot;
                        rot=0;
                        if (cc) my_style.ylabel=dupstr(cc); break;
              case 'W': if (x<0) x*=-72/2.54; my_style.xsize=x;
                        if (y<0) y*=-72/2.54; my_style.ysize=y;
                        break;
              case 'M': my_style.mode=toupper(line[2]);
                        if (my_style.mode=='P') {
                          x=my_style.xsize;
                          my_style.xsize=my_style.ysize;
                          my_style.ysize=x; }
                        break;
              case 'T': my_style.thick=x;
                        if (c) my_style.framethick=y;
                        break;
              case 'F': if (!istick++) my_style.nticks=0;
                        if (my_style.nticks<5) {
                          my_style.xticks[my_style.nticks]=my_style.yticks[my_style.nticks]=x;
                          if (c) {
                            my_style.yticks[my_style.nticks]=y;
                          if (cc) my_style.tick[my_style.nticks]=atof(cc); } }
                        my_style.nticks++;
                        break; } } }

            myopenplotps(my_style.mode=='E'?"plot.eps":"plot.ps",
 			arg[1][0]=='['?arg[2]:arg[1]);
            break;

        case 'p': if (psty<0) psty++; break;
        case 'P': psty--; break;
        case 'l': if (lsty<4) lsty++; break;
        case 'L': if (lsty>=0) lsty--; break;
        case ' ': if (lsty==-1) lsty=0,psty=-4;
                  else if (psty<0) psty=lsty=0;
                  else lsty=-1,psty=-3; break;
        case '=':
#if PARM
          {
            double a=*parm[0].parm,b=*parm[1].parm,c=*parm[2].parm,d=*parm[3].parm,e=*parm[4].parm,f=*parm[5].parm,g=*parm[6].parm,h=*parm[7].parm,i=*parm[8].parm,j=*parm[9].parm;
            static double x;
            struct _Idlist_s *oldhead=_Id.head;
            _Id.head=NULL;

            prt("enter:\n\
  a..j = parameters (equivalent to A[0]..A[9])\n\
  x = auxiliary variable\n\
  method = -2: default (amoeba for t, Newton-Raphson for T)\n\
           -1: greedy, 0: amoeba, 1: conj.grad 2: Newton-Raphson 3: MC\n\
  eps = precision\n\
  D = step for numerical derivatives\n\
  maxit = minimization steps (ignored for method=-2)\n\
  nerr = to calculate errors of parameters by sampling (cf. n N)\n\
  err = -1: uses stdev of data provided (assuming 1 if not provided)\n\
        -2: as err=-1 if stdev of data available, err=0 otherwise (default)\n\
        -3: stdev of dy_i is proportional to 3rd column,\n\
            rescaled by stdev calculated from residual sum of squares\n\
            (for data with known weights~1/dy_i^2 but unknown precision)\n\
       >=0: uses stdev calc. from resid. sum sq., multiplied by y^err:\n\
         0: the same abs.err. (of stdev above) of all data\n\
         1: the same rel.err. of all data\n\
       0.5: error is proportional to y^0.5 (suited for histogram data)\n\
  end data by ;");
            getdata
	      get(x)
              get(a) get(b) get(c) get(d) get(e) get(f) get(g) get(h) get(i) get(j)
              get(method) get(maxit) get(D) get(eps) get(nerr) get(err)
              /* no get(par) because given by t/T */
              checkdata enddata
            *parm[0].parm=a,*parm[1].parm=b,*parm[2].parm=c,*parm[3].parm=d,*parm[4].parm=e,*parm[5].parm=f,*parm[6].parm=g,*parm[7].parm=h,*parm[8].parm=i,*parm[9].parm=j;
            _Id.head=oldhead;
          }
#endif
        case '\\': case '|': psty=lsty=0; break;
        case ':': case ';': psty=0; lsty=1; break;
        case 'z': case 'o': lbl=(lbl+1)%3; break;
        case 'Z': case 'O': lbl=(lbl+2)%3; break;

        case 18: /* ^R */
          key=0;
          goto round;

        case 'R': key=1;
      round:
          roundrg(&rg.x,&rg.X,key);
          roundrg(&rg.y,&rg.Y,key);
        case 12: /* ^L */
        case 'r': break;

        case '!': {
          char line[ARGLEN];
          int rflag=0;

          if ( (fig_def=fopen("fig.def","rt")) ) {
            fgets(line,ARGLEN,fig_def);
            if ( (ch=strpbrk(line,"\t\n ")) ) *ch=0; }
          fig_dat=fopen(fig_def?line:"fig.dat","wt");
          if (!fig_dat) { fputc('\n',stderr); break; }

          fprintf(fig_dat,"#");
          loop (iarg,0,narg) fprintf(fig_dat," %s",arg[iarg]);
          fprintf(fig_dat, "\n# Win Xlbl Ylbl Frame Lwrite Placewr Size Thick/pt Over -draw  up .point Col H\n");
          if (fig_def) do {
            if (fgets(line,ARGLEN,fig_def)) {
              if (line[0]=='.') break;
              fputs(line,fig_dat);
              if (line[0]=='r') rflag++; }
            else {
              fclose(fig_def); fig_def=NULL; }
            } while (fig_def);
          if (!rflag)
            fprintf(fig_dat,"r %.16g %.16g %.16g %.16g\n",rg.x,rg.X,rg.y,rg.Y);
          fprintf(fig_dat, ". end of header\n");
          break; }

#if PARM
        case '/': if (lastparm>=0 && lastparm<PARM) parm[lastparm].step*=.5;
                  continue;
        case '*': if (lastparm>=0 && lastparm<PARM) parm[lastparm].step*=2;
                  continue;
#endif

        case '+': x=(rg.X-rg.x)*0.2; rg.X-=x; rg.x+=x;
                  x=(rg.Y-rg.y)*0.2; rg.Y-=x; rg.y+=x; break;
        case '-': x=(rg.X-rg.x)*0.5; rg.X+=x; rg.x-=x;
                  x=(rg.Y-rg.y)*0.5; rg.Y+=x; rg.y-=x; break;

        case 'X': x=(rg.X-rg.x)*0.2; rg.X-=x; rg.x+=x; break;
        case 'x': x=(rg.X-rg.x)*0.5; rg.X+=x; rg.x-=x; break;
        case 'x'&31:
          rg.x=fabs(rg.x); rg.X=fabs(rg.X);
          if (rg.X>rg.x) rg.x=-rg.X; else rg.X=rg.x,rg.x=-rg.x;
          break;

        case WHEELFORWARD:
        case 'Y': x=(rg.Y-rg.y)*0.2; rg.Y-=x; rg.y+=x; break;
        case WHEELBACKWARD:
        case 'y': x=(rg.Y-rg.y)*0.5; rg.Y+=x; rg.y-=x; break;
        case 'y'&31:
          rg.y=fabs(rg.y); rg.Y=fabs(rg.Y);
          if (rg.Y>rg.y) rg.y=-rg.Y; else rg.Y=rg.y,rg.y=-rg.y;
          break;

        case ('M'|128): /* RIGHT cursor */
          x=(rg.X-rg.x)*0.1; rg.X+=x; rg.x+=x; break;
        case ('K'|128): /* LEFT cursor */
          x=(rg.X-rg.x)*-0.0618; rg.X+=x; rg.x+=x; break;
        case ('H'|128): /* UP cursor */
          x=(rg.Y-rg.y)*0.1; rg.Y+=x; rg.y+=x; break;
        case ('P'|128): /* DOWN cursor */
          x=(rg.Y-rg.y)*-0.0618; rg.Y+=x; rg.y+=x; break;


        case 23: /* ^W = PGUP */
        case ('I'|128): lbly=20; lbldy=12; goto updn;
        case 22: /* ^V =PGDN */
        case ('Q'|128): lbly=getmaxy()-20; lbldy=-12;
      updn: lblx=21+lblx%(getmaxx()-100);
            break;
        case 4: case 8: case 127: case ('S'|128): /* DEL, backspace */
          lbldy=0; break;

        case '{': if (perc) percnr--; break;
        case '}': if (perc) percnr++; break;

        case '[': difcol=-1; goto plus;
        case ']': difcol=1;
      plus:
          for (f=head; f; f=f->next)
            if (!f->fixcoly) f->coly[0]+=difcol;
          break;

        default: if (isdigit(key)) {
          static char xxx[2];

          if (key=='0') xxx[0]=extch()*10,xxx[0]+=extch()+('@'-11*'0');
          else xxx[0]=key+'@'-'0';

          for (f=head; f; f=f->next)
            if (!f->fixcoly) f->coly=xxx; }
#if PARM
       else if (key>=('A'&31) && key<('A'&31)+PARM) {
         lastparm=key-('A'&31);
         *parm[lastparm].parm=0;
         parm[lastparm].on=1; }
       else if (key>='a' && key<'a'+PARM) {
         lastparm=key-'a';
         *parm[lastparm].parm-=parm[lastparm].step;
         parm[lastparm].on=1; }
       else if (key>='A' && key<'A'+PARM) {
         lastparm=key-'A';
         *parm[lastparm].parm+=parm[lastparm].step;
         parm[lastparm].on=1; }
#endif
        } }
    HIDEMOUSE
    }

 end:
  closegraph();
  {
    char *c=getenv("LAZYX11");
    if (!c || *c!='-') {
      prt("x-range = [%.9g %.9g] = %.9g",rg.x,rg.X,rg.X-rg.x);
      prt("y-range = [%.9g %.9g] = %.9g",rg.y,rg.Y,rg.Y-rg.y);
#if PARM
      loop (i,0,PARM) if (parm[i].on) prt("%c=%.12g",i+'a',*parm[i].parm);
#endif
  } }

 shut:
  if (isstdin) unlink("-");
  for (rgfn=rgfn0; rgfn; rgfn=rgfn->next) remove(rgfn->name);

  return 0;
}
