static int indx;

static char *next(void) /********************************************* next */
{
  char *c=strtok(NULL," \t\n");

  if (!c) ERROR(("bad line %d",indx))

  return c;
}

void readMOL(char *fn) /******************************************* readMOL */
/***
  reads *.mol file in the MEDIT format.
  fn must have extension .mol
***/
{
  char *tok,*c;
  static char resnm[4];
  int i,j,nnbr,inbr,nbr,ic,testnc=0,ndonor=0,nacceptor=0;
  struct collist_s *cl;

  file=fopen(fn,"rt");
  if (!file) {
    prt("WARNING: %s: no mol-file\n\
*** the configuration will be shown with dots (x) or white spheres\n\
*** remedy hints: molcfg, cook -y# (will call molcfg), bonds",fn);
    ns=1;
    allocarrayzero(site,ns);
    site[0].id="sphere";
    site[0].type="X";
    site[0].molcol='H';
    site[0].r=50; /* diameter = 1 A */
    nbond=1;
    if (option('i')<=0) nbond=1;
    else nbond=option('i');
    allocarray(bond,nbond);
    bond[0][0]=bond[0][1]=0;

    nc=testnc=1; }

  else while (mygetline()) if ( (tok=strtok(li," =\t\n")) ) {

    if (!strcmp(tok,"parameter_set")) {
      tok=strtok(NULL," =\t\n");
      parameter_set=strdup(tok); }
    else if (!strcmp(tok,"number_of_atoms")) {
      tok=strtok(NULL," =\t\n");
      ns=atoi(tok);
      prt("! number_of_atoms = %i",ns);
      allocarrayzero(site,ns); /* must :=0 - used later */ }

    else if (!strcmp(tok,"atoms")) {
      if (ns<=0) ERROR(("%s: undefined or bad number_of_atoms",fn))

      if (option('i')<=0) nbond=ns/8+8; /* first guess: will grow on demand */
      else nbond=option('i');
      allocarray(bond,nbond);

      loop (i,0,ns) {
        if (!mygetline()) ERROR(("%s: missing lines",fn))
        tok=li;
        site[i].mark=0;
        if (*tok=='*') { tok++; site[i].mark=4; }
        indx=atoi(strtok(tok," \t"));
        if (i!=indx) ERROR(("%s: atoms are not contiguously numbered, consider command:\n\
***   mol2mol NAME +a NEWNAME",fn))

        tok=next();
        site[i].id=strdup(tok);
        if (strlen(tok)>3 && memcmp(resnm,tok,3)) {
          memcpy(resnm,tok,3); resnm[3]=0;
          printf("\r%d:%s",i,resnm); }

        /* id: mark if connect to backbone */
        if (abs(option('b'))==1) {
          if (strstr(aminoacids,resnm)) {
            /* is aminoacid */
            for (c=tok+3; (*c>='0' && *c<='9'); c++); /* skip res. number */
            if (*c>='a' && *c<='z') c++;              /* skip chain letter */
            /* mark backbone C or CA or N (1=not draw) */
            site[i].c=strcmp(c,"C") && strcmp(c,"CA") && strcmp(c,"N"); }
          else {
            /* was not aminoacid - mark if -b, not mark if -b-1 */
            site[i].c=option('b')<0; } }

        site[i].r=16; /* means no color mark found */
        for (cl=collisthead; cl; cl=cl->next)
          if (cl->pattern && strstr(tok,cl->pattern))
            site[i].r=cl->colno;

#ifdef SHELL
        /* protein & draw marked */
//        site[i].shell=!memcmp(tok,"HOH",3) || strchr(tok,'w') ? 1 : 4;
        site[i].shell= strstr(tok,shellkey)?1:4;
#endif
        tok=next(); /* type */
        loop (j,0,i) if (!strcmp(tok,site[j].type)) {
          site[i].type=site[j].type;
          goto tdone; }
        site[i].type=strdup(tok);
       tdone:

#ifdef SHELL
        if (strchr(tok,'O')) site[i].shell |= 2; /* water O marked */
#endif
#ifdef GOLFROMMOL
        site[i].molcol=tok[0];
#endif
        /* H-bonds */
        if (!strcmp(tok,donor)) {
          site[i].mark|=1;
          ndonor++; }
        if (!strcmp(tok,acceptor)) {
          site[i].mark|=2;
          nacceptor++; }

        tok=next(); /* charge */
#ifdef CHARGES
        site[i].charge=atof(tok);
#endif
        next();
        nnbr=atoi(next());
        loop (inbr,0,nnbr) {
          nbr=atoi(next());
#ifdef SHELL
          if (site[i].shell==3) wmol[inbr]=nbr-i;
#endif
          loop (ic,0,nc)
            if (bond[ic][0]==nbr && bond[ic][1]==i) {
              testnc++; goto cont; }
          if (nc>=nbond) nbond=enlarge(&bond,nbond,nbond*3/2+4);
          bond[nc][0]=i; bond[nc][1]=nbr;
          nc++;
        cont:; }
        if (nnbr==0 && option('f')) {
          if (nc>=nbond) nbond=enlarge(&bond,nbond,nbond*3/2+4);
          bond[nc][0]=i; bond[nc][1]=i;
          nc++; testnc++; } } } }

  if (file) fclose(file);

#ifdef SHELL
  if (abs(wmol[0]+wmol[1])>3) {
    if (isL && option('j')>0) ERROR(("unsupported water"))
    wmol[0]=wmol[1]=0; }
  prt("\nwater mol: H1-O=%d H2-O=%d",wmol[0],wmol[1]);
#endif

  _n

#ifdef CHARGES
  loop (i,0,ns) {
    if (negA!=NULL || posA!=NULL) {
      int gc=groupcharge(i);

      if (negA) if (gc<=option('n')) site[i].r=negC;
      if (posA) if (gc>=option('p')) site[i].r=posC;

      if (option('e')) { /* the same for all neighbors */
        int j;

        loop (j,0,nc) {
          int b0=bond[j][0],b1=bond[j][1];

          if (i==b0 || i==b1) {
            gc=groupcharge(i==b0?b1:b0);
            if (negA) if (gc<=option('n')) site[i].r=negC;
            if (posA) if (gc>=option('p')) site[i].r=posC; } } } } }
#endif

  if (nc!=testnc) ERROR(("%s: [%d:%d] bad bonds",fn,nc,testnc))

  loop (ic,0,nc) {
    i=bond[ic][0]; nbr=bond[ic][1];
    bond[ic][2] = colortab[
                           site[i].r & site[nbr].r & 16 ? 0 :
                           site[i].r&16 ? site[nbr].r : site[i].r].turbocolor; }

  prt("%s: %d sites, %d bonds",fn,ns,nc);
  if (ndonor+nacceptor)
    prt("%s: %d donors (%s) and %d acceptors (%s)",
         fn, ndonor,    donor,  nacceptor,    acceptor);

  /* bond==bond0 if not h-bonds */
  bond0=bond; nc0=nc; nbond0=nbond;
  bondeqbond0=1;
  ncfix=nc;
}

void mol2mol(int newns) /********************************************* newns */
{
  int i,j,n,newnc;
  site_t *oldsite=site;
  bond_t *oldbond=bond;

  allocarrayzero(site,newns*sizeof(site_t));
  loop (i,0,newns) site[i]=oldsite[i%ns];
  free(oldsite);

  allocarray(bond,nbond);

  newnc=0;
  n=(newns+ns-1)/ns;
  loop (i,0,n)
    loop (j,0,nc)
      if (oldbond[j][0]+i*ns<newns && oldbond[j][1]+i*ns<newns) {
        if (newnc>=nbond) nbond=enlarge(&bond,nbond,nbond*3/2+4);
        bond[newnc][0]=oldbond[j][0]+i*ns;
        bond[newnc][1]=oldbond[j][1]+i*ns;
        bond[newnc][2]=oldbond[j][2];
        newnc++; }
  free(oldbond);

  ns=newns;
  nc=newnc;

  /* cf. readMOL above */
  bond0=bond; nc0=nc; nbond0=nbond;
  bondeqbond0=1;
  ncfix=nc;
}

void PDBbackbone(void) /**************************************** PDBbackbone */
{
  int i0=0,ic,j;

  loop (ic,0,nc)
    if (!site[bond[ic][0]].c && !site[bond[ic][1]].c) {
      bond[ic][2]&=255;
      loop (j,0,3) bond[i0][j]=bond[ic][j];
      i0++; }
  nc=i0;
}

void backbone(void) /****************************************** backbone */
/* warning: readGOL must be called after backbone */
{
  int *tree,i=0,i0,j=0,ic,t=0,ii;

#ifdef DEBUG
  out=fopen("show.dbg","wt");
#endif

  prts("looking for the backbone...");
  alloc(tree,ns*sizeof(int));
  memset(tree,0,ns*sizeof(int));
  loop (i,0,ns) site[i].c=1;

 again:
    i=j;
    tree[i]=++t;
 again2:
#ifdef DEBUG
    put2(i,t)
#endif
    i0=-1;
    loop (ic,0,ns) {
      if (bond[ic][0]==i) j=bond[ic][1];
      else if (bond[ic][1]==i) j=bond[ic][0];
      else continue;
      /* (j) is neighbour of (i) */
      if (!tree[j]) goto again;
#ifdef DEBUG
      else if (tree[j]<t-1) prt("%i %i: cycle",i,j);
#endif
      else if (tree[j]==t-1) i0=j; }

    i=i0;
    if (--t && i0>=0) goto again2;

#ifdef DEBUG
  loop (i,0,ns) prt("%4i : %i",i,tree[i]);
#endif

  i0=ns-1;
  t=tree[i0];
  if (!t) {
    /* no begin-end backbone: try find max from beginning */
    t=0; i0=0;
    loop (i,0,ns) if (tree[i]>t) { t=tree[i]; i0=i; } }

#ifdef DEBUG
  put2(i0,t0)
#endif

/* mark bonds+sites of main chain */
  while (--t) {
    loop (ic,0,nc) {
      i=bond[ic][0]; j=bond[ic][1];
      if (i>=0 && j>=0)
        if ( (i==i0 && tree[ii=j]==t) || (tree[ii=i]==t && j==i0) ) {
          bond[ic][2]|=256;
          site[i].c=0; site[j].c=0;
          i0=ii;
          break; } } }

#ifdef DEBUG
  loop (ic,0,nc) prt("%4i - %-4i  %i",bond[ic][0],bond[ic][1],bond[ic][2]);
#endif

  /* remove other bonds */

  i0=0;
  loop (ic,0,nc)
    if (bond[ic][2]&256) {
      bond[ic][2]&=255;
      loop (j,0,3) bond[i0][j]=bond[ic][j];
      i0++; }
  nc=i0;

#ifdef DEBUG
  loop (ic,0,nc) prt("%4i - %i",bond[ic][0],bond[ic][1]);
  fclose(out);
#endif

  free(tree);
}

void hbonds(void) /************************************************** hbonds */
/* calculate hydrogen bonds (from bond[ncfix] to bond[nc]) */
{
  int i,j,k;
  float x,rr;
  static int lastb;
  if (hbdist==0) return;

#if 0
  for (i=0; i<ns; i+=4) fprintf(stderr,"%d: %d/%d %d/%d %d/%d %d/%d\n",
                                i/4,
                                site[i].c,site[i].r,
                                site[i+1].c,site[i+1].r,
                                site[i+2].c,site[i+2].r,
                                site[i+3].c,site[i+3].r);
#endif

  nc=ncfix;
  loop (i,0,ns) if (site[i].mark==1 && site[i].c<NOCOL*COLLEN)
    loop (j,0,ns) if (site[j].mark==2 && site[j].c<NOCOL*COLLEN) {
      rr=0;
      loop (k,0,3) {
        x=cfg[i][k]-cfg[j][k];
        if (isL) if (L[k]) {
          while (x>Lh[k]) x-=L[k];
          while (x<-Lh[k]) x+=L[k]; }
        rr+=Sqr(x);
        if (rr>hbdistq) goto cont; }
      /* already a bond? */
      loop (k,0,nc)
        if (bond[k][0]+bond[k][1]==i+j && abs(bond[k][0]-bond[k][1])==abs(i-j)) goto cont;
      if (nc>=nbond) nbond=enlarge(&bond,nbond,nbond*5/4+4);
      bond[nc][0]=i;
      bond[nc][1]=j;
      bond[nc][2]=hbcolor;
      nc++;
     cont:; }
  if (lastb!=nc-ncfix)
    prt("%d hydrogen bond(s) found (|%s-%s|<%g)",nc-ncfix,acceptor,donor,hbdist);
  lastb=nc-ncfix;
}
static int athelp,dathelp=10;

void H(char *k,char *msg)
{
  setcolor(palcolor(YELLOW)); outtextxy(4,athelp,k);
  setcolor(palcolor(LIGHTCYAN)); outtextxy(100,athelp,msg);
  athelp+=dathelp;
}

void readGOL(void) /*********************************************** readGOL */
/***
  reads *.gol file
  fn must have extension .gol
  warning: readGOL must be called after backbone
***/
{
  char *tok,*fn=Fn("gol");
  int i,nsgol,c;
  static int GOLread=0;

  if (GOLread || option(']')==0) return;

  file=fopen(fn,"rt");
  if (!file) {
#ifdef GOLFROMMOL
    fprintf(stderr,"%s: no goal file, using atom info\n",fn);
    loop (i,0,ns) {
      /* table in blendmed.c differs in LIGHTGRAY x LIGHTCYAN for Carbon */
                     /*          111111*/
      static         /*0123456789012345*/
      char colors[17]="xxxxxxx  NXCOMSH";
                     /*BBGCRMBGGBGCRMYW*/
                     /*klryearrrlryeaeh*/

      static unsigned char rads[16]={0,0,0,0,0,0,0,0,145,128,145,140,110,103,132,70};

      /* see also blendmed.c, function atomcolor() */

      char ch=site[i].molcol;
      int col;

      loop (col,7,16) if (ch==colors[col]) break;
      if (col==16) col=LIGHTMAGENTA;
      if (ch=='P') col=BROWN; /* = ORANGE, phosphorus default */
      /* TO BE OPTIMIZED: REPEATED BELOW */
      loop (c,0,NOCOL-1)
        if (col==colortab[c].turbocolor) break;
      if (!(site[i].r&16)) c=site[i].r;
      site[i].r=rads[col]*!site[i].c;
      /* note that sites to remove have been marked by backbone() in site[i].c */
      site[i].c=1+c*COLLEN; }
#else
    ERROR(("%s: no gol-file",fn))
    loop (i,0,ns) site[i].r=100,site[i].c=11;
#endif
  }
  else {
    /* reading the GOL-file */
    mygetline();
    nsgol=atoi(li);
    if (nsgol!=ns) prt("%d site(s) found in %s (expected %d): will be %s",
      nsgol,fn,ns,nsgol<ns?"repeated":"truncated");

    loop (i,0,min(ns,nsgol)) {
      if (!mygetline()) Error("gol file too short");
      tok=strtok(li," \t\n");
      /* GRAY==CYAN for Carbons */
      if (!strcmp(tok,"GRAY")) tok="CYAN";
      if (!strcmp(tok,"ORANGE")) tok="N";
      loop (c,0,NOCOL-1)
        if (colortab[c].col[0]==tok[0]) goto colfound;
      c=NOCOL-1;
    colfound:
      if (!(site[i].r&16)) c=site[i].r;
      site[i].r=(100*atof(strtok(NULL," \t\n"))+0.5)*!site[i].c;
      /* note that sites to remove have been marked by backbone() in site[i].c */
      site[i].c=1+c*COLLEN; }
    if (nsgol<ns) loop (i,nsgol,ns) {
      site[i].r=site[i-nsgol].r;
      site[i].c=site[i-nsgol].c; } }

  GOLread=1;
}

static void floodmark(int center) /****************************** floodmark */
{
  int n,i;

  site[center].mark|=4;

  circle(xy0[center][0],xy0[center][1],2);

  loop (n,0,nc) {
    if (bond[n][0]==center) if (!(site[i=bond[n][1]].mark&4)) floodmark(i);
    if (bond[n][1]==center) if (!(site[i=bond[n][0]].mark&4)) floodmark(i); }
}

int markcolor;

static int colormark(int i) /************************************ colormark */
/* returns the color of site i, or the mark color if marked */
{
  if (site[i].mark&4) return markcolor;
  else return site[i].c;
}
