/* Exp-6 potential with optional r^-12 term, Busing form

    u_rep(r) = f (rho_i+rho_j) exp[ (R_i+R_j-r)/(rho_i+rho_j) ]
    u(r) = u_rep(r) - C_i*C_j/r^6 [ + D_i*D_j/r^12 ]

    where f=0.05 e^2/A

  Parameter correspondence in the par-file:
   alpha   -- unused
   EvdW    -- f*rho_i
   RvdW    -- R_i
   parm[0] -- C_i
   parm[1] -- D_i

  NOTE:
  - rho_i<0 is interpreted so that rho_ij=0 for any atom j, which is useful,
    e.g., for a bare charge
  - EvdW_i=0 (rho_i=0) and R_i=0 still can combine with other atoms

  The site parameters (X_i, X={rho,R,C,D}) appearing in the table are
  components, not the site-site parameters X_ii (which appear in the
  nbfix table)

  Version:
    #define REP12: the extra repulsive term  D_ij/r^12  is added

  Compatible with:
    not POLAR
    POLAR&1==0 (standard linear polarizability)
    POLAR&1==1 (shell-core model, cf. Busing version)
*/

#ifndef SITE_SITE
#  define SITE_SITE

#  ifdef REP12
#    define SS_TABLE "Busing+12"
#  else /*# REP12 */
#    define SS_TABLE "Busing"
#  endif /*#!REP12 */

#  include "simopt.h"
#  include "units.h"
#  include "int4.h"

/*** exp6 tables and combining rules ***/
#  ifdef REP12
#    define SS_PARMS      2
typedef struct pairaux_s {
  real eps,RR,rhorho,f_rhorho,C,C6,D,D12;
} pairaux_t;
#  else /*# REP12 */
#    define SS_PARMS      1
typedef struct pairaux_s {
  real eps,RR,rhorho,f_rhorho,C,C6;
} pairaux_t;
#  endif /*#!REP12 */

#  include "ssdecl.h"

#  ifdef BLEND
#define obfuscate 332.063709
#    define BUSING_f (0.05*obfuscate) /* 0.05 e^2/A^2 in kcal/mol */
#  else /*# BLEND */
#    define BUSING_f (0.05*Sqr(electron)) /* 0.05 e^2/A^2 in prog. units */
#  endif /*#!BLEND */

/* Urep'', used by blend, r=sqrt(rr) must be passed */
#define SS_ffrep if (ss->a.eps==0) ffrep=0; else {     \
  y=exp((ss->a.RR-r)/ss->a.rhorho); \
  ffrep=ss->a.f_rhorho*y; }

/* site-site repulsive energy and forces */
#define SS_MEASURE_rep if (ss->a.eps==0) Urep=frep=0; else {  \
  x=sqrt(rr); /* blend: double eval, cook: ??? */   \
  y=exp((ss->a.RR-x)/ss->a.rhorho);  \
  Urep=ss->a.eps*y;  \
  frep=BUSING_f*y/x; }
/* site-site repulsive forces only */
#define SS_NOMEASURE_rep if (ss->a.eps==0) frep=0; else { \
  x=sqrt(rr); /* double eval? */   \
  y=exp((ss->a.RR-x)/ss->a.rhorho);  \
  frep=BUSING_f*y/x; }

/* site-site energy and forces (called after SS_MEASURE_rep) */
#define SS_MEASURE { double zz; \
  z=Cub(rr);                 \
  zz=ss->a.D/Sqr(z);           \
  z=ss->a.C/z;                 \
  U=Urep-z+zz;  \
  f=frep+(12*zz-6*z)/rr; }
/* site-site forces only (called after SS_NOMEASURE_rep) */
#define SS_NOMEASURE(OP) { \
  z=Cub(rr); \
  f OP frep+(ss->a.D12/z-ss->a.C6)/(rr*z); }

#endif
