/* 
  cluster analysis
  #included from sfdx.c #ifdef CLUSTERS
*/
#define PERMUTETYPE int
#include "permute.c"

#ifndef BONDHIST
/* # of previously broken or created bonds (from a molecule to a molekule)
   at least 2 needed for reliable definition of `really new' bonds */
#define BONDHIST 32
#endif

static int cmpint(const void *a,const void *b)
{
  if (*(int*)a<*(int*)b) return -1;
  else if (*(int*)a>*(int*)b) return 1;
  else return 0;
}

static double cluster_cq;
static int isincq(real *r1,real *r2)
{
  double rr;
  vector dr;

  NearestImage(0,cluster_cq) return 1; else return 0;
}

static struct nbrs_s {
  struct nbrs_s *next;
  int n; } 
    **nbrs, /* nbrs[i] is the head of the list of neighbors of molecule i */
    **last; /* last timestep */

static void destroylast(void) /********************************* destroylast */
{
  int n;
 
  if (last) {
    struct nbrs_s *lst,*llst;
 
    loop (n,0,No.N) 
      for (lst=last[n]; lst; lst=llst) {
        llst=lst->next; free(lst); }
 
    free(last); }
 else
   ERROR(("internal"))
}

static int cluster_k; /* frame (timestep) for bonds, integer 0,1, */

static struct bondhist_s {
  struct bondhist_s *next;
  float t; /* time */
  int k; /* frame index from 0 */
  int x; /* 1: creation, 0: breakage; 2:flag of one create-break-... series*/
  int n; /* neighbor */
  } **bondhist; /* bondhist[i] = head of the list of created or broken bonds 
                   for molecule i */

static float tstart,tfinish; /* patchy way how to get time */

static int configurations,clusters;
static int *color; /* all sites i in a cluster have the same color[i] */
static FILE *clustcp;
static char (*golcolors)[64];

typedef struct cluster_s {
  struct cluster_s *next;
  char *name;  /* only if it is a named clusted */
  int col;     /* only if it is a named clusted, column # in SIMNAME.cl.cpa */
  int count;   /* how many times found in one config (active only if name) */
  int total;   /* how many times found total */
  int nm;      /* # of molecules */
  int version; /* 1,2,3... for clusters with the same stoichometry */
  int *stoich; /* stoichometry, [nspec] */
  int **nbrs;  /* array[nm] of neighbor lists;
                  each neighbor list starts with its size */
  float (*r)[3]; /* first cluster positions (of site[0]) stored */
  } cluster_t;

static cluster_t *clhead,*scl;
static int maxnbrs=8,*rennbrs,nnbr;
static int *ren,maxcluster=12;

int cmpcluster(const cluster_t *c,const cluster_t *cl) /********* cmpcluster */
/* returns 0 if the clusters are identical (stoichometry and topology) 
   and 1 otherwise
   also considers all clusters larger than maxcluster molecules with
   the same stoichometry as identical (return 0)
   warning: does not define ordering - not to be used for sorting */
{
  int n,i;
  int *clnbrs,*cnbrs;
 
  /* compare cluster size */
  if (c->nm!=cl->nm) return 1;
 
  /* compare stoichiometry */
  if (memcmp(c->stoich,cl->stoich,nspec*sizeof(cl->stoich[0]))) return 1;
 
  /* too large clusters considered identical */
  if (c->nm>maxcluster) return 0;
 
  /* compare topology */
 
  initpermutes0(cl->nm,ren,nspec,cl->stoich);
 
  do { /* all permutations of molecules of the same type in cluster cl */
 
    /* we try to match permuted cl : c */
    loop (n,0,cl->nm) {
 
      /* mol.# n in cl is matched to mol.# ren[n] in c */
      clnbrs=cl->nbrs[n];
      nnbr=clnbrs[0];
      cnbrs=c->nbrs[ren[n]];
 
      /* compare # of neighbors */
      if (nnbr!=cnbrs[0]) goto nextperm;
 
      if (maxnbrs<nnbr) {
        /* update auxiliary array of renumbered neighbors if has grown */
        free(rennbrs);
        allocarray(rennbrs,nnbr+1);
        maxnbrs=nnbr; }
 
      /* compare neighbors, in standard order */
      loopto (i,1,nnbr) rennbrs[i]=ren[clnbrs[i]];
      if (nnbr>1) qsort(rennbrs+1,nnbr,sizeof(rennbrs[0]),cmpint); 
      if (memcmp(rennbrs+1,cnbrs+1,nnbr*sizeof(rennbrs[0]))) goto nextperm;
 
      } /* n */
 
    /* now c == renumbered cl */
    return 0;
 
  nextperm:; 
    } while (permutes(ren));
 
  return 1;
}

int cmpstoich(const cluster_t *c,const cluster_t *cl) /*********** cmpstoich */
/* cluster ordering according to stoichiometry
   warning: topologically different clusters are not distinguished
   to be used for sorting for `nice' print only
 */
{
 /* compare cluster size */
 if (c->nm<cl->nm) return -1;
 if (c->nm>cl->nm) return 1;

 /* compare stoichiometry */
 return memcmp(c->stoich,cl->stoich,nspec*sizeof(cl->stoich[0]));
}

void prtcluster(FILE *removescript,cluster_t *c) /*************** prtcluster */
{
  int i,j,n,k,sp;
  char pref=' ';
  char formula[64],*f;
  double charge=0;
 
  formula[0]=0;
 
  if (!removescript) prt_("%5d ", c->count);
  prt_("%7d",c->total);
 
  loop (i,0,nspec) if (c->stoich[i]) {
    sprintf(strend(formula),"%c%s",pref,spec[i]->name);
    pref='.';
    if (c->stoich[i]!=1) sprintf(strend(formula),"%d",c->stoich[i]); 
    charge+=spec[i]->charge*c->stoich[i]; }
 
  if (c->version)
    if (c->version<27) sprintf(strend(formula),"%c",'a'-1+c->version);
    else sprintf(strend(formula),"-%d",c->version);
 
  prt_("%-12s ",formula);
  charge/=electron;
  if (fabs(charge)<1e-6) prt_(" 0");
  else prt_("%+2.0f",charge);
 
  if (c->name) prt_(" %d:%s ",c->col,c->name);
  else prt_("    ");
  if (c->nm<=maxcluster) loop (i,0,c->nm) loopto (j,1,c->nbrs[i][0])
    if ( (n=c->nbrs[i][j]) >=i)
      prt_(" %d-%d", i,n);
  _n
 
  if (removescript) if (c->r) {
    FILE *mol,*plb,*gol=NULL;
    static float hdr[2];
 
    f=strend(formula);
    strcpy(f,".mol");
    mol=fopen(formula+1,"wt");
    if (removescript) fprintf(removescript,"rm%s",formula);
    if (golcolors) {
      strcpy(f,".gol");
      gol=fopen(formula+1,"wt");
      fprintf(gol,"!\n!\n%d\n",c->nm);
      if (removescript) fprintf(removescript,"%s",formula); }
    strcpy(f,".plb");
    plb=fopen(formula+1,"wt");
    if (removescript) fprintf(removescript,"%s\n",formula);
    
    hdr[0]=c->nm;
    fwrite(hdr,2,4,plb);
    fwrite(c->r,c->nm,12,plb);
    fclose(plb);
 
    fprintf(mol,"cluster");
    if (c->name) fprintf(mol," %s",c->name);
    *f=0;
    fprintf(mol,"%s\n\n\
number_of_atoms=%d\n\n\
atoms\n",formula,c->nm);

    sp=k=0;
    loop (i,0,c->nm) {
      char *nm;
 
      while (k>=c->stoich[sp]) sp++,k=0;
      k++;
      
      if (gol) fprintf(gol,"%s\n",golcolors[sp]);
      nm=spec[sp]->name;
      fprintf(mol,"%2d %d-%s %s 0 0",i,i,nm,nm);
 
      loopto (j,0,c->nbrs[i][0])
        fprintf(mol," %d",c->nbrs[i][j]);
      fprintf(mol,"\n"); }
 
    fclose(mol);
    if (gol) fclose(gol); }
}

static void addnbr(int n1,int n2) /********************************* addnbr */
/* `bond' n1--n2 is added to the list of neighbors of both n1 and n2 */
{
  struct nbrs_s *nbr;

  allocone(nbr);
  nbr->n=n2;
  nbr->next=nbrs[n1];
  nbrs[n1]=nbr;
}

static int mark;

static void findcluster(int n)
{
  struct nbrs_s *nbr;
 
  color[n]=mark;
  for (nbr=nbrs[n]; nbr; nbr=nbr->next)
    if (!color[nbr->n]) findcluster(nbr->n);
}
 
int getsitetype(char *nm)
{
  int i;

  loop (i,0,nsites) if (!strcmp(nm,sitedef[i].name)) return i;

  ERROR(("%s is unknown site type",nm))
  
  return -1;
}

static int spectype(char *c)
{
  int i;

  loop (i,0,nspec) if (!strcmp(c,spec[i]->name)) return i;
  ERROR(("no species (molecule) %s",c))
 
  return 0;
}

/*****************************************************************************/
/*               chemical format - modified from blendmed.c                  */
/*****************************************************************************/

const char SEP[]=" -/\\=|+x\t\n";

static struct coor_s { int st,n,x0,x1,y; } *coor;

static int ns;
static char *clustername;
static int siten(int x,int y) /*************************************** siten */
{
  int i;

  loop (i,0,ns)
    if (y==coor[i].y)
      if (coor[i].x0<=x && x<=coor[i].x1) return coor[i].n;
  ERROR(("%s: bad bond in rel.line=%d, col.=%d",clustername,y,x))
  
  return 0;
}

static char **screen;
static int si;

static void findnbr(int x,int y,int c,int dx,int dy) /************** findnbr */
{
  int *nbr,from,to;

  if (screen[y][x]!=c) return;
 
  do {
    x+=dx; y+=dy;
    } while (screen[y][x]==c || screen[y][x]=='x' || screen[y][x]=='+');
  to=siten(x,y);
  from=coor[si].n;
  if (from>=ns || to>=ns) ERROR(("internal"))
  allocarray(nbr,scl->nbrs[from][0]+2);
  copy(nbr,scl->nbrs[from],(scl->nbrs[from][0]+1)*sizeof(nbr[0]));
  nbr[++nbr[0]]=to;
  free(scl->nbrs[from]);
  scl->nbrs[from]=nbr;
}

cluster_t *readCHE(FILE *f) /*************************************** readCHE */
/* cf. readCHE() in blend/blendmed.c */
{
  int nx,ny,i,is,x,st;
  char *c;
  int *stsum;
  cluster_t *cl;
  char line[256];
 
  struct line_s {
    struct line_s *next;
    char *l; } *head=NULL,*l;
  
  alloconezero(cl); 
  scl=cl;
  allocarrayzero(cl->stoich,nspec);
 
  rallocarrayzero(stsum,nspec);
 
  /***
    read and store to a list
    # of sites calculated
    # of lines & columns calculated
  ***/
  nx=ny=ns=0;
  while (fgets(line,256,f) && line[0]!='\n') 
    if (!strchr("!#",line[0])) {
      rallocone(l);
      l->l=strdup(line);
      if (!l->l) ERROR(("strdup"))
      l->next=head;
      head=l; 
      Max(nx,strlen(line));
      
      c=strtok(line,SEP);
      while (c) {
        if (isalpha(c[0])) ns++;
        c=strtok(NULL,SEP); }
    
      ny++; }
 
  if (option('v')&4) prt("cluster %d sites  %d lines  %d columns",ns,ny,nx);
  cl->nm=ns;
 
  nx+=2;
  
  ralloc(screen,(ny+2)*sizeof(*screen));
  
  ralloc(screen[0],nx); memset(screen[0],' ',nx);
  ralloc(coor,sizeof(struct coor_s)*ns);
 
  allocarrayzero(cl->nbrs,ns);
 
  is=0;
  for (i=ny,l=head; i>=1; i--,l=l->next) {
    if (!l) ERROR(("internal"))
    ralloc(screen[i],nx);
    memset(screen[i],' ',nx);
    if (strlen(l->l)>=nx) ERROR(("internal"))
    strcpy(screen[i]+1,l->l);
 
    c=strtok(l->l,SEP);
    while (c) {
      if (isalpha(c[0])) {
        if (is>=ns) ERROR(("internal"))
        coor[is].x0=(int)(c-l->l)+1;
        coor[is].x1=(int)(c-l->l)+strlen(c);
        coor[is].y=i;
        st=spectype(c);
        coor[is].st=st;
        coor[is].n=cl->stoich[st]++;
        is++; }
      c=strtok(NULL,SEP); } }
 
  if (l || is!=ns) ERROR(("internal"))
 
  loop (is,1,nspec) stsum[is]=stsum[is-1]+cl->stoich[is-1];
  loop (is,0,ns) coor[is].n+=stsum[coor[is].st];
 
  ralloc(screen[ny+1],nx); memset(screen[ny+1],' ',nx);
 
  /*** finding bonds ***/
  loop (si,0,ns) alloconezero(cl->nbrs[si]);
 
  loop (si,0,ns) {
 
    /* shorter sides */
    findnbr(coor[si].x0-1,coor[si].y,'-',-1,0);
    findnbr(coor[si].x1+1,coor[si].y,'-',1,0);
    
    /* longer sides + corners */
    loopto (x,coor[si].x0,coor[si].x1) {
      findnbr(x-1,coor[si].y-1,'\\',-1,-1);
      findnbr(x,coor[si].y-1,'|',0,-1);
      findnbr(x+1,coor[si].y-1,'/',1,-1);
      
      findnbr(x+1,coor[si].y+1,'\\',1,1);
      findnbr(x,coor[si].y+1,'|',0,1);
      findnbr(x-1,coor[si].y+1,'/',-1,1); } }
 
  loop (si,0,ns) 
    if (cl->nbrs[si][0]>1)
      qsort(cl->nbrs[si]+1,cl->nbrs[si][0],sizeof(cl->nbrs[0][0]),cmpint); 
 
  release(stsum);
  return cl;
} /* readCHE */

static double **dist;
static int namedclusters;

void readclusterdef(FILE *f) /******************************* readclusterdef */
{
  char line[256];
  char nm1[16],nm2[16];
  double d;
  int nd=0,i;
  FILE *oldf=NULL;
 
  if (!f) ERROR(("def file"))
 
  if (!rennbrs)
    allocarray(rennbrs,maxnbrs+1); /* permanent, growing on demand */
    
  for (;;) {
    char *tok;
 
    while (!fgets(line,256,f)
           || ( (tok=strtok(line," \t\n=")) && 
                (!strcmp(tok,"end") || !strcmp(tok,".")) )) {
      if (!oldf) goto prtinfo;
      f=oldf; oldf=NULL; }
 
    if (!line[0] || strchr("!#\n",line[0])) continue;
 
    else if (line[0]=='$' && line[1]=='i') {
      char *fn=line+2;
 
      if (oldf)
        WARNING(("nested $i in reading cluster definitions: returns only once"))
      if (!*fn) fn=strtok(NULL," \t\n=");
      if (!fn) ERROR(("$i missing file name"))
      prt("opening include file \"%s\"",fn);
      oldf=f;
      f=fopen(fn,"rt");
      if (!f) {
        ERROR(("no include file \"%s\"",fn))
        f=oldf; } }
    
    else if (!strcmp("off",tok)) {
      if (dist) free(dist); /* leak! */ 
      break; }
 
    else if (!strcmp("clusters",tok)) clusters++;
 
    else if (!strcmp("configurations",tok)) configurations++;
 
    else if (!strcmp("colors",tok)) {
      int i;
 
      allocarray(golcolors,nspec);
      loop (i,0,nspec) strcpy(golcolors[i],"WHITE 1"); 
      while (fgets(line,256,f)) {
        char *tok=strtok(line," \n\t");
        int sp;
 
        if (!tok) break;
        sp=spectype(tok);
        tok=strtok(NULL,"\n");
        if (tok) strcpy(golcolors[sp],tok); } }
 
    else if (!strcmp("bonds",tok)) {
      allocarray(dist,nsites);
      loop (i,0,nsites) allocarrayzero(dist[i],nsites);
  
      while (fgets(line,256,f)) {
        int sp1,sp2;
        i=sscanf(line,"%s%s%lf",nm1,nm2,&d);
        if (i<=0) break;
        if (i!=3) ERROR(("%sATOM ATOM DIST expected",line))
        sp1=getsitetype(nm1);
        sp2=getsitetype(nm2);
        
        dist[sp1][sp2]=dist[sp2][sp1]=d*d; 
        nd++; } }
 
    else if (!strcmp("maxcluster",tok)) {
      maxcluster=atoi(strtok(NULL," \t\n="));
      if (maxcluster<1 || maxcluster>1024)
        ERROR(("maxcluster out of bounds")) }
 
    else if (!strcmp("cluster",tok)) {
      cluster_t *cl,*c;
      char *name=strdup(strtok(NULL," \t\n="));
      int inm;
 
      clustername=name; /* static, for err msg */
      cl=readCHE(f);
      namedclusters++;
 
      allocarray(ren,cl->nm);
      cl->name=name; 
      inm=namedclusters;
      for (c=clhead; c; c=c->next) {
        inm--;
        if (!cmpcluster(c,cl))
          ERROR(("named clusters #%d=%s and #%d=%s are equivalent\n\
*** (check whether maxcluster is large enough)",
                 inm,c->name,namedclusters,cl->name)) }
      free(ren);
 
      cl->next=clhead;
      clhead=cl; } 
    }
 
 prtinfo:
  prt("cluster analysis: %d distances, maxcluster=%d, %d named clusters",
      nd,maxcluster,namedclusters);
  /* warnings and opening clustcp postponed to analyzeclusters() 
     because readclusterdef() can be called twice */
}

void analyzeclusters(ToIntPtr A,int frame) /**************** analyzeclusters */
{
  int i,i1,i2,n,n1,n2,sp1,sp2,ns1,ns2,imark,nbonds=0;
  molecule_t *mn1,*mn2,*mn;
  siteinfo_t *si1,*si2;
  vector *r1,*r2;
  vector LL,xL,lastL;
  int k;
  cluster_t *c;
  static int pass;
 
  if (!iscube()) ERROR(("analyzeclusters: cube expected"))
 
  loop (k,0,3) LL[k]=xL[k]=1,lastL[k]=box.L[k];
 
  if (!dist) {
    if (!pass++) prt("no cluster analysis");
    namedclusters=0; 
    return; }
  if (maxcluster<0 || maxcluster>32) WARNING(("suspicious maxcluster"))
 
  if (tau.P)
    loop (k,0,3) {
      LL[k]=1./Sqr(box.L[k]);
      xL[k]=box.L[k];
      box.L[k]=1;
      box.Lh[k]=0.5; }
 
  rallocarrayzero(color,No.N);
  allocarrayzero(nbrs,No.N); /* to be renamed to last and freed later */
  
  loop (n1,0,No.N) {
    mn1=molec+n1;
    sp1=mn1->sp;
    r1=rof(mn1,A->rp);
    ns1=mn1->ns;
    si1=spec[sp1]->si;
 
    loop (n2,0,n1) {
      mn2=molec+n2;
      sp2=mn2->sp;
      ns2=mn2->ns;
      r2=rof(mn2,A->rp);
      si2=spec[sp2]->si;
 
      loop (i1,0,ns1) loop (i2,0,ns2)
        /* WARNING: cube */
        if ( (cluster_cq=dist[si1[i1].st][si2[i2].st]*LL[0]) ) if (isincq(r1[i1],r2[i2])) {
          addnbr(n1,n2); addnbr(n2,n1);
 	 nbonds++;
          goto nextmols; }
 
      nextmols:; } }
 
  /* all neighbors nbrs[]-> known now */
 
  mark=0;
  loop (n,0,No.N)
    if (!color[n]) {
      mark++;
      findcluster(n); }
 
  if (configurations) {
    /* - make a cfg-molecule with sites #0 in molecules connected 
       - make a plb-file with sites L-shifted (may fail!) */
    FILE *mol=fopen(Fn(string("%d.mol",frame)),"wt");
    FILE *plb=fopen(Fn(string("%d.plb",frame)),"wb");
    FILE *gol=NULL;
    int is=0;
    float r[DIM],(*r0)[DIM];
 
    if (golcolors) gol=fopen(Fn(string("%d.gol",frame)),"wb");
    alloc(r0,mark*4*DIM);
    loop (i,0,mark) r0[i][0]=-9e9;
 
    r[0]=No.s;
    r[1]=box.L[0]; /* CUBE only!!! */
    if (fabs(box.L[0]-box.L[1])>1e-6 || fabs(box.L[0]-box.L[2])>1e-6) 
      ERROR(("cluster: implementation limitation: cube only"))
    if (tau.P) r[1]=lastL[0];
    fwrite(r,4,2,plb);
 
    if (gol) fprintf(gol,"!sphere#1\n\
! %s\n\
%d\n",lastFn,No.s);

    fprintf(mol,"frame %d\n\n\
number_of_atoms = %d\n\n\
atoms\n",frame,No.s);

    loop (n,0,No.N) {
      int ns=cfg[n].ns,i;
      struct nbrs_s *nbr;
      int col=color[n];
 
      if (col<0 || col>mark) ERROR(("mark=%d color[%d]=%d",mark,n,col))
      col--;
 
      mn=molec+n;
 
      if (gol) fprintf(gol,"%s\n",golcolors[mn->sp]);
      
      nnbr=0;
      for (nbr=nbrs[n]; nbr; nbr=nbr->next) nnbr++;
      
      loop (i,0,ns) {
        fprintf(mol,"%4d %2s-%dc%*s %-2s 0 0 %d",
                is,
                sitedef[spec[mn->sp]->si[i].st].name,
 	       col,4-strlen(string("%d",col)),"",
                sitedef[spec[mn->sp]->si[i].st].name,
                nnbr);
 
        VV(r,=rof(mn,A->rp)[i])
 
        if (i) 
          fprintf(mol," 0");
        else
          for (nbr=nbrs[n]; nbr; nbr=nbr->next) 
            fprintf(mol," %d",nbr->n);
 
        if (r0[col][0]<-8e8) {
          copy(r0[col],r,4*DIM); }
        else {
          int k;
 
#ifndef FREEBC
          loop (k,0,DIM) {
            while (r[k]-r0[col][k]> box.Lh[k]) r[k]-=box.L[k];
            while (r[k]-r0[col][k]<-box.Lh[k]) r[k]+=box.L[k]; }
#endif
          }
        VV(r,*=xL)
        fwrite(r,4,DIM,plb);

        nnbr=0; 
        is++; }
      fprintf(mol,"\n"); }

    fclose(mol); 
    fclose(plb); 
    if (gol) fclose(gol);
    free(r0); }

  loopto (imark,1,mark) {
    cluster_t *cl;
    int inm=0;

    alloconezero(cl);
    loop (n,0,No.N) cl->nm+=color[n]==imark;
   
    allocarrayzero(cl->stoich,nspec);
    allocarrayzero(cl->nbrs,cl->nm);
    allocarray(ren,No.N);
 
    loop (n,0,No.N) if (color[n]==imark) {
      ren[n]=inm++;
      cl->stoich[cfg[n].sp]++; }
 
    if (clusters) allocarray(cl->r,inm);
 
    inm=0;
    loop (n,0,No.N) if (color[n]==imark) {
      struct nbrs_s *nbr;
      vector *r=rof(cfg+n,A->rp);
 
      nnbr=0;
      for (nbr=nbrs[n]; nbr; nbr=nbr->next) nnbr++;
 
      allocarray(cl->nbrs[inm],nnbr+1);
      cl->nbrs[inm][0]=nnbr;
      nnbr=0;
      for (nbr=nbrs[n]; nbr; nbr=nbr->next) cl->nbrs[inm][++nnbr]=ren[nbr->n];
      if (nnbr>1) qsort(cl->nbrs[inm]+1,nnbr,sizeof(cl->nbrs[0][0]),cmpint);
      
      if (cl->r) {
        VV(cl->r[inm],=r[0])
#ifndef FREEBC
        if (inm) {
          int k;

          loop (k,0,DIM) {
            while (cl->r[inm][k]-cl->r[0][k]<-box.Lh[k]) cl->r[inm][k]+=box.L[k];
            while (cl->r[inm][k]-cl->r[0][k]> box.Lh[k]) cl->r[inm][k]-=box.L[k]; } }
#endif
        }

      inm++; }

    if (cl->r) loop (n,0,inm) VV(cl->r[n],*=xL)
     
    /* .. now cl is a `standard' image of the cluster, unique but permutation */
    /* array ren[] will re-used */
 
    /* finding whether cl matches any of the clusters already detected */
    for (c=clhead; c; c=c->next) 
 
      if (!cmpcluster(c,cl)) {
        c->count++;
        c->total++;
        loop (i,0,cl->nm) free(cl->nbrs[i]);
        free(cl->nbrs);
        free(cl->stoich);
        if (cl->r) {
          if (!c->r) {
            allocarray(c->r,c->nm);
            loop (i,0,c->nm) VV(c->r[ren[i]],=cl->r[i]) }
          free(cl->r); }
        free(cl);
        goto match; }
 
    cl->count=cl->total=1;
    cl->next=clhead;
    clhead=cl;
  match:; 
    free(ren); }
 
  {
  /* sort clusters with respect to stoichiometry */
    cluster_t cl00,*cl0=&cl00,*aux;
    int changed;
 
    cl0->next=clhead;
 
    if (cl0->next) do {
      changed=0;
      for (c=cl0; c->next->next; c=c->next)
        if (cmpstoich(c->next,c->next->next)>0) {
          aux=c->next->next;
          c->next->next=aux->next;
          aux->next=c->next;
          c->next=aux;
          changed++; }
      } while (changed); 
    clhead=cl0->next;
  }
 
  /* mark clusters with the same stoichiometry */
  if (clhead) for (c=clhead; c->next; c=c->next)
    if (cmpstoich(c,c->next)==0) {
      cluster_t *cc,*zero=NULL;
      int maxversion=0;
 
      for (cc=c; cc && cmpstoich(cc,c)==0; cc=cc->next) {
        Max(maxversion,cc->version);
        if (!cc->version) zero=cc; }
      if (zero) zero->version=maxversion+1; }
 
  /* number named clusters
     (=columns in SIMNAME.cl.cpa, used as info in .sfd only) */
  if (clhead) {
    int n=0;
 
    for (c=clhead; c; c=c->next) if (c->name) c->col=++n; }
 
  if (pass==0) {
    int l,i;

    pass++;
    /* open clustcp if necessary */
    if (namedclusters) clustcp=fopen(Fn("cl.cpa"),"wt");
 
    if (clustcp) loop (l,0,2) {
      fprintf(clustcp,"#");
      i=1;
      for (c=clhead; c; c=c->next) 
        if (c->name)
          if (l) fprintf(clustcp," %s",c->name); 
          else fprintf(clustcp," %*d",strlen(c->name),i++);
      if (l) fprintf(clustcp," REST  bonds\n");
      else fprintf(clustcp," %4d  bonds\n",i); } }
 
  {
    int rest=0;
   
    for (c=clhead; c; c=c->next) {
      if (option('v')&4) prtcluster(NULL,c);
      if (clustcp)
        if (c->name) fprintf(clustcp," %*d",strlen(c->name),c->count);
        else rest+=c->count;
      c->count=0; }
    if (clustcp) fprintf(clustcp," %4d",rest);
    fprintf(clustcp," %6d\n",nbonds);
  }

  release(color);  

  /*** break bond analysis ***/

  if (last) {
    int n;
    struct bondhist_s *bh;

    if (!bondhist) {
      allocarrayzero(bondhist,No.N);
      tstart=t; }
    tfinish=t;
 
    loop (n,0,No.N) {
      struct nbrs_s *nbr,*lst;
      
      for (nbr=nbrs[n]; nbr; nbr=nbr->next) {
        for (lst=last[n]; lst; lst=lst->next)
          if (nbr->n==lst->n) goto contnbr;
 
        /* created bond to a neighbor found */
        if (option('v')&4) prt("created bond: k=%d %d-%d",k,n,nbr->n);
 
        allocone(bh);
        bh->t=t;
        bh->k=cluster_k;
        bh->n=nbr->n;
        bh->x=1;
        bh->next=bondhist[n];
        bondhist[n]=bh;
        
       contnbr:; }
 
      for (lst=last[n]; lst; lst=lst->next) {
        for (nbr=nbrs[n]; nbr; nbr=nbr->next)
          if (lst->n==nbr->n) goto contlst;
 
        /* broken bond to a neighbor found */
        if (option('v')&4) prt("broken bond: k=%d %d-%d",k,n,lst->n);
 
        allocone(bh);
        bh->t=t;
        bh->k=cluster_k;
        bh->n=lst->n;
        bh->x=0;
        bh->next=bondhist[n];
        bondhist[n]=bh;
 
       contlst:; } }

    destroylast();
    cluster_k++; }

  last=nbrs;

  VV(box.L,=lastL)
  VV(box.Lh,=0.5*box.L)
}

void bonddynamics(void) /************************************** bonddynamics */
{
  int n,i,j;
  
  typedef int (*cp_t)[BONDHIST+1];
  /* created[k][0]: created
     created[k][1]: created with create-break-... removed
     created[k][2]: created with create-(other)-break .. removed ... */
  cp_t created,broken;
 
  if (!dist) return;
 
  if (!bondhist) return;
 
  destroylast(); /* not needed */
 
  prt("create/break bond convergence profiles from t=%.3f to %.3f (k=%d)",
      tstart,tfinish,cluster_k);
 
  allocarrayzero(created,cluster_k);
  allocarrayzero(broken,cluster_k);
 
  loopto (i,0,BONDHIST)
    loop (n,0,No.N) {
      struct bondhist_s *bh,*c=NULL;
 
      for (bh=bondhist[n]; bh; bh=bh->next) bh->x &= 1;
 
      for (bh=bondhist[n]; bh; bh=bh->next) if (!(bh->x&2)) {
        int x=bh->x,nflip=1;
 
        bh->x |= 2;
        /* this loop empty (no check) for i=0 
 	  for i>0, finding create-break... or break-create... sequences
 	  separated by at most i-1 other bond changes */
        for (c=bh->next,j=i; c && j; c=c->next,j--) if (c->n == bh->n) {
          x ^= 1;
          if (c->x != x)
            ERROR(("INTERNAL (n=%d i=%d c->t=%f c->n=%d)",n,i,c->t,c->n))
          c->x |= 2;
 	 j=i+1; /* .. looks strange, but not thet there's j-- immediately */
          nflip++; }
 
        if (nflip&1) {
          /* odd number of flips, e.g. create-break-create */
          j=0;
          for (c=bh; c; c=c->next)
            if (c->n == bh->n)
              if (j++==nflip/2) break;
 
          if (!c) ERROR(("internal"))
 
          if ( (option('v')&4) && nflip>1)
            prt("i=%d %d-%d %s %d flips center=%d %f",
                i,n,c->n,bh->x&1 ? "created" : "broken",nflip,c->k,c->t);
        
          if (c->k<0 || c->k>=cluster_k) 
            ERROR(("internal: c->k=%d not in range [0,%d]",c->k,cluster_k))
          (bh->x&1 ? created : broken)[c->k][i]++; } } }
 
  loop (n,0,2) {
    char *info=n?"created":"broken";
    FILE *f=fopen(Fn(n?"created.cpa":"broken.cpa"),"wt");
    cp_t cp=n?created:broken;
 
    if (!f) ERROR(("open %s",lastFn))
    fprintf(f,"# t 2*%s[0] ... 2*%s[%d]\n",info,info,BONDHIST);
    loop (i,0,cluster_k) {
      fprintf(f,"%7.3f",tstart+i*(tfinish-tstart)/(cluster_k-1));
      loopto (j,0,BONDHIST) {
        fprintf(f," %d",cp[i][j]);
        if (j%BONDHIST) StaSet(0,2,2,0);
        else StaSet(0,lag.err,2,lag.n);
        StaAdd(string("%s[%d]",info,j),cp[i][j]*0.5); }
      fprintf(f,"\n"); }
  fclose(f);
  free(cp); }
}
