/*
  Lagrangian constraint and SHAKE
  global #defines:
    SHAKE: 4 versions (since V2.6d should be #defined)
    POLAR: selffield calculations
    SHEAR: shear viscosity support
    VERLET: with SHAKE: kin.energy algorithm
  local #defines:
    PRTKIN PRECOND DEBUG
  removed (see old+misc/scf+lambda.c):
    LAMBDA
*/

/*
#define DEBUG
  debug energy terms, pressure, scaling */

#include "ground.h"
#define SDS
#include "alloc.h"
#include "statics.h"
#include "simglob.h"
#include "simcg.h"
#include "norm.h"
#include "constrd.h"
#include "forces.h" /* the same for LINKCELL */
#include "ewald.h" /* POLAR: epspol */
#include "cputime.h"
#include "rndgen.h"
#include "units.h"
#include "simils.h" /* Fn() */
#include "asksig.h"
#include "simdef.h" /* because of sitedef_t */
#ifdef LINKCELL
#  include "linklist.h"
#endif /*# LINKCELL */

#ifdef POLAR
#  include "intermac.h"
#  if POLAR&2
#    include "log1.c"
#  endif /*# POLAR&2 */
#  ifndef FREEBC
#    include "elst.h"
#  endif /*# FREEBC */
#endif /*# POLAR */

#ifdef SHEAR
#  include "units.h"
#endif /*# SHEAR */

#ifdef ANCHOR
void normalizer(vector cm)
/* local version of normalize for 1 vector, see also normalize in norm.c */
{
#  ifndef FREEBC
  if (cm[0]<-box.Lh[0]) cm[0]+=box.L[0]; else if (cm[0]>box.Lh[0]) cm[0]-=box.L[0];
  if (cm[1]<-box.Lh[1]) cm[1]+=box.L[1]; else if (cm[1]>box.Lh[1]) cm[1]-=box.L[1];
#    ifndef WALL
  if (cm[2]<-box.Lh[2]) cm[2]+=box.L[2]; else if (cm[2]>box.Lh[2]) cm[2]-=box.L[2];
#    endif /*# WALL */
#  endif /*# FREEBC */
}
#endif /*# ANCHOR */

#if PRESSURETENSOR&PT_VIR
static double PTvirc[PT_DIM];  /* of constraint forces */
#endif /*# PRESSURETENSOR&PT_VIR */

static double blogT(double E,double T) /****************************** blogT */
/* 1/2*ln(E/kT), but always kept in [-1,1] */
{
  const double exp2=7.38905609893065022723;

  if (E>T*exp2) return 1;
  if (E*exp2<T) return -1;

  return log(E/T)/2;
}

static void measureconstraints(void) /******************* measureconstraints */
/* finish pressure calculations and constraints statistics */
{
  int i;

  StaSet(0,lag.err,2,lag.n);
  StaAdd("Pvirc [Pa]",En.virc*(Punit/DIM)/box.V);
  StaAdd("Eel [K]",En.el);
  StaAdd("Pvir pair [Pa]",En.vir*(Punit/DIM)/box.V);
  En.vir+=En.virc+En.el;

#ifdef POLAR
  En.vir-=En.self*2; /* see also VVV(En.Pvir,-=rpol[i],*fpol[i]) below */
#  if POLAR&32
  En.self+=En.fqself;
#  endif /*# POLAR&32 */
  StaAdd("Eself [K]",En.self);
  /* 9/2001: -En.self*2 added instead of the virial of forces on aux
     charges that incorrectly did not include Ewald corrections.
     WARNING: not tested for axial and saturated polarizability !!! */
#endif /*# POLAR */


#if PRESSURETENSOR&PT_KIN
  {
    static int pass=0;
    double x=SUM(En.Pkin)/2-En.kin/2; /* still doubled */

    if (!pass && fabs(x)>1000) {
      prt("WARNING: Ekin=%g K  tr(Pkin)/2=%g K  dif=%g K\n\
          (Pkin uncorrected here; more warnings suppressed)",
          En.kin/2,SUM(En.Pkin)/2,x);
      pass=1; }
  }
  loop (i,0,PT_DIM) En.Pkin[i]*=Punit*En.Pkinq/box.V; // /DIM removed (BUG in 2.4m)
#  if PRESSURETENSOR&PT_MOL
  {
    static int pass=0;
    double x=SUM(En.PKin)/2-En.kin_tr/2; /* still doubled */

    if (!pass && fabs(x)>1000) {
      prt("WARNING: Ekininter=%g K  tr(PKin)=%g K  dif=%g K\n\
          (Pkin uncorrected here; more warnings suppressed)",
          En.kin_tr/2,SUM(En.PKin)/2,x);
      pass=1; }
  }
  loop (i,0,PT_DIM) En.PKin[i]*=Punit*En.Pkinq/box.V; // /DIM removed
#  endif /*# PRESSURETENSOR&PT_MOL */
#endif /*# PRESSURETENSOR&PT_KIN */

#if PRESSURETENSOR&PT_VIR
  loop (i,0,PT_DIM) {
    En.Pvir[i]+=PTvirc[i];
    En.Pvir[i]*=Punit/box.V; // /DIM removed
#  if PRESSURETENSOR&PT_KIN
    En.Ptens[i]=En.Pvir[i]+En.Pkin[i];
#  endif /*# PRESSURETENSOR&PT_KIN */
  }

  {
    static int pass=0;
    double x=SUM(En.Pvir)/DIM-En.vir*(Punit/DIM)/box.V; // /DIM added

    StaSet(0,lag.err,2,lag.n);
    StaAdd("Pvirerr [Pa]",x);

    if (!pass && fabs(x)>1000) {
      prt("WARNING: virial/3V=%g (%g Pa pressure-equivalent)\n\
         tr(Pvir)=%g (%g Pa)  dif=%g (%g Pa)\n\
         (more warnings suppressed; averages will be compared, too)",
          En.vir*(1./DIM)/box.V,En.vir*(Punit/DIM)/box.V,
          SUM(En.Pvir)/Punit/DIM,SUM(En.Pvir)/DIM,x/Punit,x); // /DIM added 2x
      pass=1; }
  }
#  ifdef SLAB
    StaSet(0,lag.err,2,lag.n);
    StaAdd("Pvir xy-zz [Pa]",
           En.Ptv=(En.Pvir[0]+En.Pvir[1]-2*En.Pvir[2])/DIM); // /DIM added
#    if PRESSURETENSOR&PT_KIN
    StaAdd("Pt [Pa]",
           En.Pt=((En.Pvir[0]+En.Pvir[1]-2*En.Pvir[2])
                 +(En.Pkin[0]+En.Pkin[1]-2*En.Pkin[2]))/DIM); // /DIM added
#    endif /*# PRESSURETENSOR&PT_KIN */
#  endif /*# SLAB */
    StaSet(0,2,2,lag.n);
#endif /*# PRESSURETENSOR&PT_VIR */

#if PRESSURETENSOR&PT_ANY
    loop (i,0,PT_DIM) {
      char *xx="xx\0yy\0zz\0yz\0zx\0xy"+i*3;

#  if PRESSURETENSOR&PT_VIR
      StaAdd(string("Pvir.%s [Pa]",xx),En.Pvir[i]);
#  endif /*# PRESSURETENSOR&PT_VIR */
#  if PRESSURETENSOR&PT_KIN
      StaAdd(string("Pkin.%s [Pa]",xx),En.Pkin[i]);
#  endif /*# PRESSURETENSOR&PT_KIN */
#  if PRESSURETENSOR&PT_MOL
      StaAdd(string("PKin.%s [Pa]",xx),En.PKin[i]);
#  endif /*# PRESSURETENSOR&PT_MOL */
#  if PRESSURETENSOR&PT_MOM
      if (i<3) {
        StaAdd(string("PKin2.%s [Pa^2]",xx),En.PKin2[i]*Sqr(Punit/box.V));
        StaAdd(string("PKin3.%s [Pa^3]",xx),En.PKin3[i]*Cub(Punit/box.V)); }
#  endif /*# PRESSURETENSOR&PT_MOM */
#  if (PRESSURETENSOR&PT_ANY) == PT_ANY
      StaAdd(string("Ptens.%s [Pa]",xx),En.Ptens[i]);
#  endif /*# (PRESSURETENSOR&PT_ANY) == PT_ANY */
    }
#endif /*# PRESSURETENSOR&PT_ANY */

  if (No.f_in) En.T_in=En.kin_in/No.f_in;
  if (No.f_tr>0) En.T_tr=En.kin_tr/No.f_tr;
} /* measureconstraints */

/*
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%                             constraint dynamics                             %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Formulae:
% ^^^^^^^^^
% M[a,b] = SUM{i} grad[i]c[a].grad[i]c[b] / m[i]
%
% G[a] = SUM{i} p[i].grad[i]c[a] / m[i]         (Hamilton, removed in V 2.6d)
%
% G[a] = SUM{i} F[i].grad[i]c[a] / m[i]
%      + SUM{i,j} V[i].grad[i]grad[j]c[a].V[j]  (Lagrange formalism)
%
% where constraint c[a] = (r[ia]-r[ja])**2/2 - BondLength**2/2
% ( ia = si[a].pair[0], ja = si[a].pair[1], BondLength**2 = si[a].bondq )
%
% F[i] = -grad[i]U is force, V[i] is velocity, . is dot (scalar) product
%
% Eq. for Lagrange multipliers: (note: g(Lagrange) = dg(Hamilton)/dt )
% ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
% SUM{b} M[a,b] g[b] = G[a]
%
% Eqs. of motion:
% ^^^^^^^^^^^^^^^
% m[i] dr[i]/dt = p[i] - SUM{a} g[a]grad[i]c[a]
% dp[i]/dt      = F[i] + SUM{a}SUM{j}g[a]grad[i]grad[j].dr[i]/dt (Hamilton)
%
% m[i] ddr[i]/dt/dt = F[i] - SUM{a} g[a]grad[i]c[a]              (Lagrange)
%
% Using matrix M to correct constraints: (see Lcorrect)
% ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
% Dr[i] = SUM{b} e[b] 1/m[i]*grad[i]c[b]
% where  e  solves
% SUM{b} M[a,b] e[b] = -c[a]
%
% Velocity constraints in the Lagrangian formalism:
% Dv[i] = SUM{b} ve[b] 1/m[i]*grad[i]c[b]
% where  ve  solves
% SUM{b} M[a,b] ve[b] = -SUM{i}v[i]*grad[i]c[a]
%
% Thermostats (Lagrange only):
% ^^^^^^^^^^^^^^^^^^^^^^^^^^^^
% Friction:
%   ddr[i]/dt/dt -= v[i] ln(Tkin/T)/2/tau.T
% Nose:
%   ddr[i]/dt/dt -= v[i] dlogs/dt
%   ddlogs/dt/dt = (Tkin/T-1)/tau.T^2
% Since ddr[i]/dt/dt is a `friction' force, the corresponding term appears
% also in G[]
%
% #define PRECOND :
%   uses pre-conditioned matrix M for conjugate gradients:
%     SUM{b} M*[a,b] g*[b] = G*[a]
%   where M*=diag.M.diag, G*=diag.G and the solution is g=diag.g*
%   ( . is matrix multiplication)
%   and diag=diag(M)^(-1/2)
%   It appears that about 2 iterations are spared but the control operations
%     cost about the same.
%   Not implemented for constraints corrections (option -c4)
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
*/

/*** Lagrange equations of motion *********************** constraintdynamics */
void constraintdynamics(ToIntPtr B, ToIntPtr A, ToIntPtr V)
{
  int n,a,b,nc,ns,sp,i,j,k;
  static int irhs=0;
  vector *v,*f,*r,*ra=NULL,*rp=A->rp,*rpvel=NULL;
#ifdef PRECOND
  double *diag;
#endif /*# PRECOND */
  molecule_t *mn;
  vector auxv;
  /* warning: even #ifdef FLOAT, the constraint calculations are in double */
  double *g=NULL,*aux,*G=NULL;
  double **gg=NULL;
  double z, Vlogs=0, Vlogs_mol=0;
  M_t *M,*Msp;
  siteinfo_t *si;
  vector v_mol;
  double m_mol,mi;

  En.virc=0;

#ifdef ANCHOR
  if (option('k')==0) {
    static int pass=0;
    /* NOTE: 1st pass not reported because called to obtain initial forces */
    if (pass++)
      ERROR(("ANCHOR by constraints (-k0) not supported with Gear+Lagrange\n\
*** use SHAKE (-m2) instead")) }
#endif /*# ANCHOR */

  if (No.pred) allocarray(gg,No.pred);

  if (No.maxc) {
    rallocarray(ra,No.maxc); /* = r[i]-r[j] for constraint a=(i,j) */
    rallocarray(G,No.maxc);  /* rhs of constraint equation */
#ifdef PRECOND
    rallocarray(diag,No.maxc);
#endif /*# PRECOND */
    if (No.pred==0) rallocarrayzero(g,No.maxc); }

  /* ! all kin.energies (En/kin, En.kin_tr, En.kin_in) are doubled here !*/
  switch (thermostat) {
    case T_NOSE: /* Nose */
      Vlogs=V->logs; break;
    case T_BERENDSEN: /* friction thermostat: must know En.kin=twice kin.en. in advance! */
      if (En.kin==0) Vlogs=0;
      else Vlogs=blogT(En.kin/No.f,T)/tau.T;
      break;
    case T_LANGEVIN: case T_LANGEVIN_CM:
      Vlogs=1./tau.T;
      break;
    case T_TR: /* friction thermostat - intermolecular only */
      Vlogs_mol=blogT(En.kin_tr/No.f_tr,T)/tau.T;
      break;
    case T_IN: /* friction thermostat - intramolecular only */
      Vlogs=blogT(En.kin_in/No.f_in,T)/tau.T;
      break;
    case T_FRICTIONS: { /* decoupled friction thermostats, T_tr/T_in=T_tr_in */
      double T_tr=(No.f_in+No.f_tr)*T/(T_tr_in*No.f_in+No.f_tr);
      double T_in=(No.f_in+No.f_tr)*T/(No.f_in+No.f_tr/T_tr_in);

      Vlogs=blogT(En.kin_in/No.f_in,T_in)/tau.T;
      Vlogs_mol=blogT(En.kin_tr/No.f_tr,T_tr)/tau.T; } }

  Vlogs_mol-=Vlogs;
  rpvel=V->rp;

  /*** constraints: computing M and P ***/
  loop (n,FROM,No.N) { /* <<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<< */
    mn=molec+n;
    sp=mn->sp;
    ns=mn->ns;
    si=spec[sp]->si;
    r=rof(mn,rp);
    v=rof(mn,V->rp);
    f=rof(mn,B->rp);

    if ( (nc=mn->nc) ) {

      if (No.pred) {
        loop (k,0,No.pred) gg[k]=gs[k]+mn->ig;
        g=gg[0]; }

      M=Msp=Ms[option('c')&1 ? n : sp];

      loop (a,0,nc) {
        i=si[a].pair[0]; j=si[a].pair[1];
        VVV(ra[a],=r[i],-r[j])

        M->Mab = (si[i].imass+si[j].imass)*SQR(ra[a]); /* a==b */
        while ( (b=(++M)->b) < a )
          M->Mab = M->imass*SCAL(ra[a],ra[b]);

        VVV(auxv, =si[i].imass*f[i], -si[j].imass*f[j])
        G[a] = SCAL(auxv,ra[a]);
        VVV(auxv, =v[i], -v[j])
        G[a] += SQR(auxv);
        if (tau.T) G[a] -= 2*Vlogs*SCAL(auxv,ra[a]);

        /*** predictor of Lagrange multipliers g ***/
        if (No.pred>0) {
          z=0;
          loop (i,0,No.pred) z+=binom[i]*gg[i][a];
          gg[No.pred-1][a]=z; }
      } /* a */

      if (No.pred) g=gg[No.pred-1];
      else memset(g,0,nc*sizeof(g[0]));

#ifdef PRECOND
      /* find the diagonal */
      if (option('c')&1) Error("-c&1 not allowed with PRECOND");
      /* ...should pass diag to Lcorrect (norm) and use it in the same way */
      M=Msp;
      loop (a,0,nc) {
        diag[a]=1./sqrt(M->Mab); /* a==b */
        while ( (b=(++M)->b) < a ); }

      /* calculating diag.M.diag and diag.P */
      M=Msp;
      loop (a,0,nc) {
        M->Mab=1; /* a==b */
        G[a]*=diag[a];
        g[a]/=diag[a]; /* to fix the predictor of g */
        while ( (b=(++M)->b) < a ) M->Mab*=diag[a]*diag[b]; }
#endif /*# PRECOND */

      /*** solving the eqs. for Lagrange multipliers ***/
      constrit[sp].nit[LAGR_MULT] += omega==0
        ? conjgrad  (nc,g,Mvec,Msp,G, irhs<No.pred && eps>=1?1e-12:eps)
        : directiter(nc,g,Msp,G,omega,irhs<No.pred && eps>=1?1e-12:eps);

      //      prt("%g %g %g  %g %g %g Gg",VARG(G),VARG(g));

      // DEBUG
      //       if (n==0) prt_("\n gamma");
      //       loop (a,0,nc) prt_(" %.12g",g[a]);
      //       if (n!=0) loop (a,0,6) prt_("  %g %g %g",VARG(cfg[0]->rp[a]));

#ifdef PRECOND
      loop (a,0,nc) g[a]*=diag[a];
#endif /*# PRECOND */

    } /* if (nc) */

    /*** equations of motion ***/
    loop (a,0,nc) {
      i=si[a].pair[0]; j=si[a].pair[1];

      /*** forces equiv to constraint motions, virial ***/
      VV(auxv, =g[a]*ra[a])
      VV(f[i], -=auxv)
      VV(f[j], +=auxv)
      En.virc -= g[a]*si[a].bondq;
#if PRESSURETENSOR&PT_VIR
      En.Pvir[0] -= g[a]*ra[a][0]*ra[a][0];
      En.Pvir[1] -= g[a]*ra[a][1]*ra[a][1];
      En.Pvir[2] -= g[a]*ra[a][2]*ra[a][2];
#  if PRESSURETENSOR&PT_OFF
      En.Pvir[3] -= g[a]*ra[a][1]*ra[a][2];
      En.Pvir[4] -= g[a]*ra[a][0]*ra[a][2];
      En.Pvir[5] -= g[a]*ra[a][0]*ra[a][1];
#  endif /*# PRESSURETENSOR&PT_OFF */
#endif /*# PRESSURETENSOR&PT_VIR */
    } /*a*/

    /*** thermostats & acceleration from forces (kin. energy now in rhs) ***/

    if (thermostat==0 || tau.T==0)
      /*** no thermostat ***/
      loop (i,0,ns)
        VO(f[i], *=si[i].imass)

    else if (thermostat<=T_NOSE)
      /*** Nose, NPT, or simple Berendsen (friction) thermostat ***/
      loop (i,0,ns)
        VVV(f[i], =si[i].imass*f[i], -Vlogs*v[i])

    else if (thermostat==T_LANGEVIN)
      /*** atom-based Langevin thermostat ***/
      /* WARNING: OK with long tau.T only */
      loop (i,0,ns) {
        double x=sqrt(2*T*si[i].imass/(tau.T*h)); /* TO BE OPTIMIZED */

        VVVO(f[i], =si[i].imass*f[i], -Vlogs*v[i], +x*rndgauss()) }

    else if (thermostat==T_LANGEVIN_CM) {
      /*** molecule-based Langevin thermostat ***/
      /* WARNING: OK with long tau.T only */
      double x=sqrt(2*T/(spec[sp]->mass*tau.T*h)); /* TO BE OPTIMIZED */
      vector rf;

      VO(rf,=x*rndgauss())
      loop (i,0,ns)
        VVVV(f[i], =si[i].imass*f[i], -Vlogs*v[i], +rf) }

    else if (thermostat>=T_TR) {
      /*** decoupled inter/intra thermostats ***/
      VO(v_mol,=0)
      m_mol=0;
      loop (i,0,ns) {
        mi=si[i].mass;
        m_mol+=mi;
        VV(v_mol,+=mi*v[i]) }

      VO(v_mol,/=m_mol) /* velocity of CM of a molecule */

      loop (i,0,ns)
        VVVV(f[i], =si[i].imass*f[i], -Vlogs*v[i], -Vlogs_mol*v_mol) }

  } /* n >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>> */

  if (No.pred) {
    /*** dirty trick: pointer shuffling ***/
    k=No.pred-1; aux=gs[k];
    while (k) { gs[k]=gs[k-1]; k--; }
    gs[0]=aux;
    free(gg); }

  if (No.maxc) release(ra);

  /* final energies and statistics: repeated in Shake */
  box.V=PROD(box.L);
  if (measure==1) measureconstraints();

  /* NB: En.Pkinq=1 assumed for NPT/MTK */
  /* this En.Pcfg not needed anyway since Gear+NPT/MTK not implemented (yet) */
  En.Pcfg = ((En.kin*En.Pkinq+En.vir)/DIM+En.corr/box.V)/box.V;

  En.T=En.kin/No.f;
  /* ! for historical reasons, En.kin_tr and En.kin_in are still doubled! */
  En.kin/=2;
#ifdef POLAR
  En.pot += En.self;
#endif /*# POLAR */
  En.pot += En.el; // HINT: this should be moved to forces()
  En.U = En.Unc = En.kin+En.pot;
#ifndef FREEBC
  En.U += En.corr/box.V;
#endif
  En.tot = En.U; /* cutoff corrections added in V2.7 because of the barostat */

  /*.....if (measure) fprintf(stderr,"%12.2f %12.2f %12.2f %12.2f %12.2f %12.2f\n", En.tot,En.kin,En.pot,En.el,En.intra,En.self);*/

  if (thermostat==T_NOSE) {
    En.ext = No.f*T*(A->logs+Sqr(tau.T)/2*Sqr(Vlogs));
    En.tot += En.ext;
    En.logs=A->logs;
    En.dlogs=Vlogs;
    B->logs = (En.T/T-1)/Sqr(tau.T); }
#ifdef PRTKIN
  prt_("T_in = %.2f",En.kin_in/No.f_in);
  prt("   T_tr = %.2f",En.kin_tr/No.f_tr);
#endif /*# PRTKIN */

  irhs++;

#ifdef COOK
  if (FROM) /* to keep 1st FROM molecules on place */
    loop (n,0,FROM) {
      mn=molec+n;
      ns=mn->ns;
      memset(rof(mn,V->rp),0,ns*sizeof(vector));
      memset(rof(mn,B->rp),0,ns*sizeof(vector)); }
#endif /*# COOK */

  CPUtime("constr");

#ifdef POLAR
  if (option('p')/10%10==0) {
    if (init>=0 && init<3)
      ERROR(("Gear + Car-Parrinello NOT SUPPORTED"))
    else
      WARNING(("Gear + Car-Parrinello: BRUTE FORCE INITIALIZATION"))

    loop (n,0/*FROM ?*/,No.N) {
      mn=molec+n;
      sp=mn->sp;
      ns=mn->ns;
      si=spec[sp]->si;
      r=polarrof(mn,rp);
      v=polarrof(mn,V->rp);
      f=polarrof(mn,B->rp);
      loop (i,0,ns) VVVO(r[i],=v[i],=f[i],=0); } }
#endif /*# POLAR */

} /* constraintdynamics */


/*
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%                                   SHAKE                                     %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
*/

#ifndef SHAKE
#  error SHAKE is not #defined, should be one of {-1,-2,1,2}
#endif /*# SHAKE */

void Shake(double eps, double prob) /* =============================== Shake */
/*
  Verlet integration (equivalent to leap-frog)
  with SHAKE constraint dynamics for bonds and similar external constraints
  eps = accuracy (if eps<1)
      = # of Shake iteration sweeps over all sites (if eps>=1, |SHAKE|=1 only)
  omega (see constrd.omega) = relaxation parameter (omega=1 is direct iteration)
  data:
    rof(molecule *,cfg[0]->rp) = r(t)
    rof(molecule *,cfg[1]->rp) = r(t)-r(t-h) = h*v(t-h/2)
    rof(molecule *,cfg[2]->rp) = forces; used to keep r(t+h) during shaking

  POLAR:
    polarrof(molecule *,cfg[0]->rp) = dr(t) (rel.aux.site of ind. dipole)
    polarrof(molecule *,cfg[1]->rp) = dr(t-h)
    polarrof(molecule *,cfg[2]->rp) = used in forces calculation

  Compile-time versions:
    SHAKE>0: simplified formula, equivalent upto h^2,
             usu more efficient for complex systems
    SHAKE<0: `exact' formula using the scalar product,
             perhaps better for diatomics
    |SHAKE|=1: update in sweeps (generally better)
    |SHAKE|=2: update for unprecise bonds only
               (better for e.g. mixture of complex molecule + diatomics)

  History:
    2012 MTK barostat
    2011 Langevin (?)
    2010 Nose-Hoover, several VERLET versions
    2008 pressure tensor, automatic omega optimization
    2003 ANCHOR constraints, |SHAKE|=1 only
    2001 Maxwell/Andersen thermostat
    1995 revisited
    1992 first version
*/
{
  vector rij,pij;
  int al,n,i,j,sp,nc,ns=-1,it,maxit0=-16*log(eps*0.1);
  molecule_t *mn;
  vector *r,*p,*rp=cfg[0]->rp,*r1,*rp2=cfg[2]->rp,*v;
  siteinfo_t *si;
  double rxp,bondq,z,ze,mi,mj,sq;
  vector v_mol;
  vector v_mol1,v_mol2;
  double m_mol;
  double hh=h*h,Thh=T*hh,hhh=hh/2,hscale;
  double omega; /* = omegac/2 */
  int CMbasedthermostat=thermostat==T_ANDERSEN_CM || thermostat==T_MAXWELL_CM;

  /* Nose-Hoover; arrays are allocated forever
     indices: [0] = logs
              [1,2,3] = lambda (vector)
              [RPOFFSET..] = rp (RPOFFSET=4)
  */
  static double **Vhist; /* Vhist[0]=X(t)-X(t-h) (pointer passed); X=logs,lambda,r
                            Vhist[1]=X(t-h)-X(t-2h) ... */
  static int k; /* predictor extra length, actual value, growing up to No.pred */
  static double *b; /* coefficients */
  static double *Vpred; /* Vpred[0]=h*d X(t)/dt, predicted */
  static double Vfactor;
  static vector *vpred; /* vpred=pointer to predicted h*v[0] */
  static vector scale={1,1,1};

#if SHAKE==1 || SHAKE==-1
  int ifit=eps>=1;
  int maxit = (int)eps;
  double maxsq;
#else /*# SHAKE==1 || SHAKE==-1 */
  int *moved=NULL,*moving,done;
#endif /*#!SHAKE==1 || SHAKE==-1 */

#ifdef ANCHOR
  vector cf={0,0,0};
  int notinplace=0;
#endif /*# ANCHOR */


  t+=h;
  En.virc=0;

  if (thermostat==T_NPT)
    /* NB: box.L[j] is the box throughout cook
       cfg[0]->lambda is here calculated, integrated, and converted back to box.L[j] */
    loop (j,0,DIM) cfg[0]->lambda[j]=log(box.L[j]);
  VV(box.Lh,=0.5*box.L)
  box.V=PROD(box.L);

  box.cq=Sqr(box.cutoff);
  En.kin=En.kin_tr=0;
  if (thermostat==T_NPT) measure=1; /* virial pressure needed */

  if (option('c')&2) {
    En.r1=constrainterror(cfg[0],cfg[1]);
    En.v1=vconstrainterror/h; }
  depend_r(cfg[0],0);
#ifdef POLAR
  /* pred1.c is old version with -@%=A[0] -_%=A[1]: WARNING - options changed */
#  include "pred2.c"


  scf.nit=0;
  if (option('p')/10%10==0)
    mechpolar(cfg[2],cfg[0]);
  else do { /* iterate self-field */
#endif /*# POLAR */

    zeroEn();
    forces(cfg[2],cfg[0]);

#ifdef POLAR
  } while (!selffield(cfg[2],cfg[0],scf.eps,scf.omega,1));

  StaAdd("selffield iter",En.polnit=scf.nit);
  StaAdd("selffield maxdr",En.polmaxdr=scf.maxdr);
#endif /*# POLAR */

#if PRESSURETENSOR&PT_VIR
  if (measure) memset(PTvirc,0,sizeof(PTvirc));
#endif /*# PRESSURETENSOR&PT_VIR */

  depend_f(cfg[0],cfg[2]);

#ifdef SHEAR
  Shear(cfg[2],cfg[0],cfg[1],h);
#endif /*# SHEAR */

  /*** independent pressure measurement via <dU/dV>
       and errors in induced dipoles: pass 1 */
  if (measure) measureP(1);


  if (thermostat==T_NOSE || thermostat==T_NPT) {
    static int nose_k=-1,lastk=-1;

    if (nose_k>=0 && No.pred!=nose_k)
      ERROR(("SHAKE+Nose velocity predictor length unexpectedly changed (%d->%d)",nose_k,No.pred))

    if (nose_k<0 && No.pred==0)
      WARNING(("SHAKE+Nose without a predictor (option -p0) is imprecise"))
    nose_k=No.pred;

    if (!Vpred) {
      allocarray(b,No.pred+1); /* predictor coefficients */
      allocarray(Vpred,No.eq); /* predicted value */
      vpred=(vector*)(Vpred+RPOFFSET); /* atom velocities (->rp) start here */
      allocarray(Vhist,No.pred+1); /* the history */
      loopto (j,1,No.pred) allocarray(Vhist[j],No.eq);
      k=0; /* no history known now */ }

    Vhist[0]=&cfg[1]->logs; /* pointer passed - not allocated */

    /* generate predictor coefficients */
    if (k!=lastk) {
      b[0]=(double)(2*k+1)/(k+1);
      loop (j,0,k) b[j+1]=-b[j]*(double)(k-j)/(k+j+2);
      prt("velocity predictor for Verlet+SHAKE with Nose-Hoover, k=%d (max %d):",k,No.pred);
      loopto (j,0,k) prt("  b[%d]=%.15g",j,b[j]); }
    lastk=k;

    /* the predictor */
    loop (i,0,No.eq) {
      Vpred[i]=0;
      loopto (j,0,k) Vpred[i]+=b[j]*Vhist[j][i]; }

    if (thermostat==T_NOSE)
      Vfactor=Vpred[0];
    else if (thermostat==T_NPT) {
      /* predicting the change of background scaling (constraints do not change) */
      /* lambda[0] = Vpred[1], lambda[1] = Vpred[2], lambda[2] = Vpred[3] */
      if (rescale&RESCALE_PT) {
        /* anisotropic scaling */
        loop (j,0,3) if (rescale&(1<<j)) {
          if (k>1) scale[j]=exp(2*Vhist[0][j+1]-Vhist[1][j+1]);
          else scale[j]=exp(Vhist[0][j+1]); }
        else
          Vpred[j+1]=0,scale[j]=1;
        Vfactor=Vpred[0]+(Vpred[1]+Vpred[2]+Vpred[3])/No.f; }
      else {
        /* isotropic scaling  */
        Vfactor=Vpred[0]+No.NPT*Vpred[1];
        if (k>1) scale[0]=scale[1]=scale[2]=exp(2*Vhist[0][1]-Vhist[1][1]);
        else scale[0]=scale[1]=scale[2]=exp(Vhist[0][1]); } }

    /* increase k up to No.pred */
    k=min(No.pred,k+1);

    /* history shift */
    for (j=k; j; j--)
      loop (i,0,No.eq) Vhist[j][i]=Vhist[j-1][i]; }
#if VERLET==4
  else
    ERROR(("implementation limitation\n\
*** version VERLET==4 (predicted velocities) requires a Nose-Hoover thermostat"))
#endif /*# VERLET==4 */



  /* <<<<< loop over all molecules w. Verlet+SHAKE starts here <<<<< */

  loop (n,FROM,No.N) {
    mn=molec+n;
    sp=mn->sp;
    r=rof(mn,rp);
    p=rof(mn,rp2);
    v=rof(mn,vpred);
    r1=rof(mn,cfg[1]->rp);
    si=spec[sp]->si;
    nc=mn->nc;
    omega=constrit[sp].omega; /* HALF omegac, the relaxation parameter */

#if SHAKE>0 /* simplified and in general more efficient algorithm */
#  define CALCULATE_RXP rxp=bondq;
#else /* more complicated algorithm using angle between p & r */ /*# SHAKE>0 */
#  define CALCULATE_RXP rxp=SCAL(rij,pij) \
         if (rxp<bondq*1e-6) ERROR(( \
           "constraint failure\n" \
           "it=%d %d[%d-%d]:(%.3f %.3f %.3f)->(%.3f %.3f %.3f)", \
           it,n,i,j,pij[0],pij[1],pij[2],rij[0],rij[1],rij[2]));
#endif /*#!SHAKE>0 */

#if SHAKE==1 || SHAKE==-1 /* constraints updated in sweeps */
    ns=mn->ns;

#  ifdef ANCHOR
#    define FORCEUNIT (massunit*lengthunit/Sqr(timeunit))
#    define TORQUEUNIT (massunit*Sqr(lengthunit/timeunit))
#    define PRINTANCHOR(U,V) if (anchor.f) fprintf(anchor.f," %12g %12g %12g ", \
  (U)*V[0],(U)*V[1],(U)*V[2])
   /* measure force on center-of-mass and torque
      set zeros to cf of groups
      (independent of anchoring): */
#    include "anchorm.c"
#  endif /*# ANCHOR */

/* Verlet step (for 1 molecule) + possible Andersen/Maxwell thermostat */
#  include "verlet.c"

    if (!ifit) maxit=maxit0*(nc+4);

    it=0;
    do { it++;
      maxsq=0;

#  ifdef ANCHOR
     /* anchor site: */
#    include "anchors.c"
     /* anchor CM of group(s): */
#    include "anchorg.c"
#  endif /*# ANCHOR */

      loop (al,0,nc) {
        i=si[al].pair[0]; j=si[al].pair[1];
        sq=bondq=si[al].bondq;
        VVV(pij,=p[i],-p[j])
        VV(pij,*=scale) // INEFFICIENT - should optimize
        sq -= SQR(pij);
        Max(maxsq,fabs(sq/bondq))

        VVV(rij,=r[i],-r[j])
        CALCULATE_RXP
        mi=si[i].imass; mj=si[j].imass;
        sq *= omega/(mi+mj);
#  include "virc.c"
        sq /= rxp;
        VO(rij,*=sq)
        VV(p[i],+=mi*rij)
        VV(p[j],-=mj*rij) } /*al*/

      } while ((ifit || maxsq>eps) && it<maxit);

#  ifdef ANCHOR
    /* constrain CM, axes, and print site/cm constraint force */
#    include "anchora.c"
#  endif /*# ANCHOR */

   if (!ifit && it==maxit) {
     int ii;

     ERROR(("Shake: species=%d molecule=%d: too many iterations (%d)\n\
*** maxsq=%g eps=%g omega/2=%g\n\
*** (selecting (i)gnore in interactive run will dump the molecule)",
            mn->sp,n,it,
            maxsq,eps,omega))
     prt("%d\n",mn->ns);
     loop (ii,0,ns) prt("%4s %g %g %g",sitedef[spec[mn->sp]->si[ii].st].name,VARG(r[ii])); }

#else /*# SHAKE==1 || SHAKE==-1 */
/* SHAKE=+-2: info about moved sites is kept, eps<1 only */
#  ifdef ANCHOR
#    error ANCHOR not supported with this version of SHAKE
#  endif /*# ANCHOR */

    if (ns<mn->ns) {
      ns=mn->ns;
      if (moved!=NULL) { free(moved); free(moving); }
      allocarray(moved,ns);
      allocarray(moving,ns); }

/* Verlet step (for 1 molecule) + possible Andersen/Maxwell thermostat */
#  include "verlet.c"

    loop (i,0,ns) { moving[i]=0; moved[i]=1; }

    done=it=0;
    do {
      done=1;

      if (it>maxit0*(nc+4)) {
        int ii;

        ERROR(("Shake: species=%d molecule=%d: too many iterations (%d)\n\
*** maxsq=%g eps=%g omega/2=%g\n                                        \
*** (selecting (i)gnore in interactive run will dump the molecule)",
               mn->sp,n,it,
               maxsq,eps,omega))
        prt("%d\n",mn->ns);
        loop (ii,0,ns) prt("%4s %g %g %g",sitedef[spec[mn->sp]->si[ii].st].name,VARG(r[ii])); }

      loop (al,0,nc) { it++;
        i=si[al].pair[0]; j=si[al].pair[1];

        if (moved[i] || moved[j]) {

          sq=bondq=si[al].bondq;
          VVV(pij,=p[i],-p[j])
          VV(pij,*=scale) // INEFFICIENT - should optimize
          sq -= SQR(pij);

          if (fabs(sq) > eps*bondq) {

            VVV(rij,=r[i],-r[j])
            CALCULATE_RXP
            mi=si[i].imass; mj=si[j].imass;
            sq *= omega/(mi+mj);
#  include "virc.c"
            sq /= rxp;
            VO(rij,*=sq)
            VV(p[i],+=mi*rij)
            VV(p[j],-=mj*rij)

            moving[i]=moving[j]=1; done=0; }
          } /* i and j moved */
        } /*a*/
      loop (i,0,ns) { moved[i]=moving[i]; moving[i]=0; }
      } while (!done);

/*.....if (n==0) put(it)*/

#endif /*#!SHAKE==1 || SHAKE==-1 */

/*
  kinetic energy, pressure tensor, and final cfg[0], cfg[1] update
  four versions using different formulas for velocity
*/
#define V1 r1[i] /* [r(t)-r(t-h)]/h */
#if VERLET==0
#  include "shakev0.c"
#elif VERLET==1
#  include "shakev1.c"
#elif VERLET==2
#  include "shakev2.c"
#elif VERLET==3
#  include "shakev3.c"
#elif VERLET==4
#  include "shakev4.c"
#elif VERLET==5
#  include "shakev5.c"
#elif VERLET==9
#  include "shakev9.c"
#else
#  error "wrong value of compile-time switch VERLET (one of 0,1,2,3,4,5,9 expected)"
#endif
#undef V1

    if (nc) constrit[sp].nit[COR_BONDS] += it;
  } /*n*/
  cfg[0]->dep=0;
  /* >>>>>* loop over all molecules w. Verlet+SHAKE ends here >>>>> */

#ifdef POLAR
#  include "shakecp.c"
#endif /*# POLAR */

#if SHAKE==2 || SHAKE==-2
  if (moved!=NULL) { free(moved); free(moving); }
#endif /*# SHAKE==2 || SHAKE==-2 */

  /* TWICE the respective kinetic energies: */
#if VERLET==3
  hscale = 0.5/hh;
#elif VERLET==1
  hscale = 0.25/hh;
#else
  hscale = 1./hh;
#endif

  En.kin *= hscale;
  En.kin_tr *= hscale;

#if VERLET==2
#  define HSCALE (i<DIM?hscale:hscale/2)
#else /*# VERLET==2 */
#  define HSCALE hscale
#endif /*#!VERLET==2 */

#if PRESSURETENSOR&PT_KIN
  loop (i,0,PT_DIM) En.Pkin[i]*=HSCALE;

#  if VERLET==30
  /* cannot happen... */
  hscale*=4;
#  endif /*# VERLET==30 */

#  if PRESSURETENSOR&PT_MOL
  loop (i,0,PT_DIM) En.PKin[i]*=HSCALE;
#    if PRESSURETENSOR&PT_MOM
  loop (i,0,DIM) {
    En.PKin2[i]*=Sqr(hscale);
    En.PKin3[i]*=Cub(hscale); }
#    endif /*# PRESSURETENSOR&PT_MOM */
#  endif /*# PRESSURETENSOR&PT_MOL */
#endif /*# PRESSURETENSOR&PT_KIN */

  En.kin_in=En.kin-En.kin_tr;

#ifdef SHEAR
  /*
    systematic flow kinetic energy removed
    NOTE: in some cases, the rest of the kinetic energy is calculated later
  */
  En.kin-=En.kinshear;
  En.kin_tr-=En.kinshear;
#endif /*# SHEAR */

#ifdef PRTKIN
  prt_("T_in = %.2f",En.kin_in/No.f_in);
  prt("   T_tr = %.2f",En.kin_tr/No.f_tr);
#endif /*# PRTKIN */

  En.T=En.kin/No.f; /* NB: En.kin still doubled here ! */

  if (measure) measureP(2);

  /* SHAKE bond length correction -> virial of constraint forces */
  En.virc/=hh;
#if PRESSURETENSOR&PT_VIR
  loop (n,0,PT_DIM) PTvirc[n]/=hh;
#endif /*# PRESSURETENSOR&PT_VIR */

  if (measure==1) measureconstraints();

  if (thermostat==T_NPT) {
    vector oldlambda,ddlambda={0,0,0};

    if (rescale&RESCALE_PT) {
      /* NPT: 3 general (dependent or independent) scalings */
      En.ext = No.fx*T*cfg[0]->logs;

      /* Vpred[0] could be improved using newly calc. vel. */
      En.kinT=No.M_Th/hh*Sqr(Vpred[0]);
      En.kinP=No.M_Ph/hh*(Sqr(Vpred[1])+Sqr(Vpred[2])+Sqr(Vpred[3]));

      /* Verlet (=leap-frog here) for logs; NB: En.kin is doubled */
      cfg[1]->logs += hh*(En.kin + 2*En.kinP - No.fx*T)/No.M_T;
      cfg[0]->logs += cfg[1]->logs; /* only good for the total Hamiltonian */

      En.ext += En.kinT + En.kinP + box.V*(P/Punit);

      //     prt("%g %g %g %g %g %g  %g %g %g XXX",En.kin/2/119.8,En.kinT/119.8,En.kinP/119.8,En.pot/119.8,P/Punit*box.V/119.8,No.fx*T*cfg[0]->logs/119.8,box.L[0]/3.821984,box.L[1]/3.821984,box.L[2]/3.821984);
      //      prt("%g %g %g %g %g %g   XXX",VARG(3.3753914e-08*En.Pkin),VARG(3.3753914e-08*En.Pvir));

      En.logs = cfg[0]->logs;
      En.dlogs = cfg[1]->logs/h;

      /* NB: En.Pkinq here for consistency, however, should be 1 for NPT/MTK */
      /* En.Ptens is in Pa and without cutoff corrections */

      VV(oldlambda,=cfg[0]->lambda)
#if PRESSURETENSOR&PT_VIR
      loop (j,0,3) if (rescale&(1<<j))
        ddlambda[j]=hh*(box.V/Punit*(En.Ptens[j]-P)+En.kin/No.f+En.corr/box.V)/No.M_P - Vpred[0]*Vpred[j+1];
#endif
      if (rescale&RESCALE_XisYisZ) {
        ddlambda[0]=ddlambda[1]=ddlambda[2]=(ddlambda[0]+ddlambda[1]+ddlambda[2])/3;
        /* not needed, but increases precision */
        cfg[1]->lambda[0]=cfg[1]->lambda[1]=cfg[1]->lambda[2]=(cfg[1]->lambda[0]+cfg[1]->lambda[1]+cfg[1]->lambda[2])/3; }
      else if (rescale&RESCALE_XisY) {
        ddlambda[0]=ddlambda[1]=(ddlambda[0]+ddlambda[1])/2;
        /* not needed, but increases precision */
        cfg[1]->lambda[0]=cfg[1]->lambda[1]=(cfg[1]->lambda[0]+cfg[1]->lambda[1])/2; }
      loop (j,0,3) if (rescale&(1<<j)) {
        cfg[1]->lambda[j] += ddlambda[j];
        cfg[0]->lambda[j] += cfg[1]->lambda[j];
        scale[j]=exp(cfg[0]->lambda[j]-oldlambda[j]); }
      //      else cfg[1]->lambda[j]=0,scale[j]=1;

      box.V=rescalecfg(cfg[0],(rescale&RESCALE_XYZ)|RESCALE_L,scale[0],scale[1],scale[2]); }

    else { /* NPT: isotropic scaling */
      En.ext = No.fx*T*cfg[0]->logs;

      /* Vpred[0] could be improved using newly calc. vel. */
      En.kinT=No.M_Th/hh*Sqr(Vpred[0]);
      En.kinP=No.M_Ph/hh*Sqr(Vpred[1]);

      /* Verlet (=leap-frog here) for logs; NB: En.kin is doubled */
      cfg[1]->logs += hh*(En.kin + 6*En.kinP - No.fx*T)/No.M_T;
      cfg[0]->logs += cfg[1]->logs; /* only good for the total Hamiltonian */

      En.ext += En.kinT + En.kinP + box.V*(P/Punit);

      En.logs = cfg[0]->logs;
      En.dlogs = cfg[1]->logs/h;

      /* NB: En.Pkinq=1 assumed for NPT/MTK */
      En.Pcfg = ((En.kin*En.Pkinq+En.vir)/DIM+En.corr/box.V)/box.V;

      z=cfg[0]->lambda[0];

      cfg[1]->lambda[1]=
      cfg[1]->lambda[2]=
      cfg[1]->lambda[0] += hh*(box.V*(En.Pcfg-P/Punit)+En.kin/No.f)/No.M_P - Vpred[0]*Vpred[1];

      //    prt("%g %g %g TEST",(En.vir-En.virc)/3/box.V*Punit/29626194,En.vir/3/box.V*Punit/29626194,En.virc/3/box.V*Punit/29626194);

      cfg[0]->lambda[1]=
      cfg[0]->lambda[2]=
      cfg[0]->lambda[0] += cfg[1]->lambda[0];

      scale[2]=
      scale[1]=
      scale[0]=exp(cfg[0]->lambda[0]-z);

      /* NB: also box.L rescaled */
      box.V=rescalecfg(cfg[0],RESCALE_XYZ|RESCALE_L,scale[0],scale[1],scale[2]); } }

  else if (thermostat>=T_TR) {
    /* decoupled friction thermostats */
    z=ze=0;

    switch (thermostat) {
      case T_TR: /* friction thermostat - intermolecular only */
        ze=blogT(En.kin_tr/No.f_tr,T)*(h/tau.T);
        break;
      case T_IN:
        z=blogT(En.kin_in/No.f_in,T)*(h/tau.T);
        break;
      case T_FRICTIONS: {
        double T_tr=(No.f_in+No.f_tr)*T/(T_tr_in*No.f_in+No.f_tr);
        double T_in=(No.f_in+No.f_tr)*T/(No.f_in+No.f_tr/T_tr_in);

        z=blogT(En.kin_in/No.f_in,T_in)*(h/tau.T);
        ze=blogT(En.kin_tr/No.f_tr,T_tr)*(h/tau.T); } }

    ze-=z;
    loop (n,FROM,No.N) {
      si=spec[molec[n].sp]->si;
      r1=rof(&molec[n],cfg[1]->rp);
      ns=molec[n].ns;
      VO(v_mol,=0)
      m_mol=0;
      loop (i,0,ns) {
        mi=si[i].mass;
        m_mol+=mi;
        VV(v_mol,+=mi*r1[i]) }

      VO(v_mol,/=m_mol) /* h*velocity of CM of a molecule */

      loop (i,0,ns) VVV(r1[i], -=z*r1[i], +ze*v_mol) } }

  else if (thermostat==T_BERENDSEN) {
#ifdef SLAB
    if (slab.qT) {
      /* separate Berendsen (friction) thermostat for a slab */
      z=exp(blogT(En.T,T)*-h/tau.T);
      ze=exp(blogT(En.T,slab.T)*-h/tau.T);
      loop (n,FROM,No.N) {
        r1=rof(&molec[n],cfg[1]->rp);
        loop (i,0,molec[n].ns) {
          double Z=rof(&molec[n],cfg[0]->rp)[i][2]/box.L[2];
#  ifndef WALL
          Z=fmod(Z,1);
#  endif /*# WALL */
          if (Z>=slab.Tz0 && Z<slab.Tz1) VO(r1[i],*=ze)
          else VO(r1[i],*=z) } } }
    else
#endif /*# SLAB */
    {
      /* Berendsen (friction) thermostat */
      z=exp(blogT(En.T,T)*-h/tau.T); /* z=pow(En.T/T,-h/2/tau.T); */
      loop (n,FROM,No.N) {
        r1=rof(&molec[n],cfg[1]->rp);
        loop (i,0,molec[n].ns) VO(r1[i],*=z) } } }

  else if (thermostat==T_NOSE) {
    real ldlogs=cfg[1]->logs;

    /* Verlet (=leap-frog here) for logs */
    En.ext = cfg[0]->logs;
    cfg[1]->logs += hh*(En.T/T-1)/Sqr(tau.T);
    cfg[0]->logs += cfg[1]->logs; /* only good for the total Hamiltonian */

#if VERLET==0
    En.ext += Sqr(tau.T*ldlogs/h)/2;
#elif VERLET==1
    En.ext += Sqr(tau.T*(ldlogs+cfg[1]->logs)/h)/8;
#elif VERLET==2
    En.ext += Sqr(tau.T/h)*ldlogs*cfg[1]->logs/2;
#elif VERLET==3
    En.ext += Sqr(tau.T/h)*(Sqr(ldlogs)+Sqr(cfg[1]->logs))/4;
#elif VERLET==4
    En.ext += Sqr(tau.T*Vpred[0]/h)/2;
#elif VERLET==5
    En.ext += Sqr(tau.T/h)*(5./24*(Sqr(ldlogs)+Sqr(cfg[1]->logs))+1./12*ldlogs*cfg[1]->logs);
#elif VERLET==9
    En.ext += Sqr(tau.T*cfg[1]->logs/h)/2;
#endif
    En.ext *= No.f*T;
    En.logs = cfg[0]->logs;
    En.dlogs = cfg[1]->logs/h; }

  En.kin /= 2;

#ifdef POLAR
  En.pot += En.self;
#endif /*# POLAR */
  En.pot += En.el; // HINT: this should be moved to forces()

  En.U=En.Unc=En.kin+En.pot;
#ifndef FREEBC
  En.U+=En.corr/box.V;
#endif
  En.tot=En.U+En.ext; /* cutoff corrections added in V2.7 because of the barostat */

  //  if (measure) prt("%.9g %.9g %.9g %.9g NOSE",En.kin,En.pot,No.f*T*cfg[0]->logs,No.f*T*Sqr(tau.T*Vpred[0]/h)/2);

  if (measure) if (!finite(En.tot))
    ERROR(("numeric problem: En.kin=%g En.pot=%g",En.kin,En.pot))

#ifdef ANCHOR
  if (anchor.f) {
    if (notinplace) {
      fprintf(anchor.f," not in place or too large force/torque");
      fprintf(stderr,"ANCHOR: not in place or too large force/torque (%d times)\n",notinplace); }
    fprintf(anchor.f," %g\n",t); }
#endif /*# ANCHOR */

  CPUtime("Verlet+Shake");
} /* Shake */

#ifdef POLAR
#  include "scf.c"
#endif /*# POLAR */

#ifdef SHEAR
void Shear(ToIntPtr B, ToIntPtr A, ToIntPtr VH,double H) /************ Shear */
/*
  (1) adds forces f_i[x] = f_i[y] = m_i*Kf*cos(2 PI z_i/L[z])
      where Kf=Cf/sqrt(2)  (because forces are in x and y directions)
  (2) calculates sumfc = SUM_i f_i[x] cos(2 PI z_i/L[z])
  (3) calculates sumpc = SUM_i m_i (v_i[x]+v_i[y]) cos(2 PI z_i/L[z])
      and Kv = sqrt(2)*Cv = sumpc*Kf/sumfc
  (4) records  num = sumfc*No.mass*L[2]/(2*PI*PI*L[0]*L[1])
      and      den = sumpc
      Then, <viscosity> = <num>/<den>
  (5) also Cv=sumpc/sumfc*Cf/2 is recorded in the convergence profile
      (amplitude of the velocity profile)
  (6) Twice the systematic kinetic energy contribution from the flow is
      En.kinshear = Kv^2*sumfc/2/Kf

  NOTES:
  * VH is a product of velocities and H (normally H=1, for Shake H=h)
  * For Shake, the velocities are by h/2 delayed (does not matter here)
  * Shear should be included before any thermostat (the systematic term
      En.Ekinshear has to be subtracted form the total [and
      intermolecular] energy)
*/
{
  molecule_t *mn;
  siteinfo_t *si;
  int ns,sp,n,i;
  vector *r,*f,*v;
  double
    cosz,m,fx,
    rho,
    phi,Kf,
    aux,
    sumfc,sumpc,sumfx;
  static int pass;

  if (shear==0) return;
  if (shear<0) ERROR(("negative shear not supported"))

  phi=2*PI/box.L[2];
  if (fabs(box.L[1]/box.L[0]-1)>1e-6)
    ERROR(("Shear implementation limitation: L[0]!=L[1]"))

  /* eta*dT/dt/rho^2 in program units: */
  rho=No.mass*rhounit/(box.L[0]*box.L[1]*box.L[2]);
  aux=shear/Sqr(rho)*(massunit/Pow5(lengthunit)*Sqr(timeunit));

  /* Kf=Cf/sqrt(2) because in x and y directions */
  Kf=PI*sqrt(2*aux*No.f/(Cub(box.L[2])*box.L[0]*box.L[1]));

  if (pass==0) {
    double Cf=Kf*sqrt(2.);
    double viscunit=Punit*timeunit;

    pass++;
    put3(shear,Kf,Cf);
    put2(viscunit,No.mass);
    put2(box.L[2],box.L[0]*box.L[1]*box.L[2]);
    prt("expected heating rate in prog.units [K/ps, etc.]: dE/dt = %g/eta",
        sqr(Cf*No.mass*box.L[2]/4/PI)/(box.L[0]*box.L[1]*box.L[2])); }

  /*
     NOTE to the estimate of Cf [Kf]: we use the `kinetic' estimate of
     heat capacity = No.f/2 (No.f would be probably better)
     because the same estimate is used in friction thermostats;
     (the time-scaling errors cancel)
  */

  sumfc=sumpc=sumfx=0;

  /* shear force + measurements */
  loop (n,FROM,No.N) {
    mn=molec+n;
    ns=mn->ns;
    sp=mn->sp;
    si=spec[sp]->si;
    f=rof(mn,B->rp);
    r=rof(mn,A->rp);
    v=rof(mn,VH->rp);
    loop (i,0,ns) {
      m=si[i].mass;
      cosz=cos(phi*r[i][2]);
      fx=Kf*(m*cosz);
      sumfx+=fx;
      f[i][0]+=fx; f[i][1]+=fx;
      sumfc+=fx*cosz;
      sumpc+=(v[i][0]+v[i][1])*(m*cosz); } }

  /* correction to total acceleration in x,y directions
     (exact momentum conservation) */
  fx=sumfx/No.mass;
  loop (n,FROM,No.N) {
    mn=molec+n;
    ns=mn->ns;
    sp=mn->sp;
    si=spec[sp]->si;
    f=rof(mn,B->rp);
    loop (i,0,ns) {
      m=si[i].mass;
      f[i][0]-=fx*m; f[i][1]-=fx*m; } }

  sumpc/=H;

  if (measure) {
    StaSet(0,2,2,lag.n);
    StaAdd("shear:num",No.mass*box.L[2]/(2*PI*PI*box.L[0]*box.L[1])*sumfc);
    StaAdd("shear:den",sumpc); }
  En.Cv=sumpc/sumfc*Kf*0.7071067811865476;
  En.kinshear=Sqr(En.Cv)*sumfc/Kf; /* twice! */

  /* kinetic T contribution equivalent to global cos-like velocity profile */
  StaSet(0,2,2,0);
  StaAdd("Tkinshear",En.kinshear/No.f);
}
#endif /*# SHEAR */

struct constrd_s constrd;

void measureP(int pass) /***************************************** measureP */
/*
   Measures pressure and via selected diagonal terms of the pressure tensor
   using box rescaling (virtual volume change) in selected coordinates
   (see constrd.mode).  If densprof.slab&2 (=constrd.mode&RESCALE_SLAB),
   then x,y,z are rescaled so that volume does not change (cf. Gloor GJ,
   Jackson G, Blas FJ, de Miguel E: "Test-area simulation method for the
   direct determination of the interfacial tension of systems with continuous
   or discontinuous potentials", JCP, 123, 134703 (2005))
   POLAR: polar precision epspx must be small because of numerical derivative
     taken; also measures error of the self-field
   NOTES:
   - constrd.dV is in units of V (relative change)
   - uses two evaluations of forces (sparing one would mean to include this
     code to the place where the forces are normally calculated or exporting
     them which would be mess)
   - must work in 2 passes because Ekin need not be known while pass 1:
     measureP(1) : when positions are known
     measureP(2) : when Ekin is known
*/
{
#ifndef FREEBC
  static ToIntPtr locA=NULL,locB,locRun,locEx;
  static int warn;
#  ifdef POLAR
  int i,n;
  double Polstderr,Polmaxerr;
  FILE *frun=NULL,*fex=NULL,*ferr=NULL;
#  endif /*# POLAR */
  static double Pkin,vir;
  double Ekin,Ep,Em;

  box.V=PROD(box.L);

  En_t En0;
  vector L0,QM0;

#  ifdef DEBUG
  En_t Enp;
#  endif /*# DEBUG */

  if (!constrd.dV) return;

  box.cq=Sqr(box.cutoff);

  if (pass==1) {
    /* numerical derivative dU/dV */

    if (No.c && !(rescale&RESCALE_CM) && !warn++)
      WARNING(("%d constrained bonds and atom-based rescaling:\n\
*** pressure by virtual volume change will be wrong\n\
*** - for small molecules use CM-based rescaling: rescale=8|(more options)\n\
*** - for large molecules use -u9999 (vibrating bonds) and check the thermostat",No.c))

    En0=En;
    VV(L0,=box.L)
    VV(QM0,=Q->M)

    if (!locA) {
      sdsalloc(locA,cfg[0]->size);
      sdsalloc(locRun,cfg[0]->size); /* polar part would be enough...*/
      sdsalloc(locEx,cfg[0]->size); /* polar part would be enough...*/
      sdsalloc(locB,cfg[0]->size); }

    sdscopy(locA,cfg[0]);
    sdscopy(locRun,cfg[0]); /* polar part would be enough...*/

    rescalecfg(locA,constrd.mode,exp(constrd.dV/No.ncoord),0,0);
    if (!(constrd.mode&RESCALE_CM)) depend_r(locA,1); /* meaningfull for Rowlinson only */

#  ifdef POLAR
    scf.nit=0;
    do { /* iterate self-field */
#  endif /*# POLAR */
      zeroEn();
      forces(locB,locA);
#  ifdef POLAR
    } while (!selffield(locB,locA,scf.epsx,scf.omegax,0));
    StaAdd("selffieldx iter",scf.nit);
    StaAdd("selffieldx maxdr",scf.maxdr);

    Ep=En.pot+En.el+En.self;
    sdscopy(locEx,locA); /* polar part would be enough...*/
#  else /*# POLAR */
    Ep=En.pot+En.el;
#  endif /*#!POLAR */

#  ifdef DEBUG
    Enp=En;
#  endif /*# DEBUG */
    rescalecfg(locA,constrd.mode,exp(-2*constrd.dV/No.ncoord),0,0);
    if (!(constrd.mode&RESCALE_CM)) depend_r(locA,1); /* meaningfull for Rowlinson only */

#  ifdef POLAR
    scf.nit=0;
    do { /* iterate self-field */
#  endif /*# POLAR */
      zeroEn();
      forces(locB,locA);
#  ifdef POLAR
    } while (!selffield(locB,locA,scf.epsx,scf.omegax,0));
    StaAdd("selffieldx iter",scf.nit);
    StaAdd("selffieldx maxdr",scf.maxdr);

    Em=En.pot+En.el+En.self;
    Polmaxerr=Polstderr=0;

    if (option('v')&32) {
      char *fmode=init<2 || (init_append&2) ? "at" : "wt";

      frun=fopen(Fn("run.pol"),fmode);
      fex=fopen(Fn("ex.pol"),fmode);
      ferr=fopen(Fn("err.pol"),fmode);
      fprintf(frun,"\
#    t=%.4f  Running induced dipole moments in program units (0.01175014D)\n\
#    dip_x              dip_y              dip_z               |dip|\n",t);
      fprintf(fex,"\
#    t=%.4f  Iterated induced dipole moments in program units (0.01175014D)\n\
#    dip_x              dip_y              dip_z               |dip|\n",t);
      fprintf(ferr,"\
#    t=%.4f  Errors in induced dipole moments in program units (0.01175014D)\n\
#    dip_x              dip_y              dip_z               |dip|\n",t); }

    loop (n,FROM,No.N) {
      molecule_t *mn=molec+n;
      int ns=mn->ns;
      siteinfo_t *si=spec[mn->sp]->si;
      vector *rpolRun=polarrof(mn,locRun->rp);
      vector *rpolEx=polarrof(mn,locEx->rp);
      vector *rpolA=polarrof(mn,locA->rp);
      vector dd;
      double f;

      loop (i,0,ns) if (si[i].chargepol) {
        VV(rpolEx[i],+=rpolA[i])
        VO(rpolEx[i],/=2)
        VVV(dd,=rpolRun[i],-rpolEx[i])
        f=SQR(dd)*si[i].qqpol;
        if (frun) fprintf(frun,"%18.12f %18.12f %18.12f  %18.12f\n",
                          rpolRun[i][0]*si[i].chargepol,
                          rpolRun[i][1]*si[i].chargepol,
                          rpolRun[i][2]*si[i].chargepol,
                          sqrt(SQR(rpolRun[i])*si[i].qqpol));
        if (fex) fprintf(fex,"%18.12f %18.12f %18.12f  %18.12f\n",
                         rpolEx[i][0]*si[i].chargepol,
                         rpolEx[i][1]*si[i].chargepol,
                         rpolEx[i][2]*si[i].chargepol,
                         sqrt(SQR(rpolEx[i])*si[i].qqpol));
        if (ferr) fprintf(ferr,"%18.12f %18.12f %18.12f  %18.12f\n",
                          dd[0]*si[i].chargepol,
                          dd[1]*si[i].chargepol,
                          dd[2]*si[i].chargepol,
                          sqrt(f));
        Polstderr+=f;
        Max(Polmaxerr,f) } }

    if (frun) { fprintf(frun,"\n"); fclose(frun); }
    if (fex) { fprintf(fex,"\n"); fclose(fex); }
    if (ferr) { fprintf(ferr,"\n"); fclose(ferr); }

    StaSet(0,lag.err,2,lag.n);
    Polstderr=sqrt(Polstderr/No.pol);
    Polmaxerr=sqrt(Polmaxerr);
    StaAdd("Polar stderr",Polstderr);
    StaAdd("Polar maxerr",Polmaxerr);
#  else /*# POLAR */
    Em=En.pot+En.el;
#  endif /*#!POLAR */

#  ifdef DEBUG
    prt("DEBUG NB: E=En.pot+En.el+En.self, En.vir is without elst energy");
    prt("DEBUG V          =%16.6f",box.V);
    prt("DEBUG Punit/3/V  =%16.6f",Punit/3/box.V);
    //    prt("DEBUG r[0]       =%10.6f %10.6f %10.6f",VARG(rof(cfg,cfg[0]->rp)[0]));
    prt("DEBUG dE/dV      =%16.6f",(Ep-Em)/(2*constrd.dV));
    prt("DEBUG dEn.el/dV  =%16.6f",(Enp.el-En.el)/(2*constrd.dV));
    prt("DEBUG dEn.pot/dV =%16.6f",(Enp.pot-En.pot)/(2*constrd.dV));
    //    prt("DEBUG Ep        =%16.6f",Ep);
    //    prt("DEBUG Em        =%16.6f",Em);
    prt("DEBUG av.E       =%16.6f",(Em+Ep)/2);
    //    prt("DEBUG Enp.pot   =%16.6f",Enp.pot);
    //    prt("DEBUG En.pot    =%16.6f",En.pot);
    prt("DEBUG av.En.pot  =%16.6f",(En.pot+Enp.pot)/2);
    //    prt("DEBUG Enp.el    =%16.6f",Enp.el);
    //    prt("DEBUG En.el     =%16.6f",En.el);
    prt("DEBUG av.En.el   =%16.6f",(En.el+Enp.el)/2);
    prt("DEBUG av.En.vir  =%16.6f",(En.vir+Enp.vir)/2);
    //    prt("DEBUG av.En.sum  =%16.6f",(En.el+En.pot+En.self+Enp.el+Enp.pot+Enp.self)/2);
#    ifdef POLAR
    prt("DEBUG dEn.self/dV=%16.6f",(Enp.self-En.self)/(2*constrd.dV));
 /*.....   prt("DEBUG both*3:=%16.6f",-(En.el+En.self-Enp.self-Enp.el)/(2*constrd.dV)*3);*/
    prt("DEBUG av.En.self =%16.6f",(En.self+Enp.self)/2);
#    endif /*# POLAR */
#  endif /*# DEBUG */

    En=En0;

    VV(Q->M,=QM0)
    vir=-(Ep-Em)/(2*constrd.dV);

#  ifdef POLAR
    En.Polstderr=Polstderr;
    En.Polmaxerr=Polmaxerr;
#  endif /*# POLAR */

    VV(box.L,=L0)
    VV(box.Lh,=0.5*box.L) }

  else if (pass==2) {
    /* kinetic contribution */
    if (constrd.mode & RESCALE_CM) Ekin=En.kin_tr*En.PkinqCM;
    else Ekin=En.kin*En.Pkinq;

    if (constrd.mode & RESCALE_SLAB) Pkin=0;
    else Pkin=Ekin/(DIM*box.V)*Punit;

#  if PRESSURETENSOR & PT_KIN && PRESSURETENSOR & PT_MOL
    /*
       center-of-mass rescaling: use pressure tensor component, not total Ekin
       (is the same on average anyway...)
    */
    /* UNFINISHED - probably not important */
#  endif /*# PRESSURETENSOR & PT_KIN && PRESSURETENSOR & PT_MOL */

    En.PdVnc=Pkin+vir/box.V*Punit;
    En.PdV=Pkin+vir/box.V*Punit+constrd.Pcorr;
    StaSet(0,lag.err,2,lag.n);

    StaAdd(constrd.PdVname, En.PdV);
    StaSet(0,lag.err,2,2);
    StaAdd("PdVkin [Pa]",Pkin);
    StaAdd("dVvir",vir);
    StaAdd("PdVvir [Pa]",vir/box.V*Punit);

#  ifdef DEBUG
    prt("DEBUG En.PdV     =%16.6f",En.PdV);
    prt("DEBUG PdVkin     =%16.6f",Pkin);
    prt("DEBUG dVvir      =%16.6f",vir);
    prt("DEBUG PdVvir     =%16.6f",vir/box.V*Punit);
#  endif /*# DEBUG */
  }
  else
    ERROR(("measureP: bad pass=%d",pass))
#endif /*# FREEBC */
}

#ifdef POLAR
void measureepspol(void) /************************************ measureepspol */
/* Ewald needed! */
{
#  ifndef FREEBC
  static ToIntPtr locA=NULL,locB;
  vector M0,dM;
  int i,nit0;
  double E,aux;
  static int phase; /* x,-x,y,-y,z,-z */
  static int stride=0; /* not every cycle */

  /* save previous state */
  struct Eelst_s Eext0;
  En_t En0;
  vector QM0;

  if (scf.Estride && stride++%scf.Estride) return;

  En0=En;
  Eext0=Eext;
  VV(QM0,=Q->M)

  box.cq=Sqr(box.cutoff);

  if (!locA) {
    sdsalloc(locA,cfg[0]->size);
    sdsalloc(locB,cfg[0]->size); }

  sdscopy(locA,cfg[0]);

  scf.nit=0;
  scforces(locB,locA);
  nit0=scf.nit;

  VVV(M0,=dM,=Q->M)

  E=Eext.Eepspol*(1-2*(phase&1)); /* in prog. units, w. sign */

  Eext.E[phase/2]+=E;
  Eext.isE=1;

  scf.nit=0;
  scforces(locB,locA);

  VVV(dM,=Q->M,-M0)
  aux=4*PI/PROD(box.L)/E;
  if (option('v')&4) prt("%g %g %g  %c E=%g p.u. it=%d,%d HIGHFREQ",
      dM[0]*aux,dM[1]*aux,dM[2]*aux,
      'x'+phase/2,
      E,nit0,scf.nit);

  StaSet(0,2,2,0);
  StaAdd("hf:scf.nit(0)",nit0);
  StaAdd("hf:scf.nit(E)",scf.nit);

  StaSet(0,2,2,lag.n);
  En.chi_hf=dM[phase/2]*aux;
  StaAdd("high-frequency chi",En.chi_hf);
  aux=4*PI/(3*T*PROD(box.L));
  StaAdd("total chi",En.chi_hf+aux*SQR(M0));

  phase=(phase+1)%6;
  Eext=Eext0;
  /* to fix total chi (original Eext) */
  if (Eext.isE && scf.Estride>1) {
    aux=sqrt(aux);
    putv(Eext.E)
    loop (i,0,3) if (Eext.E[i]) StaAdd(string("xchi_%c",'x'+i),aux*M0[i]); }
  En=En0;
  VV(Q->M,=QM0)
#  endif /*# FREEBC */
}
#endif /*# POLAR */

#if 1
#include "jacobi.c"
#include "jacobins.c"

typedef struct sort_s {
  double l;
  int i;
} sort_t;

static int sortnmf(const void *a,const void *b)
{
  sort_t *A=(sort_t *)a;
  sort_t *B=(sort_t *)b;
  if (A->l<B->l) return -1;
  else return A->l>B->l;
}

/*                dr  eps  ampl modes frames zero key*/
#ifdef FREEBC
struct nm_s nm = {0, 1e-9, 0.3, 0,    7,     6,   0};
#else
struct nm_s nm = {0, 1e-9, 0.3, 0,    7,     3,   0};
#endif

/* no constraints, standard Jacobi for a symmetrized matrix */
#include "nm.c"
#include "nmc.c"

#endif
