/* (normally include file for simgear.c, declarations in simgear.h) */
/* see also gear2.c */
/* does not contain optimized code for GearOrder=4 */

#define MaxGearOrder 6

real GearCoef[MaxGearOrder-2][MaxGearOrder] = {
/*3*/ {0.0,      1.0,      1.0},
/*4*/ {1.0/6,    5.0/6,    1.0, 1.0/3},
#ifdef VELOCITY
/*5*/ {19.0/90,  0.75,     1.0, 0.5,     1.0/12},
/*6*/ {3.0/16,   251.0/360,1.0, 11.0/18, 1.0/6,  1.0/60},
#else
/*5*/ {19.0/120, 0.75,     1.0, 0.5,     1.0/12},
/*6*/ {3.0/20,   251.0/360,1.0, 11.0/18, 1.0/6,  1.0/60},
#endif
  };

void Gear2pol(int neq,int nself,int predorder,int predlength,ToIntPtr *a)
/* 
   neq=number of normal equations (excl. nself)
   nself=number of variables to predict (normally, 3*polar sites) 
   predorder = predictor order; 1=no prediction, 2=ASPC, 3=3rd order...
   predlength = additional predictor length (k for ASPC)
*/
{
  int i,j,k,bin;
  int history=predorder+predlength; /* history length used */
  static int knownhistory; /* at start less may be known */

#define MAXHISTORY 3 /* history of the predictor by max 3 longer */
  static real A[3][MAXHISTORY+1][7]={ /* A[predorder-2][predlength] */
    /* A[0]: predorder=2 (ASPC) */
    { {2,   -1},
      {2.5, -2,     0.5},
      {2.8, -2.8,   1.2,    -0.2},
      {3,   -24./7, 27./14, -4./7, 1./14} },
    /* A[1]: predorder=3 */
    { {3,    -3,      1},
      {2.4,  -1.2,    -0.8,  0.6},
      {1.94, -0.46/3, -1.18, 0.06, 1./3},
      {0 /* n.a. */} }, 
    /* A[2]: predorder=4 */
    { {4,     -6,     4,     -1},
      {10./3, -10./3, 5./3,  -2./3},
      {2.71,  -1.34,  -1.74, 1.16,  0.71, -0.5},
      {2.15,  -0.05,  -1.80, -0.30, 0.95, 0.55, -0.5 } } };

  real hh2=h*h/2,s,d,ss,*gc,*r,*hdr,*pred,*rhspred;
#ifdef VELOCITY
  real *preddr;
#endif

  static real *sf[MaxGearOrder+MAXHISTORY],*sfpred;
  static int nself0=-1;

  if (GearOrder<3 || GearOrder>MaxGearOrder) ERROR(("GearOrder=%d",GearOrder))
  gc=GearCoef[GearOrder-3];
  
  r=D(a[0]);
  hdr=D(a[2]);
  pred=D(a[GearOrder]);
#ifdef VELOCITY
  preddr=D(a[GearOrder+2]);
#endif
  rhspred=D(a[GearOrder+1]);

  /* Gear predictor */
  loop (i,0,neq) {
    s = 0;
    for (j=0; j<GearOrder; j++) s += (D(a[j]))[i];
    pred[i] = s;
#ifdef VELOCITY
    s = 0;
    loop (j,1,GearOrder) s += j*(D(a[j]))[i];
    preddr[i] = s/h;
#endif
    }

  /* predictor - self-consistent field */

  if (predorder>4 || predorder<1 || predlength>4 || predlength<0 || A[predorder-2][predlength]==0)
    ERROR(("polar predictor order=%d, predictor length=%d: no such method",predorder,predlength))

  if (predorder==1) return; /* no prediction */

  if (nself0>=0) if (nself!=nself0)
    ERROR(("# of polar sites changed from %d to %d",nself0,nself))

  loop (k,0,GearOrder) sf[k]=(D(a[k]))+neq;
  sfpred=pred+neq;
  if (knownhistory==0) {
    WARNING(("this module was changed, but not debugged"))
    loop (k,GearOrder,history) allocarrayzero(sf[k],nself);
    knownhistory=GearOrder;
    if (knownhistory<predorder)
      ERROR(("polar: requested predictor order (%d) > Gear order (%d)",predorder,GearOrder))
    nself0=nself; }

  loop (i,0,nself) {
    sfpred[i]=0;
    loopto (k,0,knownhistory)
      sfpred[i] += sf[k][i]*A[predorder-2][knownhistory-predorder][k]; }

  /* r.h.s. */
  t += h;

#ifdef VELOCITY
  rhs(a[GearOrder+1],a[GearOrder],a[GearOrder+2]);
#else
  rhs(a[GearOrder+1],a[GearOrder]);
#endif

  /* Gear corrector */
  loop (i,0,neq) {
    ss = hh2*rhspred[i];
    d = ss-hdr[i];
    loop (j,3,GearOrder) d -= j*(j-1)/2*(D(a[j]))[i];
    r[i] = pred[i] + d*gc[0];
#ifdef VELOCITY
    (D(a[1]))[i] = h*preddr[i] + d*gc[1];
#else
    s = d*gc[1];
    loop (j,1,GearOrder) s += j*(D(a[j]))[i];
    (D(a[1]))[i]=s;
#endif
    loop (j,3,GearOrder) {
      s = d*gc[j]; bin = 1;
      loop (k,j,GearOrder) {
	s += (D(a[k]))[i]*bin;
	bin = bin*(k+1)/(k+1-j); }
      (D(a[j]))[i] = s; }
    hdr[i] = ss; }

  /* store history : self-field */
  for (k=knownhistory-1; k>0; k--)
    copyarray(sf[k],sf[k-1],nself);
  copyarray(sf[0],sfpred,nself);

  knownhistory=min(knownhistory+1,history);
}

/*
MAPLE source to obtain all the 3,4-order predictor coefficients:
# A0 = A_1 of the paper, etc.
P:=3*d0-3*d1+d2+A0*(d0-3*d1+3*d2-d3)+A1*(d1-3*d2+3*d3-d4)+A2*(d2-3*d3+3*d4-d5);
A1:=0;
A2:=0;
A0:=-0.6; P;
A0:=-1.06; A1:=-1./3; P;

P:=4*d0-6*d1+4*d2-d3
   +A0*(d0-4*d1+6*d2-4*d3+d4)
   +A1*(d1-4*d2+6*d3-4*d4+d5)
   +A2*(d2-4*d3+6*d4-4*d5+d6);
A1:=0;
A2:=0;
A0:=-2/3; P;
A0:=-1.29; A1:=-0.5; P;
A0:=-1.85; A1:=-1.45; A2:=-0.5; P;
based on paper:
J. Kolafa
Numerical integration of equations of motion with a self-consistent field given by an implicit equation
Mol. Simul. 18, 193-212 (1996)
*/
