/*
  elementary site-site interactions: Lennard-Jones + r-space Ewald
  - nonpolar version
  - #included by interpot.c
  - used by the the all-pair version and the configuration initializer
    (the linked-cell list version uses ljqqnp.c instead, however, this
    module is still needed for the configuration initializer)

  WARNING: SS_NOMEASURE_rep, SS_MEASURE_rep not supported!

  NOTES:
    LJ cutoff must not be greater than the r-space Ewald cutoff
    virial is defined as  r*force = -r*u'(r)
    only non-electrostatic part of energy is returned by the functions, the
      r-space part of the electrostatic energy is summed to En.el
    virial of electrostatic forces is equal to the electrostatic energy and
       will be added to the total virial with the k-space part included
    only LJM and LJQQM needed with the link-cell method
*/

#ifdef SS_MEASURE_rep
#  error "SS_MEASURE_rep not supported in nonpolar version"
#endif /*# SS_MEASURE_rep */

#ifdef DUMP
/* to be turned on globally, affects internp, intrapol */
double Uelst0;
#define DUMP0 Uelst0=ENEL;
#define DUMP(A,B,C,D) dumpss(A,B,C,D);
void dumpss(char *info,vector r1,vector r2,double u)
{
  prt("%-7s: %3d-%-3d LJ=%-12g QQ=%-12g  %.7f %.7f %.7f  %.7f %.7f %.7f",
      info,
      (vector*)r1-cfg[0]->rp,(vector*)r2-cfg[0]->rp,
      u,ENEL-Uelst0,
      VARG(r1),VARG(r2));
}

#else
#define DUMP(A,B,C,D) /*void*/
#define DUMP0 /*void*/
#endif

double LJM(sitesite_t *ss,vector r1,vector r2,vector f1,vector f2 GLOB)
/* LJ site-site interaction, measurements incl.                     **** LJM */
{
  double U=0,rr,f,x,y,z;
  real fdr;
  rdf_t *ssrdf;
  vector dr;

  DUMP0

  NearestImage(0,box.cq) {
    if ( (ssrdf=ss->rdf) ) {
      int i=(int)(sqrt(rr)*ssrdf->grid);
      if (i<ssrdf->nhist) (ssrdf->hist[i])++; }
#ifdef NIBC
    SS_MEASURE
    ENVIR+=rr*f;
    MADDFORCES(f1,f2,f,dr)
#else /*# NIBC */
    if (rr < ss->C2q) {
      if (rr < ss->C1q)
        SS_MEASURE
      else {
        x=rr-ss->C2q; U=x*x*ss->A; f=ss->A4*x; }
      ENVIR+=rr*f;
      MADDFORCES(f1,f2,f,dr) }
#endif /*#!NIBC */
    }
  DUMP("LJM",r1,r2,U)
  MARK_PATCH
  return U;
} /* LJM */

void   QQM(sitesite_t *ss,vector r1,vector r2,vector f1,vector f2,double qq GLOB)
/* real-space charge-charge interactions, measurements incl.       ***** QQM */
{
  double rr,f;
  real fdr;
  rdf_t *ssrdf;
  ermacrodcl
  vector dr;

  DUMP0
  NearestImage(,box.cq) { SG
    if ( (ssrdf=ss->rdf) ) {
      int i=(int)(sqrt(rr)*ssrdf->grid);
      if (i<ssrdf->nhist) (ssrdf->hist[i])++; }
#ifdef COULOMB
    ENEL+=erud(rr)*qq; f=byerd*qq;
#else /*# COULOMB */
    ENEL+=f=qq/sqrt(rr); f/=rr;
#endif /*#!COULOMB */
    MADDFORCES(f1,f2,f,dr) }

  DUMP("QQM",r1,r2,0)
  MARK_PATCH
} /* QQM */

void rdfM(sitesite_t *ss,vector r1,vector r2) /************************ rdfM */
/* measuring rdf of sites not having q-q nor LJ (e.g. some water models) */
{
  double rr;
  rdf_t *ssrdf;
  vector dr;

  NearestImage(,box.cq)
    if ( (ssrdf=ss->rdf) ) {
      int i=(int)(sqrt(rr)*ssrdf->grid);
      if (i<ssrdf->nhist) (ssrdf->hist[i])++; }

} /* rdfM */

double LJQQM(sitesite_t *ss,vector r1,vector r2,vector f1,vector f2,double qq GLOB)
/* LJ + real-space charge-charge interactions, measurements incl. **** LJQQM */
{
  double U=0,rr,f,x,y,z;
  real fdr;
  rdf_t *ssrdf;
  ermacrodcl
  vector dr;

  DUMP0
  NearestImage(0,box.cq) { SG

    if ( (ssrdf=ss->rdf) ) {
      int i=(int)(sqrt(rr)*ssrdf->grid);
      if (i<ssrdf->nhist) (ssrdf->hist[i])++; }

#ifdef NIBC
    SS_MEASURE
    ENVIR+=rr*f;
#else /*# NIBC */
    f=0;
    if (rr < ss->C2q) {
      if (rr < ss->C1q)
        SS_MEASURE
      else {
        x=rr-ss->C2q; U=x*x*ss->A; f=ss->A4*x; }
      ENVIR+=rr*f; }
#endif /*#!NIBC */
#ifdef COULOMB
    ENEL+=erud(rr)*qq; f+=byerd*qq;
#else /*# COULOMB */
    ENEL+=x=qq/sqrt(rr); f+=x/rr;
#endif /*#!COULOMB */
    MADDFORCES(f1,f2,f,dr) }

  DUMP("LJQQM",r1,r2,U)
  MARK_PATCH
  return U;
} /* LJQQM */

#ifndef LINKCELL

#  ifdef COOK

double LJQQ14M(sitesite_t *ss,vector r1,vector r2,vector f1,vector f2,double qq GLOB)
                                                                   /* LJQQM */
/* as above with factor14 - requires special treatment */
{
  double U=0,rr,f,x,y,z,sr;
  real fdr;
  rdf_t *ssrdf;
  ermacrodcl
  vector dr;

  DUMP0
  NearestImage(0,box.cq) { SG
    f=0;
    sr=sqrt(rr);
    if ( (ssrdf=ss->rdf) ) {
      int i=(int)(sr*ssrdf->grid);
      if (i<ssrdf->nhist) (ssrdf->hist[i])++; }
#    ifdef NIBC
    SS_MEASURE
    ENVIR+=rr*f;
#    else /*# NIBC */
    if (rr < ss->C2q) {
      if (rr < ss->C1q)
        SS_MEASURE
      else {
        x=rr-ss->C2q; U=x*x*ss->A; f=ss->A4*x; }
       ENVIR+=rr*f; }
#    endif /*#!NIBC */

#    ifdef COULOMB
#      if COULOMB==2 || COULOMB==0
    qq*=factor14; ENEL+=erud(rr)*qq; f+=byerd*qq;
#      else /*# COULOMB==2 || COULOMB==0 */
    ENEL+=(erud(rr)+(y=factor14_1/sr))*qq; f+=(byerd+y/rr)*qq;
#      endif /*#!COULOMB==2 || COULOMB==0 */
#    else /*# COULOMB */
    ENEL+=x=factor14*qq/sr; f+=x/rr;
#    endif /*#!COULOMB */

    MADDFORCES(f1,f2,f,dr) }

  DUMP("LJQQM14",r1,r2,U)
  MARK_PATCH
  return U;
} /* LJQQ14M */

void LJQQ14(sitesite_t *ss,vector r1,vector r2,vector f1,vector f2,double qq)
                                                                   /* LJQQM */
/* as above with factor14 - requires special treatment */
{
  double rr,f,x,y;
  real fdr;
  ermacrodcl /* warning: byerd not needed */
  vector dr;

  NearestImage(,box.cq) { SG

#    ifdef COULOMB
#      if COULOMB==2 || COULOMB==0
    f=erd(rr)*factor14*qq;
#      else /*# COULOMB==2 || COULOMB==0 */
    f=(erd(rr)+factor14_1/(rr*sqrt(rr)))*qq;
#      endif /*#!COULOMB==2 || COULOMB==0 */
#    else /*# COULOMB */
    f=factor14*qq/(rr*sqrt(rr));
#    endif /*#!COULOMB */

#    ifdef NIBC
    SS_NOMEASURE(+=)
#    else /*# NIBC */
    if (rr < ss->C2q) {
      if (rr < ss->C1q)
        SS_NOMEASURE(+=)
      else {
        x=rr-ss->C2q; f+=ss->A4*x; } }
#    endif /*#!NIBC */
    ADDFORCES(f1,f2,f,dr) }
} /* LJQQ14 */

#    ifdef COULOMB

double XQQM(vector r1,vector r2,vector f1,vector f2,double qq GLOB)  /* XQQM */
/* corrections to 1-2 and 1-3 interactions because of Ewald, measure */
{
#      if COULOMB==2 || COULOMB==0 /* cutoff electrostatics */
  return 0;
#      else /*# COULOMB==2 || COULOMB==0 */
  real fdr;
  double U=0,rr,f,x;
  /*rdf_t *ssrdf;*/
#        if COULOMB>=-1
  double y;
  ermacrodcl
#        endif /*# COULOMB>=-1 */
  vector dr;

  NearestImage(0,box.cq) { SG
    x=sqrt(rr);
    /* ?! do not calculate rdf */
#        if COULOMB==-2
    ENEL+=exacterud_sqrt(rr,&f)*qq; f*=qq;
#        else /*# COULOMB==-2 */
    ENEL+=(erud(rr)-(y=1/x))*qq; f=(byerd-y/rr)*qq;
#        endif /*#!COULOMB==-2 */
    MADDFORCES(f1,f2,f,dr) }
  MARK_PATCH
  return U;
#      endif /*#!COULOMB==2 || COULOMB==0 */
} /* XQQM */

void XQQ(vector r1,vector r2,vector f1,vector f2,double qq) /********** XQQ */
/* corrections to 1-2 and 1-3 interactions because of Ewald, no measure */
/* parameter ss never used ! */
{
#      if COULOMB==2 || COULOMB==0 /* cutoff electrostatics */
  return;
#      else /*# COULOMB==2 || COULOMB==0 */
  double rr,f;
  real fdr;
#        if COULOMB>=-1
  ermacrodcl /* warning: byerd not needed */
#        endif /*# COULOMB>=-1 */
  vector dr;

  NearestImage(,box.cq) { SG
#        if COULOMB==-2
    exacterud_sqrt(rr,&f); f*=qq;
#        else /*# COULOMB==-2 */
    f=(erd(rr)-sqrt(rr)/Sqr(rr))*qq;
#        endif /*#!COULOMB==-2 */
    ADDFORCES(f1,f2,f,dr) }

  return;
#      endif /*#!COULOMB==2 || COULOMB==0 */
} /* XQQ */

#    endif /*# COULOMB */

#  endif /*# COOK */

void LJQQ(sitesite_t *ss,vector r1,vector r2,vector f1,vector f2,double qq)
                                                                 /**** LJQQ */
/* LJ + real-space charge-charge interactions, no measurements */
{
  double rr,f,x,y;
  real fdr;
  ermacrodcl /* warning: byerd not needed */
  vector dr;

  NearestImage(,box.cq) { SG
#  ifdef COULOMB
    f=erd(rr)*qq;
#  else /*# COULOMB */
    f=qq/(rr*sqrt(rr));
#  endif /*#!COULOMB */
#  ifdef NIBC
    SS_NOMEASURE(+=)
#  else /*# NIBC */
    if (rr < ss->C2q) {
      if (rr < ss->C1q)
        SS_NOMEASURE(+=)
      else {
        x=rr-ss->C2q; f+=ss->A4*x; } }
#  endif /*#!NIBC */
    ADDFORCES(f1,f2,f,dr) }

  return;
} /* LJQQ */

void   QQ      (vector r1,vector r2,vector f1,vector f2,double qq) /**** QQ */
/* added 10/96 */
{
  double rr,f;
  real fdr;
  ermacrodcl /* warning: byerd not needed */
  vector dr;

  NearestImage(,box.cq) { SG
#  ifdef COULOMB
    f=erd(rr)*qq;
#  else /*# COULOMB */
    f=qq/(rr*sqrt(rr));
#  endif /*#!COULOMB */
    ADDFORCES(f1,f2,f,dr) }

  return;
}

void LJ(sitesite_t *ss,vector r1,vector r2,vector f1,vector f2) /******* LJ */
/* LJ site-site interaction, no measurements */
{
  double rr,f,x,y;
  real fdr;
  vector dr;

  NearestImage(,ss->C2q) {
#  ifdef NIBC
    SS_NOMEASURE(=)
#  else /*# NIBC */
    if (rr < ss->C1q)
      SS_NOMEASURE(=)
    else {
      x=rr-ss->C2q; f=ss->A4*x; }
#  endif /*#!NIBC */
    ADDFORCES(f1,f2,f,dr) }

  return;
} /* LJ */

#endif /*# LINKCELL */

