#define VERSION "2.7k"

#if defined(LINKCELL) && defined(FREEBC)
#  error "LINKCELL not supported for FREEBC"
#endif /*# defined(LINKCELL) && defined(FREEBC) */

#include "ground.h"
#define SDS
#include "alloc.h"
#include <time.h>

#include "options.h"
#include "optlist.c"

#include <signal.h>

#include "varfile.h"
#include "statics.h"
#include "linregr.h"

#include "rndgen.h"
#include "simglob.h"

#ifndef PARALLEL
#  define PARALLEL 0
#endif /*# PARALLEL */

#include "asksig.h"

#include "units.h"
#include "simgear.h"
#include "simcg.h"

#include "interpot.h"
#include "simdef.h"

/*
   %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
   %                       Ewald summation & erfc                          %
   %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
*/

#ifdef COULOMB
#  include "ewald.h"
#  include "elst.h"
#endif /*# COULOMB */

#include "simpot.h"
#include "siminit.h"
#ifdef DIHEDRAL
/* WARNING this is something very obsolete... */
#  include "dihedral.h"
#endif /*# DIHEDRAL */
#include "simmeas.h"

#include "setss.h" /* widomrdf() needed */

#include "norm.h"
#ifdef LINKCELL
#  include "linklist.h"
#endif /*# LINKCELL */

#include "simils.h"
#include "rhs.h"
#include "constrd.h"
#include "cputime.h"

#include "sfdx.h"

#ifdef MARK
#  include "mark.h"
#endif /*# MARK */

static char *email=NULL,*host,*arg0;
static FILE *lock;
static void removelock(void) /*********************************** removelock */
/*
  removes the lock file SIMNAME.loc
  registered by atexit() => to be called at exit()
*/
{
  if (lock) fclose(lock);
  remove(Fn("loc"));

  //  if (sig!=4 && sig!=2 && sig!=15)
  /* do not mail if stopped by .stp, kill-2, kill-15 */
  if (email)
    system(string("echo \"%s %s %s on %s finished\" | mail -s \"%s_finished\" %s",
                  arg0,simils.sysname,simils.simname,host,simils.simname,email));
}

static int justnow(double dt,double half) /************************* justnow */
{
  dt=fabs(dt);
  if (dt==0) return 0;

  return (int)((t+half)/dt+1e4) > (int)((t-half)/dt+1e4);
}

static int SI=1; /* 1 if SI units, 0 if kcal/mol etc. */
static double SIE=1;

static void StaAddE(const char *name,double x) /******************** StaAddE */
{
  char nm[128];

  strcpy(nm,name);
  if (SI) strcat(nm," [J/mol]");
  else strcat(nm," [kcal/mol]");

  StaAdd(nm,x/SIE);
}

int main(int narg,char **arg)
{
  static int
    noint=10,     /* number of steps in 1 cycle = frequency of measurements */
    no=10000,     /* # of cycles in 1 main cycle */
    nomax=0,      /* max # of cycles from start of measurements (init>=1) */
    initno=0,     /* option -NUMBER (overrides no in the 1st cycle) */
    quit=0,       /* quit=1 causes program stop (deprecated) */
    key=0,        /* "command" */
    sort=0,       /* sort by CM, coordinate: 1=x, etc. (deprecated) */
    fracpass=0,   /* turned off by any option q and in further cycles */
    pass=0,
    CPnbit=0,     /* resolution (1/2^CPnbit) for packed convergence profiles */
    rhoreached,   /* switch to check whether rho was reached */
    iarg,nsteps,initvel=0,plbopened=0,issta=0,frame0,expectedbuffer;

  static volatile int
    icyc=0,iint=0;

  static double
    stop=0,       /* max time from simulation start (init>=2) */
    Emax=1e4,     /* max energy of inserted molecule - see initcfg */
    drmax=0.125,  /* max displacement in one integration step (init>2 only)
                     drmax<0: applies for any init */
    bulkmodulus=0, /* for NPT */
    maxscale=1.03,/* max scaling of box, applies with tau.P, tau.R, (tau.rho) */
    pins=0.01,    /* limit probability of insertions - see initcfg */
    initrho=0,    /* initial density - see initcfg; 0=rho */
    dV=0,         /* volume-change and surface-change, cf. constrd.dV */
    DT,           /* cycle in s */
    minmass=0,    /* to check h */
    LJcutoff=-3;  /* Lennard-Jones cutoff - see comments in simdef.c! */

  static struct {
    double
      grid,       /* intervals/1A; negative:s-s rdf's merged; 0:none */
      cutoff;     /* range of rdf */
    int
      onefour;    /* 1 if 1-4 are to be included to RDF's, 0 otherwise */
  } rdf = {0,0,1};

  static struct {
    double
      prt, /* frequency of printing t,T,E,... to out [in ps] (default=every) */
      cfg, /* save configuration frequency [in ps] - see savecfg */
      plb; /* frequency of saving playback file */
  } dt = { 1e-4,0,0 };

  static vector shift; /* added to cfg. before run, then :=0 */
  static vector Eelst; /* external elst field, in V/m */

#ifdef WIDOM
  static struct {
    int spreal; /* for scaled particle (id change): the real species */
    int sp; /* species to insert or be changed to virtually */
    int n; /* # of attempted insertions per cycle [WALL: and slab]; negative: excl. wall-mol pot. */
#  ifdef SLAB
    int mode; /* also symmetrized profile */
    double z0,z1,dz; /* define slabs */
#  endif /*# SLAB */
    double corr; /* cutoff correction without division by V */
  } widom={-1,0,0,2};
#endif /*# WIDOM */

#ifdef XSECTION
  static struct {
    double grid;   /* grid of the mesh, in points per A */
    double RvdW;   /* van der Waals radius of the test particle, additive */
    double Rscale; /* multiply van der Waals radii; default=2^(-1/6) */
    int mode;      /* 0..4, see sfdx.c */
    int freq;      /* how often to calculate, in # of cycles (1=every cycle) */
    int cp;        /* -1: averaged xsec in CP, >=0 : 1 mol value */
    int ns;        /* max # of sites incl. in Xsec calculations; -1=all */
    int sizelimit; /* max size of the grid array */
    int maxsize;   /* that one actually encountered */
  } xs={10.0,1.4,0.89089871814033927,0,0,0,-1,64*1024*1024,0};
#endif /*# XSECTION */

#ifdef RGYR
  static struct {
    int end[2];   /* indices of `ends' */
    int cp;       /* -1: averaged Rgyr and end-to-end in CP, >=0 : 1 mol.val. */
  } rg={{0,-1},0};
#endif   /*# RGYR */

#ifdef DIHHIST
  static struct {
    double res;  /* resolution, in deg */
    int grid;   /* # of intervals / 360 deg */
    int cp,dcp;
  } dih={0,0,-1,0};
#endif /*# DIHHIST */

#ifdef DISTURB
  double disturb[2];  /* random perturbation of cfg. - see disturb() */
#endif /*# DISTURB */

  vector Pscale={1,1,1};
  double RvdW=0; /* particle size to adjust: see tau.R, tau.i */
  double lastEtot,halftcyc;

  char *mark;
  double starttime,stoptime,stop_start; /* see cputime.h, #define CHEAPTIME */
  static double rho=0; /* density [kg/m3] */
  static double refrho;
  static vector L; /* box size (in input data) */
  static double cutoff; /* cutoff (in input data) */
  static int corr=3+16;
    /* 1=cutoff corrections included in final results, not in P for the barostat
       2=cutoff corrections included (also in barostat)
       4=kinetic pressure finite size correction (atom-based)
       8=kinetic pressure finite size correction (molecule-based)
      16=4+8 used if not NPT/MTK */

  FILE *indef; /* to reread the def-file */

  /* keys for ID="mnemonics" on data and help moved to mainhlp.c */
#define INFO "*** " PROJECT " *** V" VERSION " *** (c) J.Kolafa 1991-2012 ***"
#include "mainhlp.c"

  out=stderr;

  /*** option analysis ***/
  arg0=arg[0];
  loop (iarg,1,narg)
    if (arg[iarg][0]=='-')
      if (arg[iarg][1]>='0' && arg[iarg][1]<='9')
        initno=atoi(arg[iarg]+1);
      else {
        if (arg[iarg][1]=='@') email=arg[iarg]+2;
        getoption(arg[iarg])
        if (arg[iarg][1]=='q') fracpass++; }
#if CHECKHEAP==-2
    /* to debug memory problems: option +AllocRange */
    else if (arg[iarg][0]=='+') {
/*.....    AllocSizeLim=atoi(arg[iarg]+1);*/
      AllocTrace++;
      AllocRange=atoi(arg[iarg]+1); }
#endif /*# CHECKHEAP==-2 */
    else {
      if (strlen(arg[iarg])>4 && !strcmp(strend(arg[iarg])-4,".plb")) simils.plbname=arg[iarg];
      else if (!simils.sysname) simils.sysname=arg[iarg];
      else if (simils.simname)
        ERROR(("third argument (no -option nor .plb), max two (SYSNAME and SIMNAME) allowed"))
      else simils.simname=arg[iarg]; }

  if (option('p')<0) ERROR(("negative option -p no longer supported"))
  No.pred=option('p')%10;
  if (No.pred==9) No.pred=option('m');

  if (option('s')==1) option('s')=256;

#if PARALLEL
  if (option('i')==-9) option('i')=1+(!option('s'));
  if (option('i')==0) WARNING(("option -i0 for PARALLEL"))
#  if PARALLEL==1 || PARALLEL==3
  if (sizeof(struct pll_global_s)%CACHELINE) DISASTER(("sizeof(pll_global_s)=%d is not an integer multiple of the cache line=%d\n\
*** adjust padding in `struct pll_global_s' in sim/interpot.h and recompile!",sizeof(struct pll_global_s),CACHELINE))
#  endif /*# PARALLEL==1 || PARALLEL==3 */
#else /*# PARALLEL */
  if (option('i')==-9) option('i')=2*!option('s');
#endif /*#!PARALLEL */

  if (option('f')) {
    //    if (!el.sf) el.sf=1; /* compatibility */
#ifdef POLAR
    option('a')=0;
    prt("option -a0 set because of -f");
#endif /*# POLAR */
  }

  if (!(option('m')==2) && option('c')&1) epsc=0.05;

  /* install SIGINT (2) and SIGTERM (15) signal handler */
  if (option('i')>=0) {
    signal(SIGINT,asksig);
    /* SIGTERM not for MPI ...? */
    signal(SIGTERM,asksig); }

  simils.rotatefrom=abs(option('j'));
  if (option('j')<0) option('j')=0;

  initscroll(option('s')*1024);

  if (!simils.simname && simils.sysname) {
    prt("WARNING: SIMNAME not specified, SIMNAME=SYSNAME will be used.");
    simils.simname=simils.sysname; }

  if (simils.simname) {
    lock=fopen(Fn("loc"),"r+");
    out=stdout;
    if (lock) {
      if (option('o')) ERROR(("Lock file \"%s\" exists\n\
(\"cook %s\" may be running or has not been stopped properly)",
            lastFn,simils.simname))
      else
        fprintf(stderr,"WARNING: Lock file exists but is ignored because of -o0\n"); }
    else
      lock=fopen(Fn("loc"),"w");

    atexit(removelock);
    remove(Fn("sig"));

    if (option('s')) {
      if (option('b')<0) option('b')=1; /* default beep level */
      out=stdout; }
    else {
      char *ext=option('f') ? "sfd" : "prt";
      char eee[12]="S";

      if (option('g') && option('f')) {
        ext=eee; sprintf(ext+1,"%d",option('g')); }
      else
        backup(ext);
      out=fopen(Fn(ext),"wt"); } }

  if (out==NULL) Error("open .prt");

  prts(INFO);

  prts_("***");

#ifdef FREEBC
  prts_(" FREEBC");
#endif /*# FREEBC */

#ifdef NIBC
  prts_(" NIBC");
#endif /*? FREEBC */ /*# NIBC */

#ifdef COULOMB
  prt_(" COULOMB=%d",COULOMB);
#endif /*# COULOMB */

#ifdef POLAR
  prt_(" POLAR=%d",POLAR);
#endif /*# POLAR */

#if PARALLEL
  prt_(" PARALLEL=%d",PARALLEL);
#endif /*# PARALLEL */

#ifdef REDUNITS
  prts_(" REDUNITS");
#endif /*# REDUNITS */

#ifdef PRESSURETENSOR
  prt_(" PRESSURETENSOR=%d",PRESSURETENSOR);
#endif /*# PRESSURETENSOR */

#ifdef SLAB
  prt_(" SLAB=%d",SLAB);
#endif /*# SLAB */

#ifdef LINKCELL
  prts_(" LINKCELL");
#endif /*# LINKCELL */

  prt_(" VERLET=%d",VERLET);
  prt(" SHAKE=%d",SHAKE);

  { /* options, sex, environment */
    int Int=1,i;
    char *end=(char*)&Int;
    char *user=getenv("USERNAME");
    host=getenv("HOSTNAME");

    if (!host) host=getenv("HOST");
    if (!host) host="UNKNOWN";
    if (!user) user=getenv("USER");
    if (!user) user="UNKNOWN";

    prt("*** %s endian short=%d int=%d long=%d pointer=%d ***",
        *end==1?"little":*end==0?"BIG":"?",
        sizeof(short),sizeof(int),sizeof(long),sizeof(void*));

    loop (i,0,narg) {
      prt_("%s%c",arg[i]," \n"[i==narg-1]);
      fprintf(stderr,"%s%c",arg[i]," \n"[i==narg-1]); }
    prt("HOST=%s USER=%s PWD=%s",host,user,getenv("PWD"));
    fprintf(stderr,"HOST=%s USER=%s PWD=%s\n",host,user,getenv("PWD"));
    prt("Edited arg list:");

    if (simils.sysname) prt_("%s ",simils.sysname);
    if (simils.simname) prt_("%s",simils.simname);
    if (simils.simname || simils.sysname) _n

    loop (i,0,32) {
      if (i==17) _n
      prtoption(i,optionlist) }
    _n
  }

  if (!simils.simname || !simils.sysname) ERROR(("no SIMNAME specified"))

  /* remove the stp-file so that the simulation is not immediately stopped */
  remove(Fn("stp"));
  _n

#if PARALLEL
  if (option('i')==0) WARNING(("option -i0 for PARALLEL"))
  {
    char *env=getenv("NSLOTS");

    if (!env) ERROR(("parallel: undefined number of threads\n\
*** (no environment variable NSLOTS)"))
    No.th=No.lcth=atoi(env);
  }
#endif /*# PARALLEL */

  rndinit(7L/*low quality enough*/,option('z'));

  /*
     open sysname.def - to be reread in pass 0 below
     here only equalize read because : must be known in advance,
     so must save in
  */
  if ( !(in=fopen(Fn("def"),"rt")) ) {
    /* simils.simname.def not found: trying sysname.def */
    char *swapname=simils.simname;

    simils.simname=simils.sysname; /* swap sysname/simils.simname temporarily */
    in=fopen(Fn("def"),"rt");
    simils.simname=swapname; /* normal sysname/simils.simname again */
    if (!in) ERROR(("no %s.def nor %s.def found",simils.sysname,simils.simname)) }

  _Id.ignoreid=1; /* NB: no check of data! */
  getdata
    get(equalize)
    get(no) get(noint) get(h)
    get(E) get(rho) getvec(L,,DIM)
    get(tau.T) get(T)
  enddata
  _Id.ignoreid=0; /* check of data again */

  rewind(in);
  indef=in;

#ifdef COOK
  /* see also mainble.c */
  mark=simils.simname; simils.simname=simils.sysname; /* swap sysname/simname temporarily */
  in=fopen(Fn("ble"),"rt");
  simils.simname=mark; /* normal sysname/simname again */
  if (!in) Error("open system ble-file");
  minmass=readblend();
  fclose(in);
#else /*# COOK */
  convertsiteunits(); /* just writes info #ifdef REDUNITS */
#endif /*#!COOK */

  in=indef; /* for pass=0, rewound above */

  while (sig<2) { /* <<< run <<< */
    sig= -1;

    checkranges(1);

    /* getdata : twice for the 1st time to read .def */
    /* NB: .def already partially read because of equalize */
    do {

#ifdef COOK
      int *N,i;
      if (pass==0) {
        allocarray(N,nspec);
        loop (i,0,nspec) N[i]=spec[i]->N; }
#endif /*# COOK */

      if (pass==1) {
#ifdef COOK
        loop (i,0,nspec) spec[i]->N=N[i];
        free(N);
        if (option('k')>0) initfix(Fn("fix"));
        initNo(); /* now we know everything to call it */

        checkfixed();
#endif /*# COOK */
        /* close .def file, open .get/.G# file or use stdin */
        fclose(in);
        if (option('s')) in=stdin;
        else {
          char ext[12]="get";

          if (option('g')) sprintf(ext,"G%d",option('g'));
          in=fopen(Fn(ext),"rt"); }
        if (in==NULL) ERROR(("open %s",lastFn)) }

      if (pass>=1) {
        _n putline('%',79);
        if (option('s')) prt("Enter data in form `VARIABLE=EXPRESSION\', `?\' for help, `;\' to start simulation:"); }

      getdata
        /* auxiliary variables for calculations */
        static double
          aux=0, a=0, b=0, c=0, x=0, y=0, z=0;
        static int
          i=0,j=0,k=0,n=0;

        if (pass==0) {
          getvec(N,,nspec)
          get(equalize) }
        get(el.centroid)
        get(AllocSizeLim)
#ifdef FREEBC
        get(center.K) get(center.r0)
#else /*# FREEBC */
        get(el.epsk) get(el.epsr)
        get(el.alpha) get(el.kappa)
        get(el.epsinf) get(el.test) get(el.diag) get(el.diff)
        get(el.sf)
        get(el.rplus) get(el.rshift)
        get(el.sat)
        get(tau.sat)

        getvec(center.K,,DIM)
        getvec(center.r0,,DIM)
        get(el.Perr)
#  ifdef SLAB
        getvec(slab.n,,CENTERSLAB)
        getvec(slab.ns,,CENTERSLAB)
        getvec(slab.z,,CENTERSLAB)
        getvec(slab.Kz,,CENTERSLAB)
        getvec(slab.z0,,CENTERSLAB)
        getvec(slab.z1,,CENTERSLAB)
#    if SLAB & 2
        get(cleave.sigma)
        get(cleave.K)
        get(cleave.n)
        get(cleave.z0)
        get(cleave.z1)
        get(cleave.init)
#    endif /*# SLAB & 2 */
#  endif /*# SLAB */
#endif /*#!FREEBC */
        getvec(center.cmK,,DIM)
        get(center.cmn)
        getvec(Eelst,,DIM)
        get(dV)
        getkey(SI,yesno)
        getkey(init,initkey)
        get(mirror) getkey(sort,sortkey)
        getkey(pins,pinskey) get(initrho) getkey(MC,yesno)
        get(E) get(rho) getvec(L,,DIM)
        getvec(shift,,DIM)
        get(tau.E) get(tau.rho)
        getkey(rescale,rescalekey)
        get(tau.T) get(T) get(initvel)
        getkey(thermostat,thermostatkey)
        get(T_tr_in)
        get(tau.P) get(P) get(maxscale) get(bulkmodulus)
        get(tau.R) get(tau.i)
        getvec(load.n,,DIM)
        get(load.N)
        getvec(load.L,,DIM)
        getkey(load.tr,loadtrkey)
        getkey(load.zero,loadzerokey)
        get(nm.method) get(nm.dr) get(nm.eps) get(nm.key) get(nm.zero) 
        get(nm.frames) get(nm.modes) get(nm.ampl) 
#ifdef DISTURB
        getvec(disturb,,2)
#endif /*# DISTURB */
        get(omega) get(omegac)
        get(eps) get(epsc)
        get(no) get(noint) get(h)
        get(nomax) get(stop)
        get(dt.prt) get(dt.cfg) get(dt.plb)
        get(lag.err) get(lag.n)
        get(lag.cond)
#if PRESSURETENSOR&PT_OFF
        get(lag.visc)
#endif /*# PRESSURETENSOR&PT_OFF */
        get(Emax) get(drmax)
        get(rdf.grid) get(rdf.cutoff) get(rdf.onefour)
        get(el.grid) get(el.minqq) get(el.slab)
        get(cutoff) get(LJcutoff) get(corr)
        get(CPnbit)
#ifdef POLAR
        get(scf.eps) get(scf.epsq) get(scf.omega) get(scf.maxit)
        get(scf.epsx) get(scf.omegax)
        get(scf.E) get(scf.Estride)
        get(tau.dip)
#endif /*# POLAR */
#ifdef SHEAR
        get(shear)
#endif /*# SHEAR */
#ifdef LINKCELL
        getvec(No.cell,,DIM)
        get(No.percell)
        get(box.over14) get(box.rmin)
#if PARALLEL==1
        get(No.lcth)
#endif
#endif /*# LINKCELL */
#ifdef WALL
        /* structure wall defined in simdef */
        get(wall.rho) get(wall.minz) get(wall.g) get(wall.scalez)
        get(wall.n) get(wall.rep)
#endif /*# WALL */
        get(No.first)
        get(drift) get(conserved)
        getkey(quit,yesno)
#ifdef WATERPLUS
        {
#  include "watercut.h"
          get(cut.from) get(cut.to) }
#endif /*# WATERPLUS */
        if (quit==-4) {
          prt("WARNING: quit=\"shell\"=-4 is deprecated, use key=\"shell\"=8 instead");
          quit=0;
          key=8; }

        getkey(key,keykey)
        if (key) {
          prt("executing key=%d:",key);
          switch (key) {
            case 1: case 2: case 3:
            case -1: case -2: case -3:
              sortmolecules(key);
              break;
            case 4:
              system(string("showcp -p99 %s",simils.simname));
              break;
            case 5:
              system(string("show %s",simils.simname));
              break;
            case 6:
              system(string("rdfg %s -g -p",simils.simname));
              break;
            case 7:
              system(string("rdfg %s -c -p",simils.simname,simils.simname));
              break;
            case 8: {
              /* the shell (DOS,EMX removed...) */
              int ierr;

              fprintf(stderr,"\n *** Type `exit' to return to %s! ***\n",arg[0]);
              ierr=system(getenv("SHELL")?getenv("SHELL"):"bash");
              prt("shell return code=%d",ierr); }
              break;
            case 9: goto TheEnd;
          }

          key=0; }

        get(cache)
#ifdef DIHHIST
        get(dih.res) get(dih.grid) get(dih.cp) get(dih.dcp)
#endif /*# DIHHIST */

#ifdef WIDOM
        get(widom.sp) get(widom.spreal) get(widom.n)
#  ifdef SLAB
        get(widom.z0) get(widom.z1) get(widom.dz) get(widom.mode)
#  endif /*# SLAB */
#endif /*# WIDOM */

#ifdef SLAB
        get(slab.grid) get(slab.zmax)
        get(slab.sp) get(slab.mode)
        get(slab.T) get(slab.Tz0) get(slab.Tz1)
#endif /*# SLAB */

#ifdef XSECTION
        get(xs.grid) get(xs.RvdW) get(xs.Rscale) get(xs.mode) get(xs.freq)
        get(xs.cp) get(xs.ns) get(xs.sizelimit)
#endif /*# XSECTION */

#ifdef RGYR
        getvec(rg.end,,2) get(rg.cp)
        get(lag.v)
#endif /*# RGYR */

        /* auxiliary variables */
        get(aux)
        get(a) get(b) get(c)
        get(x) get(y) get(z)
        get(i) get(j) get(k)
        get(n)

        checkdata
      enddata
      _n

      onefourinrdf=rdf.onefour;

#ifdef CLUSTERS
      /* reads both .def and .get */
      if (option('f') && !option('g')) readclusterdef(in);
#endif /*# CLUSTERS */

      if (quit>0) {
        prt("WARNING: quit>0 is deprecated, use key=\"quit\" or key=9 instead");
        goto TheEnd; }
    } while (pass++==0);

    if (option('y')>0 && (init<0 || init>1) && !option('f')) makemolgol(option('y'));
    init_append=0;

    if (tau.E!=0 && thermostat) {
      WARNING(("tau.E!=0 and thermostat: thermostat turned off"))
      thermostat=0; }

    setdrift();

    prt("minimum atom mass=%g g/mol => suggested timestep h in (%.4f,%.4f) ps",
        minmass, 0.0005*sqrt(minmass), 0.0015*sqrt(minmass));
    if (h<0.0005*sqrt(minmass))
      prt("WARNING: the timestep h=%g is probably unnecessarily short",h);
    if (h>0.0025*sqrt(minmass))
      WARNING(("the timestep h=%g is probably too long\n\
*** bad energy conservation or crash may occur",h))
    if (h>0.0015*sqrt(minmass)) prt("\
WARNING: the timestep h=%g is probably too long\n\
         bad energy conservation or crash may occur",h);

    if (rho==0 && L[0]*L[1]*L[2]==0)
      if (init>=3) ERROR(("undefined box: set box (L[0],L[1],L[2]) or density (rho)"))

    No.ncoord="\0\1\1\2\1\2\2\3"[rescale&RESCALE_XYZ];

    underline("rescaling of the simulation box");
    prt("applies if tau.P (barostat), tau.rho, or dV (virtual volume change)");
    prt("rescale = %d = 0x%x = %s",rescale,rescale,int2sumbin(rescale));
#if (PRESSURETENSOR&PT_ANY) != PT_ANY
    if (rescale&RESCALE_PT)
      ERROR(("Pressure tensor components are requested but this version of cook\n\
*** was not compiled so. Use at least PRESSURETENSOR=3."))
#endif /*# (PRESSURETENSOR&PT_ANY) != PT_ANY */
    if (rescale&RESCALE_XYZ) {
      prt("configuration rescaling in these %d coordinate(s): %s\n\
based on %s",
        No.ncoord,
        prtxyz(rescale&RESCALE_XYZ),
        rescale&RESCALE_PT?"pressure tensor components":"isotropic virial pressure"); }
    else {
      prt("no configuration rescaling allowed, dV=tau.P=tau.rho=0 set\n\
*** (check rescale if this is not what you want)");
      dV=tau.P=tau.rho=0; }

    if (dV<0) {
      ERROR(("dV<0 (molecule-based rescaling) no longer supported\n\
*** use dV>0 and rescale=8|(more options) or rescale=\"[options]CM\""))
      dV=fabs(dV);
      rescale|=RESCALE_CM; }

    prt("configuration rescaling will be based on %s",
      rescale&RESCALE_CM ? "molecular centers-of-mass" : "sites");
    _n

    constrd.dV=dV;
    if (dV) {
      if (thermostat>=T_NPT)
        WARNING(("barostat based on the virial pressure, not virtual volume"))
      if (tau.P<0)
        ERROR(("every-cycle barostat (tau.P<0) is not supported with pressure by virtual volume change (dV)\n\
*** use tau.P>0"))
      else if (tau.P>0)
        WARNING(("the friction barostat will use the virtual volume change-based pressure")) }
    constrd.PdVname = rescale&RESCALE_CM ? "PdVmol [Pa]" : "PdVatom [Pa]";

    if (thermostat>=T_ANDERSEN && thermostat<=T_MAXWELL_CM) {
#if VERLET==1 || VERLET==2
      WARNING(("kinetic energy inacccurate with Andersen/Maxwell thermostat\n\
*** recompile with VERLET=3 (recommended) or 0 (less accurate)"))
#endif /*# VERLET==1 || VERLET==2 */
      if (option('m')>2)
        WARNING(("Gear integration inacccurate with Andersen/Maxwell thermostat")) }

    if (thermostat>=T_NPT) {
      if (option('m')!=2) ERROR(("NPT via Nose/MTK not implemented with Gear")) }

    if (tau.T && !thermostat)
      prt("WARNING no thermostat: nonzero tau.T=%g ignored\n\
(if you mean friction thermostat as in old cook, use thermostat=\"friction\")",tau.T);
    if (tau.T==0 && thermostat>0 && thermostat<T_NPT) {
      ERROR(("thermostat=%d was specified but tau.T=0. There is no default anymore!\n\
*** Hints for tau.T:\n\
***   0.1--0.3  Nose, NPT\n\
***   0.05--0.2 Berendsen/frictions start\n\
***   0.5--2    Berendsen/frictions productive runs\n\
***   0.2--2    Andersen,Maxwell\n\
***   >1 Langevin\n",thermostat))
      thermostat=0; }

    if (tau.P==0 && thermostat>=T_NPT) {
      ERROR(("thermostat=NPT (i.e., thermostat+barostat) specified but tau.P=0"))
      thermostat=0; }

    /* total charge check delayed to be interactive */
    if (fracpass==0) {
      int sp, nfrac=0;
      double charge;

      fracpass++;
      loop (sp,0,nspec) {
        charge=spec[sp]->charge/electron;
        if (fabs(fmod(fabs(charge+0.5),1)-0.5) > 1e-5) {
          nfrac++;
          prt("charge =%10.6f e of species %d is fractional",charge,sp); } }
      if (nfrac)
        WARNING(("\
%d molecule(s) with fractional charges\n\
(you may need blend -_PREC, check also mol-file)",nfrac)) }

    if (initno) {
      prt("no overriden by option -%d",initno);
      no=initno;
      initno=0; }

    underline("units");
#ifdef REDUNITS
    prt("! WARNING: reduced units used even if real units printed !");
#else /*# REDUNITS */
    if (SI) {
      SIE=1;
      prt("energy in statistics and convergence profile in J/mol"); }
    else {
      SIE=kcal;
      prt("energy in statistics and convergence profile in kcal/mol"); }
    prt("if no units are given (as for Etot), energy is in Kelvin");
#endif /*#!REDUNITS */

    sig=0;

#ifdef DIHHIST
    if (dih.cp<-1) dih.cp=-1;
#endif /*# DIHHIST */

#ifdef COULOMB
#  if COULOMB<=-1
    if (el.kappa<=0) if (!option('f')) WARNING(("el.kappa=%g is suspicious for Ewald",el.kappa))
#  else /*# COULOMB<=-1 */
    if (el.kappa) WARNING(("el.kappa=%g is suspicious for cutoff electrostatics",el.kappa))
#  endif /*#!COULOMB<=-1 */
#endif /*# COULOMB */

    GearOrder=option('m');

#ifdef POLAR
    if ((option('p')/10)%10==0) if (tau.dip==0)
      ERROR(("Car-Parrinello-like method and tau.dip=0"))
    if (abs(option('^'))>99) option('^')=abs(option('^'))%100;
    if (scf.omega==-9) { /* scf.omega not specified in data */
      if (option('^'))
        scf.omega=option('^')/100.;
      else
        /* no -^ nor scf.omega: defaults
           - for -m2 (ASPC) the default is postponed to pred2.c
           - for -m>2 (full-iteration) the default is scf.omega=0.95
        */
        if (option('m')!=2) scf.omega=0.95; }
    else {
      /* scf.omega specified in data */
      option('^')=100*scf.omega+0.5; /* not to be set to default in pred2.c */ }
#endif /*# POLAR */

    putkey(init,initkey)
    if (init>=3 && init<10) {
      putkey_(pins,pinskey) put2(initrho,Emax) }
    put2(el.Perr,el.epsinf)
    put2(rho,tau.rho)
    put2(E,tau.E)
    put_(T) put_(tau.T) putkey(thermostat,thermostatkey)
    if (thermostat>=T_TR && thermostat<=T_FRICTIONS) put(T_tr_in)
    put2(P,tau.P)
    if ((tau.P!=0 || tau.rho!=0) && tau.R!=0)
      ERROR(("intermolecular parameter adjustement requires NVT"))

#ifdef WALL
    put3(wall.n,wall.rho,wall.minz)
    put2(wall.g,wall.scalez)
#endif /*# WALL */
    if (rdf.grid) put2(rdf.grid,rdf.cutoff)

#ifdef DIHHIST
    if (dih.res!=0) {
      if (dih.grid)
        WARNING(("max one of dih.grid, dih.res may be specified (dih.grid will be ignored)"))
      dih.grid=(int)(360./dih.res+0.5);
      if (fabs(360./dih.res/dih.grid-1)>1e-6)
        WARNING(("dih.res rounded to %f",360./dih.grid));
      dih.res=0; }
#endif /*# DIHHIST */

/* rdf is parallelized with PARALLEL==1, dih is serial (intramolecular) */
#if PARALLEL==3
    if (rdf.grid) {
    prt("WARNING: rdf not parallelized, all measurements serial (check noint)");
    No.measureserial++; }
#  ifdef DIHHIST
    if (dih.grid) {
      WARNING(("WARNING: dih not parallelized, all measurements serial (check noint)"))
      No.measureserial++; }
#  endif /*# DIHHIST */
#endif /*# PARALLEL==3 */

    if (tau.P*tau.rho) {
      WARNING(("both tau.rho and tau.P: tau.P replaced by zero"))
      tau.P=0; }

    /* put(omega) */
    if (!(option('c')&1)) put_(omegac)
    put2(epsc,eps)
    put3(no,noint,h)
    No.dt=noint*h;
    DT=No.dt*timeunit;
    prt("cycle (of noint MD steps): dt=%g s",DT);
    put3(dt.prt,dt.cfg,dt.plb)
    put2(lag.err,lag.n)
#if PRESSURETENSOR&PT_OFF
    put_(lag.visc)
#endif /*# PRESSURETENSOR&PT_OFF */
#ifdef RGYR
    put_(lag.v)
#endif /*# RGYR */
    put(lag.cond)
    StaSet(DT,lag.err,2,lag.n);
    put3(cutoff,rdf.cutoff,LJcutoff)
    if (LJcutoff==0) WARNING(("LJcutoff=0 (no Lennard-Jones interaction)"))

#ifdef RGYR
    put3(rg.cp,rg.end[0],rg.end[1])
#endif /*# RGYR */
#ifdef XSECTION
    prt("\
xs.grid=%g  xs.RvdW=%.8f  xs.Rscale=%.8f\n\
xs.mode=%d  xs.freq=%d xs.ns=%d  xs.sizelimit=%d",
xs.grid,    xs.RvdW,      xs.Rscale,
xs.mode,    xs.freq,   xs.ns,    xs.sizelimit);
    if (xs.sizelimit>AllocSizeLim) {
ERROR(("xs.sizelimit=%d exceeds AllocSizeLim=%d\n\
*** (if you are sure, increase also AllocSizeLim)",xs.sizelimit,AllocSizeLim))
      AllocSizeLim=xs.sizelimit; }
#endif /*# XSECTION */
    put(cache)

    /*** repeatable initializations ***/
    ralloc(mark,1);

#ifndef COOK
    initspecies(); /* init spec + calculates No */
#endif /*# COOK */

    underline("box size setting and determination of cutoff");
    refrho=setL(L,rho);
    VV(box.L,=L)

    prt("reference box = %.8f %.8f %.8f A",VARG(L));
    prt("reference rho = %.4f kg/m3",refrho);
    prt("> The reference box and density are derived from input rho and L[].\n\
> They are used for setting cutoff and Ewald parameters el.alpha, el.kappa\n\
> (if these are not given explicitly).\n\
> The actual box may differ:\n\
> * in NPT simulation (see tau.P and rescale)\n\
> * if other box is loaded and change is not allowed (see tau.rho and rescale)");

    if ((init>=0 && init<3) || option('f'))
      loadcfg(-1,&RvdW); /* 1st pass */
    else if (init<0)
      readplayback(-init,0); /* L and NS only */
    else if (init>10)
      readasc(0,0); /* L only */
    else {
      if (initrho && fabs(initrho/refrho-1)>5e-16) {
        prt("Initial density (initrho) differs from the reference density.");
        setL(box.L,initrho); }
      else
        prt("Initial density is given by rho/L specified"); }

    /* now: box.L=actual box loaded or to be initialized,
            L=reference or target box */

    VV(simils.changeL,=load.L) /* ugly control for historical reasons */
    replaceL(L);
    /* VO(load.L,=0) after initreplicate */
    rhoreached=0;

    if (fabs(box.L[0]/L[0]-1)>5e-16
     || fabs(box.L[1]/L[1]-1)>5e-16
     || fabs(box.L[2]/L[2]-1)>5e-16) {
      prt("WARNING: The actual (loaded/initial) box size differs from the reference size.");
      if (tau.P) {
        if (box.L[0]/L[0]-1<0 || box.L[1]/L[1]-1<0 || box.L[2]/L[2]-1<0)
          prt("WARNING: tau.P specified and the reference box is larger than actual\n\
(this is often benign, but sometimes may compromise accuracy of electrostatics)");
        }
      else if (tau.rho==0 || (rescale&7)==0) {
        prt("WARNING: Box size change is not allowed because tau.rho==0 or (rescale&7)==0\n\
==> the reference box will be replaced by the actual box (and kept constant)");
        VV(L,=box.L) }
      else if (tau.rho) {
        rhoreached++;
        prt("Since tau.rho is specified, the box will be gradually rescaled until the\n\
reference box is reached."); } }
    if (tau.P)
      prt("Since tau.P is specified, the box size will fluctuate; the reference box\n\
will be used for parameter setting.");

    VV(box.Lh,=0.5*box.L)

    box.cutoff=initcutoff(cutoff,L); /* set cutoff; needs No, L */

    if (rdf.cutoff>box.cutoff)
      WARNING(("rdf.cutoff=%g > cutoff=%g",rdf.cutoff,box.cutoff))
    if (rdf.cutoff==0) rdf.cutoff=fmin(box.cutoff,10);
    initrdf(rdf.grid,rdf.cutoff,Fn("s-s"));
#ifdef DIHHIST
    /* only histograms initialized - control structures already initialized */
    initdihhist(dih.grid);
#endif /*# DIHHIST */
#ifdef HARMONICS
    initharmonics(rdf.grid,rdf.cutoff,T,refrho);
#endif /*# HARMONICS */

    constrd.mode=rescale+RESCALE_L; /* force box rescale */

#ifdef SLAB
    if (slab.mode && !slab.grid)
      WARNING(("slab.mode specified, but slab.grid not given:\n\
*** surface tension and cutoff corrections will not be calculated\n"))
    if (slab.mode&2) constrd.mode|=RESCALE_SLAB;
    slab.qT=0;
    if (thermostat && slab.T>=0) {
      slab.qT=sqrt(slab.T/T);
      prt("separate thermostat T=%g for slab z in [%g,%g)",
          slab.T,slab.Tz0,slab.Tz1);
      if (slab.Tz0>=slab.Tz1) WARNING(("this range is empty"))
      if (GearOrder>2 || !strchr("\1\3\4\5\6",thermostat)) /* T_BERENDSEN, T_ANDERSEN .. T_MAXWELL_CM */
        WARNING(("slab thermostat not supported for -m%d and thermostat=%d",GearOrder,thermostat))
        if (thermostat==T_BERENDSEN) WARNING(("The slab-based Berendsen thermostat may not work as you expect,\n\
*** please consult the manual.")) }

    initdpr(slab.grid,slab.zmax,slab.sp,Fn("dpr"));
#endif /*# SLAB */

    initss(LJcutoff);
    initpot();
    initmolecules(); /* here cfg allocated */
#ifdef METAL
    eldensities(NULL); /* rho's are 0 for initcfg */
    eldens.rref=cfg[0]->rp; /* will be used by initcfg */
#endif
    initconstrit(omegac);
#ifdef WALL
    initwall();
#endif /*# WALL */

#include "mainextf.c"

    if (No.c && (thermostat==T_ANDERSEN || thermostat==T_MAXWELL))
      WARNING(("Constraint dynamics and atom-based Maxwell/Andersen thermostat:\n\
*** The results will be likely severely affected.\n\
*** Consider center-of-mass-based Andersen and Maxwell thermostats instead\n\
*** (thermostat=\"AndersenCM\" or thermostat=\"MaxwellCM\")."))

    En.corr=cutcorr(LJcutoff); /* called because of counting rdf->npairs */

    /* pressure finite-size corrections */
    underline("kinetic pressure corrections");
    prt("\
In some boundary conditions, momenta and angular momenta are conserved.  The\n\
kinetic temperature is always calculated from the number of not-conserved\n\
degrees of freedom, but the kinetic (ideal-gas) part of pressure requires the\n\
number of sites or molecules (the latter applies for the center-of-mass based\n\
virtual volume change method).  The corrections fulfill these conditions.");
    if (corr&16) {
      corr=corr&3;
      if (thermostat<T_NPT) corr|=12;
      prt("corr=%d set to be consistent with the barostat",corr); }
    if (corr&4) {
      En.Pkinq=(double)No.f4P/No.f;
      if (En.Pkinq!=1)
        prt("Correction factor in site-frame = %g",En.Pkinq); }
    else {
      En.Pkinq=1;
      if (No.f4P!=No.f && thermostat<T_NPT)
        WARNING(("You opted for ignoring the kinetic (ideal gas) pressure correction\n\
*** in the site-frame")) }
    if (corr&8) {
      En.PkinqCM=(double)No.f4P_tr/No.f_tr;
      if (En.PkinqCM!=1)
        prt("Correction factor in molecule-frame = %g",En.PkinqCM); }
    else {
      En.PkinqCM=1;
      if (No.f4P!=No.f && rescale&8 && dV!=0)
        WARNING(("You opted for ignoring the kinetic (ideal gas) pressure correction in\n\
*** the molecule frame (used in the virtual volume change method)")) }

    prt("Correction factors: site: %.8f, CM: %.8f\n",En.Pkinq,En.PkinqCM);
    if (thermostat==T_NPT)
      if (En.Pkinq!=1) WARNING(("Kinetic pressure corrections should not be used with the Nose-MTK barostat\n\
*** Hint: use corr=3\n\
*** NB: the default is set at start, so if you switch to thermostat=\"NPT\"\n\
*** you should also set corr=3"))

#ifdef WIDOM
    if ( (widom.n || widom.spreal>=0) && thermostat==0)
      WARNING(("Widom insertion or scaling requires a thermostat because\n\
*** temperature T is used to calculate the insertion probability"))
    widom.corr=Widomcutcorr(widom.spreal,widom.sp,LJcutoff);
    if (widom.spreal>=0 && widom.n)
      ERROR(("Widom insertion (widom.n>0) and Widom scaling (widom.spreal>=0) are exclusive"))
#endif /*# WIDOM */

#if defined(TWODIM) || defined(FREEBC)
    En.corr=0; /* not supported */
#endif /*# defined(TWODIM) || defined(FREEBC) */

    prt("Mass of the whole configuration in program units: ");
    put2(No.mass,No.free_mass)
    prt("mass = %.8g g/mol = %.8g kg", No.mass*munit,No.free_mass*massunit);
    prt("rho = %.12g / V  (for rho in kg m^-3 and V in A^3)", rhounit*No.mass);
    _n

    if (option('f')) {
      /* autodiffusion and structure factor from stored cfgs */
      double d=dt.plb;
      char *info="plb";

      if (option('f')<0) d=dt.cfg,info="cfg";
      if (d<=0)
        ERROR(("read plb mode (-f): dt.%s=%g, but a positive value is expected\n\
*** Hint: it is best to specify dt.%s in %s\n\
*** (%s is not read if -g, e.g., if called from plb2diff)",
               info,d,info,Fn("def"),Fn("get")))
      diffSF(d,h*noint,no,init,nspec);
      goto TheEnd; }

    if (option('m')==1 && init>=0) init=-1;

    initgammas();

    if (init==0) {
      StaLoad(Fn("sta"));
      issta++;
      if (nomax>0) if (StaN("Tkin")>=nomax) {
        fprintf(stderr,"already nomax=%d cycles has finished - stop\n",nomax);
        prt("already nomax=%d cycles has finished - stop\n",nomax);
        exit(1); } }
    LRFree();

    if (init==0) loadrdf(rdf.grid);
#ifdef DIHHIST
    if (init==0) loaddih(dih.grid);
    if (dih.dcp) dihcp=fopen(Fn("dcp"),init<2?"at":"wt");
#endif /*# DIHHIST */
#ifdef HARMONICS
    if (init==0) loadharmonics(rdf.grid,rdf.cutoff);
#endif /*# HARMONICS */
#ifdef SLAB
    if (init==0) loaddpr(slab.grid);
#endif /*# SLAB */


/* SIMNAME.000, ... to plb conversion, no simulation (obsolete) */
    if (init==999) {
      int i;

      openplayback(option('y'),0);
      loopto (i,1,no) {
        loadcfg(i,&RvdW); /* 1st pass */
        loadcfg(i,NULL);  /* 2nd pass */
        if (simils.frommol) ERROR(("truncated %s - some molecules undefined",lastFn));
        writeplayback(); }
      return 0; }

    /* fix sites/axes by SHAKE constraint, measure forces/torques */
#ifdef ANCHOR
    initanchor(Fn("fix"),init,drmax);
#endif /*# ANCHOR */

    /*** initialize/load configuration ***/
    if (init>=3) {
      if (PROD(load.n)!=1 || load.N || SUM(load.L) || load.tr)
        WARNING(("load.n[], load.N, load.L[], load.tr: specified but ignored because init>=3"))
#ifdef COULOMB
      if (init<10) initrspace(option('v')&1 ? el.grid : -el.grid);
      /* why here? - cf. Ewaldtest() below... */
#endif /*# COULOMB */
      if (init>=10) {
        readasc(0,1);
        if (init>=11) readasc(1,1);
        if (init>=12) readasc(2,1);
        if (init>12) ERROR(("init=%d not implemented",init)) }
      else if (init==5)
        initcryst((int)pins,Emax);
      else {
        if (MC && option('y')) {
          openplayback(option('y'),0);
          plbopened=1; }
        initcfg(pins,Emax,
                option('y')?dt.plb==0?1:(int)(fabs(dt.plb)/h+0.5):0); } }
    else {
      int repl=initreplicate();

      if (init>=0) {
        loadcfg(-1,NULL); /* 2nd pass */
        if (repl>1 && simils.frommol) ERROR(("cannot replicate and add molecules at once"))
        if (init==2) t=0;
        if (stop>0 && t>stop-h/2) sig=3;
        if (simils.frommol) {
#ifdef COULOMB
          initrspace(el.grid); /* why here? - cf. Ewaldtest() below... */
#endif /*# COULOMB */
          initcfg(0,Emax,0);
          simils.frommol=0; } }
      else {
        if (plbopened && simils.plbname) ERROR(("internal: attempt to open the same playback file for reading and writing"))
        readplayback(-init,1);
        prt("initialized from playback frame %d, continuing as with init=2",
            frame0=-init);
        init=2; }
      replicatecfg();
      VO(load.n,=1)
      VO(load.L,=0)
      load.N=load.tr=0; }


    /* nullify some coordinates - only once */
    zerocfg(load.zero); load.zero=0;

    /* to suppress undefined/badly defined logs in case of thermostat changes */
    if (thermostat!=T_NOSE && thermostat<T_NPT) {
      int k;

      loop (k,0,GearOrder) if (cfg[k]->logs) {
        prt("NB: Nose extended variable^(%d):=0 because other thermostat",k);
        cfg[k]->logs=0; } }

    /* to suppress undefined/badly defined lambda in case of barostat changes */
    if (thermostat<T_NPT) {
      int k;

      loop (k,0,GearOrder) if (cfg[k]->logs) {
        prt("NB: barostat extended variable:=0 (%d deriv)",k);
        VO(cfg[k]->lambda,=0) } }

#if defined(SLAB) && SLAB & 2
    if (cleave.init&1) {
      cleave.init&=0x7ffffffe;
      cleave.z[0]=cleave.z0;
      cleave.z[1]=cleave.z1;
      prt("cleave.z0=%.9f cleave.z1=%.9f copied from data to cleave.z[0,1]\n\
cleave.init bit 0 unset (not to copy next time)"); }
    if (cleave.n==2) {
      if (fabs(cleave.z[0]-cleave.z[1])<1e-6)
        WARNING(("cleave.n==2 and cleave.z[0] == cleave.z[1] (with precision 1e-6)"))
      if (cleave.z[1]<cleave.z[0])
        ERROR(("cleave.n==2 and cleave.z[1]<cleave.z[0]")) }
#endif /*# defined(SLAB) && SLAB & 2 */

    if (init<0 || init>1) {
      En.tot=3e33; /* = will be ignored in the statistic of dEtot^2 */
      /* for Nose: restart En.tot */
      cfg[0]->logs=0; }

    /* fix sites by springs, do not measure force */
    if (option('k')>0) {
      makefixa();
      loadfixa();
      savefixa(); }

    if (initvel) Maxwell(0,initvel,1);
    initvel=0; /* just once */

    if (option('h')) {
      int valence=abs(option('h'));

      while (valence) {
        prt("WARNING: %d %d-groups or neighbors centered\n",
            centergroups(cfg[0],valence%10), valence%10);
        valence/=10; }
      if (option('h')>0) option('h')=0; }

    box.V=PROD(box.L);

#ifdef LINKCELL
    lcsetup(L);
#  ifdef WALL
    if (wall.n<2)
      WARNING(("LINKCELL && wall.n=1 may cause error out of box for z>=L"))
#  endif /*# WALL */

#endif /*# LINKCELL */

#if PARALLEL
    initparallel();
#  if PARALLEL==1 || PARALLEL==3
    parallelizerdf(1);
#  endif /*# PARALLEL==1 || PARALLEL==3 */
#endif /*# PARALLEL */
    addshift(shift);
    if (init>=2) normalize(-2);
    putv(L)
    putv(box.L)
    put2(box.cutoff,box.V)

    /* init=0,1: apppend playback; init=2,3: new playback */
    if (option('y')) {
      if (!plbopened) openplayback(option('y'),init<2&&init>=0);
      if (init>=2 && dt.plb>=0) writeplayback(); /* t=0: do'nt write for dt.plb<0*/ }

    constrd.Pcorr=En.corr/Sqr(box.V)*Punit;
    underline("cutoff corrections");
    prt("Lennard-Jones cutoff=%g %s ==> cutoff correction for homogeneous fluid\n\
En.corr=%.9g K A^3 (Ecorr=En.corr, Pcorr=En.corr/V^2)\n\
prog.units: Ecorr=%g K  Pcorr=%g p.u.",
        fabs(LJcutoff),LJcutoff>0?"A":"sigma",En.corr,En.corr/box.V,En.corr/Sqr(box.V));
    prt("Ecorr=%g %s, Pcorr=%g Pa",
        En.corr/box.V*Eunit/SIE,SI?"J/mol":"kcal/mol",constrd.Pcorr);
    if ((corr&3) == 2 ) ERROR(("implementation limitation: corr=2 not supported, consult the manual!"))
    prt("Because of corr=%d, the above corrections are:\n\
  %sincluded in P, pressure tensor, and E\n\
  %sincluded in P used in the barostat",corr,
        corr&1?"":"NOT ",
        corr&2?"":"NOT ");
    if (!(corr&1)) constrd.Pcorr=En.corr=0;
#ifdef SLAB
    if (slab.grid)
      prt("\n\
>>> SLAB version hack: For compatibility, the above corrections are included\n\
>>> which is incorrect in slab geometry. Correct corrections will be calculated\n\
>>> later from the density profile (you need to set slab.mode&1).");
#endif /*# SLAB */

    halftcyc=h*noint/2;

#ifdef COULOMB
    underline("initialization of electrostatic forces");
    if (Ewaldtest(L)) goto TheEnd;
    prt("Ewald set final values: cutoff=%g el.alpha=%g el.kappa=%g",
                             box.cutoff,   el.alpha,   el.kappa);
//    // DEBUG
//    {int i;
//      double u,byerd,rr;
//
//      loop (i,0,1000) {
//        rr=i*.01;
//        u=exacterud_sqrt(rr,&byerd);
//        prt("%.2f %.16g %.16g exacterud_sqrt",rr,u,byerd); }
//      exit(0);
//    }
#endif /*# COULOMB */

    if (quit<0) quit=debugcfg(quit);

#ifdef MARK
    /* Not tested recently! */
    mark_do(no);
    /* Everything suppressed for !MARK */
#else /*# MARK */

    if (init%10>=3) {
      /* thus, forces are not calculated for init=10 */
      measure=1; /* debugging only! */
      /* necessary rhs if Verlet? */
#  ifdef POLAR
      { /* OOPS! scf.omega may be -9 here to be set later.... */
        double o=scf.omega;
        scf.omega=0.95;
#if POLAR&32
        initfqcharges();
#endif
        rhs(cfg[2],cfg[0],cfg[1]);
        scf.omega=o; }
#  else /*# POLAR */
      rhs(cfg[2],cfg[0],cfg[1]);
#  endif /*#!POLAR */
      rescalecfg(cfg[2],RESCALE_XYZ,h*h/2,0,0); /* for Gear only? */

#  ifdef POLAR
      /*.....  memset(polarrof(molec,cfg[2]->rp),0,No.s*sizeof(vector));*/
      {
        int i;
        loop (i,1,GearOrder)
          copy(polarrof(molec,cfg[i]->rp),polarrof(molec,cfg[0]->rp),No.s*sizeof(vector));
      }
#  endif /*# POLAR */
    }

#  if 0
#    include "foolproo.c"
#  endif /*# 0 */

    /* columns 4,5 of the convergence profile */
    if (tau.P==0 && tau.rho==0 && tau.R==0)
      initCP(no,CPnbit,dV==0?(thermostat==T_NOSE?"Ep+k":No.first?"Ep0":"Ein"):constrd.PdVname,"P",corr);
    else
      initCP(no,CPnbit,tau.R?"RvdW":"rho",dV==0?"P":constrd.PdVname,corr);

    if (tau.R==0) {
      if (RvdW!=0) prt("RvdW=%g in cfg-file found but ignored (ble-file value applies)\n\
RvdW=0 was set for a subsequent write to the cfg-file\n\
so that next time tau.R is specified, RvdW from the ble-file will be used",RvdW);
      RvdW=0; }
    else
      if (RvdW==0) {
        RvdW=sitedef[tau.i].LJ[0].RvdW;
        prt("RvdW adjustment, tau.i=%d: RvdW=0 (loaded or initialized) replaced by force field value %g",tau.i,RvdW); }
      else {
        prt("RvdW adjustment, tau.i=%d: RvdW=%g loaded from %s",tau.i,RvdW,Fn("cfg"));
        sitedef[tau.i].LJ[0].RvdW=RvdW; }

#  ifdef DIHHIST
    gauchetrans(-3);
#  endif /*# DIHHIST */

    starttime=inittime(option('t')?8:0);
    nsteps=0;

#  ifdef POLAR
    underline("polar iterations");
    if (scf.eps>2e33) scf.eps=1;
    put3(scf.eps,scf.omega,scf.maxit)
    scf.eps0=scf.eps;
    if (scf.epsq>=1) ERROR(("scf.epsq=%g is invalid",scf.epsq))
    if (scf.eps>0) {
      if (scf.epsq>0) scf.eps *= 1-scf.epsq;
      if (scf.epsx>2e33) scf.epsx=scf.eps0*0.001; }
    else {
      scf.epsq=0;
      if (scf.epsx>2e33) scf.epsx=(int)(scf.eps*2-1); }
    put(scf.epsq)
    put2(scf.epsx,scf.omegax)
    if (tau.dip) prt("WARNING: tau.dip=%g: Car-Parrinello-like (extender Lagrangian) is deprecated",tau.dip);
#  endif /*# POLAR */

    {
      char *st=myctime(starttime);

      if (st[24]=='\n') st[24]=0;
      /* this string used by cutprt.c - do not change! */
      fprintf(stderr,"%d cycle%s started at %s\n",no,"s"+(no==1),st);
      prt("\n*** %d cycle%s started at %s",no,"s"+(no==1),st);
    }

    if (option('b')<0) {
      /* for batch mode -s0 only - default -b */
      option('b')=100000*abs(option('b'))/(No.s*noint)+1;
      prt("-b%d: cycles to check %s",option('b'),Fn("stp")); }

    fflush(out);

    if ((drift&DRIFT_WHEN)==DRIFT_START) {
      drift&=DRIFT_DEPEND;
      prt("no CM/v/angular drift corrected anymore because drift&%d specified\n\
final drift = %d = %s",DRIFT_START,drift,int2sumbin(drift)); }

    /* <<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<< */
    for (icyc=0; icyc<no && sig==0; icyc++) {

#  ifdef DISTURB
      Disturb(0,disturb[0]);
      Disturb(1,disturb[1]);
#  endif /*# DISTURB */

#  ifdef DIHF
      /* WARNING this is something very obsolete... */
      /* warning : memset for doubles generally not portable */
      memset(dihf,0,sizeof(dihf));
      memset(dihfcov,0,sizeof(dihfcov));
#  endif /*# DIHF */

      lastEtot=En.tot;

      if (option('d')==2) distancecheck();

      if (option('m')==1) {
        /* read playback instead of simulating */
        readplayback(icyc+frame0,1);
        prt("frame %d/%d read",icyc+frame0,no);
        icyc+=noint-1; }

      else loopto (iint,1,noint) {

        if (option('s') && option('b')>3)
          fprintf(stderr,"%d/%d %d/%d t=%f   \r",iint,noint,icyc,no,t);

        /* friction-like isobaric ensemble: Pscale kept during cycle */
        if (thermostat<T_NPT && tau.P>0)
          box.V=rescalecfg(cfg[0],rescale|RESCALE_L,Pscale[0],Pscale[1],Pscale[2]);

        /* adjusting an R-like parameter */
        if (tau.R>0) {
          sitedef[tau.i].LJ[0].RvdW/=Pscale[0];
          setss(&sstab[tau.i][tau.i],tau.i,tau.i,LJcutoff,0); }

        if (option('d')>2) distancecheck();

        if (sig==5) break;

        measure=iint==noint;

        if (option('m')==2 && thermostat==T_LANGEVIN_CM) ERROR(("not implemented yet"))
        if (option('m')==2 && thermostat==T_LANGEVIN) ERROR(("not implemented yet"))

        if (option('m')==2) Shake(epsc,
          thermostat==T_MAXWELL||thermostat==T_MAXWELL_CM ? (double)justnow(tau.T,h/2) :
          thermostat==T_ANDERSEN||thermostat==T_ANDERSEN_CM ? h/tau.T : 0);
        else {
#  ifdef POLAR
            Gear2pol(No.eq,No.s*DIM,(option('p')/10)%10,(option('p')/100)%10,cfg);
#  else /*# POLAR */
            Gear(No.eq,cfg);
#  endif /*#!POLAR */
            CPUtime("Gear"); }

#  if defined(COULOMB) && COULOMB<=-1
        /* Ewald summation */
        if (Q) {
          StaSet(DT,lag.err,2,lag.n+2);
          StaAdd("Mx",Q->M[0]);
          StaAdd("My",Q->M[1]);
          StaAdd("Mz",Q->M[2]);
          StaAdd("M^2",SQR(Q->M)); }
#  endif   /*# COULOMB */

        nsteps++;

        if (!(option('m')==2)) {
          int n;
          vector *rp=cfg[0]->rp;
          vector *rpvel=cfg[1]->rp;

          if (thermostat==T_ANDERSEN) Maxwell(-1,No.N,h/tau.T);
          if (thermostat==T_ANDERSEN_CM) MaxwellCM(h/tau.T);
          if (thermostat==T_MAXWELL && justnow(tau.T,h/2)) Maxwell(-1,No.N,1);
          if (thermostat==T_MAXWELL_CM && justnow(tau.T,h/2)) MaxwellCM(1);

          if (measure && option('c')&4) {
            En.r2=constrainterror(cfg[0],cfg[1]);
            En.v2=vconstrainterror; }

          if (!(option('c')&1)) {
            loop (n,FROM,No.N) if (molec[n].nc)
              constrit[molec[n].sp].nit[COR_BOTH] +=
                Scorrect(molec+n,rof(molec+n,rp),rpvel,epsc); }
          else
            Lcorrect(cfg[0],cfg[1]);
          CPUtime("corr"); }

        normalize(-1);

        if (measure && option('c')&8) {
          En.r3=constrainterror(cfg[0],cfg[1]);
          En.v3=vconstrainterror; }

        CPUtime("norm");

        if (option('y')) if (justnow(dt.plb,h/2)) writeplayback();

#  ifdef POLAR
        if (scf.epsq>0) scf.eps = scf.eps*scf.epsq + (1-scf.epsq)*scf.eps0;
#  endif /*# POLAR */

        if (No.c) doconstrit(omegac);
        if ((drift&DRIFT_WHEN)==DRIFT_STEP) removedrifts(option('v')&64);

        init_append|=1; } /* iint */

      if (!finite(En.tot)) {
        sig=2;
        WARNING(("En.tot is infinite or not a number => stop as if Ctrl-C")) }

      if (No.c) { /* constraint errors */
        int i,sp;
        /* constraint iterations reported every cycle even if known every step */
        /* ... not for SHAKE */
        loop (i,0,COR_LAST) loop (sp,0,nspec) if (constrit[sp].nit[i])
          StaAdd(string("constrit%d spec%d",i,sp),
            (double)constrit[sp].nit[i]/spec[sp]->N);
        if (En.v1) StaAdd("v constr err before",En.v1);
        if (En.v2) StaAdd("v constr err mid",En.v2);
        if (En.v3) StaAdd("v constr err after",En.v3);
        if (En.r1) StaAdd("constr err before",En.r1);
        if (En.r2) StaAdd("constr err mid",En.r2);
        if (En.r3) StaAdd("constr err after",En.r3); }

      if (drift&DRIFT_DEPEND) depend_r(cfg[0],0); /* new in 2.0b */

      box.V=PROD(box.L);
#  ifdef FREEBC
      box.V=9e9; // UGLY!
#  endif /*# FREEBC */

      /* pressure [Pa] */
      /* virial-based pressure w/o cutoff corrections (elst virial=energy) */
      En.Pnc=((En.kin*2*En.Pkinq+En.vir)/DIM)/box.V*Punit;
      /* as above, w. cutoff corrections added */
      En.P=((En.kin*2*En.Pkinq+En.vir)/DIM+En.corr/box.V)/box.V*Punit;
      /* ... should be identical to En.Pcfg*Punit */

#  if (PRESSURETENSOR&PT_ANY) == PT_ANY
      /* pressure w. cutoff corrections based on the directly calculated pressure tensor */
      En.trPt=SUM(En.Ptens)/3+En.corr/(box.V*box.V)*Punit;
#  endif
      if (nomax>0) {
        if ((issta?StaN("Tkin")+1:1)>=nomax) sig=3; }

      if (stop>0)
        if (t>stop-h/2) sig=3;

      if (option('w')>1) expectedbuffer+=16;
      if (expectedbuffer>7500) {
        expectedbuffer=0;
        waitfordiskspace(64);
        fflush(out); }

/*.....  if (!option('g') || out!=stdout)*/
      if (icyc==0 || icyc+1==no
          || sig
          || justnow(dt.prt,halftcyc))

        if ( (option('v')&1) && option('m')!=1) {

          if (option('w')>1) expectedbuffer+=88;

#  ifdef POLAR
#    ifdef REDUNITS
          if (icyc==0)
            header("    t         T       Epot       Etot      rho      P     polerr  polit  cerr1 ");
          /* never active...*/
          prt("%8.3f %7.4f %9.2f %12.4f %7.4f %8.3f%10.2e%10.2e",
              t,    En.T,En.pot,En.tot,No.mass/box.V,
              En.P,En.polstderr,En.polnit,En.r1);
#    else /*# REDUNITS */
          if (icyc==0) {
            if (tau.R) header("   t/ps    Tkin/K   Epot/K    Etot/K     RvdE/A  P/MPa  polerr polit  cerr1 ");
            else header("   t/ps    Tkin/K   Epot/K    Etot/K  rho/kgm^-3 P/MPa  polerr polit  cerr1 "); }
          if (tau.R) prt("%9.3f %6.1f %9.0f %12.2f %7.4f %8.3f%9.2e%2d%10.2e",
              t,    En.T,En.pot,En.tot,sitedef[tau.i].LJ[0].RvdW,
              En.P/1e6,En.polstderr,En.polnit,En.r1);
          else prt("%9.3f %6.1f %9.0f %12.2f %7.2f %8.3f%9.2e%2d%10.2e",
              t,    En.T,En.pot,En.tot,No.mass/box.V*rhounit,
              En.P/1e6,En.polstderr,En.polnit,En.r1);
#    endif /*#!REDUNITS */
#  else /*# POLAR */
          double pcer=En.r1+En.r2+En.r3;

          if (pcer>0) pcer=-log10(pcer);
          else pcer=9.99;
          if (icyc==0) {
            prt("\nNOTE: pcer = -log10(summary constraint error), recommended pcer>6");
            if (T) prt_("NOTE: the graph of Tkin is in the range of [0,2*T]=[0,%g] K",T*2);
            else prt_("NOTE: the graph of Tkin is in the range of [0,0.2] K");
#    ifdef REDUNITS
            header("    t         T       Epot       Etot      rho      P  pcer0-------Tkin------2T");
#    else /*# REDUNITS */
            if (tau.R) header("   t/ps    Tkin/K   Epot/K    Etot/K     RvdE/A  P/MPa pcer0-------Tkin------2T");
            else header("   t/ps    Tkin/K   Epot/K    Etot/K  rho/kgm^-3 P/MPa pcer0-------Tkin------2T");
#    endif /*#!REDUNITS */
          }

#    ifdef REDUNITS
          prt_("%8.3f %7.4f %9.2f %12.4f %7.4f %6.3f%5.2f",
              t,    En.T,En.pot,En.tot,No.mass/box.V*rhounit,En.P,pcer);
#    else /*# REDUNITS */
          /*** standard nonpolar prt ***/
          if (tau.R) prt_("%9.3f %6.1f %9.0f %12.2f %7.4f %6.1f%5.2f",
               t,    En.T,En.pot,En.tot,sitedef[tau.i].LJ[0].RvdW,
               En.P/1e6,pcer);
          else prt_("%9.3f %6.1f %9.0f %12.2f %7.2f %6.1f%5.2f",
               t,    En.T,En.pot,En.tot,No.mass/box.V*rhounit,
               En.P/1e6,pcer);
#    endif /*#!REDUNITS */
          graph(En.T/(T+0.1*(T==0))*0.5,20);

          if (!(option('m')==2))
            if (En.r1>=0 || En.r2>=0 || En.r3>=0) prt(
             "                                                 %10.2e%10.2e%10.2e",
             En.v1,En.v2,En.v3);

#    ifdef erminmax
          if (!(option('m')==2))
            prt("%8.5f %8.5f",sqrt(Erfc.min),sqrt(Erfc.max));
          Erfc.min=3e33; Erfc.max=0;
#    endif /*# erminmax */
#  endif /*#!POLAR */
        }

#  ifndef FREEBC
        {
          double minLh=fmin(fmin(box.Lh[0],box.Lh[1]),box.Lh[2]),perc;
          static double oldperc;
          static int info=1;

          if (box.cutoff>minLh*1.005) {
            perc=(box.cutoff/minLh-1)*100;
            if (info)
              prt("\
Explanation: The box size is variable in NPT simulation while the cutoff is\n\
  constant. If the cutoff is inappropriately set (e.g., to minimum L/2 at\n\
  simulation start which is the default), it may happen that the cutoff\n\
  exceeds minL/2. Overshooting by less than 0.5 %% is acceped without warning.\n\
  More than a few %% is inaccurate and not acceptable in productive runs");
            if (perc>oldperc)
              WARNING(("cutoff=%f exceeding minL/2=%f by %.2f%%\n\
*** (more warnings for cutoff exceeding minL/2 by more than %.2f%%)",
                        box.cutoff,minLh,perc,oldperc=perc*1.2))
            info=0; }
         }
#  endif /*# FREEBC */

      if (sig==5) break; /* 2nd attempt */

      depend_r(cfg[0],0); /* correct r for measurements (e.g., dipole moment) */

      { /*** statistics and convergence profile ***/
        struct ssd_s *ssd;
#  ifdef FREEBC
        double Epot=En.pot*Eunit;
#  else /*# FREEBC */
        double Epot=(En.pot+En.corr/box.V)*Eunit;
#  endif /*#!FREEBC */
#  ifdef RGYR
        double endtoend,Rgyr;
#  endif /*# RGYR */
#  ifdef WIDOM
        double Wid;
#  endif /*# WIDOM */
#  ifdef XSECTION
        static double Xsec;

        if (xs.freq && icyc%xs.freq==0) {
          StaSet(DT*abs(xs.freq),lag.err,2,lag.n);
          if (xs.mode&8)
            Xsec=cXsection(xs.grid,xs.mode&7,xs.RvdW,xs.Rscale,
                           xs.freq<0,xs.cp,xs.ns,xs.sizelimit,&xs.maxsize);
          else         
            Xsec=mXsection(xs.grid,xs.mode&7,xs.RvdW,xs.Rscale,
                           xs.freq<0,xs.cp,xs.ns,xs.sizelimit,&xs.maxsize);
          StaSet(DT,lag.err,2,lag.n); }

#  endif /*# XSECTION */

#  ifdef RGYR
        Rgyr=measureplus(&endtoend,rg.end,rg.cp);
#  else /*# RGYR */
        measureplus();
#  endif /*#!RGYR */

        bounddr(drmax);
        if ((drift&DRIFT_WHEN)==DRIFT_CYCLE) removedrifts(option('v')&64);

        if (option('m')!=2) {
          /*** independent pressure measurement via <dU/dV> */
          /* SHAKE: ==> moved to constrd.c, Shake() */
          measureP(1);
          measureP(2); }

#ifdef POLAR
        if (scf.E) measureepspol();
#endif

        ssdistance();

/*** WARNING: watch the former place of this part*/
#  ifdef WIDOM
        if (widom.spreal>=0) Wid=XWidom(widom.spreal,widom.sp);
#    ifdef SLAB
        else Wid=Widom(widom.sp,widom.n,widom.z0,widom.z1,widom.dz,widom.mode);
#    else /*# SLAB */
        else Wid=Widom(widom.sp,widom.n,widom.corr);
#    endif /*#!SLAB */
#  endif /*# WIDOM */

#  ifdef DIHHIST
        StaSet(DT,lag.err,2,lag.n);
        gauchetrans(-2);
#  endif /*# DIHHIST */

#  ifdef HARMONICS
        measureharmonics();
#  endif /*# HARMONICS */

#  ifdef SLAB
        measuredpr();
#  endif /*# SLAB */

        if (CPbuf) {
          float *CPrec=CPbuf+icyc*NCP;
          int count=0;

          memset(CPrec,0,NCP*sizeof(CPrec[0]));
#  define recordCP(X) { if (count<NCP) CPrec[count++]=(float)(X); }
          /* order of recordCP must match CPinfo in simmeas.c */
          /*column (index is by 1 less)*/
          /*1*/ recordCP(En.tot)
          /*2*/ recordCP(En.T)
          /*3*/ recordCP(Epot/SIE)

          /* columns 4,5 of the convergence profile */
          if (tau.P==0 && tau.R==0 && tau.rho==0) {
            if (dV==0) {
              if (thermostat==T_NOSE || thermostat==T_NPT) recordCP(En.U*Eunit/SIE)
              else if (No.first) recordCP(En.pot0*Eunit/SIE)
              else recordCP(En.intra*Eunit/SIE) }
            else
              recordCP(En.PdV)
            recordCP(En.P) }
          else {
            if (tau.R) recordCP(sitedef[tau.i].LJ[0].RvdW)
            else recordCP(rhounit*No.free_mass/box.V);
            if (dV==0) recordCP(En.P)
            else recordCP(En.PdV) }
          /*6*/ recordCP(En.T_in)
          /*7*/ recordCP(En.T_tr)
#  ifdef COULOMB
          /*8*/ if (Q) {
                  if (Eext.isE<0) recordCP(Q->M[Eext.isE+3])
                  else recordCP(sqrt(SQR(Q->M))) }
                else count++;
#  endif /*# COULOMB */
#  ifdef SHEAR
          /*9*/ recordCP(En.Cv)
#  endif /*# SHEAR */
#  ifdef POLAR
          /*10*/ recordCP(En.self)
#  endif /*# POLAR */
#  ifdef DIHHIST
          /*11*/ recordCP(gauchetrans(dih.cp))
          if (dihcp) fprintf(dihcp,"\n");
#  endif /*# DIHHIST */
#  ifdef RGYR
          /*12*/ recordCP(Rgyr)
          /*13*/ recordCP(endtoend)
#  endif /*# RGYR */
#  ifdef XSECTION
          /*14*/ recordCP(Xsec) /* warning: need not be at the same frequency! */
#  endif /*# XSECTION */
#  ifdef WIDOM
          /*15*/recordCP(Wid)
#  endif /*# WIDOM */

            // moved to constrd.c :
            //          En.logs=cfg[0]->logs;
            //          En.dlogs=cfg[1]->logs;

            // big fixed in 2.7f: was En.pot instead of En.U
          En.Hnc=En.U+En.Pnc/Punit*box.V;
          En.H=En.U+En.corr/box.V+En.P/Punit*box.V;

#  if (PRESSURETENSOR&PT_ANY) == PT_ANY
          En.Hz=En.U+En.Ptens[2]/Punit*box.V;
#  endif /*# (PRESSURETENSOR&PT_ANY) == PT_ANY */

          /* variables recorded in SIMNAME.cpi (incl. distances) */
          looplist (ssd,ssd0) {
            if (ssd->indx[0]==-1) recordCP(*ssd->u.q*Eunit/SIE)
            else if (ssd->indx[0]==-2) recordCP(*ssd->u.q)
            else recordCP(ssd->u.dist); } }

        StaSet(DT,lag.err,2,lag.n);
        StaAdd("Tkin",En.T); issta=1;
        StaAdd("Tin",En.T_in);
        StaAdd("Ttr",En.T_tr);
        StaAddE("Epot",Epot);
        StaAddE("Ein",En.intra*Eunit);
        if (No.first)
          StaAddE(string("Epot0 (<%d)",No.first),En.pot0*Eunit);
        StaSet(DT,2,2,0);
        StaAdd("virial",En.vir);
        StaSet(DT,lag.err,2,lag.n);
        StaAdd("P [Pa]",En.P);
        if (dV) StaAdd("P-PdV [Pa]",En.P-En.PdV);
        if (tau.P) StaAdd("V",box.V);
        if (tau.R) StaAdd("RvdW",sitedef[tau.i].LJ[0].RvdW);

#  if PRESSURETENSOR&PT_OFF
        if (lag.visc) {
          StaSet(DT,lag.visc,2,0);
          StaAdd("Ptyz [Pa]",En.Ptens[3]);
          StaAdd("Ptzx [Pa]",En.Ptens[4]);
          StaAdd("Ptxy [Pa]",En.Ptens[5]); }
#  endif /*# PRESSURETENSOR&PT_OFF */

        StaSet(DT,lag.err,2,lag.n);
        StaAdd("Etot [K]",En.tot);
        StaAddE("internal energy",En.U*Eunit);
        StaAddE("enthalpy",En.H*Eunit);
        if ((thermostat==1 || thermostat>2) && tau.E==0)
          LRAdd("Etot",1,t,En.tot);
        if (lastEtot<2.999e33) {
          lastEtot -= En.tot;
          StaSet(DT,2,2,0);
          StaAdd("|dEtot|",fabs(lastEtot));
          StaAdd("dEtot^2",lastEtot*lastEtot); }
#  if (PRESSURETENSOR&PT_ANY) == PT_ANY
        StaSet(DT,lag.err,2,lag.n);
        StaAdd("P(Ptens) [Pa]",En.trPt=SUM(En.Ptens)/3+En.corr/(box.V*box.V)*Punit);
        StaSet(DT,2,2,0);
        StaAdd("P(Ptens)-P [Pa]",En.trPt-En.P);
        if (dV) StaAdd("P(Ptens)-PdV [Pa]",En.trPt-En.PdV);
#  endif /*# (PRESSURETENSOR&PT_ANY) == PT_ANY */

        StaSet(DT,lag.err,2,lag.n);
        looplist (ssd,ssd0) if (ssd->stat) {
          if (ssd->indx[0]==-1) StaAddE(ssd->name,*ssd->u.q*Eunit);
          else if (ssd->indx[0]==-2) StaAdd(ssd->name,*ssd->u.q);
          else StaAdd(ssd->name,ssd->u.dist); } }

#  ifdef DIHF
#    include "dihf.c"
#  endif /*# DIHF */

      if (dt.cfg!=0) {
        int n=(t+halftcyc)/dt.cfg;

        if ( n > (int)((t-halftcyc)/dt.cfg) ) {
          stoptime=mytime();
          if (tau.R) RvdW=sitedef[tau.i].LJ[0].RvdW;
          savecfg(n,(int4)stoptime,&RvdW); } }

      if (tau.E!=0) {
        double Etot=En.tot;
        if (tau.P) Etot+=P/Punit*box.V; /* enthalpy in K */
        rescalecfg(cfg[1],RESCALE_XYZ,exp(halftcyc*(E-Etot)/En.kin/tau.E),0,0); }

      StaSet(DT,2,2,0);
      if (drift&7) StaAdd("CM drift",En.cmdrift);
      if (drift&7*8) StaAdd("v drift",En.vdrift);
      if (drift&7*64) StaAdd("angular drift",En.angdrift);

#include "mainresc.c"

      // ??? why this ???
      //      V=L[0]*L[1]*L[2];

      init_append|=2;

      if (!option('s') && option('b')) if (icyc%option('b')==0) {
        /* raises sig if SIMNAME.stp exists */
        FILE *stp=fopen(Fn("stp"),"rt");

        if (stp) {
          fprintf(stderr,"%s exists, interrupting\n",lastFn);
          sig=4;
          fclose(stp); } }

    } /*** icyc (# of cycles=no, or until ^C) ***/
    /* >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>> */

    if (option('v')&1) header("");
    if (option('d')) distancecheck();

#  ifdef POLAR
    if (scf.epsq>0) {
      put3(scf.epsq,scf.eps,scf.eps0)
      scf.eps=scf.eps0; scf.epsq=-fabs(scf.epsq); }
#  endif /*# POLAR */

    underline("sweep timing");
    prt("planned sweep (one ';' in the data) = %d cycles by %d steps by %g ps",
        no,noint,h);

    stoptime=mytime();
    stop_start=stoptime-starttime;

    prt_("stopped at real time = %.4f s = %s",stoptime,myctime(stoptime));
    prt("duration %.4f s = %.3f min = %.5f h",stop_start,stop_start/60.,stop_start/3600.);
    prt("stopped at MD time t=%.5f ps",t);
    prt("sweep MD time = %.11g ps = %d MD steps (%d cycles by %d steps)",
        nsteps*h,nsteps,icyc,noint);
    if (nsteps && stop_start>0)
      prt("speed = %g cycles/min = %g cycles/hour = %g ns/day",
          60.0/((double)stop_start/nsteps*noint),
          3600.0/((double)stop_start/nsteps*noint),
          86.4*h/((double)stop_start/nsteps));

    fprintf(stderr,"%d of %d*%d steps (t=%.3f ps) stop at %s",
            nsteps,icyc,noint,t,myctime(stoptime));

    if (option('t')) {
#  ifdef CHEAPTIME
      if (nsteps) {
        int i=(int)stop_start%nsteps;

        if (i>nsteps/2) i=nsteps-i;
        if (i<=(int)sqrt(nsteps)) prt("CHEAPTIME: possible interference"); }
#  endif /*# CHEAPTIME */
      printtime(); }

#  if PARALLEL
    printpartimes();
#  endif /*# PARALLEL */
    _n

    if (option('s') && option('b')>1) fputc('\a',stderr);

    /* results of this run (one ; in data)
       NB: mainres.c contains release(mark) */

#  include "mainres.c"

#endif /*#!MARK */
    if (tau.sat<0) init=1;
    else init=0; /* default in every pass */
  } /* >>> run >>> */

 TheEnd:

 checkranges(1);

  if (sig>0) {
    if (sig==3) prt("finished because nomax or stop reached");
    else if (sig==4) prt("quitted because of %s",Fn("stp"));
    else prt("quitted with sig=%d",sig); }

  if (fclose(out)) Error("close out");
  fclose(in);

  if (option('s') && option('b')) fputc('\a',stderr);

  return 0;
}
