#include "ground.h"
#include "alloc.h"
#include "bitfile.h"
#include "pakcp.h"

static int efread(void *rec,size_t size,size_t n,FILE *f,int reverse)
{
int ret=fread(rec,size,n,f);
int i;
if (size!=4) Error("bad architecture");
if (reverse) loop (i,0,ret) endian((char*)rec+4*i);
return ret;
}

static int NCP,reverse; /* used for decompressing routines only */
static int verbose=1; /* verbose=1 requests info printed; set by getcpheader */

float *getcpheader(FILE *f,int *NCPp,int *reversep,int v) /***** getcpheader */
/*
  reads 1st record in f (packed or unpacked)
  returns:
    *NCPp and NCP 
    *reversep is flipped if reverse endian found for NCP
    the record read as float[NCP] (space is allocated)
*/
{
float hdr[2];
float *r;
int4 ncp;

verbose=v;

if (fread(hdr,sizeof(hdr),1,f)!=1) ERROR(("file too short"))

if (hdr[0]!=CPmark) ERROR(("bad header of cp file"))
ncp=*(int4*)(&hdr[1]);
if (ncp>127) {
  *reversep=!*reversep;
  if (verbose) prt("REVERSING ENDIAN");
  endian((char*)&ncp);
  if (ncp<2 || ncp>32) ERROR(("bad file or old .cp format")) }
*NCPp=(int)ncp;
NCP=*NCPp;
reverse=*reversep;
if (verbose) prt("ncp=%d",NCP);
alloc(r,ncp*sizeof(float));
*(int4*)(&hdr[1])=ncp; /* pasting back with correct endian */
copy(r,hdr,2*sizeof(float));
if (fread(r+2,sizeof(float),NCP-2,f)!=NCP-2) ERROR(("file too short"))
return r;
}

int4 appendpakcp(
  FILE *f,    /* file must be opened for writing in binary mode */
  int NCP,    /* # of CP items recorded; WARNING: shades static NCP */
  int nbit,   /* resolution is 2^nbit in the min-max range */
  void *data, /* if norec=0 then the head of rec_t* list 
                 if norec>0 then pointer to r[norec][NCP] (one contig.array) */
  int4 norec)

/*
   appends packed convergence profile to (opened) file f
   nbit : the range between min and max value is divided into 
          2^nbit subintervals
   input configuration
*/
{
int i,bcut,ibcut;
float *minr,*maxr,*r;
int4 ir,dir,*ir0,*len,*minlen,no;
int4 *bestbcut; /* int 4 because written - compatibility! */
rec_t *rec;

alloc(minr,NCP*sizeof(float));
alloc(maxr,NCP*sizeof(float));

alloc(ir0,NCP*sizeof(int4));
alloczero(bestbcut,NCP*sizeof(int4));
alloc(len,NCP*sizeof(int4));
alloczero(minlen,NCP*sizeof(int4));

loop (i,0,NCP) {
  minr[i]=3e33;
  maxr[i]=-3e33;
  minlen[i]=0x7fffffffL; }

for (no=0, rec=(rec_t*)data;
     norec ? no<norec : rec!=NULL;
     no++, rec = norec ? NULL : rec->next) {
  r = norec ? (float*)data+no*NCP : rec->r;

  loop (i,0,NCP) {
    Min(minr[i],r[i])
    Max(maxr[i],r[i]) } }

if (verbose) prt_("no=%d   ",no);
assignbit(f,'w');

/* to improve! */
loop (i,0,NCP) if (maxr[i]!=minr[i]) {
  float x=maxr[i];
  int shift=nbit;
  do maxr[i]+=(maxr[i]-minr[i])/(1<<shift); 
    while (shift-- && maxr[i]==x);
  if (++shift!=nbit)
    if (verbose) prt("! %d : [%.9g, %.9g->%.9g] (shift=%d)",
		     i,minr[i],x,maxr[i],shift); }

if (fwrite(minr,sizeof(float),NCP,f)!=NCP) Error("cannot write");
if (fwrite(maxr,sizeof(float),NCP,f)!=NCP) Error("cannot write");
if (fwrite(&no,sizeof(int4),1,f)!=1) Error("cannot write");

/* this is for all writes not through bitput */
/*.....bitfile.bytecount = 2*sizeof(float)*NCP + sizeof(int4)*(2+NCP);*/

for (ibcut=nbit-1; ibcut>=1; ibcut--) {
  memset(ir0,0,NCP*sizeof(int4));

  if (ibcut==1) {
    if (verbose) prt_("bitcut: ");
    loop (i,0,NCP) {
      if (maxr[i]==minr[i]) bestbcut[i]=0;
      if (verbose) prt_("%2d ",bestbcut[i]); }
    if (verbose) prt("/%3d",nbit); 

    if (fwrite(bestbcut,sizeof(int4),NCP,f)!=NCP) Error("cannot write");

    ir=(int4)nbit;
    if (fwrite(&ir,sizeof(int4),1,f)!=1) Error("cannot write"); }

  bcut=ibcut;
  memset(len,0,NCP*sizeof(int4));

  for (no=0, rec=(rec_t*)data;
       norec ? no<norec : rec!=NULL;
       no++, rec = norec ? NULL : rec->next) {
    r = norec ? (float*)data+no*NCP : rec->r;

    loop (i,0,NCP) if (maxr[i]!=minr[i]) {
      if (ibcut==1) bcut=(int)bestbcut[i];
      ir=(int4)((r[i]-minr[i])/(maxr[i]-minr[i])*(1<<nbit)+0.5);

      /* fix rounding problems ... */
      if (ir<0 || ir>=(1<<nbit)) {
        if (ir<0) ir=0;
        if (ir>=(1<<nbit)) ir=(1<<nbit)-1; }

      dir=ir-ir0[i]+(1<<(bcut-1));
      if (dir>=0 && dir<(1<<bcut)) {
        len[i]+=1+bcut;
        if (ibcut==1) {
          putbit(1,1); putbit(dir,bcut); } }
      else {
        len[i]+=1+nbit;
        if (ibcut==1) {
          putbit(0,1); putbit(ir,nbit); } }
      ir0[i]=ir; }
    }
  if (ibcut>1) {
    loop (i,0,NCP) {
/*.....      prt_("%5d ",len[i]);*/
      if (len[i]<minlen[i]) {
        minlen[i]=len[i];
        bestbcut[i]=bcut; } }
/*.....    _n*/
    }
  }

flushbit();
fwrite(&bitfile.checksum,sizeof(bitfile.checksum),1,f);
if (verbose) prt("bitfile.bytecount=%ld",(long)bitfile.bytecount);

free(minlen);
free(len);
free(bestbcut);
free(ir0);

free(maxr);
free(minr);

return bitfile.bytecount;
}


/* %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
  Reading packed CP.  Use as follows:

  FILE *pakcp=fopen(NAME,"rb");
  int NCP;
  float *r=getcpheader(pakcp,&NCP,&reverse); // r[NCP] gets allocated 

  // now r[NCP] is the file header 

  while (getpakcp(pakcp)) {
    while (nextcprec(r)) {
      // now r[NCP] is one record (records with CPmark included!)
      }
    endgetpakcp(pakcp); }
*/

static float *minr,*maxr,*markrec;
static int4 *ir,*bestbcut;
static int bcut,nbit;
static int4 no;

int nextcprec(float *r) /***************************************** nextcprec */
/*
  gets next record of packed CP file
  returns 1 on success,
  0 if eof or CPmark encouneterd
*/
{
int i;

if (markrec) {
  copy(r,markrec,NCP*sizeof(float));
  free(markrec); 
  return 1; }

if (--no<0) return 0;

loop (i,0,NCP)
  if (maxr[i]-minr[i]) {
    bcut=(int)bestbcut[i];
    if (getbit(1)) ir[i]+=getbit(bcut)-(1<<(bcut-1));
    else ir[i]=getbit(nbit);
    r[i]=minr[i]+(float)ir[i]/(1<<nbit)*(maxr[i]-minr[i]); }
  else 
    r[i]=minr[i];

return 1;
}

int getpakcp(FILE *f) /******************************************** getpakcp */
/*
  begins reading one CP series (`no' records, i.e. between CPmark's)
*/
{
int i;
int4 ibc;

assignbit(f,'r');
if (feof(f)) return 0;

alloc(minr,NCP*sizeof(float));
alloc(maxr,NCP*sizeof(float));

alloczero(bestbcut,NCP*sizeof(int4));
alloczero(ir,NCP*sizeof(int4));

if (efread(minr,sizeof(float),NCP,f,reverse)!=NCP) return 0;

if (minr[0]==CPmark) {
  markrec=minr; /* pointer assignment !!! */
  alloc(minr,NCP*sizeof(float));
  if (efread(minr,sizeof(float),NCP,f,reverse)!=NCP) return 1; }

if (efread(maxr,sizeof(float),NCP,f,reverse)!=NCP) Error("cannot read");

/*.....loop (i,0,NCP) prt("%2d %g %g",i,minr[i],maxr[i]);*/
 
loop (i,0,NCP) if (maxr[i]<minr[i]) Error("max<min");
loop (i,0,NCP) if (fabs(maxr[i])>3e33 || fabs(minr[i])>3e33)
  Error("too big number");

if (efread(&no,sizeof(no),1,f,reverse)!=1) Error("cannot read");

if (efread(bestbcut,sizeof(int4),NCP,f,reverse)!=NCP) Error("cannot read");

if (efread(&ibc,sizeof(int4),1,f,reverse)!=1) Error("cannot read");
nbit=(int)ibc;

if (verbose) {
  prt_("bitcut: ");
  loop (i,0,NCP) prt_("%2d ",bestbcut[i]);
  prt("/%3d  no=%d",nbit,no); }

/* WARNING: this is symptomatic ad hoc bug fix ... 
there is rather something wrong in WRITING -- extra 4B */
if (no==1) efread(&i,sizeof(i),1,f,reverse);

return 1;
}

void endgetpakcp(FILE *f) /************************************* endgetpakcp */
/*
  ends reading one CP series (`no' records, i.e. between CPmark's)
*/
{
unsigned4 checksum=0;

flushbit();

if (bitfile.bytecount) {
  fread(&checksum,sizeof(checksum),1,f);
  if (reverse) endian((char*)&checksum);

  if (checksum!=bitfile.checksum)
    ERROR(("checksum %lu, expected %lu",
	   (long unsigned) bitfile.checksum,checksum)) 
  if (verbose) prt("bitfile.bytecount=%ld",(long)bitfile.bytecount); }

if (bestbcut) {
  free(bestbcut);
  free(ir);
  free(maxr);
  free(minr); }
}

void endian(char *a) /********************************************** endian */
{
char x;
x=a[0]; a[0]=a[3]; a[3]=x;
x=a[1]; a[1]=a[2]; a[2]=x;
}

