/*
   %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
   %                        some global variables                          %
   %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
*/

#include <stdio.h>
#include "simglob.h"

pot2_t ***pot;    /* array [nspec][nspec] of pointers to functions computing
                     INTERmolecular interactions */

#ifdef LINKCELL
pot1_t  **pot1;   /* array [nspec] of pointers to functions computing
                     INTRAmolecular interactions */
#endif


specinfo_t **spec;/* array [nspec] of pointers to species definitions */
M_t **Ms;         /* array [nspec or N] of pointers to matrices M */

molecule_t *molec;/* array [N] of molecule descriptors */
ToIntPtr cfg[9];  /* [MaxGearOrder+3] */

/* warning: even #ifdef FLOAT, the constraint calculations are in double */
double **gs;      /* array [No.pred]:
                     tables of Lagrange multipliers g to predict their values
                     NOTE: pointers are NOT constant because swaping of cor-
                     responding arrays is implemented by pointer interchange */
double *binom;    /* coefficients of this predictor */

rdf_t ***rdf; /* table of pointers to rdf's
                 note: also in each sitesite_t */

int init=0;       /* INITIAL KEY :
__init____.CFG___.CP____.STA___t____logs
    0     read  append  read  read  read
    1     read  append  new    0    read
    2     read  rewrite new    0     0
    3     init  rewrite new    0     0  (sequential version only)
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/
int init_append; /* = init but :=0 after one step (to control appending) */

No_t No;          /* # of sites, constraints, degrees of freedom etc. */

double t;         /* MD time */
#ifdef COOK
real   h=0.001;   /* time step */
#else
real   h=0.002;   /* time step */
#endif
//vector L,Lh;      /* size of the periodic box, Lh=L/2 */
double equalize=0; /* masses equalization factor; 0=no equalization */

double eps=1e-6;  /* accuracy for computing Lagrange multipliers
                     if >=1 then # of iteration steps
                     (was 1e-5 prior V 2.0d) */
double omega=0;   /* not implemented */

double epsc=1e-6; /* accuracy for correcting constraints
                     if epsc>=1, then # of iteration steps
                     (default is changed for -c4);
                     accuracy for Shake (was 1e-5 prior V 2.0d) */
double omegac=-1.2;/* relaxation parameter for 'correct'
                     relaxation parameters for Shake */

#ifdef POLAR
             /* eps,epsq, omega, epsx, omegax, E, Estride, maxit ... */
struct scf_s scf={1,0.8, -9,     3e33,   0.95,   0, 0,       30};
#endif

#ifdef REDUNITS
double T=1;       /* Nose|friction temperature (if tau.T) */
double P=1;       /* pressure (friction, if tau.P) */
#else
double T=300;     /* Nose|friction temperature (if tau.T) */
double P=1e5;     /* pressure (friction, if tau.P) */
#endif

En_t En;          /* energies, virial, pressure tensor, constraint errors.. */
int GearOrder;    /* see also option('m') */
int rescale=RESCALE_XYZ|RESCALE_CM;

struct lag_s lag={32,15
#ifdef RGYR
                  ,3,0x7fffffff
#endif
}; /* err,nblocks,[dim,nv,v]... */

double E=0;       /* total energy in Eunit to be kept constant */
double T_tr_in=1; /* ratio T_tr/T_in for thermostat=13 */
int thermostat=0; /* see simglob.h */
tau_t tau;        /* thermostat, volume change,... */

int measure;      /* 0: only forces required
                     1: also energy etc.
                     2: special for calculating # of pairs (DEBUG only) */

/*** Ewald related global variables - see  comments in ewald.c and erfc.c ***/
#ifdef LINKCELL
struct box_s box;
#else
struct box_s box = { 3, 1e-4 };
#endif

        /* double...................................................  int............................ */
        /* alpha,kappa, epsr,epsk, minqq,epsinf,diff,Perr,rplus, sat, rshift, grid,test,diag,centroid */
#if COULOMB==2 || COULOMB==0                                             
struct el_s el={0.7,0,   0,   0,   1,    1,     0.05,1e7, 2.5,   0,   0,      256, -10, 0,   1};
#elif COULOMB==3                                                             
struct el_s el={0.2,0,   0,   0,   1,    1,     0.05,1e7, 2.5,   0,   0,      256, -10, 0,   1};
#elif COULOMB==-1                                                            
struct el_s el={0.2,0.2, 0.05,0.5, 1,    3e33,  0.05,1e6, 2.5,   0,   3,      512, -10, 0,   1};
#elif COULOMB==-2                                                            
struct el_s el={0.2,0.2, 0.05,0.5, 1,    3e33,  0.05,1e6, 2.5,   0,   3,      256, -10, 0,   1};
#else
#error "unsupported COULOMB (one of -1,-2,0,2,3 expected)"
#endif

int cache=64;     /* block size for pair sums:
                     cache=1: simplest pair forces in 2 nested loops
                        (good for small systems)
                     cache>0: blocked (4 nested loops total)
                        (large systems of small molecules)
                     cache<=0: recursive triangulation
                        (large systems of polyatomic molecules) */

/* Q_t *Q; is in ewald.h and ewald.c */

#ifdef POLAR
int polar_off;    /* offset of auxiliary charges in the units of bytes */
#endif

#ifdef SHEAR
double shear=0;   /* max gradient of acceleration for cos(z) method */
#endif

fixsites_t *fixsites0;
ToIntPtr fixa;

int drift=DRIFT_AUTO; /* controls how numerical drift of CM, momentum, ang.mom. is fixed */
int conserved=-1; /* number of conserved degrees of freedom */

/* int norm; OLD: 0: default according to b.c. and fixsites0
             1: always remove drift
            -1: never remove drift of momentum
            0,2,-2: the same for rotations and angular momentum */

struct center_s center;

#ifdef SLAB
struct slab_s slab = { 0,0,0x7fffffff,0,-1,0.25,0.75 };

#if SLAB & 2
struct cleave_s cleave = { 0,0,-0x7fffffff };
#endif
#endif

struct Eelst_s Eext;

#ifdef ANCHOR
struct anchor_s anchor;
#endif

int onefourinrdf=0; /* control whether 1-4 included in RDF's */
