#ifndef _SIMGLOB_
#  define _SIMGLOB_

#  ifndef SHAKE
#    define SHAKE 1
#  endif /*# SHAKE */

/*
#  define SERIAL
for debugging only - serial loop instead of starting threads in parallel */

#  include "simopt.h"

#  if defined(FREEBC)+defined(NIBC)+defined(COULOMB) != 1
#    error Exactly one of {FREEBC,NIBC,COULOMB} must be #defined. Check simopt.h!
#  endif

#  ifdef GOLD
#    define WALL
#  endif /*# GOLD */

/* do not forget to #include "simopt.h" into simulation and project-specific
modules that do not #include "simglob.h" ! */

#  include "vector.h"

#  ifdef FLOAT
#    define realfmt "%f"
#  else /*# FLOAT */
#    define realfmt "%lf"
#  endif /*#!FLOAT */

#  ifndef PARALLEL
#    define PARALLEL 0
#  endif /*# PARALLEL */

#  if PARALLEL
#    include "sys4par.h"
#  endif /*# PARALLEL */

#  if defined(WIDOM) && defined(FREEBC)
#    error WIDOM requires periodic boundary conditions but FREEBC selected
#  endif

/* PRESSURETENSOR: sum of flags: */
#  define PT_VIR 1 /* virial (=configurational + constraint) part */
#  define PT_KIN 2 /* kinetic part */
#  define PT_ANY 3 /* virial or kinetic */
#  define PT_OFF 4 /* also off-diagonal terms (with PT_VIR and PT_KIN) */
#  define PT_MOL 8 /* also molecular-based (center-of-mass), modifies PT_KIN */
#  define PT_MOM 16 /* SPECIAL: 2nd and 3rd moments of center-of-mass energy */
#  ifndef PRESSURETENSOR
#    define PRESSURETENSOR 0
#  else /*# PRESSURETENSOR */
#    if PRESSURETENSOR == PT_OFF
#      error PRESSURETENSOR&PT_OFF (off-diagonal) and none of PRESSURETENSOR&PT_VIR (virial) and PRESSURETENSOR&PT_KIN (kinetic)
#    endif /*# PRESSURETENSOR == PT_OFF */
#    if PRESSURETENSOR == PT_MOL
#      error PRESSURETENSOR&PT_MOL (off-diagonal) and not PRESSURETENSOR&PT_KIN (kinetic)
#    endif /*# PRESSURETENSOR == PT_MOL */
#    if PRESSURETENSOR&PT_OFF
#      define PT_DIM (2*DIM)
#    else /*# PRESSURETENSOR&PT_OFF */
#      define PT_DIM (DIM)
#    endif /*#!PRESSURETENSOR&PT_OFF */
#    if (PRESSURETENSOR & PT_MOM) && !(PRESSURETENSOR&PT_MOL)
#      error PRESSURETENSOR: higher moments of CM velocity requested without CM velocity (PT_KIN)
#    endif /*# (PRESSURETENSOR & PT_MOM) && !(PRESSURETENSOR&PT_MOL) */
#  endif /*#!PRESSURETENSOR */

#  ifdef COOK
#    include "intrapot.h"
#  endif /*# COOK */

/*
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    %                     basic control structures                      %
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

Species are numbered by integers from zero.
Information on each species is stored in a structure of type `specinfo_t' which
contains number of sites, masses and charges of sites, number of constraints
and corresponding sites and lengths.  These structures are of variable length
and there is an array of pointers 'spec' to these structures.
Example:  spec[1]->si[2].charge  is charge of site 2 of species 1

Any of N molecules (numbered from 0 to N-1) is described by a structure of
type `molecule_t' which contains the species number, number of sites and
constraints, information where to find the physical configuration.  There is
an array of these molecule descriptors called `cfg'.

The physical configuration is stored in a special contiguous array of vectors
which is a part of a self-defined structure (`SDS', see "alloc.c") of type
`ToInt'.  There are several such self-defined structures containing higher
derivatives etc. as required by the Gear predictor-corrector method for solving
the differential equations.

The input information on interaction sites is stored in array of type
`sitedef_t' - see "cook/simdef.h"

9/96: dependants:
   these are auxiliary sites whose positions are defined as linear
   combinations of other sites.
   Let
     depend_t *d;
     struct depitem_s *dep=d->dep;

   then
     r[d] =       SUM       r[dep[j].i] * dep[j].w
                j<d->n

   and analogously for the forces
9/2010: dependants extended to out-of-plane and code changed
*/

#  ifdef COOK
/* WARNING: this differs from blend: NBFIX solved there */
enum exc_e { END=-1,EXCL=0,ONEFOUR=1 };

typedef struct exception_s {
  int indx;   /* atom index */
  enum exc_e type;
} exception_t;

#  endif /*# COOK */

typedef int pair_t[2];

typedef struct {
  real mass;      /* mass of site [extended! - more efficient, more memory] */
  real imass;     /* 1/mass of site */
  real charge;    /* charge of site */
  real bondq;     /* squared lengths of constraints (bonds) */
  pair_t pair;    /* site1 and site2 of constraint */
#  ifdef POLAR
  real chargepol; /* charge of auxiliary site to mimic polarisation */
  real qqpol;     /* chargepol^2 */
  real alpha_qq;  /* alpha/chargepol^2 where alpha=polarisability */
#    if POLAR&2
  real Esat;      /* saturation energy */
#    endif /*# POLAR&2 */
#    if POLAR&1
  real qkappa;    /* chargepol*kappa */
#    endif /*# POLAR&1 */
#define SHELL1  1 /* (POLAR&1) 1st of the shell-core pair (anion,kappa!=0) */
#define SHELL2  2 /* (POLAR&1) 2nd of the pair (cation, repulsive); param. rep */
#define CHARGED 4 /* charged atom (nonzero central charge) */
#define DRUDE   8 /* polarizable: nonzero DRUDE charge */
#define FQ     32 /* polarizable: fluctuating charge */
  int qtype;  /* 0 = uncharged nonpolarizable atom, otherwise sum of keys */
#    if POLAR&8
  real alphazz_qq;/* alphazz/chargepol^2 (z direction) */
  int tozz;       /* -1: no axial polarizability,
                     >=0: atom defining the direction */
#    endif /*# POLAR&8 */
#  endif
  int st;         /* site type - not in old versions! */
#  ifdef COOK
  exception_t *exc;  /* pointer to the list of exceptions */
#  endif /*# COOK */
#  ifndef FREEBC
  real sfweight;  /* atom weight for structure factor, see also sitedef_t */
#  endif /*# FREEBC */
} siteinfo_t;

struct depitem_s {
  int i;   /* independent site in species */
  real w;  /* weight */ };

typedef struct depend_s {
  struct depend_s *next;   /* next in the list */
  int indx;                /* the dependent site */
  int n;                   /* number of reference sites (parents) */
  enum deptype_e { DEP_OLDM, DEP_M, DEP_R, DEP_L , DEP_N} type;
  /* DEP_OLDM = old syntax of DEP_M; No.depend[0] = total # of all dependants
     DEP_M = "Middle" (linear combination) dependants
     DEP_R = "Rowlinson", site perpendicular to a (generally flexible) triangle
     DEP_L = "Lone", general rigid triangle, dependant out of plane 
     NB: order significant, operators < > used */
  struct depitem_s dep[3]; /* [n]
                              indices and weights of the parents
                              assumed: SUM dep[].w = 1 */
  /* DEP_L, for n=3 only (wz for DEP_R):
     local coord. system = (X,Y,Z), Z=X x Y (x=cross product)
     f_x=X.f, etc. */
  vector x;                /* X from 3 sites: X=sum_i x[i] r[i] */
  vector y;                /* Y from 3 sites: Y=sum_i y[i] r[i] */
  real wz;                 /* z-component of the site; DEP_R: LO distance */
  vector tx;               /* tx[a]*f_x added to f[a]_z */
  vector ty;               /* ty[a]*f_y added to f[a]_z */
} depend_t;

#  if defined(POLAR) && POLAR&32
/* fluctuating charge for water */
typedef struct fq4_s {
  struct fq4_s *next;
  int indx[4]; /* sites in order: H H' L L' */
  double AHH,ALL,AHL; /* interaction constants */
  double AHL2; /* 2 AHL */
  double H1H1,H1H2,L1L1,L1L2,HL,LH; /* Phi->q constants */
} fq4_t;
#  endif /*# defined(POLAR) && POLAR&32 */

typedef struct {
  int ns;           /* # of sites */
  int nc;           /* # of constraints */
  int N;            /* # of molecules */
  int config;       /* set if abs init config (from blend) */
#  ifdef COOK
  int intra;        /* if the molecule has at least one bond,angle,torsion */
  int pot;          /* 0=default, 3=TIP3P, 4=TIP4P, 5=ST2 (optimized code) */
  bond_t *bond;     /* bonds [external constraints] table head */
  angle_t *angle;   /* angle table head */
  torsion_t *dihedral,*improper,*aromatics; /* heads of torsion tables */
  depend_t *dependants; /* table of lin. dependent sites */
#    if defined(POLAR) && POLAR&32
  fq4_t *fq4;       /* fluctuating charge */
#    endif /*# defined(POLAR) && POLAR&32 */
  double zero_energy; /* const to be added to the pot energy */
  double charge;    /* molecular charge */
  double mass;      /* molecular mass */
  char *name;       /* species name */
#  endif /*# COOK */
  siteinfo_t si[1]; /* [max{ns,nc}] */
} specinfo_t;

typedef struct {
  int sp;    /* species */
  int ns;    /* # of sites */
  int nc;    /* # of constraints */
  int ir;    /* relative address of 1st site in rp - in bytes (use rof) */
  int ig;    /* relative address of 1st g[amma] in gs */
#  ifdef ANCHOR
  int anchor;  /* OR of the following bits: */
                         /* MEASURE and PRINT (no constraint): */
#    define ANCHOR_r   0x01  /* position of center-of-mass */
#    define ANCHOR_v   0x02  /* velocity of center-of-mass */
#    define ANCHOR_x   0x04  /* acceleration of center-of-mass */
#    define ANCHOR_f   0x08  /* force to center-of-mass */
#    define ANCHOR_m   0x10  /* momentum of force (torque) to center-of-mass */

                         /* CONSTRAIN ("anchor" = keep constant): */
#    define ANCHOR_s  0x100  /* site (at r0) */
#    define ANCHOR_c  0x200  /* center-of-mass */
#    define ANCHOR_i  0x400  /* principial axis of the inertia tensor
                            (parallel to vector axis) */
#    define ANCHOR_a  0x800  /* axis given by center-of-mass and site */
#    define ANCHOR_p 0x1000  /* axis given by a pair of sites */
#    define ANCHOR_t 0x2000  /* axis perpendicular to a plane of 3 sites */
#    define ANCHOR_g 0x4000  /* center-of-mass of group(s) */
#    define ANCHOR_pos (ANCHOR_s|ANCHOR_c)
                         /* constrain any position BUT group CM */
#    define ANCHOR_axis (ANCHOR_i|ANCHOR_a|ANCHOR_p|ANCHOR_t)
                         /* constrain any axis */
  struct anchorgroup_s {
    struct anchorgroup_s *next;
    vector r0; /* the anchor point */
    vector cf; /* constrain force measured here */
    int ns;
    int site[1]; /* [ns] */ } *group;
  vector r0;   /* the anchor point for ANCHOR_{s,c} */
  int iaxis[3];/* site numbers for ANCHOR_{i,a,p,t}; iaxis[0] for ANCHOR_s */
  vector axis; /* the axis direction (normalized; ANCHOR_{i,a,p,t} only) */
#  endif /*# ANCHOR */
} molecule_t;

/***
    Type of function computing forces between m1 and m2.
    It should return energy.
    Forces are summed to f1 and f2.
***/
typedef double pot2_t(
  vector f1[/*m1->ns*/], vector f2[/*m2->ns*/],
  molecule_t *m1, molecule_t *m2,
  vector *rp);

/***
    Type of function computing intramolecular forces of molecule m and eventually
    correcting for wrong site-site forces.  It should return energy.
    Forces are summed to f.
    Used if pairs are calculated by the link-cell method.
***/
typedef double pot1_t(
  vector f[/*m1->ns*/],
  molecule_t *m,
  vector *rp);

#  ifndef LINKCELL
#    define pot_t pot2_t
#  endif /*# LINKCELL */

#define RPOFFSET 7 /* @(rp) - @(logs), or 1(logs) + 3(lambda) +3(aux) */

typedef struct /*SDS*/ {
  int  size;    /* whole struct in bytes */
  int  dep;     /* set if dependants calculated, unset if cfg changes */
  real logs;    /* log of the Nose variable s */
  vector lambda;/* log(L): NEW since V2.7a 
                   active for derivatives, cfg[0].lambda is always derived from box.L*/
  vector offdiag;/* reserved, NEW since V2.7a */
  vector rp[1]; /* contiguous array of No.s vectors (POLAR: 2*No.s) */
} *ToIntPtr,ToInt;
/* WARNING: better DO NOT change this struct, dirty tricks used inside cook
   -- logs is the 1st real variable to be integrated
   -- reals must follow until the end of the structure
   -- MUST begin with 2 ints */

/***
    Macro  rof(M,RP)  returns the address of array of vectors containing
    positions of sites of molecule *M in array RP of all sites and momenta
    (typically RP=a[i]->rp).  In other words,  rof(M,RP)[i]  is a vector
    of 3 Cartesian coordinates of the i-th site.
    Macro  pof(M,RP)  does the same for momenta.
    POLAR: see also polarrof
***/
#  define rof(M,RP) ((vector*)((char*)(RP)+(M)->ir))

/* sparse matrix M for one molecule: see simcg.c */
typedef struct {
  int b;          /* 2nd subscript, b<=a */
  real imass;     /* 1/mass of the common site - not def for a==b */
  real Mab;       /* values of M[a,b] = SUM{i} grad[i]c[a].grad[i]c[b] / m[i] */
} M_t;

typedef struct {
  double mass;    /* total mass */
  double free_mass; /* mass in all movable molecules */
  double charge;  /* total charge */
  double eps;     /* machine precision */
  double epsq;    /* squared charge conventional precision */
  double dt;      /* h*noint */
  double NPT;     /* 1+3/No.f */
  double M_T ;    /* mass of thermostat (Nose+MTK) */
  double M_Th;    /* 1/2 of the mass of thermostat (Nose+MTK) */
  double M_P;     /* the mass of barostat (Nose+MTK) */
  double M_Ph;    /* 1/2 of the mass of barostat (Nose+MTK) */
#  ifdef LINKCELL
  double percell; /* av. # of sites/cell */
  int cell[DIM];  /* # of cells per box */
  int occup;      /* report cell occupancies up to occup (occup=1 -> guess) */
#  endif /*# LINKCELL */
#  ifdef POLAR
  double A;       /* sum of all polarizabilities (in A^3), without 4PI */
  int pol;        /* # of dipole sites (used for Car-Parrinello) */
  int free_pol;   /* as above, in movable molecules
                     BUG: fixed molecules are not polarizable */
#  endif /*# POLAR */
#  if PARALLEL
  int th;         /* number of threads */
  int lcth;       /* number of threads for linked-cell */
  pthread_t *thread; /* [No.th]: pthread-style threads */
  pthread_attr_t thread_attr; /* thread attributes */
  int measureserial; /* force measurements in serial (void now) */
#  endif /*# PARALLEL */
  int N;          /* total # of molecules */
  int s;          /* total number of sites (sum over all molecules) */
  int c;          /* total # of constraints */
  int free_N;     /* # of movable molecules */
  int free_s;     /* # of sites in movable molecules */
  int free_c;     /* # of constraints in movable molecules */
  int free_depend;/* total # of dependants in movable molecules */
  int maxc;       /* max # of constraints / molecule */
  int maxs;       /* max # of sites / molecule */
  int f;          /* effective # of degrees of freedom */
  int fx;         /* NPT: f+extra degrees of freedom */
  int conserved;  /* # of conserved degrees of freedom excl. Etot (momentum + angular momentum components) */
  int ncoord;     /* # of rescalable coordinates */
  int f4P;        /* # of degrees of freedom excl. conserved quantities */
  int f_tr;       /* effective # of intermolecular degrees of freedom */
  int f4P_tr;     /* # of intermolecular degrees of freedom excl. conserved quantities */
  int f_in;       /* effective # of intramolecular degrees of freedom */
  int nreal;      /* # of reals in vectors (POLAR: 2*) */
  int eq;         /* # of equations of motion */
  int first;      /* # of molecules to be included in En.first */
  int pred;       /* option('p')%10 : predictor of constraints or velocity */
  int depend[DEP_N]; /* total # of dependants of given type */
  int ion;        /* number of ions */
} No_t;
extern No_t No;   /* # of sites, constraints, degrees of freedom etc. */

typedef struct {
  double T;      /* Nose correlation time or friction thermostat time */
  double E;      /* correlation time to keep energy constant */
  double rho;    /* typical time for density changes to reach given density */
  double P;      /* isobaric friction-like correlation time */
#  ifdef POLAR
  double dip;    /* for Lagrange (Car-Parrinello) mechanical dipoles */
#  endif /*# POLAR */
  double R;      /* correlation time to change parameter LJ.RvdW in setss(0,i,i) */
  double sat;    /* for automatic saturation calculation */
  int i;         /* site (both sites must be the same) */
} tau_t;

/*** En_t: summary of cycle statistics ***/
typedef struct {
  double pot;      /* potential energy, as sum of bond,LJ,el etc. */
  double kin;      /* kinetic energy (first calculated as 2*kin.en., later
                      halved to get the true kin.en.) */
  double tot;      /* total Hamiltonian (pot+kin+zero_energy+Nose degrees+corr) */
  double el;       /* electrostatic energy; for Ewald both k- and r-space */
  double corr;     /* LJ cutoff correction: Ecorr=corr/V, Pcorr=corr/V^2 */
  double Pkinq;    /* finite size correction for the kinetic component of P, site-based */
  double PkinqCM;  /* finite size correction for the kinetic component of P, CM-based */
  double virc;     /* virial of constraint forces */
  double LJ;       /* Lennard-Jones energy (sum over all pairs) */
  double bonded;   /* sum of bond+angle+torsion terms */
  double intra;    /* site-site intramolecular terms (?) */
  double fix;      /* extra terms to fix some atom positions */
  double pot0,el0,LJ0,bonded0;  /* as above - No.first molecules */
  double potX,elX,LJX;          /* as above - all but No.first molecules */
  double LJmn;     /* mol-mol LJ energy, passed from charmm2 */
  double kin_tr;   /* 2*translational kin. energy (from center-of-mass) */
  double kin_in;   /* 2*intramolecular+rotatioal energy */
#  ifdef POLAR
  double self;     /* self energy */
  double chi_hf;   /* optical frequency susceptibility by elst field sampling */
#    if POLAR&32
  double fqself;   /* self energy - fluctuating charge (NB: not in virial) */
#    endif
  double kinpol,Tpol; /* mech.dip.: kin.energy and temperature */
  double polmaxerr,polstderr,polmaxdr;
                   /* errors during predictor/corrector or iterations */
  double Polmaxerr,Polstderr;
                   /* errors from measureP(); use short dV */
  int polnit;
#  endif /*# POLAR */
  double vir;      /* virial of force */
  double Pcfg;     /* virial pressure [p.u.], with cutoff correction */
  double P;        /* virial pressure [Pa], with cutoff correction */
  double Pnc;      /* pressure, without cutoff correction */
  double U;        /* Internal energy = Epot + Ekin + Ecorr (no Nose etc.) [J/mol] */
  double Unc;      /* Internal energy = Epot + Ekin (no cutoff corr, Nose etc.) */
  double H;        /* enthalpy, with cutoff correction */
  double Hnc;      /* enthalpy, without cutoff correction */
  double PdV;      /* pressure from virtual V change [Pa] */
  double PdVnc;    /* pressure from virtual V change, without cutoff correction */
  double T;        /* kinetic temperature */
  double kinT;     /* kin. energy of the thermostat degree of freedom (sometimes) */
  double kinP;     /* kin. energy of the barostat degree of freedom */
#  if PRESSURETENSOR&PT_VIR
  double Pvir[PT_DIM];  /* xx yy zz [yz zx xy] (virial part) [Pa] */
#  endif /*# PRESSURETENSOR&PT_VIR */
#  if PRESSURETENSOR&PT_KIN
  double Pkin[PT_DIM];  /* xx yy zz [yz zx xy] (kinetic part - site-based) [Pa] */
#  endif /*# PRESSURETENSOR&PT_KIN */
#  if PRESSURETENSOR&PT_MOL
  double PKin[PT_DIM];  /* xx yy zz [yz zx xy] (kinetic part - CM-based) [Pa] */
#  endif /*# PRESSURETENSOR&PT_MOL */
#  if PRESSURETENSOR&PT_MOM
  double PKin2[DIM];  /* SPECIAL: PKin^2 (to calc. 2nd moment = kurtosis of v) [Pa] */
  double PKin3[DIM];  /* SPECIAL: PKin^3 (to calc. 3rd moment) [Pa] */
#  endif /*# PRESSURETENSOR&PT_MOM */
#  if PRESSURETENSOR
  double Ptens[PT_DIM];  /* xx yy zz [yz zx xy] total [Pa] */
#  endif /*# PRESSURETENSOR */
#  if (PRESSURETENSOR&PT_ANY) == PT_ANY
  double trPt;     /* SUM(En.Ptens)/3+En.corr/V/V*Puni
                      =P with correction, to put to .cp
                      NB: for cutoff electrostatics, the standard pressure uses Eelst=virial  */
  double Hz;       /* uncorrected enthalpy with uncorrected Pzz */
#  endif /*# (PRESSURETENSOR&PT_ANY) == PT_ANY */
  double T_in;     /* kin. T from intramolecular motions and rotations*/
  double T_tr;     /* kin. T from intermolecular motions (translations) */
  double r1;       /* error of constraints of predictor */
  double r2;       /* error of constraints after integration step */
  double r3;       /* error of constraints (of true configuration) */
  double v1,v2,v3; /* errors of velocity constraints */
#  ifdef SHEAR
  double Cv;       /* SUM v_i[xy] cos(2PIz/L) / SUM cos^2(2PIz/L) */
  double kinshear; /* kin E of shear flow */
#  endif /*# SHEAR */
  double cmdrift;   /* CM position */
  double vdrift;    /* velocity of CM */
  double angdrift;  /* angular velocity */
  vector J;         /* current density in SI [A/m2] */
  vector M;         /* dipole moment of the box (determined outside Ewald) */
  double logs,dlogs; /* a[0]->logs, a[1]->logs */
  double ext;       /* extended degrees of freedom (Nose etc.) kin+pot energy */
#ifdef RGYR
  vector RGm;       /* inertia tensor, diagonalized */
#endif
#  ifdef SLAB
  double zshift;    /* slab position by autocenter */
  double Pt;        /* tangential pressure tensor (Ptxx+Ptyy-2*Ptzz)/3 [Pa] */
  double Ptv;       /* virial part of tangential pressure tensor (Pvxx+Pvyy-2*Pvzz)/3 [Pa] */
#    if SLAB & 2
  double Wcleave; /* dG (rev. work) per d sigma and unit area */
  double zcleave; /* actual positions, in the units of box.L[2] */
  double fcleave[2]; /* cleaving force */
#    endif /*# SLAB & 2 */
#  endif /*# SLAB */
} En_t;

#  include "rdf.h"

/****** miscellaneous ******/
/* for some strange compatibility reasons,
   we do not #include "options.h" here */
#  ifndef option
extern int optionlist[32];
#    define option(X) optionlist[X & 31]
#  endif /*# option */

#  ifdef COOK
#    define FROM option('j')
#  else /*# COOK */
#    define FROM 0
#  endif /*#!COOK */

/*
   %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
   %                     declarations of global variables                  %
   %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
*/

extern
pot2_t ***pot;       /* array [nspec][nspec] of pointers to functions computing
                        INTERmolecular interactions */

#  ifdef LINKCELL
extern
pot1_t  **pot1;      /* array [nspec] of pointers to functions computing
                        INTRAmolecular interactions */
#  endif /*# LINKCELL */

extern
specinfo_t **spec;   /* array [nspec] of pointers to species definitions */

extern
M_t **Ms;            /* array [nspec or N] of pointers to matrices M */

extern
molecule_t *molec;   /* array [N] of molecule descriptors */
extern
ToIntPtr cfg[9];     /* [MaxGearOrder+3] */
extern
ToIntPtr fixa;       /* fixed sites (by harmonic springs) */

/* warning: even #ifdef FLOAT, the constraint calculations are in double */
extern
double **gs;         /* [option('p')] tables of Lagrange multipliers g to
                        predict their values
                     NOTE: pointers are NOT constant because swaping of cor-
                     responding arrays is implemented by pointer interchange */
extern
double *binom;       /* [option('p')] coefficients of this predictor */

extern
rdf_t ***rdf;        /* table of pointers to rdf's
                        see rdf.h for rdf_t
                        note: also in each sitesite_t sstab[][] (see setss) */

extern int init;        /* INITIAL KEY :
__init____.cfg___.cp____.sta .rdf .dih__logs
    0     read  append      read       read
    1     read  append      new        read
    2     read  rewrite     new         0
   >=3    init  rewrite     new         0  (sequential version only)
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/
extern int init_append; /* = init but :=0 after one step (to control appending) */

extern double t;     /* MD time */
extern real h;       /* time step */
extern double equalize; /* masses equalization factor; 0=no equalization */

extern double eps;   /* accuracy for computing Lagrange multipliers
                      if >=1 then # of iteration steps */
extern double omega; /* not implemented */

extern double epsc;  /* accuracy for correcting constraints
                      if epsc>=1, then # of iteration steps
                      (default is changed for -c4) */
extern double omegac;/* relaxation parameter for 'correct' */

#  ifdef POLAR
extern struct scf_s {
  /* public (in data): */
  double eps;   /* eps>0: accuracy for self-field (induced dipole moments)
                        in program units (p.u.= 0.00245 eA = 0.01175 D)
                   eps<0: |eps| iterations */
  double epsq;  /* eps enhancement factor for simulation (re)start
                   (close to 1:more, epsq<=0:off) */
  double omega; /* relaxation paramater for self-field iterations,
                   see also option -^ (default = max. stability or 0.95) */
  double epsx;  /* as above, for P by virtual volume change and similar
                   testing purposes (default 9e9 => epsp*(1-epspq)/10 */
  double omegax;/* as above, for P by virtual volume change and similar
                   testing purposes; default = 0.95  */
  double E;     /* for determination of epspol, in V/m (see also Eext) */
  int Estride;  /* epspol sampled once a Estride cycles only (0=1=every) */
  int maxit;    /* max # of self-field iterations until WARNING
                   (for positive eps,epsx) */
  /* private: */
  double maxdr; /* Drude: max amplitude */
  double eps0;  /* for start (see epsq) */
  double maxerr;/* max dipole error passed from selffield; orig. maxerrold */
  double rate;  /* convergence rate (if defined) */
  int nit;      /* running number of iterations */
} scf;
#  endif /*# POLAR */

extern double T;     /* thermostat temperature (if tau.T) [K] */
extern double P;     /* barostat pressure [Pa] */

extern En_t En;      /* energies, virial, pressure tensor, constraint errors.. */

extern int GearOrder;/* see also option('m') */

#define RESCALE_X        1
#define RESCALE_Y        2
#define RESCALE_Z        4
#define RESCALE_XYZ      7
#define RESCALE_CM       8 /* center-of-mass based rescaling */
#define RESCALE_PT      16 /* use PRESSURETENSOR (components) */
#define RESCALE_XisY    32 /* modifier of RESCALE_PT (xscale=yscale) */
#define RESCALE_XisYisZ 64 /* uniform (xscale=yscale=zscale) */
#define RESCALE_H      128 /* rescale also logs,lambda */
#define RESCALE_L      256 /* rescale also box */
#define RESCALE_WALL   512 /* WALL version only */
#define RESCALE_SLAB  1024 /* SLAB version only (virtual area change) */

#define RESCALE_CLEAVE  2048 /* z-box barostat w. cleaving boundaries */
extern int rescale;     /* what is allowed to rescale (with tau.P, tau.rho, dV):
                           sum of the flags above */

extern struct lag_s {
  int err;           /* lag for error analysis of important variables */
  int n;             /* number of blocked calculations: max.block size is 2^n */
#  ifdef RGYR
  int dim;           /* how many coordinates recorded (3 for vx,vy,vz) */
  int nv;            /* how many molecules */
  int v;             /* lag for velocity-velocity time autocorr. functions */
#  endif /*# RGYR */
  int cond;          /* lag for current time autocorr.f. (conductivity) */
#  if PRESSURETENSOR&PT_OFF
  int visc;          /* lag for Ptxy,... time autocorr.f. (viscosity) */
#  endif /*# PRESSURETENSOR&PT_OFF */
} lag;

extern double E;     /* total energy in [J/mol] (=Eunit) to be kept constant */
extern tau_t tau;    /* relaxation times for thermostat, volume change,... */
extern double T_tr_in; /* ratio T_tr/T_in for thermostat=13 */
extern int thermostat;
/* NOTE: do not change - relations < and > used, strchr used */
#  define T_BERENDSEN   1 /* Berendsen (simple friction) thermostat */
#  define T_NOSE        2 /* Nose-Hoover; with Verlet it uses predicted velocities */
#  define T_ANDERSEN    3 /* randomly selected atoms updated */
#  define T_MAXWELL     4 /* periodically all atoms updated at once */
#  define T_ANDERSEN_CM 5 /* randomly selected molecules (center-of-mass) updated */
#  define T_MAXWELL_CM  6 /* periodically all molecules updated at once */
#  define T_LANGEVIN    7 /* atom-based Langevin thermostat */
#  define T_LANGEVIN_CM 8 /* CM-based Langevin thermostat */
#  define T_TR         11 /* friction only translations */
#  define T_IN         12 /* friction only intramolecular velocities */
#  define T_FRICTIONS  13 /* combined 11,12: see T_tr_in */
#  define T_NPT        22 /* Nose+MTTK */

extern int measure;  /* 0: only forces required
                        1: also energy etc.
                        2: special for calculating # of pairs (DEBUG only) */

/*** Ewald related global variables - see  comments in ewald.c and erfc.c ***/
/* they are used also for not-Ewald sums, so they are here... */
/* WARNING: change initialization as well! */
extern struct el_s {
  real alpha;      /* Ewald separation parameter in units of 1/A;
                      also cutoff electrostatic parameter */
  real kappa;      /* k-space cutoff in form kappa=K[i]/L[i] */
  real epsr,epsk;  /* expected r- and k-space errors in forces, in K/A */
  real minqq;      /* minimum expected charge-charge separation
                      (to test Drude and to report error of the erfc approximation) */
  real epsinf;     /* dielectric constant at infinity
                      (must be very big if free charges are present */
  real diff;       /* max rel.change of L before a warning about "unchanged K" */
  real Perr;       /* pressure tensor threshold (basic; WARNING for 10*Perr) */
  real rplus;      /* additional range to erfc splines (former ERFCPLUS)
                      needed for optimized water models, 
                      does not hurt if unnecessarily large */
  real sat;        /* target saturation (see tau.sat) */
  int rshift;      /* &1: eru(r)=erfc(r)/r spline shifted to avoid jump at cutoff
                      &2: as above for the derivative, erd(r) 
                      &4: prints eru(r), erd(r), d eru(r)/dr verbosely on init */
  int grid;        /* # of grid points per unity of squared distance */
  int test;        /* test module switch (former etest) */
  int diag;        /* whether to include diagonal correction */
  int centroid;    /* 1=dipole moments (of ions) are w.r.t charge centroid */
  int sf;          /* structure factor switch:
                      1 = sphericalized (cube only)
                      3 = 3D */
  int slab;        /* slab correction (former el.epsinf=-1), see
                      In-Chul Yeh and Max L. Berkowitz, JCP 111, 3155 (1999) */

  /* private: */
  int sfsize;      /* to check overflow */
  struct sfr_s {   /* for sphericalized (radial) structure factor */
    int nk;           /* # of independent vectors */
    real q;           /* squared structure factor */
  } *sfr;
  struct sf3d_s {  /* for 3D structure factor */
    int k[DIM];       /* k-vector */
    real q;           /* squared structure factor */
  } *sf3d;
  double sumM;     /* sum of molecule dipole moments = max.cell moment
                      for saturation autoset, one cfg value (cf. Eext.sumM)
                      POLAR: includes induced moments */
  double xinf;     /* 1/(el.epsinf*2+1) */
} el;

extern struct box_s {
#  ifdef LINKCELL
  real over14;         /* factor by which the estimate of change of the max
                        1-2,1-3,1-4 distance is multiplied.
                        Must be >1; the higher, the safer
                        negative: also print distances and speeds
                        1.5 is pessimistic for proteins,
                        but just enough for water */
  real max14v;  /* max (safe) speed of max14 change from step to step */
  real max14;   /* max 14 found in the sweep */
  real rmin;    /* minimum x,y,z of (any site of) a molecule - user value */
  real r0;      /* minimum x,y,z of (any site of) a molecule - used */
  real r0limit; /* to warn only once */
#  endif /*# LINKCELL */
  vector L;     /* size of the periodic box;
                   it's important that L is real, not double, for automatic
                   float<->double conversion in loadcfg() */
  vector Lh;    /* Lh=L/2 */
  real V;       /* Lx*Ly*Lz */
  real cutoff;  /* r-space cutoff */
  real cq;      /* =cutoff^2 */
} box;

extern int cache;    /* molecules in block for r-space sum */

/* Q_t *Q; is in ewald.h and ewald.c */

#  ifdef POLAR
extern
int polar_off; /* offset of auxiliary charges in the units of bytes */
#    define polarrof(M,RP) ((vector*)((char*)(RP)+(M)->ir+polar_off))
#  endif /*# POLAR */

#  ifdef SHEAR
extern double shear; /* max gradient of acceleration for cos(z) method */
#  endif /*# SHEAR */

typedef struct fixsites_s {
  struct fixsites_s *next;
  int from,to; /* C-style range of sites: [from,to) */
  int isr; /* whether r applies */
  vector r; /* position */
} fixsites_t;
extern fixsites_t *fixsites0;

extern int drift;
#  define DRIFT_X 1 /* center-of-mass */
#  define DRIFT_Y 2
#  define DRIFT_Z 4
#  define DRIFT_VX 8 /* momentum (velocity) */
#  define DRIFT_VY 16
#  define DRIFT_VZ 32
#  define DRIFT_AX 64 /* angular momentum drift */
#  define DRIFT_AY 128
#  define DRIFT_AZ 256
#  define DRIFT_WHEN 1536
#  define DRIFT_CYCLE 0 /* every cycle, init, load, save */
#  define DRIFT_START 512 /* init, load only, then cancel */
#  define DRIFT_STEP  1024 /* every step, init, load, save */
#  define DRIFT_SAVE  1536 /* init, load, save */
#  define DRIFT_DEPEND 2048 /* dependants every step (former norm&4) */
#  define DRIFT_AUTO 4096 /* auto set */
/* WARNING: do not change numbers - coded sometimes directly by numbers */
extern int conserved; /* number of conserved degrees of freedom */

extern struct center_s {
#  ifdef FREEBC
  double r0,K;
#  else /*# FREEBC */
  int on;              /* flag: 2=to box center, 1=z-pos */
  vector K,K2;         /* force constant to center-of-box */
  vector r0;           /* range (force starts at center+-r0) */
#  endif /*#!FREEBC */
  vector cmK;          /* global force of the center-of-mass to (box) center */
  double cmmass;       /* mass of cmn molecules (from FROM) */
  int cmn;             /* # of molecules affected; 0=all but FROM */
} center;

#ifdef SLAB
extern struct slab_s {
  double grid; /* histogram bins/A */
  double zmax; /* range is [0,zmax] if zmax given, or [0,Lz] scaled to [0,1] */
  int sp;      /* species to autocenter the slab:
                  negative: all species 0..|sp| (incl.)
                  non-negative: species sp
                  out of range: no autocenter */
  int mode;    /* 1: calculate surface tension incl. cutoff corr.
                  2: with dV, forces const-volume scaling (test area method) */
  /* slab thermostat */
  double T;       /* temperature in [Tz0,Tz1]: negative = off */
  double Tz0,Tz1; /* range for T */
  double qT;      /* private: sqrt(slab.T/T); 0=off */
  /* forces to organize the slab: */
#      define CENTERSLAB 10 /* number of force terms */
  int n[CENTERSLAB];      /* how many molecules is affected (input) */
  int nn[CENTERSLAB];     /* molecules affected [range) */
  int ns[CENTERSLAB];     /* affected sites: 0: all, >0: i<ns only, <0: i>=|ns| */
  double z[CENTERSLAB];   /* center relatively to Lh[2] */
  double z0[CENTERSLAB];  /* radius of zero-pot. range */
  double Kz[CENTERSLAB];  /* z-force const, for molecules up to n[] */
  double Kz2[CENTERSLAB]; /* z1=0: 2*Kz, z1!=0: step */
  double z1[CENTERSLAB];  /* for slab/outside bias function (input) */
  double dz[CENTERSLAB];  /* z1-z0, for z1!=0 */
} slab;

#  if SLAB & 2
extern struct cleave_s {
  /* cleaving potential (is a bell-like function): */
  int init;       /* if set then copy z0,z1 to z[] */
  int n;          /* number of cleaving forces: 0,1,2 */
  int i;          /* site susceptible to the cleaving potential;
                     i<0: all sites 0..|i| (incl.) are susceptible */
  double sigma;   /* half-width of the cleaving potential, in A */
  double K;       /* 1/2 of force, in K/A */
  double Ksigmah; /* cleave*sigma/2 */
  double z[2];    /* z-positions (used + stored in .cfg), in units of Lz */
  double z0,z1;   /* z-positions (initial only: see cleave.init) */
} cleave;
#  endif
#endif

extern struct Eelst_s {
  int isE;             /* 1 if there is general nonzero E, -3=x, -2=y, -1=z */
  vector E;            /* in prog. units */
  double Eepspol;      /* for epspol, in prog. units */
  double sumM;         /* sum of molecule dipole moments passed (incl. induced)
                          = max. dipole moments of the cell, averaged value
                          cf. el.sumM (for 1 cfg) */
} Eext;

#  ifdef ANCHOR
extern struct anchor_s {
  FILE *f; /* output file of anchor forces .anc */
  double rr; /* max displacement, squared */
  double cos,sin; /* max cos,sin of rot angle */
} anchor;
#  endif /*# ANCHOR */

extern int onefourinrdf; /* control whether 1-4 included in RDF's */

#  if PARALLEL
  /* usage of the pthread-based parallel macro:

     PARALLELIZE(FUNCTION)

     will call option('\\')-times function FUNCTION, which is passed
     the thread number in the form of void* casted int*
     (thread 0 is the caller)

     Example:

     void *FUNCTION(void *arg)
     {
       int ithread=*(int*)arg;
       fprintf(stderr,"thread %d started\n",ithread);
       ...
       return arg;
     }
  */

#    define PARALLELIZE(FUNC,NTHREADS) { \
  int ith; char *rc; \
  loop (ith,1,NTHREADS) \
    if (pthread_create(No.thread+ith,&No.thread_attr,FUNC,(void*)(No.thread+ith))) \
      ERROR(("cannot create thread %d",ith)) \
  ith=0; \
  FUNC((void*)No.thread); \
  loop (ith,1,NTHREADS) \
    if (pthread_join(No.thread[ith],(void**)&rc)) \
    ERROR(("thread %d failed",ith)) }

#    ifdef SERIAL
/* serial version for debugging: */
#      define PARALLELIZE(FUNC,NTHREADS) { \
  int ith; char *rc; \
  loop (ith,0,NTHREADS) FUNC((void*)(No.thread+ith)); }
#    endif /*# SERIAL */

#  endif /*# PARALLEL */

#endif /*# _SIMGLOB_ */ /*# missing #if */
