/* some extra forces */

#include "ground.h"
#define SDS
#include "alloc.h"
#include <time.h>
#include "simglob.h"
#include "norm.h"
#include "ewald.h"
#include "forces.h"
#include "xforces.h"
#include "simmeas.h"
#include "cputime.h"
#include "intermac.h" /* because of fixsites */
#include "statics.h" /* cleaving */

#ifdef  WALL /* brrrrrrrrrr! */
#  include "interpot.h"
#  include "simdef.h"
#  include "simpot.h"
#  include "units.h"
#endif /*#  WALL */

void fixforces(ToIntPtr B, ToIntPtr A) /************************** fixforces */
{
  vector *rpB=B->rp,*rpA=A->rp,*rpf;
#ifdef GOLD
  double xxyy;
#endif /*# GOLD */
  real *r1,*r2,*f,rr;
  vector dr;
  int i;
  fixsites_t *fix;
  double U=0,K=option('k');

  if (!fixa) return;
  rpf=fixa->rp;

  for (fix=fixsites0; fix; fix=fix->next)
    loop (i,fix->from,fix->to) {
      r1=rpA[i];
      r2=rpf[i];
      NearestImage(,box.cq) {
        f=rpB[i];
        if (measure) U+=K*rr;
        VV(f,-=K*dr) }
      else {
        static int count=0;
        double KK=K*box.cq/rr; /* smaller force outside cutoff sphere */

        f=rpB[i];
        /* U wrong for this reduced force */
        VV(f,-=KK*dr)
        if (count<10) {
          WARNING(("too far from the fix location (more than 10 warnings will be suppressed)\n\
*** fix=(%g %g %g), real=(%g %g %g)",
                   VARG(r2),VARG(r1)))
            count++; } }
      }

  U/=2;
  if (measure && option('v')&4) prt("fixE=%g", U);

  En.pot += En.fix=U;
}

void centerforces(ToIntPtr B, ToIntPtr A) /******************** centerforces */
/*
   additional forces, originally "to the center of the box"
*/
{
  if (center.cmn) {
    int ns,i,k,n;
    vector CM,K2;
    double mi;
    molecule_t *mn;
    siteinfo_t *si;
    vector *r,*f;
    static int warn=1;

    VO(CM,=0)

    if (center.cmn>No.N) center.cmn=No.N;

    loop (n,FROM,center.cmn) {
      mn=molec+n;
      ns=mn->ns;
      r=rof(mn,A->rp);
      f=rof(mn,B->rp);
      si=spec[mn->sp]->si;

      loop (i,0,ns) {
        mi=si[i].mass;
        VV(CM,+=mi*r[i]) } }

    VO(CM,/=center.cmmass)
#ifndef FREEBC
    VV(CM,-=box.Lh) /* in periodic b.c. centereded to the center of the slab */
#endif /*# FREEBC */

    loop (k,0,DIM) En.pot+=center.cmK[k]*Sqr(CM[k]);
    VV(K2,=2/center.cmmass*center.cmK)

    loop (n,FROM,center.cmn) {
      mn=molec+n;
      ns=mn->ns;
      r=rof(mn,A->rp);
      f=rof(mn,B->rp);
      si=spec[mn->sp]->si;

      loop (k,0,DIM) if (center.cmK[k])
        loop (i,0,ns)
          f[i][k]-=K2[k]*si[i].mass*CM[k]; }
    }

#ifdef FREEBC
  /*
    force to the center of coordinate system - Lagrange only
      WARNINGS:
      this code is repeated in the initializer simils.c (?)
  */
  if (center.K) {
    double r0q=Sqr(center.r0);
    int n,ns,i,sp;
    vector *r,*f;
    molecule_t *mn;

    loop (n,FROM,No.N) {
      mn=molec+n;
      sp=mn->sp;
      ns=mn->ns;
      r=rof(mn,A->rp);
      f=rof(mn,B->rp);

      if (r0q) loop (i,0,ns) {
        double rr=SQR(r[i]);

        if (rr>r0q) {
          double R=sqrt(rr);

          En.pot+=center.K*Sqr(R-center.r0);
          R=(center.K*2)*(R-center.r0)/R;
          VV(f[i],-=R*r[i]) } }
      else loop (i,0,ns) {
        En.pot+=center.K*SQR(r[i]);
        VV(f[i],-=(center.K*2)*r[i]) } } }
#else /*# FREEBC */
  /*
    force to the center of the box
  */

  if (center.on) {
    int ns,i,k,n,sp;
    molecule_t *mn;
    vector *r,*f;

#  ifdef SLAB
#    if SLAB & 2
    int ifrom,ito,icleave;
    En.Wcleave=En.fcleave[0]=En.fcleave[1]=0;
#endif
#endif

    loop (n,FROM,No.N) {
      mn=molec+n;
      sp=mn->sp;
      ns=mn->ns;
      r=rof(mn,A->rp);
      f=rof(mn,B->rp);

      /* forces to box center */
      if (center.on&2) loop (i,0,ns) loop (k,0,DIM) {
        double dr=r[i][k]-box.Lh[k];

        if (fabs(dr)>center.r0[k]) {
          if (dr<0) dr+=center.r0[k]; else dr-=center.r0[k];
          En.pot+=center.K[k]*Sqr(dr);
          f[i][k]-=center.K2[k]*dr; } }

#  ifdef SLAB
      /* z (slab) forces */
#    if SLAB & 2
      /* cumbersome code: */
      if (cleave.i<0) ifrom=0,ito=abs(cleave.i);
      else ifrom=cleave.i,ito=cleave.i;
      if (ito>=ns) ito=ns-1;

      loop (icleave,0,cleave.n) {
        loopto (i,ifrom,ito) {
          double zc=box.L[2]*cleave.z[icleave];
          double dz=r[i][2]-zc,x,ff,uu;

          if (dz>box.Lh[2]) dz-=box.L[2];
          if (dz<-box.Lh[2]) dz+=box.L[2];
          x=dz/cleave.sigma;
          if (fabs(x)>=1) continue;
          if (fabs(x)<0.5) ff=x,uu=0.5-x*x;
          else if (x>0) ff=1-x,uu=ff*ff;
          else ff=-x-1,uu=ff*ff;
          ff*=cleave.K; /* now ff=force */
          En.fcleave[icleave]+=ff;
          f[i][2]+=ff;
          En.pot+=uu*=cleave.Ksigmah; /* now uu=cleaving potential Phi */
          En.vir+=dz*ff;
#      if PRESSURETENSOR&PT_VIR
          En.Pvir[2]+=dz*ff;
#      endif /*# PRESSURETENSOR&PT_VIR */
          En.Wcleave+=(uu+dz*ff)/(2*box.L[0]*box.L[1]*cleave.sigma); }
      }
#    endif /*# SLAB & 2 */

      if (center.on&1) loop (i,0,ns) loop (k,0,CENTERSLAB) {
        if (slab.nn[k]<=0) break;
        if (n<slab.nn[k]) {
          double dz=r[i][2]-(box.Lh[2]+slab.z[k]);

          if (fabs(dz)>slab.z0[k]
              && (slab.ns[k]==0
                  || (slab.ns[k]>0 && i<slab.ns[k])
                  || (slab.ns[k]<0 && i>=slab.ns[k]))) {

            if (dz<0) dz+=slab.z0[k]; else dz-=slab.z0[k];

#    if SLAB & 1
            if (slab.z1[k]) {
              /*                              ___     ___ _Kz2 */
              /* slab/outside bias function:     \___/    _0   */
              if (fabs(dz)>slab.dz[k])
                En.pot+=slab.Kz2[k];
              else if (fabs(dz)<slab.dz[k]/2) {
                En.pot+=slab.Kz[k]*Sqr(dz);
                f[i][2]-=2*slab.Kz[k]*dz; }
              else {
                if (dz<0) dz=-slab.dz[k]-dz; else dz=slab.dz[k]-dz;
                En.pot+=slab.Kz2[k]-slab.Kz[k]*Sqr(dz);
                f[i][2]-=2*slab.Kz[k]*dz; } }
            else
#    endif /*# SLAB & 1 */
            {
              /* harmonic or U-shaped forces */
              En.pot +=slab.Kz[k]*Sqr(dz);
              f[i][2]-=slab.Kz2[k]*dz;
            } }
          break; } }
#  endif        /*# SLAB */
    }
#  if defined(SLAB) && SLAB & 2
    StaSet(0,lag.err,2,lag.n);
    StaAdd("Wcleave",En.Wcleave);
#  endif /*# defined(SLAB) && SLAB & 2 */
  }
#endif /*#!FREEBC */
}

#ifdef WALL
void wallforces(ToIntPtr B, ToIntPtr A)  /************************** wallforces */
/*
  1/ atom-wall integrated LJ potential
  2/ GOLD version: Coulomb potential of charge vs. its inverted image
*/
{
  vector *f;
  int n;
  molecule_t *mn;
  double U=0;

  loop (n,FROM,No.N) {
    mn=molec+n;
    f=rof(mn,B->rp);
    U += mol_wall(f,mn,A->rp); }

  En.pot += U;
}
#endif /*# WALL */

#ifdef POLAR
void elstforces(ToIntPtr B, ToIntPtr A)  /********************** elstforces */
/*
  homogeneous electrostatic field Eprogunits
*/
{
  vector *f,*fpol,*r,*rpol;
  int i,n;
  molecule_t *mn;
  siteinfo_t *si;
  double U=0,q;

  if (Eext.isE) {

    loop (n,FROM,No.N) {
      mn=molec+n;
      f=rof(mn,B->rp);
      fpol=polarrof(mn,B->rp);
      r=rof(mn,A->rp);
      rpol=polarrof(mn,A->rp);
      si=spec[mn->sp]->si;
      loop (i,0,mn->ns)
#if POLAR&32
        if (si[i].qtype&32) {
          double qpol=rpol[i][0];
          VV(f[i],+=qpol*Eext.E)
          fpol[i][0]-=SCAL(Eext.E,r[i]); /* doublecheck sign! */
          U-=qpol*SCAL(Eext.E,r[i]);
        }
      else
#endif
        {
          double q=si[i].charge;
          double qpol=si[i].chargepol;
          vector rpi;

          VVV(rpi,=r[i],+rpol[i])

          VV(f[i],+=q*Eext.E)
          VV(fpol[i],+=qpol*Eext.E)

          U-=q*SCAL(Eext.E,r[i]);
          U-=qpol*SCAL(Eext.E,rpi); } }
    En.pot += U; }
}
#else /*# POLAR */
void elstforces(ToIntPtr B, ToIntPtr A)  /********************** elstforces */
/*
  homogeneous electrostatic field Eprogunits
*/
{
  vector *f,*r;
  int i,n;
  molecule_t *mn;
  siteinfo_t *si;
  double U=0,q;

  if (Eext.isE) {

    loop (n,FROM,No.N) {
      mn=molec+n;
      f=rof(mn,B->rp);
      r=rof(mn,A->rp);
      si=spec[mn->sp]->si;
      loop (i,0,mn->ns) {
        double q=si[i].charge;

        VV(f[i],+=q*Eext.E)
        U-=q*SCAL(Eext.E,r[i]); } }
    En.pot += U; }
}
#endif /*#!POLAR */
