/* \make asc2cfg

*/

#define SDS
#include "ground.h"
#include "alloc.h"
#include "varfile.h"

typedef double vector[3];

int main(int narg,char **arg)
{
  FILE *asc;
  char *fn;
  struct a_s {
    int  size;    /* whole struct in bytes */
    int  padd;    /* padded to 8 bytes */
    double logs;  /* log of the Nose variable s */
    vector lambda;/* log(box) */
    vector shape; /* (reserved for box shape) */
    vector rp[1]; /* contiguous array of all r[ns] (and p[ns]) */
  } *a[3];
  int nspec;
  struct spec_s {
    int N;           /* number of molecules */
    int ns;          /* number of sites */
  } *spec; /* [nspec] */
  struct rec_s {
    int key;
    int intval;
    vector vecval;
  } rec;
  int i,cfgkey,ns=0,size;
  int optionlist[32];
  char line[1024];

  if (narg<2) {
    fprintf(stderr,"Convert asc-image back into cfg-file. Call by:\n\
  cfg2asc NAME\n\
where\n\
  NAME.cfa will be converted into NAME.cfg\n\
  Note that missing coordinates in NAME.cfa are replaced by zero\n\
See also:\n\
  cfg2asc cfg2atm plb2cfg cfgconv\n");
    exit(0); }

  initscroll(0);

  asc=fopen(string("%s.cfa",arg[1]),"rt");
  VarOpen(fn=string("%s.cfg",arg[1]),"w");

  prt("writing %s",fn);
  fgets(line,1024,asc);
  cfgkey=atoi(line);
  VarPut(&cfgkey,sizeof(cfgkey));

  loop (i,0,32) {
    fgets(line,1024,asc);
    optionlist[i]=atoi(line+2);
    if (i!=(line[1]&31)) ERROR(("%s: option -c% expected",tolower(i+'@'))) }
  VarPut(optionlist,sizeof(optionlist));

  fgets(line,1024,asc);
  nspec=atoi(line);
  VarPut(&nspec,sizeof(nspec));

  allocarray(spec,nspec);
  loop (i,0,nspec) {
    fgets(line,1024,asc);
    sscanf(line,"%d%d",&spec[i].N,&spec[i].ns); }
  VarPut(spec,nspec*sizeof(spec[0]));

  for (;;) {
    fgets(line,1024,asc);
    if (line[0]=='.') break;

    sscanf(line,"%d%d%lf%lf%lf",&rec.key,&rec.intval,&rec.vecval[0],&rec.vecval[1],&rec.vecval[2]);
    if (rec.key==2) { ns=rec.intval; put(ns) }
    VarPut(&rec,sizeof(rec)); }

  if (!ns) ERROR(("record key=2 missing or ns=0"))

  size=sizeof(*a[0])+sizeof(vector)*(ns-1);
  put(size)
  loop (i,0,3) sdsalloczero(a[i],size);

  if (!fgets(line,1024,asc)) ERROR(("unexpected EOF while reading logs"))
  if (!strstr(line,"logs")) ERROR(("%s: logs expected",line))
  sscanf(line,"%lf%lf%lf",&a[0]->logs, &a[1]->logs, &a[2]->logs);

  if (!fgets(line,1024,asc)) ERROR(("unexpected EOF while reading ln(L)=lambda"))
  sscanf(line,"%lf%lf%lf%lf%lf%lf%lf%lf%lf",
          &a[0]->lambda[0], &a[0]->lambda[1], &a[0]->lambda[2],
          &a[1]->lambda[0], &a[1]->lambda[1], &a[1]->lambda[2],
          &a[2]->lambda[0], &a[2]->lambda[1], &a[2]->lambda[2]);


  loop (i,0,ns) {
    if (!fgets(line,1024,asc)) {
      WARNING(("unexpected EOF while reading record %d",i))
      break; }
    while (line[0]=='!') fgets(line,1024,asc);
    sscanf(line,"%lf%lf%lf%lf%lf%lf%lf%lf%lf",
           &a[0]->rp[i][0], &a[0]->rp[i][1], &a[0]->rp[i][2],
           &a[1]->rp[i][0], &a[1]->rp[i][1], &a[1]->rp[i][2],
           &a[2]->rp[i][0], &a[2]->rp[i][1], &a[2]->rp[i][2]);
  }
  VarPut(a[0],size);
  VarPut(a[1],size);
  VarPut(a[2],size);
  VarClose();

  fclose(asc);

  prt("%s written",fn);

  return 0;
}
