/* cc -O2 -o asc2plb asc2plb.c 
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#define Error(X) do { fprintf(stderr,"ERROR %s\n",X); exit(1); } while(0)

/* C-style loop (last not included) */
#define loop(_I,_From,_To) for (_I=_From;_I<_To;_I++)

/* Pascal etc. style loop (last included) */
#define loopto(_I,_From,_To) for (_I=_From;_I<=_To;_I++)

typedef float floatvector[3];

void endian(char *a) /********************************************** endian */
{
  char x;

  x=a[0]; a[0]=a[3]; a[3]=x;
  x=a[1]; a[1]=a[2]; a[2]=x;
}

int main(int narg,char **arg) /*************************************** main */
{
  FILE *plb,*pla;
  float hdr[2];
  floatvector r;
  int j,n=0,ns=0,isns=1,reverse=0,prec=0,varL,NS,iarg,ask=1,atm=0;
  char *infn=NULL,*outfn=NULL;
  char line[128];
  char *ext="plb";

  if (narg<2) {
    fprintf(stderr,"\
Convert text coordinate file to binary playback file.  Call by:\n\
  asc2plb ASC-FILE PLB-FILE [OPTIONS]\n\
  asc2plb NAME [OPTIONS]\n\
where\n\
  ASC-FILE: text image of playback file, usual .EXT=.pla,.vla; -=stdin\n\
  PLB-FILE: MACSIMUS playback file, usual .EXT=.plb,.vlb; -=stdout\n\
  NAME: ASC-FILE=NAME.pla, PLB-FILE=NAME.plb\n\
  -a: input is atm-format (repeating header NS blank, atom info as 1st column)\n\
  -eEXT: output EXT\n\
  -r: reverse endian of PLB-FILE\n\
  -y: overwite without asking\n\
  -n#: no header, #=number of sites (input=.3dt, output=.3db)\n\
See also: plb2asc plbconv plbmerge plbcut plbinfo plbfilt\n\
Deprecated old version: plbasc\n\
");
    exit(0); }

  loop (iarg,1,narg)
    if (arg[iarg][0]=='-' && arg[iarg][1])
      switch (arg[iarg][1]) {
        case 'a': atm++; break;
        case 'e': ext=arg[iarg]+2; break;
        case 'r': reverse++; break;
        case 'y': ask=0; break;
        case 'n': ns=atoi(arg[iarg]+2); isns=0; break;
        default: Error("wrong option"); }
    else {
      if (infn) Error("more than 2 file names");
      infn=outfn;
      outfn=arg[iarg]; }

  if (!outfn) Error("no file/name");

  if (!infn) {
    char *name=outfn;

    if (strlen(ext)>3) Error("EXT loger than 3 chars");
    infn=malloc(strlen(name)+5);
    sprintf(infn,"%s.pla",name);
    outfn=malloc(strlen(name)+5);
    sprintf(outfn,"%s.%s",name,ext); }
    
  if (strcmp("-",infn)) pla=fopen(infn,"rt");
  else pla=stdin;
  if (!pla) Error(infn);
  fprintf(stderr,"%s (ascii) -> %s (binary)\n",infn,outfn);

  if (strcmp("-",outfn)) {
    plb=fopen(outfn,"rb");
    if (plb && ask) {
      char line[256];

      fprintf(stderr,"File %s exists. Overwrite (y/n)? ",outfn);
      gets(line);
      if (tolower(line[0])!='y') return 1; 
      fclose(plb); }
    plb=fopen(outfn,"wb"); }
  else
    plb=stdout;

  if (!ns) {
    if (!fgets(line,128,pla)) Error("file too short");
    sscanf(line,"%d%f",&ns,&hdr[1]);
    fprintf(stderr,"ns=%d L=%f\n",ns,hdr[1]); }

  hdr[0]=ns;
  if (reverse) {
    endian((char*)hdr); endian((char*)(hdr+1)); }
  if (isns) fwrite(hdr,sizeof(float),2,plb);

  while (fgets(line,128,pla)) {
    char *tok=strtok(line," \t\n\r,");

    if (!tok) continue;

    if (atm) {
      tok=strtok(NULL," \t\n\r,"); if (!tok) continue; }

    r[0]=atof(tok); 
    tok=strtok(NULL," \t\n\r,"); if (!tok) continue;  
    r[1]=atof(tok);
    tok=strtok(NULL," \t\n\r,"); if (!tok) continue;  
    r[2]=atof(tok);
    n++;
    if (reverse) loop (j,0,3) endian((char*)(r+j));
    fwrite(r,sizeof(floatvector),1,plb); }

  fprintf(stderr,"%d vectors read, %g configuration(s) written\n",
	 n,                (double)n/(ns+(hdr[1]<0)));

  fclose(plb);
  fclose(pla);

  return 0;
}
