/* \make cfg2asc

*/

#define SDS
#include "ground.h"
#include "alloc.h"
#include "varfile.h"

typedef double vector[3];

int main(int narg,char **arg)
{
  FILE *asc;
  vector L={0,0,0};
  char *fn;
  struct a_s {
    int  size;    /* whole struct in bytes */
    int  padd;    /* padded to 8 bytes */
    double logs;  /* log of the Nose variable s */
    vector lambda;/* log(box) */
    vector shape; /* (reserved for box shape) */
    vector rp[1]; /* contiguous array of all r[ns] (and p[ns]) */
  } *a[3];
  int nspec;
  struct spec_s {
    int N;           /* number of molecules */
    int ns;          /* number of sites */
  } *spec; /* [nspec] */
  struct rec_s {
    int key;
    int intval;
    vector vecval;
  } rec;
  int i,k,cfgkey,N=0,ns=0,nsx,oldsize;
  int optionlist[32];
  double t,h,Entot;

  if (narg<2) {
    fprintf(stderr,"Convert cfg-file into asc-image. Call by:\n\
  cfg2asc NAME\n\
where\n\
  NAME.cfg will be converted into NAME.cfa\n\
See also:\n\
  asc2cfg cfg2atm plb2cfg cfgconv\n");
    exit(0); }

  initscroll(0);

  VarOpen(fn=string("%s.cfg",arg[1]),"r");
  asc=fopen(string("%s.cfa",arg[1]),"wt");
  prt("reading %s",fn);
  VarRead(&cfgkey,sizeof(cfgkey));
  fprintf(asc,"%d cfgkey (%s)\n",cfgkey,cfgkey&2?"polar":"nonpolar");

  VarRead(optionlist,sizeof(optionlist));
  loop (i,0,32) fprintf(asc,"-%c%d\n",tolower(i+'@'),optionlist[i]);

  VarRead(&nspec,sizeof(nspec));
  fprintf(asc,"%d nspec\n",nspec);                                            

  allocarray(spec,nspec);
  VarRead(spec,nspec*sizeof(spec[0]));
  loop (i,0,nspec) fprintf(asc,"%d %d N ns (of species %d)\n",spec[i].N,spec[i].ns,i);

  while (VarFile.size==sizeof(struct rec_s)) {
    VarRead(&rec,sizeof(rec));
    fprintf(asc,"%d %d %.16g %.16g %.16g key %s\n", 
            rec.key,rec.intval,rec.vecval[0],rec.vecval[1],rec.vecval[2],
            rec.key==1?"N t h En.tot" : rec.key==2?"ns L[3]" : "intval vecval[3]");

    switch (rec.key) {
      case 1: N=rec.intval;
        put2(cfgkey,N)
        t=rec.vecval[0];
        h=rec.vecval[1];
        Entot=rec.vecval[2];
        put3(t,h,Entot)
        break;

      case 2:
        ns=rec.intval;
        loop (k,0,3) L[k]=rec.vecval[k];
        prt("Ns=%d L=%g %g %g loaded",ns,L[0],L[1],L[2]);
        break;
    } }

  fprintf(asc,". end of header: cfg, vel, acc follow, VarFile.size=%u\n",VarFile.size);

  put(VarFile.size)
  nsx=(VarFile.size-sizeof(*a[0]))/sizeof(vector)+1;
  //  nsx=(VarFile.size-((char*)(a[0]->rp) - (char*)(&a[0]->size)))/sizeof(vector);
  if (cfgkey&2) {
    prt("POLAR configuration detected");
    nsx/=2; }
  if (ns && nsx!=ns) prt("WARNING: ns=%d from record length, %d from key=2: %d used",nsx,ns,nsx);
  ns=nsx;
  put(ns)
  if (ns<N) ERROR(("ns<N read from file"))

  loop (i,0,3) sdsalloczero(a[i],VarFile.size);
  loop (i,0,3) {
    oldsize=VarFile.size;
    if (VarFile.size) {
      VarRead(a[i],VarFile.size);
      if (a[i]->size!=oldsize)
        ERROR(("a[%d]: size of SDS in file (%d) does not match VarFile.size (%d)",
               i,a[i]->size,oldsize)) } }

  fprintf(asc,"%20.16f %.16g %.16g logs\n",
          a[0]->logs, a[1]->logs, a[2]->logs);
  fprintf(asc,"%20.16f %20.16f %20.16f  %.16g %.16g %.16g  %.16g %.16g %.16g  lambda=ln(L)\n",
          a[0]->lambda[0], a[0]->lambda[1], a[0]->lambda[2],
          a[1]->lambda[0], a[1]->lambda[1], a[1]->lambda[2],
          a[2]->lambda[0], a[2]->lambda[1], a[2]->lambda[2]);

  fprintf(asc,"! i  r[3]  h*velocity[3]  h^2*acceleration[3]\n");
  loop (i,0,ns)
    fprintf(asc,"%20.16f %20.16f %20.16f  %.16g %.16g %.16g  %.16g %.16g %.16g site[%d]\n",
            a[0]->rp[i][0], a[0]->rp[i][1], a[0]->rp[i][2],
            a[1]->rp[i][0], a[1]->rp[i][1], a[1]->rp[i][2],
            a[2]->rp[i][0], a[2]->rp[i][1], a[2]->rp[i][2],
            i);

  VarClose();

  fclose(asc);


  prt("%s: ascii image written",fn);

  return 0;
}
