/* make cppak

  8/2010: options improved, multiple files supported

  Compress/decompress convergence profile files (cppak0.c, 1995)

*/

#include "ground.h"
#include "alloc.h"
#include "bitfile.h"
#include "pakcp.h"

static int reverse;                /* endian change */

rec_t *rechead;

int main(int narg,char **arg) /*************************************** main */
{
  char nm[256];
  FILE *cp,*pakcp;
  char *c,*dot;
  int NCP,i,iarg;
  int nbit=15;
  int verbose=1,erase=0,overwrite=0,ignore=0;
  rec_t *rec,*last=NULL;
  float *r;
  enum mode_e { AUTO,COMPRESS,DECOMPRESS } mode=AUTO,dir=AUTO;
  char line[128];

  if (narg<2) {
    fprintf(stderr,"Lossy compression of MACSIMUS convergence profile.  Call by:\n\
  cppak [OPTIONS] FILE.EXT [[OPTIONS] FILE.EXT ..]\n\
FILES:\n\
  FILE.cp  = MACSIMUS convergence profile file\n\
  FILE.cpz = packed convergence profile file\n\
OPTIONS (apply to the rest of arguments)\n\
  -nNBITS  number of bits in [min,max] (default=%d/A)\n\
  -a       compress FILE.cp, decompress FILE.cpz (default mode)\n\
  -d       force decompress\n\
  -c       force compress\n\
  -o       overwrite the target without warning (default = ask)\n\
  -e       erase the original after (de)compression\n\
  -q       quiet\n\
  -r       reverse endian on input\n\
  -y       shorthand for -o -e -q\n\
See also:\n\
  showcp cp2cp plbpak\n",nbit);
  exit(0); }

  initscroll(0);

  loop (iarg,1,narg) 
    if (arg[iarg][0]=='-')
      switch (arg[iarg][1]) {
        case 'n': nbit=atoi(arg[iarg]+2); break;
        case 'a': mode=AUTO; break;
        case 'c': mode=COMPRESS; break;
        case 'd': mode=DECOMPRESS; break;
        case 'i': ignore=1; break;
        case 'q': verbose=0; break;
        case 'r': reverse=1; break;
        case 'e': erase=1; break;
        case 'y': erase=1; verbose=0; ignore=1;
        case 'o': overwrite=1; break; 
        default: ERROR(("%s: unnown option",arg[iarg])) }
    else {
      strcpy(nm,arg[iarg]);
      /* determine extension */
      dot=NULL;
      for (c=nm; *c; c++) if (*c=='.') dot=c;
      dir=mode;
      if (mode==AUTO) {
        if (!dot) ERROR(("%s: no extension and none of -c -d specified",arg[iarg])) 
        if (!strcmp(dot,".cp")) dir=COMPRESS;
        if (!strcmp(dot,".cpz")) dir=DECOMPRESS;
        if (dir==AUTO) ERROR(("%s: unknown extension and none of -c -d specified",arg[iarg]))  }

      if (dir==COMPRESS) {

        /*** compression ***/

        prt("compressing %s...",nm);
        cp=fopen(nm,"rb");
        if (!cp) Error("no such file");
        if (dot) strcpy(dot,".cpz"); else strcat(nm,".plz");
        if (!overwrite) {
          pakcp=fopen(nm,"rb");
          if (pakcp) {
            prt("%s exists. Overwrite (y/n)? ",nm);
            gets(line);
            if (toupper(line[0])!='Y') continue;
            fclose(pakcp); } }
        pakcp=fopen(nm,"wb");

        r=getcpheader(cp,&NCP,&reverse,verbose);
        if (fwrite(r,sizeof(float),NCP,pakcp)!=NCP)
          ERROR(("%s: cannot write",arg[iarg]))

        while (!feof(cp)) {
          for (;;) {
            i=fread(r,sizeof(float),NCP,cp);
            if (i!=NCP) {
              if (i) ERROR(("%s: bad file end",arg[iarg]))
              if (rechead) break;
              goto done; }
            if (reverse) loop (i,0,NCP) endian((char*)(r+i));
            if (r[0]==CPmark) break;
            alloc(rec,sizeof(rec_t)-sizeof(float)+NCP*sizeof(float));
            copy(rec->r,r,NCP*sizeof(float));
            if (!rechead) rechead=last=rec;
            else last=last->next=rec; }

          if (last) last->next=NULL;
          appendpakcp(pakcp,NCP,nbit,rechead,0);

          if (r[0]==CPmark)
            if (fwrite(r,sizeof(float),NCP,pakcp)!=NCP) Error("cannot write");

          /* free all */
          while (rechead) {
            rec=rechead->next; free(rechead); rechead=rec; } } }
      
      else {

        /*** decompression ***/

        prt("decompressing %s...",nm);
        pakcp=fopen(nm,"rb");
        if (!pakcp) Error("no such file");
        if (dot) strcpy(dot,".cp"); else strcat(nm,".cp");
        if (!overwrite) {
          cp=fopen(nm,"rb");
          if (cp) {
            prt("%s exists. Overwrite (y/n)? ",nm);
            gets(line);
            if (toupper(line[0])!='Y') continue;
            fclose(cp); } }
        cp=fopen(nm,"wb");
        
        r=getcpheader(pakcp,&NCP,&reverse,verbose);
        if (fwrite(r,sizeof(float),NCP,cp)!=NCP) Error("cannot write");

        while (getpakcp(pakcp)) {
          while (nextcprec(r)) 
            if (fwrite(r,sizeof(float),NCP,cp)!=NCP) Error("cannot write");
          endgetpakcp(pakcp); }
      }
      
    done:
      fclose(pakcp);
      fclose(cp);
      if (erase) {
        fflush(stdout);
        remove(arg[iarg]);
        if (verbose) prt("%s erased.\n",arg[iarg]); } }

  return 0;
}



