/* cc -o gaussian2plb gaussian2plb.c -lm
 */
#include "../gen/include.h"
typedef double real;
#include "../gen/vector3d.h"

double ftof(char *c)
/* get fortran-style number (accepts 9.9D+9 and 9.9d+9) */
{
  char nr[128],*cc;

  while (*c && *c<=' ') c++;

  if (!*c) return 0;
  cc=nr;
  while (*c>' ') {
    if (toupper(*c=='D')) *cc++='e';
    else *cc++=*c;
    c++; }
  *cc=0;

  return atof(nr);
}

int main(int narg,char **arg)
{
  FILE *in;
  char *fn;
  FILE *plb=NULL;
  float r[3];
  char line[256];
  int i,n,N=0;
  vector *center;
  vector **dr;
  double *nmf; /* [cm-1] */
  double A=1;
  int isfreq=0;

  if (narg<2) {
    fprintf(stderr,"Call by:\n\
  gaussian2plb NAME [AMPLITUDE]\n\
where AMPLITUDE is the normal mode vibration amplitude (if present)\n\
- the Gaussian output file NAME.log is read\n\
- file NAME-gauss.plb with energy minimization trajectory is generated\n\
  if NAME.mol exists, show is called w. coordinate matcher:\n\
    show NAME NAME-opt.plb -mNAME.plb:1:NAME.keep -I@\n\
- files nm####.plb are generated\n\
  if NAME.mol exists, show is called with vibration viewer:\n\
    NAME nm%%04d.plb -Iw%%\n");
    exit(0); }

  if (narg>2) A=atof(arg[2]);
  
  in=fopen(fn=string("%s.log",arg[1]),"rt");
  if (!in) Error(fn);

  /* reading NAtoms: it may appear AFTER configuration !!! */
  while (fgets(line,256,in)) 
    if (strstr(line,"NAtoms=")) {
      if (N) {
        if (N!=ftof(line+8)) Error("NAtoms changed"); }
      else {
        N=ftof(line+8);
        if (plb) Error("plb already open");
        plb=fopen(string("%s-gauss.plb",arg[1]),"wb");
        r[0]=N; r[1]=0; /* old plb-file */
	put(N)
        fwrite(r,4,2,plb);
        allocarray(center,N); 
        allocarray(dr,3*N-6);
        loop (n,0,3*N-6) allocarray(dr[n],N);
        allocarray(nmf,3*N-6); } }

  if (!N) Error("keyword \"NAtoms=\" is missing");

  rewind(in);

  while (fgets(line,256,in)) 
    if (strstr(line,"Standard orientation")) {
      fgets(line,256,in);
      fgets(line,256,in);
      fgets(line,256,in); 
      fgets(line,256,in); 
      if (line[3]!='-') Error("format");


      loop (i,0,N) {
        fgets(line,256,in);
        /*
    1          1             0        1.919787   -0.003953    0.753920
        */
        center[i][0]=ftof(line+34);
        center[i][1]=ftof(line+46);
        center[i][2]=ftof(line+58); 
        VV(r,=center[i])
        fwrite(r,4,3,plb); } }

  fclose(plb);
  
  plb=fopen(string("%s.mol",arg[1]),"rt");
  if (plb) {
    system(string("show %s %s-gauss.plb -m%s.plb:1:%s.keep -I@",
      arg[1],arg[1],arg[1],arg[1]));
    fclose(plb); }
  
  rewind(in);
  n=0;

  while (fgets(line,256,in))
    if (strstr(line,"Frequencies --")) {
      isfreq++;
      if (n>=3*N-6) Error("to many NMFs");

      /*
                     1                      2                      3
                     A                      A                      A
 Frequencies --   127.9989               147.9907               150.7587
 Red. masses --     1.0709                 1.3622                 1.0323
 Frc consts  --     0.0103                 0.0176                 0.0138
 IR Inten    --   112.2665               160.1418                59.7591
 Atom AN      X      Y      Z        X      Y      Z        X      Y      Z
   1   1     0.00   0.63   0.00    -0.23  -0.01   0.17     0.00   0.52   0.00
   2   1     0.00   0.20   0.00     0.07   0.00  -0.37     0.00  -0.32  -0.01
   3   8     0.00  -0.06   0.00     0.14   0.00   0.00     0.00   0.00   0.00
   4   1    -0.37   0.06   0.38    -0.53   0.00   0.32     0.49   0.25   0.09
   5   1     0.36   0.07  -0.37    -0.52  -0.01   0.33    -0.50   0.25  -0.08
   6   8     0.00   0.00   0.00    -0.06   0.00  -0.03     0.00  -0.04   0.00
      */
      nmf[n]=ftof(line+15);
      nmf[n+1]=ftof(line+38);
      nmf[n+2]=ftof(line+61);
      while (fgets(line,256,in)) if (strstr(line," Atom ")) break;

      loop (i,0,N) {
        fgets(line,256,in);
        dr[n][i][0]=ftof(line+10);
        dr[n][i][1]=ftof(line+17);
        dr[n][i][2]=ftof(line+24);
        dr[n+1][i][0]=ftof(line+33);
        dr[n+1][i][1]=ftof(line+40);
        dr[n+1][i][2]=ftof(line+47);
        dr[n+2][i][0]=ftof(line+56);
        dr[n+2][i][1]=ftof(line+63);
        dr[n+2][i][2]=ftof(line+70); }

      n+=3; }

  put2(n,3*N-6)

  if (isfreq) loop (n,0,3*N-6) {
    char *fn=string("nm%04d.plb",n);
    int k;

    plb=fopen(fn,"wb");
    r[0]=N; r[1]=0; /* old plb-file */
    fwrite(r,4,2,plb);

    printf("%s: %g\n",fn,nmf[n]);

    loopto (k,0,6) {
      double q=A*cos(k*PI/6);

      loop (i,0,N) {
        VVV(r,=center[i],+q*dr[n][i])
        fwrite(r,4,3,plb); } }
    fclose(plb); 
    plb=fopen(string("%s.mol",arg[1]),"rt");
    if (plb) {
      system(string("show %s nm%%04d.plb -I%%w",arg[1],arg[1]));
      fclose(plb); }
    }

  return 0;
}
