/* cc -O2 -o hbonds hbonds.c -lm
 */
#include "../gen/include.h"

int NS;
#define NB 12 /* max # of H-bonds per molecule */
int H1,H2,O,M,LL,OFF;
int *add;

int ren(int i)
/* renumber NS-site water to 3-site water */
{
  return ((i-OFF)/NS)*3+add[(i-OFF)%NS]+OFF;
}

typedef float vector[3];
vector *r;
float L[3];
double Lh[3];
FILE *plbin;
int varL,ns;


struct site_s {
  int nnbr;
  int nbr[NB]; } *s;

void addbond(i,j)
{
  if (i>=ns || j>=ns) Error("index out of range (bad OFF?, bad SITES?)");
  if (s[i].nnbr>=NB) Error("too many bonds");
  s[i].nbr[s[i].nnbr++]=j;
  if (s[j].nnbr>=NB) Error("too many bonds");
  s[j].nbr[s[j].nnbr++]=i;
}

int periodic;

double distq(int i,int j)
{
  int k;
  double rr=0;

  if (periodic) 
    loop (k,0,3)
      rr+=Sqr(fabs(fabs(r[i][k]-r[j][k])-Lh[k])-Lh[k]);
  else
    loop (k,0,3)
      rr+=Sqr(r[i][k]-r[j][k]);
    
  return rr;
}

int readplb()
{
  int k;
  
  if (varL) {
    if (fread(L,12,1,plbin)!=1) return 0;
    loop (k,0,3) Lh[k]=L[k]/2; }
  if (fread(r,12,ns,plbin)!=ns) return 0;
  
  return 1;
}

int main(int narg,char **arg)
{
  int i,j,frame;
  FILE *plb,*mol,*gol;
  float hdr[2];
  double ho,hoq;
  char *sem;

  if (narg<2) {
    fprintf(stderr,"Generate H-bonds for a model of water. Call by:\n\
  hbonds PLBFILE SITES[:OFF] [O-H distance]\n\
output = series of files hb####.{plb,mol,gol} in 3-site water format\n\
both old and new plb formats accepted, output type=input type\n\
ARGUMENTS:\n\
  SITES = string of {H,O,M,L} defining the water model\n\
  OFF = offset (# of atoms other than water at the beginning)\n\
  O-H distance defining the hydrogen-bond (default=2.44)\n\
  negative O-H distance = ignore periodic b.c.\n\
Example:\n\
  hbonds solution.plb HOMH:12 -2.4\n\
See also:\n\
  hbonds4 (old version)\n\
");
    exit(0); }

  plbin=fopen(arg[1],"rb");
  if (!plbin) Error(arg[1]);
  if (narg>3) ho=atof(arg[3]); else ho=2.44;
  periodic=ho>0;
  ho=fabs(ho);
  hoq=ho*ho;
  fread(hdr,4,2,plbin);
  ns=(int)hdr[0];
  if (ns<1 || ns>100000) Error("too many sites");
  L[0]=L[1]=L[2]=hdr[1];
  loop (i,0,3) Lh[i]=L[i]/2;
  varL=hdr[1]==-3;
  
  allocarray(r,ns);
  allocarray(s,ns);

  if ( (sem=strchr(arg[2],':')) ) {
    OFF=atoi(sem+1);
    NS=sem-arg[2]; }
  else {
    OFF=0;
    NS=strlen(arg[2]); }
    
  allocarray(add,NS);
  j=0;
  loop (i,0,NS) {
    add[i]=j;
    switch (arg[2][i]) {
      case 'H': H1=H2; H2=i; j++; break;
      case 'L': LL=i; break;
      case 'M': M=i; break;
      case 'O': O=i; j++; break; } }

  for (frame=0; readplb(); frame++) {
    
    loop (i,0,ns) s[i].nnbr=0;
    for (i=OFF; i<ns; i+=NS) {
      addbond(i+H1,i+O);
      addbond(i+H2,i+O); }

    for (i=OFF; i<ns; i+=NS)
      for (j=OFF; j<ns; j+=NS) if (i!=j) {
	if (distq(i+H1,j+O)<hoq) addbond(i+H1,j+O);
	if (distq(i+H2,j+O)<hoq) addbond(i+H2,j+O); }

    mol=fopen(string("hb%04d.mol",frame),"wt");
    gol=fopen(string("hb%04d.gol",frame),"wt");
    plb=fopen(string("hb%04d.plb",frame),"wb");
    fprintf(gol,"!\n!\n%d\n",(ns-OFF)/NS*3+OFF);
    fprintf(mol,"water HB structure\n\
parameter_set = water\n\
number_of_atoms = %d\n\
\n\
atoms\n\
! i  atom-id  a-type charge chir nbonds bound_atoms\n",(ns-OFF)/NS*3+OFF);
    hdr[0]=(ns-OFF)/NS*3+OFF;
    fwrite(hdr,4,2,plb);
    if (varL) fwrite(L,4,3,plb);
    loop (i,0,ns) {
      if (i<OFF) {
        fprintf(mol,"%3d %3d-a X 0 0 0\n", i,i);
        fprintf(gol,"GREEN 1.24\n"); }
      else if (strchr("HO",arg[2][i%NS])) {
        fprintf(mol,"%3d %3d-%c%c%03d %c 0 0 %d",
  	      ren(i),i,arg[2][i%NS],OFF?'b':'a',(ns-OFF)/NS,arg[2][i%NS],s[i].nnbr);
        loop (j,0,s[i].nnbr) fprintf(mol," %d",ren(s[i].nbr[j]));
        fprintf(mol,"\n");
        if (arg[2][i%NS]=='H') fprintf(gol,"WHITE 0.56\n");
        else fprintf(gol,"RED 1.24\n"); }
      fwrite(r[i],4,3,plb); }
    fclose(mol);
    fclose(gol);
    fclose(plb); }

  return 0;
}
