/* cc -O2 -o plb2asc plb2asc.c
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#define Error(X) do { fprintf(stderr,"ERROR %s\n",X); exit(1); } while(0)

/* C-style loop (last not included) */
#define loop(_I,_From,_To) for (_I=_From;_I<_To;_I++)

/* Pascal etc. style loop (last included) */
#define loopto(_I,_From,_To) for (_I=_From;_I<=_To;_I++)

typedef float floatvector[3];

void endian(char *a) /********************************************** endian */
{
  char x;

  x=a[0]; a[0]=a[3]; a[3]=x;
  x=a[1]; a[1]=a[2]; a[2]=x;
}

int main(int narg,char **arg) /*************************************** main */
{
  FILE *plb,*pla;
  float hdr[2];
  floatvector r,L;
  int j,n=0,ns=0,isns=1,reverse=0,is=0,varL,iarg,ask=1,omithdr=0,NS;
  char *infn=NULL,*outfn=NULL;
  char *FMT="%8.5f",*atoms=NULL,*ext="pla";
  char *boxopt=" ";
  char *at;

  if (narg<2) {
    fprintf(stderr,"\
Convert playback (and similar) file to text.  Call by:\n\
  plb2asc PLB-FILE ASC-FILE [OPTIONS]\n\
  plb2asc NAME [OPTIONS]\n\
where\n\
  PLB-FILE: MACSIMUS playback file, usual .EXT=.plb,.vlb; -=stdin\n\
  ASC-FILE: its text image, usual .EXT=.pla,.vla; -=stdout\n\
  NAME: PLB-FILE=NAME.plb, ASC-FILE=NAME.pla\n\
  -eEXT set EXT\n\
  -r: reverse endian of PLB-FILE\n\
  -y: overwite without asking\n\
  -%%FMT -f%%FMT: number format (must accept double), default=%s\n\
  -nNS: input is .3db (no header), NS=number of sites (deprecated)\n\
The default output is an ascii image of PLB-FILE. Modificators:\n\
  -aATOMS: (EXT=.atm): header of NS+box (repeats before every frame)\n\
    atom names need not be separated (AlCl=Al,Cl), pattern repeats if needed\n\
  -h: no header (of NS, EXT=.asc)\n\
  -l-: omit box info line [default=-l\" \" (plain box info)]\n\
  -l: blank line instead of box info\n\
  -lHDR: use HDR to start line with box info (e.g., -lbox)\n\
Example (print x,y,z block of last 25 atoms in format compatible with PDB)\n\
  plb2asc simul.plb - \"-%%7.3f\" | tail -n 25\n\
See also:\n\
  asc2plb plbconv plbmerge plbcut plbinfo plbfilt view3db\n\
Deprecated old version: plbasc\n\
",FMT);
    exit(0); }

  loop (iarg,1,narg)
    if (arg[iarg][0]=='-' && arg[iarg][1])
      switch (arg[iarg][1]) {
        case '%': FMT=arg[iarg]+1; break;
        case 'f': FMT=arg[iarg]+2; break;
        case 'e': ext=arg[iarg]+2; break;
        case 'r': reverse++; break;
        case 'y': ask=0; break;
        case 'l': boxopt=arg[iarg]+2; break;
        case 'h': omithdr++; ext="asc"; break;
        case 'a': atoms=arg[iarg]+2; ext="atm"; break;
        case 'n': ns=atoi(arg[iarg]+2); isns=0; break;
        default: Error("wrong option"); }
    else {
      if (infn) Error("more than 2 file names");
      infn=outfn;
      outfn=arg[iarg]; }

  if (!outfn) Error("no file/name");
  if (!infn) {
    char *name=outfn;

    if (strlen(ext)>3) Error("EXT loger than 3 chars");
    infn=malloc(strlen(name)+5);
    sprintf(infn,"%s.plb",name);
    outfn=malloc(strlen(name)+5);
    sprintf(outfn,"%s.%s",name,ext); }

  if (strcmp("-",infn)) plb=fopen(infn,"rb");
  else plb=stdin;
  if (!plb) Error(infn);
  fprintf(stderr,"%s (binary) -> %s (ascii)\n",infn,outfn);

  if (strcmp("-",outfn)) {
    pla=fopen(outfn,"rt");
    if (pla && ask) {
      char line[256];

      fprintf(stderr,"File %s exists. Overwrite (y/n)? ",outfn);
      gets(line);
      if (tolower(line[0])!='y') return 1;
      fclose(pla); }
    pla=fopen(outfn,"wt"); }
  else
    pla=stdout;

  if (!ns) {
    if (fread(hdr,sizeof(hdr),1,plb)!=1) Error("file too short");
      if (reverse) {
	endian((char*)hdr);
	endian((char*)(hdr+1)); }
      ns=hdr[0];
      fprintf(stderr,"ns=%d L=%f\n",ns,hdr[1]); }

  varL=hdr[1]<0;
  if (!varL) L[0]=L[1]=L[2]=hdr[1];
  NS=ns+varL;

  if (!omithdr && !atoms) {
    fprintf(pla,"%d ",ns);
    fprintf(pla,FMT,hdr[1]);
    fprintf(pla,"\n"); }

  while (fread(r,sizeof(floatvector),1,plb)==1) {

    if (is==0) {
      /* header */
      if ( (at=atoms) ) {
        fprintf(pla,"%d ",ns);
        fprintf(pla,"\n"); }

      /* box info line */
      if (varL) {
        /* L=r */
        if (!strcmp(boxopt,"")) goto blank; /* empty line */
        if (!strcmp(boxopt,"-")) goto omit; /* no box line */
        fprintf(pla,"%s",boxopt); /* will write L=r */ }
      else {
        if (!strcmp(boxopt,"")) fprintf(pla,"\n"); /* empty line */
        else if (!strcmp(boxopt,"-")) /* no box line */;
        else {
          /* write L */
          fprintf(pla,"%s",boxopt);
          loop (j,0,3) fprintf(pla,FMT,L[j]);
          fprintf(pla,"\n"); } } }
    else {
      if (atoms) {
        if (*at==0) at=atoms;
        if (*at) {
          fprintf(pla,"%c",*at++);
          if (islower(*at)) fprintf(pla,"%c",*at++);
          if (*at && !isalnum(*at)) at++; } }
      else
        fprintf(pla," "); }

    loop (j,0,3) {
      if (reverse) endian((char*)(r+j));
      fprintf(pla," ");
      fprintf(pla,FMT,r[j]); }
   blank:
    fprintf(pla,"\n");
   omit:
    is=(is+1)%NS;
    n++; }

finished:

  fprintf(stderr,"%d vectors read, %g configuration(s) written\n",n,(double)n/NS);

  fclose(plb);
  fclose(pla);

  return 0;
}
