/* cc -O2 -o plb2nbr plb2nbr.c -lm
 */
#include "../gen/include.h"

int main(int narg,char **arg)
{
  char line[1024];
  typedef double vector[3];
  vector *r,L;
  int i,j,N,k,frame,oldformat;
  double dist,R=1;
  FILE *mol,*gol,*plb;
  FILE *in;
  float f[3];
  int NRED=0,SCALE=1;
  static int hist[1000];

  if (narg<3) {
    fprintf(stderr,"Read plb file and sort sites to files according to # of neighbors.\n\
Call by:\n\
  plb2nbr PLBFILE DIST [DIAM [[-]NRED [SCALE]]]\n\
DIST = distance to define neighbors\n\
DIAM = diameter [default=1]\n\
NRED = # of neighbors shown RED (then, MAGENTA BLUE CYAN GREEN YELLOW WHITE)\n\
       [default=0]\n\
-NRED : nbrs<=NRED are small RED, nbrs>NRED+DN*6 are WHITE (default: modulo 7)\n\
SCALE : nbrs in [NRED,NRED+SCALE) are RED, etc. [default=1]\n\
see also: c/shownbr hbonds plb2cryst\n");
    exit(0); }

  in=fopen(arg[1],"rb");
  if (!in) Error(arg[1]);
  dist=atof(arg[2]);
  if (narg>3) R=atof(arg[3]);
  if (narg>4) NRED=atoi(arg[4]);
  if (narg>5) SCALE=atoi(arg[5]);

  fread(f,4,2,in);
  
  N=f[0];
  if (N<1 || N>1000000) Error("bad N");
  oldformat=f[1]>=0;
  if (oldformat) L[0]=L[1]=L[2]=f[1];

  allocarrayzero(r,N);

  for (frame=0;;frame++) {

    loop (i,0,1000) hist[i]=0;

    if (!oldformat) {
      if (3!=fread(f,4,3,in)) exit(0);
      L[0]=f[0]; L[1]=f[1]; L[2]=f[2]; }

    loop (i,0,N) {
      if (3!=fread(f,4,3,in)) exit(0);
      r[i][0]=f[0]; r[i][1]=f[1]; r[i][2]=f[2]; }

    mol=fopen(string("nbr%04d.mol",frame),"wt");
    fprintf(mol,"neighbors\n\
parameter_set=unknown\n\
number_of_atoms=%d\n\
atoms\n",N);

    gol=fopen(string("nbr%04d.gol",frame),"wt");
    fprintf(gol,"!\n!\n%d\n",N);

    plb=fopen(string("nbr%04d.plb",frame),"wt");
    f[0]=N; f[1]=-3;
    fwrite(f,4,2,plb);
    f[0]=L[0]; f[1]=L[1]; f[2]=L[2];
    fwrite(f,4,3,plb);

    loop (i,0,N) {
      int nnbr=-1; /* i-i will count for 1 */
      char *col="WHITE";
      double rad=R/2;

      loop (j,0,N) {
	double dd;
	if (L[0])
	  dd=Sqr(L[0]/2-fabs(L[0]/2-fabs(r[i][0]-r[j][0])))
	    +Sqr(L[1]/2-fabs(L[1]/2-fabs(r[i][1]-r[j][1])))
	    +Sqr(L[2]/2-fabs(L[2]/2-fabs(r[i][2]-r[j][2])));
	else
	  dd=Sqr(r[i][0]-r[j][0])+Sqr(r[i][1]-r[j][1])+Sqr(r[i][2]-r[j][2]);

	nnbr+=dd<Sqr(dist); }


      if (nnbr>=1000) Error("more than 1000 neighbors");
      hist[nnbr]++;

      nnbr+=1001*SCALE-abs(NRED);
      nnbr/=SCALE;
      if (NRED<0) {
	nnbr-=1001;
	if (nnbr<0) nnbr=0;
	if (nnbr==0) rad=R/20;
	if (nnbr>6) nnbr=6; }
      else nnbr%=7;
      if (nnbr<0) Error("");

      col="R\0M\0B\0C\0G\0Y\0W"+nnbr*2;
      fprintf(gol,"%s %f\n",col,rad);
      fprintf(mol,"%d HS HS-%s 0 0 0\n",i,col);
      f[0]=r[i][0];
      f[1]=r[i][1];
      f[2]=r[i][2];
      fwrite(f,4,3,plb); }

    fclose(plb);
    fclose(mol);
    fclose(gol); 

    printf("%d:",frame);
    loop (i,0,1000) if (hist[i]) printf(" %d=%d",i,hist[i]);
    printf("\n"); }

  return 0;
}
