/* cc -o plbmerge -O2 plbmerge.c
(c) J.Kolafa 11/95
05/08: append option added
06/05: new plb format (variable L[3])
11/04: more frames allowed
10/00: arbitrary selection added
08/97: header[1] printed as t0 (10/00: changed into L)
       bug fixed (FROM did not work correctly)
       -BY added
11/95: first version
*/

#include "../gen/include.h"

int main(int narg,char **arg)
{
  FILE **f;
  float r[3],hdr[2];
  int iarg,*ns,*varL,i,frame=0,NS=0,narg0=1,cat=0,nsmin=0x7fffffff;
  float L=0,L3[3],*LL;
 
  if (narg<2) {
    fprintf(stderr,"\
Merge binary playback files.  Call by:\n\
  plbmerge [-c[#]] PLB-FILE [PLB-FILE ...] > MERGED-PLB-FILE\n\
OPTION:\n\
  NONE\tmerge frames synchronously (default):\n\
\t- stops when any plb-file reaches EOF\n\
\t- output box size L is set to MAXIMUM of all plb-files\n\
\t- hint: use molcfg to create merged mol- and gol-files\n\
  -c#\ttruncate playback files to # sites and concatenate\n\
  -c\tas above, the minimum number of sites of all files is used\n\
Always generates new plb format (cf. plbconv or mergeplb for old format)\n\
Examples:\n\
  plbmerge -c run1.plb run2.plb > runs1+2.plb\n\
  plbmerge sim.p00 sim.p01 > sim.plb\n\
See also:\n\
  plbinfo plbmsd plbcut plbfilt plbtran plbrot plb2diff plbasc\n");
    exit(0); }
  
  hdr[0]=0;
  
  allocarray(f,narg);
  allocarray(ns,narg);
  allocarray(varL,narg);
  allocarray(LL,narg);
  
  if (!memcmp(arg[1],"-c",2)) NS=atoi(arg[1]+1),narg0=2,cat=1;
  
  loop (iarg,narg0,narg) {
    f[iarg]=fopen(arg[iarg],"rb");
    if (!f[iarg]) Error(arg[iarg]);
    if (2!=fread(r,4,2,f[iarg])) Error(arg[iarg]);
    fprintf(stderr,"%20s: ns=%-5.0f L=%f\n",arg[iarg],r[0],r[1]);
    varL[iarg]=r[1]==-3;
    Max(L,r[1])
    LL[iarg]=r[1];
    Min(nsmin,(int)r[0])
    hdr[0]+=r[0];
    ns[iarg]=r[0]; }

  hdr[1]=-3;

  if (cat) {
    if (NS) if (NS<nsmin) nsmin=NS;
    else Error("too many sites in option -c");
    if (nsmin<1) Error("no site!");
    hdr[0]=nsmin; }

  fprintf(stderr,"              OUTPUT: ns=%-5.0f L=%g (variable box)\n",hdr[0],hdr[1]);

  fwrite(hdr,4,2,stdout);

  if (cat) {
    int frame0=0;
    
    loop (iarg,narg0,narg) {
      for (;;) {
        int writeL=1;
  
        if (varL[iarg]) {
          if (3!=fread(L3,4,3,f[iarg])) {
            fprintf(stderr,"%s: %d frame(s) written; %d total\n",
              arg[iarg],frame-frame0,frame);
            frame0=frame;
            goto cont; } }
        else {
          L3[0]=L3[1]=L3[2]=LL[iarg]; }
        
        loop (i,0,ns[iarg]) {
          if (3!=fread(r,4,3,f[iarg])) {
            fprintf(stderr,"%s%s: %d frame(s) written; %d total\n",
              i||varL[iarg]?"unexpected EOF (truncated frame) on ":"",arg[iarg],frame-frame0,frame);
            frame0=frame;
            goto cont; }
          if (writeL) {
            fwrite(L3,4,3,stdout);   
            writeL=0; }
          if (i<nsmin) fwrite(r,4,3,stdout); }
        frame++; }
       cont:; } }

  else for (;;) {
    L3[0]=L3[1]=L3[2]=L;
    loop (iarg,narg0,narg)
      if (varL[iarg]) {
        if (3!=fread(r,4,3,f[iarg])) {
	  fprintf(stderr,"%d frame(s) written\n",frame);
	  return 0; }
	Max(L3[0],r[0])
	Max(L3[1],r[1])
        Max(L3[2],r[2]) }

    fwrite(L3,4,3,stdout);

    loop (iarg,narg0,narg) {
      loop (i,0,ns[iarg]) {
	if (3!=fread(r,4,3,f[iarg])) {
	  fprintf(stderr,"EOF on %s: %d frame(s) written\n",arg[iarg],frame);
	  return 0; }
	fwrite(r,4,3,stdout); } }
    frame++; }

  return 0;
}
