/* cc -o plbreplicate plbreplicate.c
*/

#include "../gen/include.h"

typedef float vector[3];

int perm=0;

void permute(vector v)
{
  float x;

  switch (perm) {
    case 0: break;
    case 2: x=v[0]; v[0]=v[1]; v[1]=v[2]; v[2]=x; break;
    case 4: x=v[0]; v[0]=v[2]; v[2]=v[1]; v[1]=x; break;
    default: Error("bad permutation"); }
}

int main(int narg,char **arg)
{
  FILE *f1,*f2;
  int ns,i,nx,ny,nz,ix,iy,iz;
  static float header[2];
  vector *r;
  vector L,newL;
  double Xsqueeze=0,Ysqueeze=0;

  if (narg<5) {
    fprintf(stderr,"Replicate a cell. Call by:\n\
  plbreplicate INPUT_FILE OUTPUT_FILE NX NY NZ [XSQUEEZE [YSQUEEZE]]\n\
ARGUMENTS:\n\
  NX NY NZ = replication factors; negative values determine input shuffling:\n\
  NY<0 = cyclic permutation: y -> x -> z -> y (input y becomes x, etc.)\n\
  NZ<0 = cyclic permutation: z -> x -> y -> z (input z becomes x, etc.)\n\
  INPUT_FILE = plb-file with box info (new format)\n\
  OUTPUT_FILE = the same format\n\
  XSQUEEZE = adds XSQUEEZE*z/newLz*newLx to x\n\
  YSQUEEZE = adds Ysqueeze*z/newLz*newLy to y\n\
WARNING: not suitable as cook* input if the cell contains more species\n\
See also:\n\
  cook* (load.n[] in data)   molcfg\n");
    exit(0); }

  if (!strcmp(arg[1],arg[2])) Error("INPUT_FILE = OUTPUT_FILE");

  f1=fopen(arg[1],"rb");
  if (f1==NULL) Error("no INPUT_FILE");

  f2=fopen(arg[2],"wb");
  if (f2==NULL) Error("cannot open OUTPUT_FILE");

  fread(header,sizeof(header),1,f1);
  ns=header[0];

  if (header[1]!=-3) Error("wrong header (variable box required), try plbconv");

  allocarray(r,ns);

  perm=0;
  nx=atoi(arg[3]); if (nx<0) perm+=1,nx=-nx;
  ny=atoi(arg[4]); if (ny<0) perm+=2,ny=-ny;
  nz=atoi(arg[5]); if (nz<0) perm+=4,nz=-nz;


  if (narg>6) Xsqueeze=atof(arg[6]);
  if (narg>7) Ysqueeze=atof(arg[7]);
  
  header[0]=ns*nx*ny*nz;
  if (header[0]>16777216.||header[0]<1) Error("bad number of sites (max. 16777216)");
  
  fwrite(header,sizeof(header),1,f2);
  fread(L,sizeof(L),1,f1); permute(L);

  newL[0]=L[0]*nx;
  newL[1]=L[1]*ny;
  newL[2]=L[2]*nz;
  if (ns!=fread(r,sizeof(vector),ns,f1)) Error("plb file truncated");
  loop (ix,0,ns) permute(r[ix]);

  fwrite(newL,sizeof(newL),1,f2);

  loop (ix,0,nx) 
    loop (iy,0,ny) 
      loop (iz,0,nz) 
        loop (i,0,ns) {
          vector newr;

          newr[2]=L[2]*iz+r[i][2]; 
          newr[0]=L[0]*ix+r[i][0]+Xsqueeze*newL[0]/newL[2]*newr[2];
          newr[1]=L[1]*iy+r[i][1]+Ysqueeze*newL[1]/newL[2]*newr[2];
          fwrite(newr,sizeof(newr),1,f2); }

  fclose(f2);
  fclose(f1);

  printf("! plbreplicate %s %s %d %d %d\n\
! N[0]=%.0f ! the total number of sites\n\
L[0]=%.7f\n\
L[1]=%.7f\n\
L[2]=%.7f\n",
         arg[1],arg[2],nx,ny,nz,
         header[0],newL[0],newL[1],newL[2]);


  return 0;
}
