/* cc -o plbsites plbsites.c -lm
*/
#include "../gen/include.h"

int ifollow,iframe,nfollow;

float *old,maxd=-1;
char *fmt=" %9.6f";
FILE *out=NULL;

void follow(float x,float L)
{
  if (ifollow<nfollow) {
    if (iframe) {
      while (x<old[ifollow]-L/2) x+=L;
      while (x>old[ifollow]+L/2) x-=L;
      Max(maxd,fabs(x-old[ifollow])/L) }
    old[ifollow++]=x; }
  if (out) fprintf(out,fmt,x);
}

int main(int narg,char **arg)
{
  int i,k,ns,nsmol=1,oo,from,nmol=1,off=0,n;
  float (*r)[3],hdr[2];
  float *L;
  FILE *plb;
  int varL=0;
  FILE **outfiles=NULL,*outL=NULL;
  char *fnfmt=NULL;

  if (narg<2) {
    fprintf(stderr,"Print trajectory of selected atoms in ASCII. Call by:\n\
  plbsites PLB-FILE [OPTIONS] ATOM[xyz] [ATOM[xyz] ..]\n\
OPTIONS:\n\
  -oOFFSET = number added to each ATOM [default=0]\n\
  -nNS = number of sites per molecule [default=1]\n\
  -mNM = number of molecules [default=1]:\n\
         pattern of ATOMs is repeated NM times with multiples of NS added\n\
  -fFMT = output format, default=\"%s\"\n\
  -FFMT = every atom will be printed to a separate xyz file\n\
          FMT must contain a format accepting int for the atom number\n\
          [default=stdout, one line per frame]\n\
  -l    = follow periodic boundary conditions\n\
  -lNP  = follow periodic boundary conditions for first NP atoms only\n\
ATOM = atom (site) number (from 0 to NS-1); negative number denotes box size\n\
       if none of -o,-m is given, ATOM is just plb atom number\n\
If x,y,z is given, only selected coordinates are printed (default=xyz)\n\
Bug: sites are always in order x,y,z even if different order is specified\n\
Examples:\n\
  plbsites xxx.plb -1z 33zx                # 3 columns: Lz x_33 z_33\n\
  plbsites xxx.plb -1z 33z 33x             # 3 columns: Lz z_33 x_33\n\
  plbsites xxx.plb -Fa%%03d.xyz -n5 -m20 1 # a001.xyz, a006.xyz, ..., a096.xyz\n\
See also:\n\
  plbfilt plb2plb plb2asc\n\
",fmt);
    exit(0); }

  loop (i,2,narg) if (arg[i][0]=='-') switch (arg[i][1]) {
    case 'm': nmol=atoi(arg[i]+2); break;
    case 'n': nsmol=atoi(arg[i]+2); break;
    case 'o': off=atoi(arg[i]+2); break;
    case 'l': nfollow=atoi(arg[i]+2); 
              if (nfollow==0) nfollow=-1; break;
    case 'f': fmt=arg[i]+2; break;
    case 'F': fnfmt=arg[i]+2; 
              if (!strchr(arg[i],'%')) Error("-F: no % in format");
              break;
    default: if (!isdigit(arg[i][1])) Error("bad option"); }

  plb=fopen(arg[1],"rb");
  if (!plb) Error(arg[1]);

  fread(hdr,4,2,plb);
  ns=hdr[0];
  allocarray(r,ns+1);

  if (nfollow==-1) nfollow=ns;
  if (nfollow) allocarray(old,nfollow);

  if (fnfmt) allocarrayzero(outfiles,ns);
  
  r++;
  if (hdr[1]<0) {
    if (hdr[1]==-3) varL=1;
    else Error("wrong L in plb file"); }
  else
    r[-1][0]=r[-1][1]=r[-1][2]=hdr[1];
  L=r[-1];

  for (iframe=0;;iframe++) {
    int oo=off;
    
    if (ns+varL!=fread(r-varL,12,ns+varL,plb)) break;
    ifollow=0;
    
    loop (n,0,nmol) {
      loop (i,2,narg) if (arg[i][0]!='-' || isdigit(arg[i][1])) {
        int s=atoi(arg[i]);
        int all=1;
        int site;
      
        if (s<0) s=-oo-1;
        
        if (s>=ns) Error("specified site number out of range");
        
        site=oo+s;
        if (site>=ns) Error("(calculated) site number out of range");
        if (outfiles) {
          if (site<0) {
            if (outL==NULL) outL=fopen(string(fnfmt,-1),"wt");
            out=outL; }
          else {
            if (outfiles[site]==NULL) outfiles[site]=fopen(string(fnfmt,site),"wt");
            out=outfiles[site]; } }
        else
          out=stdout;

        if (!out) Error("no output file (wrong format, too many files)");
        
        if (strchr(arg[i],'x')) follow(r[site][0],L[0]),all=0;
        if (strchr(arg[i],'y')) follow(r[site][1],L[1]),all=0;
        if (strchr(arg[i],'z')) follow(r[site][2],L[2]),all=0;
        if (all) loop (k,0,3) follow(r[site][k],L[k]); }

      if (outfiles) if (out) fprintf(out,"\n");
      oo+=nsmol; }
    if (!outfiles) printf("\n"); }

  if (nfollow) 
    fprintf(stderr,"max jump between frames = %g L%s\n",
            maxd,maxd>0.4?" ********** TOO MUCH!!! **********":"");

  fclose(plb);

  return 0;
}
