/* make rdfg
   7/2012 extended to coord.n. and Kirkwood-Buff integrals
          unused options removed/simplified
   7/1999 bug fixed: coord.number was 1/2 of correct value for the same atoms
*/

#include "ground.h"
#define SDS
#include "alloc.h"
#include "varfile.h"

#ifdef Reverse
#define VarReverse VarFile.reverse
#else
int VarReverse; /* dummy */
#endif

#include "rdf.h"

#define NFILE 4

struct outtype_s {
  int yes;
  char *ext;
  char sys[10240];
  char *info;
  char *hdr;
  char *fmt;
} outtype[NFILE] = {
  {0,".g","plot","radial distribution function","g(r)","%8.4f"},
  {0,".hist","plot","histogram (per 1 measurement)","G(r)","%9.2f"}, 
  {0,".cn","plot","running (cumulative) coordination number","N(r)","%9.2f"},
  {0,".kb","plot","running Kirkwood-Buff integral","G(r)","%8.4f"},
};

int main(int narg, char **arg)
{
  rdf_t *rdf;
  double q,g,r,coord=-1,rmin=9e99,V;
  char fnmin[256];
  int ir,iarg,plot=0,bin=0;
  enum outtype_e { G,HIST,CN,KB } iout;
  char *rfmt="%6.3f";
  char site0[8],site1[8];
  double rho0=1,rho1=1;
  long sumhist;

  static char fn[256];
  char *dot;

  initscroll(0);

  if (narg<2) {
    fprintf(stderr,"\
Calculate and plot radial distribution functions, coordination numbers, and\n\
Kirkwood-Buff integrals from cook binary files SIMNAME.rdf. Call by:\n\
  rdfg [OPTIONs] SIMNAME[.rdf] [OPTIONs]\n\
OPTIONs:\n\
  -rFMT   format for r [default=%s]\n\
  -g[FMT] calculate RDFs SIMNAME.SITE1.SITE2.g [set output format, df.=%s]\n\
  -h[FMT] histogram SIMNAME.SITE1.SITE2.hist [%s]\n\
  -c[FMT] calculate running coordination numbers SIMNAME.SITE1.SITE2.cn [%s]\n\
  -k[FMT] calc. running Kirkwood-Buff integrals SIMNAME.SITE1.SITE2.kb [%s]\n\
          (binary: print hints to get the partial molar volumes)\n\
  -p      plot the generated files (command `plot' needed)\n\
  -R      reverse endian on input\n\
Example:\n\
  rdfg simul.rdf -r%%7.4f -g -p\n\
See also:\n\
  rdfgv1  # old version compatible with cook prior V2.7f\n\
  rdfconv # conversion from old to new format\n",
            rfmt,
            outtype[G].fmt,
            outtype[HIST].fmt,
            outtype[CN].fmt,
            outtype[KB].fmt);
  exit(1); }

  loop (iarg,1,narg)
    if (arg[iarg][0]=='-') switch (arg[iarg][1]) {
        case 'R':
#ifdef Reverse
          VarReverse=1;
#else
          ERROR(("option -r (reverse endian) but no compile-time Reverse"))
#endif
          break;
        case 'r':
          rfmt=arg[iarg]+1;
          break;
        case 'g':
          outtype[0].yes++;
          if (arg[iarg][2]) outtype[0].fmt=arg[iarg]+2;
          break;
        case 'h':
          outtype[1].yes++;
          if (arg[iarg][2]) outtype[1].fmt=arg[iarg]+2;
          break;
        case 'c':
          outtype[2].yes++;
          if (arg[iarg][2]) outtype[2].fmt=arg[iarg]+2;
          break;
        case 'k':
          outtype[3].yes++;
          if (arg[iarg][2]) outtype[3].fmt=arg[iarg]+2;
          break;
        case 'p':
          plot++;
          break;
        default: ERROR(("%s is unknown option",arg[iarg])) }
    else {
      if (fn[0]) ERROR(("two file names, %s and %s\n\
*** note that the syntax of rdfg changed: to calculate rdfs and plot them, use\n\
***   rdfg %s -g -p",fn,arg[iarg],fn)) 

      strcpy(fn,arg[iarg]); }

  if (!fn[0]) ERROR(("no RDF file name"))

  dot=strend(fn);
  if (strlen(fn)>4 && !strcmp(dot-4,".rdf")) dot-=4;

  strcpy(dot,".rdf");
  VarOpen(fn,"r");
  fprintf(stderr,"rdfg:");
  while (VarFile.size) {
    int4 vs=VarFile.size,i4;
    double d8;

    if (vs==4) {
      VarRead(&i4,4); ReverseOrder(&i4,4);
      fprintf(stderr," nsites=%d",i4); }
    if (vs==8) {
      VarRead(&d8,8); ReverseOrder(&d8,8);
      fprintf(stderr," grid=%g",d8); }

    if (vs>8) {
      rdf=VarGetSds();
      if (rdf->size<sizeof(rdf_t)-2*sizeof(unsigned4) || rdf->size>8000000)
	ERROR(("rdf->size=%d is suspicious - endian?",rdf->size))

      if (VarReverse) {
	int i;

	ReverseOrder(rdf->indx,sizeof(rdf->indx[0]));
	ReverseOrder(rdf->indx+1,sizeof(rdf->indx[0]));
	ReverseOrder(rdf->ns,sizeof(rdf->ns[0]));
	ReverseOrder(rdf->ns+1,sizeof(rdf->ns[0]));
	ReverseOrder(&rdf->npair,sizeof(rdf->npair));
	ReverseOrder(&rdf->grid,sizeof(rdf->grid));
	ReverseOrder(&rdf->V,sizeof(rdf->V));
	ReverseOrder(&rdf->nmeas,sizeof(rdf->nmeas));
	ReverseOrder(&rdf->nhist,sizeof(rdf->nhist));
	loop (i,0,rdf->nhist)
	  ReverseOrder(rdf->hist+i,sizeof(rdf->hist[0])); }

      if (rdf->nhist*sizeof(rdf->hist[0])+((char*)rdf->hist-(char*)rdf) != rdf->size)
        WARNING(("This file file was likely obtained by an old version of cook (< V2.7f)\n\
*** use  rdfconv  to convert it to the new version"))

      if (rdf->npair) {

	loop (ir,0,rdf->nhist) if (rdf->hist[ir]) break;
        r=(ir+0.5)/rdf->grid;
        /* 1st nonzero g(r) */
        if (r<rmin) {
          rmin=r;
          sprintf(fnmin,"%s(%d)-%s(%d) hist[%d]=%d",
                  rdf->name[0],rdf->ns[0],
                  rdf->name[1],rdf->ns[1],
                  ir,rdf->hist[ir]); }

        if (outtype[KB].yes) bin++;

        loop (iout,0,NFILE) if (outtype[iout].yes) {
          strcpy(dot,"."); strcat(fn,rdf->name[0]);
          strcat(fn,".");  strcat(fn,rdf->name[1]);
          strcat(fn,outtype[iout].ext);
          if (plot) {
            strcat(outtype[iout].sys," ");
            strcat(outtype[iout].sys,fn); }

          out=fopen(fn,"wt");

          V=rdf->V/rdf->nmeas;
          prt("### %s ###\n# %s:%s\n\
# %ld sites of %s[%ld]  %ld sites of %s[%ld]  %.0f pairs\n\
# %ld measurements  <V>=%f  grid=%ld\n\
# number densities [A^-3]: rho%s=%g  rho%s=%g",
              outtype[iout].info,
              rdf->name[0],rdf->name[1],
              (long)rdf->ns[0],rdf->name[0],(long)rdf->indx[0],
              (long)rdf->ns[1],rdf->name[1],(long)rdf->indx[1],
              rdf->npair,
              (long)rdf->nmeas,V,(long)rdf->grid,
              rdf->name[0],rdf->ns[0]/V,
              rdf->name[1],rdf->ns[1]/V);

          if (iout==KB && rdf->indx[0]!=rdf->indx[1]) {
            strcpy(site0,rdf->name[0]);
            strcpy(site1,rdf->name[1]);
            rho0=rdf->ns[0]/V;
            rho1=rdf->ns[1]/V; }

          prt("# r       %s     hist #",outtype[iout].hdr);

          q = rdf->V/(4*PI*rdf->npair*Sqr((double)rdf->nmeas)/Cub(rdf->grid));
          sumhist=0;

          loop (ir,0,rdf->nhist) {

            if (iout>=CN) {
              r=(ir+1)/rdf->grid;
              sumhist += rdf->hist[ir];
              g=(double)sumhist/rdf->nmeas;
              if (iout==CN) {
                /* coordination number */
                if (rdf->indx[0]==rdf->indx[1])
                  coord=g*=2./rdf->ns[0];
                else {
                  coord=g/rdf->ns[1];
                  g/=rdf->ns[0]; } }
              else {
                /* Kirkwood-Buff */
                g*=rdf->V/rdf->nmeas/rdf->ns[0]/rdf->ns[1];
                if (rdf->indx[0]==rdf->indx[1]) g*=2;
                g-=4*PI/3*Cub(r); } }
            else {
              /* RDF */
              r=(ir+0.5)/rdf->grid;
              if (iout==G)g = rdf->hist[ir]*q/(ir*(ir+1.0)+1.0/3);
              else g=rdf->hist[ir]/rdf->nmeas; }

            prt_(rfmt,r);
            prt_(" ",r);
            prt_(outtype[iout].fmt,g);
            if (iout==CN) prt_(outtype[iout].fmt,coord);
            _n }
          fclose(out); }

        free(rdf); } } }

  if (plot) loop (iout,0,NFILE) if (outtype[iout].yes) {
    strcat(outtype[iout].sys," &");
    fprintf(stderr,"\n%s\n",outtype[iout].sys);
    system(outtype[iout].sys);  }

  fprintf(stderr," minr=%g for %s\n",rmin,fnmin);

  ir=0;
  loop (iout,0,NFILE) ir+=outtype[iout].yes;
  if (!ir)
    fprintf(stderr,"no output file because none of -g -h -c -k specified (rdfg w/o args for help)\n");

  *dot=0;
  if (bin==3)
    printf("Hints to determine partial molar volumes of a binary mixture\n\
WARNING: tests passed with problems - bug in the r^3 term?\n\
export a=%.9g\n\
export b=%.9g\n\
mergetab \\\n\
  %s.%s.%s.kb:1:2 \\\n\
  %s.%s.%s.kb:1=1:2 \\\n\
  %s.%s.%s.kb:1=1:2 | \\\n\
  tabproc A '(1+b*(D-C))/(a+b+a*b*(B+D-2*C))' > %s.V%s\n\
mergetab \\\n\
  %s.%s.%s.kb:1:2 \\\n\
  %s.%s.%s.kb:1=1:2 \\\n\
  %s.%s.%s.kb:1=1:2 | \\\n\
  tabproc A '(1+a*(B-C))/(a+b+a*b*(B+D-2*C))' > %s.V%s\n\
plot %s.V%s:A:'B+a*A^3' %s.V%s\n",
           rho0,rho1,
           fn,site0,site0,
           fn,site0,site1,
           fn,site1,site1,
             fn,site0,
           fn,site0,site0,
           fn,site0,site1,
           fn,site1,site1,
             fn,site1,
           fn,site0,fn,site1);

  return 0;
}
