/* make showcp
 */

/* printing MACSIMUS convergence profiles, as produced by cook* etc.

use `make' to compile showcp - see makefile

IF the first record in the .cp file starts with CPmark as the 1st
number, then the 2nd number contains NCP either in the 1st or 4th byte
(to recognize endianess), the 4th or 1st beeing zero.  The remaining
(NCP-2) 4-byte fields in the record contain max 4 letter info; the 1st
two info's are assumed to be "Etot" and "Tkin".
The minimum NCP is 2.

Old format: 5 columns without a header
*/

#include "options.h"
#define INTFMT "%d"
#define x badoption
int optionlist[32] =
/* ` a b c d e f g h i j k  l m  n  o p q r s t  u v w x y z { | } ~   */
  {x,0,0,0,1,0,1,0,0,x,x,1,29,0,15,-1,0,x,0,0,0,-1,x,x,0,x,x,x,x,x,x,x};
/* @ A B C D E F G H I J K  L M  N  O P Q R S T  U V W X Y Z [ \ ] ^ _ */
#undef x

#include "ground.h"
#include <time.h>
#include <unistd.h> /* unsigned sleep(unsigned); */
#include "cpmark.h"
#include "pakcp.h"

#include "alloc.h"
#include "varfile.h"
#include "statics.h"

#include "linregr.h"

struct opto_s {
  struct opto_s *next;
  int n;
  int col[1]; /* [n] */} *head=NULL;

void addopto(char *opt)
{
  char *c;
  int n=1;
  struct opto_s *o;

  for (c=opt; *c; c++) n+=*c==',';

  alloc(o,sizeof(*o)+(n-1)*sizeof(o->col[0]));

  o->col[0]=atoi(opt)-1;
  if (o->col[0]<0) ERROR(("-o: syntax or nonpositive arg"))
                     n=1;
  for (c=opt; *c; c++) if (*c==',') {
    o->col[n]=atoi(c+1)-1;
    if (o->col[n]<0) ERROR(("-o: syntax or nonpositive arg"))
                       n++; }
  o->next=head;
  o->n=n;
  head=o;
}

FILE *CP;
int tr=0; /* PAR compatibility */

char *simname,*plotname,*dot;

int NCP=5;
int CPL=79;   /* chars per line -1 */
int LINES=22; /* lines per page for -# options -2 */

/* pseudograph: */
#define STEPS "_-^"
#define CHARLEV 3 /* 2 or 3 */

char defCPkey[5]=  {'E',   'T',    'U',   'i',     'P'};
char deftit[5][8]= {"Etot","Tkin", "Epot","Eintra","P   "};
char *CPkey;
char (*tit)[8];
int argname;

static double sumT,sumTT;
static double *sumr,*sumrT,*drdT;
double averT;

int block,from,to,ncp,n;

int check=0, file=0;

float *r,*minr,*maxr,*firstr,*lastr;
double *sum; /* 12/2002 */
int    i,c;
char   *l; /* [CPL+1] */

static int rewound=1, packed;
static int verbose;

#define MAXNCP 100
char cols[MAXNCP];
int ncols;
unsigned char *gr[MAXNCP]; /* pseudo-graph */
static int reverse;

int eread(void) /***************************************************** eread */
/* fread CP with endian change */
{
  int i,j;
  static int next=0;

  if (rewound) {
    next=0;
    reverse=option('r');
    r=getcpheader(CP,&NCP,&reverse,verbose);

    allocarrayzero(tit,NCP);
    alloc(CPkey,NCP);
    loop (i,0,2) {
      strcpy(tit[i],deftit[i]);
      CPkey[i]=defCPkey[i]; }
    loop (i,2,NCP) {
      memcpy(&tit[i],&r[i],4);
      CPkey[i]=tit[i][0];
      if (i<NCP) loop (j,0,i) if (CPkey[j]==CPkey[i]) {
        if (i<9) CPkey[i]='0'+i+1;
        else CPkey[i]=0; } }

    {
      char key='@';
      loop (i,0,NCP) if (!CPkey[i]) {
      advancekey: key++; loop (j,0,NCP) if (CPkey[j]==key) goto advancekey;
      CPkey[i]=key; }
    }

    if (!r) Error("no r");
    if (!minr) {
      allocarrayzero(minr,NCP);
      allocarrayzero(maxr,NCP);
      allocarrayzero(lastr,NCP);
      allocarrayzero(firstr,NCP);
      allocarrayzero(sum,NCP);
      allocarrayzero(sumr,NCP);
      allocarrayzero(sumrT,NCP);
      allocarrayzero(drdT,NCP); }
    rewound=0;
    return 1; }

  if (!packed) {
    int ret=fread(r,sizeof(float),NCP,CP)==NCP;
    if (reverse) loop (i,0,NCP) endian((char*)(r+i));
    return ret; }
  else { /* packed */
   again:
    if (!next) {
      if (!getpakcp(CP)) return 0;
      else next=1; }
    if (nextcprec(r))
      return 1;
    else {
      endgetpakcp(CP);
      next=0;
      goto again; } }

  /* no return here - cannot reach */
}

static void initgr(int n)
{
  int i,j;
  
  if (option('u')) {
    alloczero(gr[n],(CPL+1)*LINES*3);
    loop (j,0,CPL)
      loop (i,0,LINES) {
        gr[n][(i*(CPL+1)+j)*3+0]=226;
        gr[n][(i*(CPL+1)+j)*3+1]=160;
        gr[n][(i*(CPL+1)+j)*3+2]=128; } }
  else {
    alloczero(gr[n],(CPL+1)*LINES);
    loop (j,0,CPL)
      loop (i,0,LINES) gr[n][i*(CPL+1)+j]=' '; }
/* ?  option('b')=(CPL-1); */
}

static void addbraille(unsigned char *grn,int x,int y)
{
  unsigned char *G=grn+((y/4)*(CPL+1)+(x/2))*3;
  int x1=x&1;
  int y3=y&3;
  
  if (G-grn>=(CPL+1)*LINES*3 || G-grn<0) ERROR(("addbraille %d",G-grn))

  if (y3==0) G[1]|=1<<x1;
  else G[2]|=1<<((3-y3)+x1*3);
}

int Pipe=0;
void PIPE(void)
{
  if (Pipe) prtc('#');
}

int main(int narg, char **arg) /************************************** main */
{
  FILE *CPA=NULL;
  static FILE *xfile[MAXNCP];
  static int iscpa;
  int pp=0,ecol1,ecol2;
  char *plotcommand;
  double DT=1;

#if CHECKHEAP==2
  AllocRange=96; AllocTrace=1;
#endif

  plotcommand=getenv("PLOTCOMMAND");
  if (!plotcommand) plotcommand="plot";

  initscroll(0);
  Pipe=1;
  PIPE();
  Pipe=0;
  prt(" SHOWCP: view and analyze MACSIMUS convergence profiles");

  /*** option analysis ***/
  loop (i,1,narg)
    if (arg[i][0]=='-') {
      if (arg[i][1]==0)
        Pipe++;
      else if (arg[i][1]>='0' && arg[i][1]<='9') {
        int n=atoi(arg[i]+1)-1;

        if (n<0 || n>=MAXNCP) Error("option -# (column) out of range");
        cols[n]++;
        Max(pp,n) }
      else {
        if (arg[i][1]=='o' && arg[i][2]!='-' && arg[i][2]!='0')
          addopto(arg[i]+2);
        getoption(arg[i])
        if (arg[i][1]=='h') DT=atof(arg[i]+2); } }
    else if (!simname) {
      char *c;
      
      alloc(simname,strlen(arg[i])+StaNameLen+7);
      *simname++='*'; /* for .cov */
      strcpy(simname,arg[i]);
      for (c=simname; *c; c++) if (*c=='.') dot=c;
      
      if (dot) {
        int unpacked=!strcmp(dot,".cp");
        
        packed=!strcmp(dot,".cpz");
        if (packed||unpacked) {
          CP=fopen(simname, "rb");
          if (!CP) ERROR(("%s not found",simname)); }
        else {
          PIPE();
          prt("%s: unexpected extension, trying to append .cp or .cpz",
              simname); } }
              
      if (!CP) {
        dot=c;
        strcpy(dot,".cp");
        PIPE();
        prt("no extension, trying %s",simname);
        CP=fopen(simname, "rb");
        if (CP) {
          FILE *tst;
          strcpy(dot,".cpz");
          tst=fopen(simname, "rb");
          strcpy(dot,".cp");
          if (tst) {
            WARNING(("both %s and .cpz files present: %s will be used",
                     simname,simname))
              fclose(tst); } }
        else {
          strcpy(dot,".cpz");
          CP=fopen(simname, "rb");
          packed=1;
          PIPE();
          prt("not found, trying %s",simname);
          if (!CP) {
            ERROR(("no %s nor .cp file",simname))
            exit(0); } } } }
    else /* !simname */
      if (!argname) argname=i;

  if (option('p')) option('a')=option('p');
  option('p')=abs(option('p'));

  if (option('o')<0) addopto("2,6,7");

  if (!simname) { fprintf(stderr,"\
Usage:\n\
  showcp [OPTIONS] {SIMNAME | SIMNAME.cp | SIMNAME.cpz} [OPTIONS] [NAME ..]\n\
FILES:\n\
  SIMNAME.cp=convergence profile file\n\
  SIMNAME.cpz=packed convergence profile file\n\
  SIMNAME=try SIMNAME.cp, then SIMNAME.cpz\n\
COLUMNS: (if none of -# and NAME is given, all columns are selected)\n\
  -#   select column number # (1st=-1, can repeat)\n\
  NAME select column of name NAME (can repeat; always -1=Etot, -2=T)\n\
  -o#[,#...] with -p: merge given columns to One plot, can repeat\n\
  -o-  all plots separate [default = -o2,6,7 = merge T,Tin,Ttr]\n\
RANGE/LAG:\n\
  -f#  From record # [default=-f1=1st record]\n\
  -t#  To record # (incl.), [default=0=last]\n\
  -b#  Block (subaverage) size for showing and additional analysis\n\
       [default=0=auto-adjust to a power of 10, or screen width with -g]\n\
  -b   the same as -b1 = no blocking (showing may be slow)\n\
  -h#  timestep (noint*h from cook run), applies to -x,-c1,-c2 [1]\n\
  -l#  Lag for time autocorrelation analysis [default=29]\n\
  -n#  number of consecutive blockings by 2 (not for -c1,2) [default=12]\n\
ACTIONS [default=print pseudographs of selected columns]:\n\
  -a   write ASCII file SIMNAME.cpa, -a-1 add column Etot-Etot0\n\
  -a - as above with output=stdout\n\
  -c#  sum of [default=-c0=one line of summary unless -l0]:\n\
       1=write files SIMNAME.NAME.tcf with time correlation functions\n\
       2=write files SIMNAME.NAME.cov with covariances\n\
       4=detailed autocorrelation analysis (not just summary)\n\
       8=autocorrelation analysis of data blocked by option -b\n\
       16=autocorrelation analysis with removed linear T-dependence\n\
  -g   print pseudoGraphs [default if columns selected and no other action]\n\
  -m   print Merged y-x pseudograph (1 line = 1 block)\n\
  -p   Plot blocked selected columns with not-constant data (forces -a)\n\
  -x   write blocked x-y files SIMNAME.NAME.xy\n\
  -e   statistical analysis and regression of the first two selected columns\n\
OPTIONS:\n\
  -k#  pseudograph info line: 1=range [default], 2=1st/last; 3=both lines\n\
  -d#  Delay # s between plots (-p) [default = 1]\n\
  -r   Reverse endian\n\
  -s#  Scroll (buffer #kB, -s=-s30)\n\
  -u#  charset for pseudographs: -u: Braille UTF-8, -u-: ASCII [df.=locale]\n\
Example (sim.cp -> sim.cpa in blocks by 10, show Etot, temperatures, P):\n\
  showcp -p -b10 sim Etot T P\n\
See also:\n\
  cp2cp cppak autocorr\n");
  exit(1); }

  if (getenv("LINES")) LINES=atoi(getenv("LINES"))-2;
  if (getenv("COLUMNS")) CPL=atoi(getenv("COLUMNS"))-1;

  option('k')&=3;
  if (option('k')==3) LINES--;
  allocarrayzero(l,CPL+1);

  if (option('u')<0) {
    char *e=getenv("LOCALE");
    
    if (e) option('u')=!!strcmp("C",e);
    e=getenv("LC_ALL");
    if (e) option('u')=!!strcmp("C",e); 
    if (option('u')<0) option('u')=1; }
      
  if (!CP) Error("file not found");

  if (option('s')==1) option('s')=80;
  if (option('s')) initscroll(option('s')*1024);
  from=option('f');
  to=option('t');
  block=option('b');

  if (!eread()) Error("empty file or bad header");
  rewind(CP);

  if (argname)
    loop (i,argname,narg) if (arg[i][0]!='-') {
      int n;

      loop (n,0,NCP) if (!memcmp(arg[i],tit[n],min(strlen(arg[i]),4))) {
        cols[n]++; 
        Max(pp,n) 
        goto found; }
      ERROR(("field \"%s\" not found",arg[i]))
   found: ; }
                
  loop (i,0,NCP) if (cols[i]) {
    ecol2=ecol1,ecol1=i;    
    ncols++; }
  if (option('e')) if (ncols!=2) Error("-e requires exactly 2 columns");

  /* no action and any column selected forces pseudograph */
  if (!(option('p') || option('e') || option('T') || option('a') || option('c')))
    if (ncols) option('g')=1;

  /* no column selected -> all columns */
  if (!ncols) loop (i,0,NCP) cols[i]++;
  ncols=NCP;

  if (option('g')) loop (i,0,NCP) if (!gr[i] && cols[i]) initgr(i);

  PIPE(); 
  prt_("from=%d to=%d (0=eof) block=%d ",from,to,block);
  /* note: because of verbose=1, next output will be ncp= from eread() */

  if (option('l')) StaSet(DT,option('l'),2,option('n'));

  PIPE();

  rewound=1;

  ncp=n=0;
  verbose=1;

  /* determining min/max and non-blocked statistics */
  while (eread()==1) {
    if (r[0]>CPmark) {
      if (++ncp>=from) {
        if (to) if (ncp>to) break;
        LRAdd("Etot",1,n,r[0]);

        if (option('l'))
          loop (i,0,NCP) if (cols[i]) StaAdd(tit[i],r[i]);

        if (option('e')) {
          LRAdd(string("%s-%s",tit[ecol1],tit[ecol2]),
                1,
                r[ecol1],r[ecol2]);
          StaAdd(string("%s-%s",tit[ecol1],tit[ecol2]),
                 r[ecol1]-r[ecol2]); }

        sumT+=r[1]; sumTT+=Sqr(r[1]);
        loop (i,2,NCP) { sumr[i]+=r[i]; sumrT[i]+=r[i]*r[1]; }

        if (n++)
          loop (i,0,NCP) {
            Min(minr[i],r[i])
            Max(maxr[i],r[i]) 
            lastr[i]=r[i]; }
        else {
          copy(firstr,r,sizeof(float)*NCP);
          copy(minr,r,sizeof(float)*NCP);
          copy(maxr,r,sizeof(float)*NCP); } } }
    else {
      unsigned char t[25],*c;

      copy(t,r+1,24); t[24]=0;
      if (reverse) /* reversing strings back */
        for (c=t; c<t+24; c+=4) endian((char*)c);
      for (c=t; *c; c++) if (*c<' ' || *c>127) { 
        strcpy((char*)t,"-"); 
        break; }
      PIPE();
      prt("%7d: %s",n,t); } }

  prt("\n# %d data records read\n",ncp);

  *dot=0; /* no extension */

  if (option('l')) {
    if (!option('c'))
      /* one line of statistics */
      StaPrintAll(NULL);
    else {
      if (option('c')&1) StaPrintAll(simname); /* TCF */
      if (option('c')&2) StaPrintAll(simname-1); /* COV */
      if (option('c')&4) StaPrintAll(""); }
    StaFree(); }
  
  rewind(CP);
  rewound=1;

  if (block==0) {
    if (option('m')) block = (n+99)/100; /* cca 100 lines */
    else if (n<=1000) block=1;
    else if (n<=10000) block=10;
    else if (n<=100000) block=100;
    else if (n<=1000000) block=1000;
    else if (n<=10000000) block=10000;
    else if (n<=100000000) block=100000;
    else block=1000000; 
    prt("### block autoset: block=%d (%d %sdata points)",block,n/block,n==1?"":"blocked "); }
      
  if (option('g'))
    if (option('u')) block = (n+CPL*2-1)/(CPL*2);
    else block = (n+CPL-1)/CPL;

  plotname=simname;

  if (option('a')) {
    strcpy(dot,".cpa");
    CPA=Pipe?stdout:fopen(simname,"wt");
    if (!CPA) {
      CPA=fopen(plotname="/tmp/showcp.cpa","wt");
      if (!CPA) ERROR(("cannot write to %s nor /tmp/showcp.cpa",simname))
      else prt("WARNING: cannot create %s, using /tmp/showcp.cpa\n\
(do not run more instances of showcp)",simname); }
    iscpa++;
    fprintf(CPA,"# block="INTFMT"\n#",block);
    loop (i,0,NCP) fprintf(CPA,"   %d:%-8s",i+1,tit[i]);
    if (option('a')<0) fprintf(CPA," %s-%s0\n",tit[0],tit[0]);
    else fprintf(CPA,"\n"); }

  PIPE();
  put(averT=sumT/n)
    _n PIPE();
  prt("colm_ky_name__________min____________max________max-min___________d/dT_");
  loop (i,0,NCP) {
    PIPE();
    prt_("%2d %c  %c %4s %13.6g  %13.6g  %13.6g",
         i+1,i+'A',CPkey[i],
         tit[i],minr[i],maxr[i],maxr[i]-minr[i]);
    if (i>=2) {
      double den=sumTT*n-Sqr(sumT);
      if (den) drdT[i]=(sumrT[i]*n-sumr[i]*sumT) / den;
      else drdT[i]=0;
      prt_("  %13.6g",drdT[i]); }
    _n }

  if (block>1) {
    PIPE();
    prt("averaged in blocks by %i",block); }

  if (option('m')) putline('=',CPL);

  ncp=n=0;
  memset(sum,0,sizeof(sum[0])*NCP);
  verbose=0;

  if (option('x')) {
    int i;
    
    loop (i,0,NCP) if (cols[i]) {
      char *c=dot,*t=tit[i];
      
      *c++='.';
      while (*t>' ') *c++=*t++;
      strcpy(c,".xy");
      xfile[i]=fopen(simname,"wt");
      if (!xfile[i]) 
        ERROR(("cannot create %s",simname))
      else {
        PIPE();
        prt("writing %s",simname); } } }

  /* T-dependence, blocked statistics, graph */
  while (eread()==1) {
    if (r[0]>CPmark) {
      if (++ncp>=from) {
        if (to) if (ncp>to) break;
        if (option('l') && option('c')&16)
          loop (i,2,NCP) if (cols[i]) {
            char st[64];
                        
            strcpy(st,"linT-"); strcat(st,tit[i]);  
                                      
            StaAdd(tit[i],r[i]-drdT[i]*(r[1]-averT)); }
        loop (i,0,NCP) sum[i] += r[i];
        if ((++n)%block==0) {
          memset(l,' ',CPL); l[CPL]=0; l[0]='|'; l[CPL-1]='|';
          loop (i,0,NCP) if (cols[i]) {
            if (block!=1 && option('c')&8) {
              char st[64];
            
              strcpy(st,"blocked-"); strcat(st,tit[i]);
              if (option('l')) StaAdd(st,sum[i]/block); }

            c=floor( (sum[i]/block-minr[i])/(maxr[i]-minr[i]+1e-30)*CPL );
            if (c<0) c=0; if (c>=CPL) c=CPL-1;
            l[c] = l[c]==' ' || l[c]=='|'  ? CPkey[i] : 'X';
            if (gr[i]) 
              if (option('u')) {
                c=(sum[i]/block-minr[i])/(maxr[i]-minr[i]+1e-30)*(LINES*4);
                Max(c,0) Min(c,LINES*4-1)
                addbraille(gr[i],(n-1)/block,c); }
              else {
                c=(sum[i]/block-minr[i])/(maxr[i]-minr[i]+1e-30)*(LINES*CHARLEV);
                Max(c,0) Min(c,LINES*CHARLEV-1)
                gr[i][(CPL+1)*(c/CHARLEV)+(n-1)/block]=STEPS[c%CHARLEV]; } }

          if (CPA) {
            static int first=1;
            static double E0;
            if (first) { E0=sum[0]/block; first=0; }
            loop (i,0,NCP) fprintf(CPA,"%11.8g ",sum[i]/block);
            if (option('a')<0) fprintf(CPA,"%11.8g\n",sum[0]/block-E0);
            else fprintf(CPA,"\n"); }

          loop (i,0,NCP) if (xfile[i]) {
            int ier;

            if (DT!=0) ier=fprintf(xfile[i],"%.8g %.8g\n",ncp*DT,sum[i]/block);
            else ier=fprintf(xfile[i],INTFMT" %.8g\n",ncp,sum[i]/block);
            if (ier<=0) ERROR(("cannot write to xy file - disk full?")) }

            if (option('m')) { PIPE(); prts(l); }

          memset(sum,0,sizeof(sum[0])*NCP); } } } }

  fclose(CP);
  if (CPA) fclose(CPA);
  loop (i,0,NCP) if (xfile[i]) fclose(xfile[i]);

  if (option('m')) putline('=',CPL);

  if (option('c')&24) {
    _n PIPE();
    StaPrintAll(option('c')&4?"":NULL); }
  
  LRPrint(' ');

  if (iscpa) {
    PIPE();
    prt("column numbers:");
    loop (i,0,NCP) prt_(" %d=%s",i+1,tit[i]);
    _n

  /*.....  if (!Pipe) prt("\ngnuplot\np \'%s\' u 1\n\nplot %s:0:1",simname,simname); */
  }

  if (option('g')) putline('=',CPL);
  loop (i,0,NCP) if (gr[i]) {
    if (maxr[i]>minr[i]) {
      if (option('u')) for (n=LINES-1; n>=0; n--) prts((char*)(gr[i]+n*(CPL+1)*3));
      else for (n=LINES-1; n>=0; n--) prts((char*)(gr[i]+n*(CPL+1))); }
    else
      prt("<<<EMPTY RANGE>>>"); 
    if (option('k')&1) putline('=',
      CPL-prt_("=== %d:%s === [%g,%g] === dif=%g === by "INTFMT" ",
                   i+1,tit[i],minr[i],maxr[i],maxr[i]-minr[i],block)); 
    if (option('k')&2) putline('=',
      CPL-prt_("=== %d:%s === 1st->last=%g->%g === dif=%g by "INTFMT" ",
                   i+1,tit[i],firstr[i],lastr[i],lastr[i]-firstr[i],block)); }
                            
  if (option('s')) {
#if defined(DOS) || defined(__EMX__)
    _n
#endif
  prt("use $COMMAND ($? for help) or ; to end");
    getdata enddata }

  if (option('p')) {
    struct opto_s *o;

    loop (i,0,NCP) if (cols[i]) {
      looplist (o,head) {
        int j;
        loop (j,0,o->n) if (o->col[j]==i) goto colfound; }
      /* not plotted for empty range unless specified in -o */
      if (maxr[i]>minr[i]) addopto(string("%d",i+1));
    colfound:; }

    loop (i,0,NCP) if (cols[i]) looplist (o,head) if (o->col[0]==i) {
      char s[1024];
      int j;

      sprintf(s,"%s \"[:][%.8g:%.8g]\" %s",
        plotcommand,
        minr[i],maxr[i],
        plotname);

      loop (j,0,o->n) sprintf(s+strlen(s),":0:%d ",o->col[j]+1);

#if defined(DOS) || defined(__EMX__)
      prt("\n\n%s = %s",s,tit[i]);
      system(s);
#else /* UNIX: starts all plots in parallel */
      { static char genv[64],nenv[256];
      int sl=option('d');
        sprintf(genv,"PLOTGEOMETRY=720x300+%d+%d",8*(i+1),30*(i+1));

      putenv(genv);
      sprintf(nenv,"PLOTNAME=%s %d ",simname,i+1);
      *dot='.';
      loop (j,0,o->n) if (o->col[j]<NCP) {
        strcat(nenv," ");
        strcat(nenv,tit[o->col[j]]); }
      putenv(nenv);
      strcat(s,"&");
      system(s);
        /*.....fprintf(stderr,"%s\n",s);*/
      sleep(sl); }
#endif
      } }

  fclose(out);

  return 0;
}
