/* cc -O2 -o smoothpl smoothpl.c -lm
*/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

/* C-style loop (last not included) */
#define loop(_I,_From,_To) for (_I=_From;_I<_To;_I++)

/* Pascal etc. style loop (last included) */
#define loopto(_I,_From,_To) for (_I=_From;_I<=_To;_I++)

void endian(char *a) /********************************************** endian */
{
char x;
x=a[0]; a[0]=a[3]; a[3]=x;
x=a[1]; a[1]=a[2]; a[2]=x;
}

typedef float vector[3];

int main(int narg, char **arg)
{
vector **r,*auxr;
double *window,w;
vector v;
int ns,s,smooth,gauss=0,i,j,no=0;
FILE *f,*ff;
int rev=0;

if (narg<4) {
  fprintf(stderr,"call by:\n\
  smoothpl IN OUT WINDOW [GAUSS] [SITES]\n\
IN     input playback file\n\
OUT    output playback file\n\
WINDOW rectangular window [1/WINDOW,...,1/WINDOW]\n\
GAUSS  # of [0.5,0.5] windows to convolute\n\
SITES  # of sites\n\
  negative WINDOW means to reverse endian on input\n\
  if SITES is given, no header is assumed, otherwise 2 floats\n");
  exit(0); }

f=fopen(arg[1],"rb");
if (!f) {
  fprintf(stderr,"no plb-file\n");
  exit(1); }

ff=fopen(arg[2],"wb");

if (!strcmp(arg[1],arg[2])) {
  fprintf(stderr,"IN and OUT files identical\n");
  exit(-1); }

smooth=atoi(arg[3]);
if (smooth<0) { rev=1; smooth=-smooth; }
if (smooth<2) {
  fprintf(stderr,"smooth must be >=2\n");
  exit(1); }

if (narg>4) gauss=atoi(arg[4]);

window=malloc((smooth+gauss)*sizeof(window[0]));
loop (i,0,smooth) window[i]=1.0/smooth;
loop (j,0,gauss) {
  window[smooth]=0;
  for (i=smooth; i>0; i--)
    window[i]=(window[i]+window[i-1])/2;
  window[0]=window[0]/2;
  smooth++; }

r=malloc(smooth*sizeof(r[0]));

loop (i,0,smooth) fprintf(stderr,"%6.3f\n",window[i]);

if (narg>5)
  ns=atoi(arg[5]);
else {
  fread(v,sizeof(float),2,f);
  if (rev) endian((char*)v);
  fwrite(v,sizeof(float),2,ff);
  ns=v[0]; }

printf("number of sites=%d\n",ns);
if (ns<1) {
  fprintf(stderr,"bad number of sites\n");
  exit(-1); }

loop (s,0,smooth) {
  r[s]=malloc(ns*sizeof(vector));
  if (!r[s]) {
    fprintf(stderr,"no heap\n");
    exit(1); }
  if (fread(r[s],sizeof(vector),ns,f)!=ns) {
    fprintf(stderr,"unexpected eof: maybe too long smooth window\n");
    exit(1); }
  printf(".");
  if (rev) loop (i,0,ns) loop (j,0,3) endian((char*)&r[s][i][j]);
  }

printf("\n");
for (;;) {
  printf("."); no++;
  if (no%50==0) printf("\n");

  loop (i,0,ns) {
    loop (j,0,3) {
      w=0;
      loop (s,0,smooth) w+=window[s]*r[s][i][j];
      v[j]=w; }
    fwrite(v,sizeof(vector),1,ff); }

  auxr=r[0];
  loop (s,1,smooth) r[s-1]=r[s];
  r[smooth-1]=auxr;

  if (fread(auxr,sizeof(vector),ns,f)!=ns) break;
  if (rev) loop (i,0,ns) loop (j,0,3) endian((char*)&auxr[i][j]);
  }

printf("\n%d smoothed cfgs\n",no);
fclose(ff);
fclose(f);
return 0;
}
