/* make spectrum
 */
#include "ground.h"
#include "alloc.h"

typedef double real;
#include "fourier.c"

#include "solve2.h"

int n,win;

double window(double x)
{
  x=sin(x/2);
  return powi(x,win);
}

double eps=1e-7,gpass;
int npass;
double A,*ypass;

double f(double a,double s)
{
  int i;
  double e,p,ee=0,ye=0,eepps=0,yepps=0,eepss=0,yepss=0;
    
  loopto (i,0,npass) {
    p=i-a;
    e=exp(-Sqr(p*s));
    ee+=e*e;
    ye+=ypass[i]*e;
    eepps+=e*e*Sqr(p*p)*s;
    yepps+=ypass[i]*e*Sqr(p*p)*s;
    eepss+=e*e*p*Sqr(s);
    yepss+=ypass[i]*e*p*Sqr(s); }

  A=ye/ee;
  gpass=A*eepss-yepss;

  return A*eepps-yepps;
}

double g(double a,double s)
{
  return gpass;
}

double peek(double *y,int n)
{
  double a=n/2,s=0.5;
  int err;

  npass=n;
  ypass=y;

  err=Solve2(f,g,
	     &a,0,n,eps,
	     &s,0,2,eps,-30,-2);

  if (err) return 0;
  else return a;
}

int main(int narg,char **arg)
{
  complex *y;
  real *sp;
  FILE *fpeek=NULL;
  char *fn;
  int iarg,ic,i,k=0,col=0,omit=0,plot=0,ndata=0,missing=0;
  double dt=0,sum=0;

  if (narg<2) {
    fprintf(stderr,"\
Spectrum (Fourier transform of data with optional window). Call by:\n\
  spectrum [OPTIONS] [FILE|-]\n\
FILES:\n\
  FILE = input file of white-separated data, #- and !- lines are ignored\n\
  FILE.sp = output spectrum\n\
            * if -h : 2 column (frequency,|amplitude|^2)\n\
            * if no -h : 1 column (|amplitude|^2)\n\
  FILE.pk = peeks analyzed (-h required)\n\
  -: input=stdin, both outputs are concatenated to stdout\n\
OPTIONS:\n\
  -c# = column of input data [default=1]\n\
  -o# = omit # lines of data from start [default=0]\n\
  -n# = # of valid data lines on input [no default]\n\
        more data are ignored, less data are padded by the average\n\
  -k# = # of output frequences [default=n/2]\n\
  -h# = time resolution [default=1], also -d#\n\
        * to get output in 1/cm, use -h(timestep/ps*0.0299792458)\n\
        * kin.energy has doubled frequences => use -h(2*timestep)\n\
  -p  = plot the output spectrum (only if FILE given)\n\
  -w# = multiply input data by bell-like window [sin(PI*i/NDATA)]^#\n\
        (large # give a Gaussian, 2 may be a good compromise)\n\
        WARNING: with -w, only a few data may be padded\n\
METHODS:\n\
  - a variant of the fast Fourier transform with arbitrary factoring is used\n\
  - may be slow if the number of data is not a product of small primes\n\
    (use -n bigger than the number of data to padd the data)\n\
  - the peek searcher uses a parabola of 4 points closest to a maximum\n\
Examples:\n\
  tab 1 1024 | tabproc \"abs(cos(A))\" | spectrum -n1024 - | plot -:0:1\n\
  showcp -a -b1 H2O; spectrum -n20000 -k5000 -h0.002 -c2 -p H2O.cpa\n\
");
  exit(0); }

  initscroll(0);

  loop (iarg,1,narg)
    if (arg[iarg][0]=='-') {
      switch (arg[iarg][1]) {
        case 0: fn=NULL; break;
        case 'c': col=atoi(arg[iarg]+2)-1; break;
        case 'o': omit=atoi(arg[iarg]+2); break;
        case 'n': n=atoi(arg[iarg]+2); break;
        case 'k': k=atoi(arg[iarg]+2); break;
        case 'w': win=atoi(arg[iarg]+2); break;
        case 'h':
        case 'd': dt=atof(arg[iarg]+2); break;
        case 'p': plot++; break;
        default: ERROR(("%s is wrong argument",arg[iarg])); } }
    else
      fn=arg[iarg];

  if (fn) {
    in=fopen(fn,"rt");
    if (!in) ERROR(("%s not found",fn))
    out=fopen(string("%s.sp",fn),"wt");
    if (dt) fpeek=fopen(string("%s.pk",fn),"wt"); }
  else {
    in=stdin;
    out=fpeek=stdout; }

  if (n<=0) ERROR(("no or bad # of data (option -n)"))
  if (!k) k=n/2;

  alloczero(y,n*sizeof(y[0]));

  loop (i,-omit,n) {
    double x=2*PI*(i+0.5)/n,yy;
    char line[1024],*tok;

    do if (!fgets(line,1024,in)) {
      if (!missing) {
        fprintf(stderr,"only %d data lines on input (%d expected)\n\
WARNING: missing data will be padded by the average %.8g\n",i+omit,n+omit,sum/ndata);
        if (win) fprintf(stderr,"ERROR: -w specified => the window may not be properly centered\n\
(only a few padded data are acceptable without a loss of precision)\n"); }        
      missing++; }
    while (strchr("!#",line[0]));

    if (i>=0) {
      if (missing) 
        y[i].re=sum/ndata;
      else {
        tok=strtok(line," \t\n\r");

        loop (ic,0,col) {
          if (!tok) ERROR(("line %d not enough columns",i))
	  tok=strtok(NULL," \t\n\r"); }

        if (!tok) ERROR(("line %d not enough columns",i))

        yy=atof(tok);
        sum+=yy;
        ndata++;
        y[i].re=yy; }
      if (win) y[i].re*=window(x); } }
 
  Fourier(y,n);
 
  sp=(real*)y;
 
  loop (i,0,k) {
    if (dt) prt_("%g ",i/(dt*n));
    sp[i]=(i>=2)*sqrt(Sqr(y[i].re)+Sqr(y[i].im));
    prt("%g",sp[i]); }

  if (fn) {
    fclose(out);
    out=fpeek; }  

  if (dt)
    loop (i,1,k-5)
      if (sp[i]<sp[i+1] && sp[i+1]<sp[i+2] && sp[i+2]>sp[i+3] && sp[i+3]>sp[i+4]) {
	real p=peek(sp+i,4);
	if (p) prt("\n%g %g\n%g %g",(i+p)/(dt*n),0.0,(i+p)/(dt*n),A); }

  if (out) fclose(out);
  
  if (plot) 
    if (dt) system(string("plot %s.sp:1:2 %s.pk",fn,fn));
    else system(string("plot %s.sp:0:1",fn));
 
  return 0;
}
