/* \make staprt

6/2009 -N changed
10/2008 bug fixed: .stx not recognized 
*/

#include "ground.h"
#include "alloc.h"
#include "statics.h"

static char how[2048];
int isname=0,iarg,files=0,tcf=0,iiarg,prec=0;

void howcpy(char *fn)
{
  char *p=NULL,*c;

  if (strlen(fn)>1024) ERROR(("too long filename"))
  if (strlen(fn)<5) ERROR(("%s: too short filename",fn))

  strcpy(how,fn);
  for (c=how; *c; c++) if (*c=='.') p=c;
  if (!p) ERROR(("%s: no extension",fn))
  if (strcmp(p,".sta")) WARNING(("%s: not standard .sta extension",fn))
  *p=0;
}

void prefix(char p)
{
  memmove(how+1,how,strlen(how)+1);
  how[0]=p;
}

char *addext(char *fn)
{ 
  char *c,*dot=NULL;
  
  for (c=fn; *c; c++) if (*c=='.') dot=c;

  if (dot && (!strcmp(dot,".sta") || !strcmp(dot,".stx") || !strcmp(dot,".sta~"))) 
    c=fn;
  else {
    alloc(c,strlen(fn)+5);
    sprintf(c,"%s.sta",fn); }
  return c;
}

void process(char *fn)
{
  char *c=addext(fn);
  
  StaLoad(c);
  
  printf("====== %s ======\n",fn);
  how[0]=0;
  if (files) howcpy(c);
  if (tcf && !files) prefix('-'); /* SHOULD NEVER HAPPEN */
  if (tcf==2) prefix('*');
  if (prec && !files && !tcf) how[0]='+';
}

char *key=NULL;
char *fmt=NULL;
char *ifmt=NULL;
double Qexp=-3e33;
double Q=1;

void one(char *fn,char *name)
{
  char *k,*c=addext(fn);
  double Q0;
  
  if (!fmt) fmt = prec ? " %18.12g" : " %9g";
  if (!ifmt) ifmt=" %7d";

  StaLoad(c);

  Q0=0;
  for (c=fn; *c; c++) if (isdigit(*c) || *c=='.' && isdigit(c[1])) {
    Q0=atof(c);
    break; }
  if (Qexp>-2e33) Q=pow(Q0,Qexp);

  for (k=key; *k; k++)
    switch (*k) {
      case 'a': printf(fmt,Q*StaMean(name)); break;
      case 'e': printf(fmt,Q*StaStdErr(name)); break;
      case 'r': printf(fmt,StaStdErr(name)/StaMean(name)); break;
      case 'd': printf(fmt,Q*sqrt(StaVar(name))); break;
      case 'v': printf(fmt,Q*StaVar(name)); break;
      case 'n': printf(ifmt,StaN(name)); break;
      case 'N': printf(" %s",name); break;
      case 'q': printf(fmt,Q0); break;
      case 'Q': printf(fmt,Q); break;
      case 'f': printf(" %s",fn); break;
      case 't': putchar('\t'); break;
      default: putchar(*k); }

  putchar('\n');
}

int main(int narg,char **arg)
{

  initscroll(0);

  if (narg<2) {
    fprintf(stderr,"Prints the contents of .sta file(s).\n\
  staprt [OPTIONS] FILE.sta [FILE.sta ...]\n\
OPTIONS:\n\
  -nNAME print statistics for given name (this option may repeat)\n\
  -NNAME the same as -nNAME -kN:aef\n\
         NAME may contain wildcard * at end, but unique match is required\n\
  -k     (w/o  -n) only 1 line summary statistics for each recorded variable\n\
  -kKEY  (with -n) only 1 line of statistics, KEY = string of:\n\
         a=average\n\
         e=standard error of the average\n\
         r=relative standard error of the average\n\
         d=standard deviation of the data = sqrt(variance)\n\
         v=variance of the data\n\
         n=number of data points\n\
         q=the value of Q\n\
         Q=the 1st number found in the filename\n\
         t=tabulator\n\
         f=file name\n\
         N=NAME (value of option -n or -N)\n\
         other=print as is\n\
  -FFMT  number format for line output (KEY=aed) [default=-F\" %%9g\"]\n\
  -IFMT  number format for line output (KEY=n) [default=-F\" %%7d\"]\n\
  -qQ    set scaling factor Q for line output (KEY=aedv) [default=1]\n\
  -Q#    Q = the 1st number found in the filename (e.g., N32 -> 32, T.2-> 0.2)\n\
         to the power of # (e.g., use -1 to divide)\n\
  -*Q -/Q --Q -+Q -^Q modify the scaling factor (not the -q@ version)\n\
  -t0    print standard statistics [default]\n\
  -t1    print time correlation functions (to FILE.NAME.tcf)\n\
  -t2    print time covariances (to FILENAME.cov)\n\
  -a     more dec.digits on output for tables (use -F for line output)\n\
FILE:\n\
  if NAME is given, extension may be omitted; then .sta is assumed\n\
Examples:\n\
  staprt -N\"P [Pa]\" water256.sta water512.sta\n\
  staprt -n\"P*\" -kaenf -q1e-6 -F\" %%7.2f\" *.sta\n\
See also: autocorr showcp\n\
");
    exit(0); }

  loop (iarg,1,narg) if (arg[iarg][0]=='-') switch (arg[iarg][1]) {
    case 'F': fmt=arg[iarg]+2; break;
    case 'I': ifmt=arg[iarg]+2; break;
    case 'N': key="N:aef";
    case 'n': isname++; break;
    case 'k': key=arg[iarg]+2; break;
    case 't': tcf=atoi(arg[iarg]+2); if (!tcf) tcf++; files++; break;
    case 'q': Q=atof(arg[iarg]+2); break;
    case 'Q': Qexp=atof(arg[iarg]+2); break;
    case '*': Q*=atof(arg[iarg]+2); break;
    case '/': Q/=atof(arg[iarg]+2); break;
    case '+': Q+=atof(arg[iarg]+2); break;
    case '-': Q-=atof(arg[iarg]+2); break;
    case '^': Q=pow(Q,atof(arg[iarg]+2)); break;
    case 'a': prec++; break;
    default: ERROR(("%s is bad option",arg[iarg])); }

  if (isname) {
    loop (iarg,1,narg) if (arg[iarg][0]=='-' && tolower(arg[iarg][1])=='n')
      loop (iiarg,1,narg) if (arg[iiarg][0]!='-') {
	if (key) 
	  one(arg[iiarg],arg[iarg]+2);
	else {
	  process(arg[iiarg]);
	  StaPrint(arg[iarg]+2,how[0]?how:""); }
	StaFree(); } }
  else 
    loop (iiarg,1,narg) if (arg[iiarg][0]!='-') {
      process(arg[iiarg]);
      if (key) StaPrintAll(NULL);
      else StaPrintAll(how[0]?how:"");
      StaFree(); }

  return 0;
}
