/* make stereo
 */

#define COLOR 2
/*
  COLOR = 0: black-and-white, each pixel 2 shade levels (0,1),
             random or text-file-defined pattern
             output tailored to HP LaserJet
             also PostScript output
  COLOR = 1: color, MCY system, each color 2 shade levels,
             random pattern
             output tailored to HP color inkjet printer 75 100 150 300 dpi
             also ppm output
  COLOR = 2: color, RGB system, each color 256 shade levels
             pattern from a ppm file
             output to a ppm file
*/

/*
                      =============================
                       STEREO (c) J. Kolafa 1994-5
                        updated (ppm format) 1998
                      =============================
DESCRIPTION
^^^^^^^^^^^
This program calculates stereograms normally with the eye cross point behind
the paper, i.e., your eyes should watch an imaginary point lying behind the
paper in approximately the same distance as the eye-paper distance (this is
for the recommended default of the x-pattern size one half of the eye's
distance).  Different people may prefer different distances or even
stereograms with the cross point above the paper.

METHOD
^^^^^^
The programs works with rows of pixels in the x-direction.  It starts from a
periodically repeating `pattern' from left (the periodicity is given by the
condition that both eyes must see the same pattern) and maps the images of
pixels on the paper with their positions on the 3D scene.  The following
figure shows how one element of the pattern denoted by A repeats both on the
paper and in the scene.

            (1)    (2)
**A*****A*****      ******A****
             *      *              scene
             A**A**A*
               ..
  -A--A--A--A-A-A-A--A--A--        paper
             .  .
            .   .
           .    .
          O     O                  eyes

It is seen that the pattern shortens if the scene is closer to the eyes; if we
go from left then a piece of the pattern disappears at point (1) and another
piece of pattern must appear at point (2).  The latter case is a source of
technical problems how to create such a pattern (see  rightedge  and
rightedgep  in input data).

VERSIONS
^^^^^^^^
If COLOR is #defined, then the stereograms are in colors in the
Yellow-Magenta-Cyan system, otherwise black-and-white.

INPUT SCENE
^^^^^^^^^^^
The input scene must be given in the form of z-buffer as a file of bytes
(=unsigned chars) scaled to the range <0,256), stored by rows.  The number of
bytes per row is  xzbuf  (default=320), the number of rows is yzbuf
(default=200).  The first row is the top row if mirror==0 (default),
otherwise the bottom row.  The value of the z-buffer=0 means the background
plane, the values grow towards the eyes (note that many ray-tracing programs
use the opposite convention - use invert=1 in this case).

INPUT PATTERN
^^^^^^^^^^^^^
(black-and-white version only)
The pattern may be generated randomly or given by a file or combined.  The
pattern file is a text file with (at least)  ypat  lines, (at least)
xpat  columns each.  Each character corresponds to one pixel.
  SPACE = white
  `.' = 7.7% gray (i.e. a larger area filled with `.' will contain dithered
        pixels looking as light gray),
  `:' = 25% gray
  anything else = black (100%)

INPUT DATA
^^^^^^^^^^
The data are in format `identifier=expression', ended by `;'.  Single `?' on
input line will print a short help, '$?' will print a short help for
scrolling.
See the default file "stereo.def" for a list of input variables.

COLORS
^^^^^^
File "stereo.col" must contain definition of colors as  ncol  lines of
Yellow Magenta Cyan Probability
where all four data are integers.  The first line is skipped.

OUTPUT
^^^^^^
If hp=1 is selected, file "hp" is created in the HP Laser Jet (CPL) format.
If hex=1 is selected, file "hex" with hexadecimal dump is created (white=1,
black=0; to be included into PostScript); this is not recommended because
STEREO provides its own dithering of gray areas.

MAKING
^^^^^^
EITHER edit makefile (write the name of the compiler as CC=..., the name of
directory containing stereo.c as D2=..., the name of directory contaning
rndgen, ground, and alloc as D1=..., and run make)
OR run something like this (assuming all sources are in the working
directory):
  cc -O2 -DSCR -o stereo stereo.c rndgen.c alloc.c ground.c -lm
(optimizing option(s) recommended, -DSCR enables interactive scrolling)

With Turbo C 2.0, use project stereo.prj (edit directories first!), use the
COMPACT model and #define from menu DOS and optionally SCR
Similarly for other DOS compilers for 16 bit code and data.

BUGS
^^^^
Pattern making algorithm is inefficient, especially in the color version.
Does not communicate with standard graphic formats.
The z-buffer format is stupid (because compatible with DOS PROSIS playback
program `show').
The order of colors YMC is wrong throught the program.

Sample file stereo.def
^^^^^^^^^^^^^^^^^^^^^^
! default file for STEREO
! all these values may be changed interactively
! remember to end the data set by a semicolon

resolution=72       ! output resolution in dpi (for hp: 75,150,300,600)
hex=0               ! to generate hex pixel map (can be pasted to PS)
hp=0                ! to generate HP LaserJet printfile
ppm=2               ! to generate ppm-file
                    ! ppm=1: pixel-pixel
                    ! ppm>1: square ppm x ppm --> 1 output pixel
                    !        (antialiasing -- avoids `layers' caused
                    !        by big pixels)
                    !        all sizes must be scaled accordingly!
                    !        recommended ppm=2 or better 3
                    ! Hint: the final stereogram can be safely converted 
                    ! to JPEG with quite a low quality (<50%)
!xpat=3.5/2.54*resolution*ppm
xpat=180            ! according to the pattern ppm file
ypat=xpat           ! pattern size in output pixels
                    ! recommended: xpat=<a bit more than eye distance>/2
                    ! (use ###/2.54*resolution to convert ### cm into pixels)
pattern=5           ! sum of: 1=random patches, 2=from file, file 4=waves
                    ! negative: flip pattern x:y (ppm only)
                    ! if from file, xpat and ypat must match
! BW only:
!p[0]=5 p[1]=2       ! random patches: initial prob of white, black,
!p[2]=3 p[3]=4       ! 1/4 gray, 1/13 gray (black-and-white version only)
mc=30               ! MC steps of pattern patch growing, for pattern&1
                    ! (there are different algorithms for different COLOR)
line= 3             ! line width (for waves or pattern file) in output pixels
                    ! (for pattern&4)
                    ! COLOR<1: line<0: pattern from file will be subtracted
wave=6              ! # of waved lines in the pattern (for pattern&4)
maxk=8              ! # of random cos terms in waves (for pattern&4)
amplitude=1         ! relative amplitude of waves  (for pattern&4)
up=1                ! floating small objects (otherwise towers)
rightedge=12        ! right edge limit to take another pattern, in output pix
rightedgep=0        ! probability of adding rand points to right edge pattern
phase=0             ! correct a bit phase shift caused by the method starting
                    ! from the left (not very good)
ypaper=480*ppm      ! picture height in output pixels (xpaper will be
                    ! calculated from z-buffer size and aspect ratio))
                    ! should be a multiple of 16
xzbuf=640 yzbuf=480 ! z-buffer sizes (in bytes) yzbuf will be scaled to ypaper
                    ! recommended if ypaper is an integer multiple of yzbuf
                    ! (use the window geometry info when runing show)
textzbuf=0          ! 0:zbuf file in binary format (xzbuf*yzbuf bytes)
                    ! 1:zbuf file in text format (xzbuf  bytes read from each
                    ! line ended by LF)
zbufd=3             ! z-buffer scaling (recommended integer): z-buffer data
                    ! (which are in principle in the units of output pixels)
                    ! will be DIVIDED by zbufd.
                    ! If the resulting 3D image is too DEEP then  zbufd  should
                    ! be INCREASED and vice versa
aspect=1            ! aspect ratio of the z-buffer (y/x pixel size)
                    ! aspect=1.2 for the standard VGA screen
mirror=0            ! 0: 1st line in the z-buffer file is TOP; 1:BOTTOM
invert=0            ! 1: z-buffer inverted (zbuffer:=255-zbuffer)
                    ! this will generate a cross-eye stereogram
xmargin=60          ! x-margin (landscape) for HP in pixels
ymargin=xmargin     ! y-margin (landscape) for HP in pixels
rndseed=0           ! seed for random numbers; 0:use time

! COLOR==1 version only:
!ncol=6            ! # of colors (=lines in "stereo.col")
!wcol=3*36         ! color of waves in the form Y+6*M+36*C (?? order)
palette=0         ! 1: palette test printed (no stereogram)

; !end of data

Sample file stereo.col
^^^^^^^^^^^^^^^^^^^^^^
Y M C prob
0 0 2 1
2 2 0 1
5 3 0 1
4 0 4 1
5 0 0 1
3 3 3 1
0 4 0 1
1 0 4 1 */

#ifndef COLOR
#error #define COLOR required
#endif
#if COLOR<0 || COLOR>2
#error bad #define COLOR
#endif

#include "ground.h"
#include "alloc.h"
#include "rndgen.h"

#define inc(X) X++ /* translated from Pascal */

int xpat=383, ypat=171; /* pattern size in printer pixels */

typedef unsigned char byte;
typedef byte rgb[3];

#if COLOR<2
byte **pat;  /* [xpat][ypat] */
#else
rgb **pat;   /* [xpat][ypat] */
#endif

int paper=4;         /* A4 paper */
int xpaper=124*24;   /* = ldraw, to be recalculated */
int ypaper=120*16;   /* = ldraw */
float aspect=1.2;
int up=1,hex=0,hp=1,ppm=0;

char fn[128];
float rightedgep=0;
unsigned mc=20;
double amplitude=1;

#if COLOR==0

#define NPAT 4
static int p[NPAT]={4,2,3,4};
byte **sheet;
#define lputpix(X,Y) sheet[X][Y/8] |= bit[Y&7]
static byte bit[8]={128,64,32,16,8,4,2,1};

#elif COLOR==1

#define NPAT 216
static int p[NPAT];
typedef byte cym[3];
byte **sheet[3];
cym **cympat;
#define lputpix(C,X,Y) sheet[C][X][Y/8] |= bit[Y&7]
static byte bit[8]={128,64,32,16,8,4,2,1};

#elif COLOR==2

rgb **sheet;
#define lputpix(X,Y,C) sheet[X][Y][0]=C[0],sheet[X][Y][1]=C[1],sheet[X][Y][2]=C[2]

#endif



void rndpat(void) /* =============================================== rndpat */
/*
   random pattern
   MC algorithm looking for a ground state is applied to a sort of
   2D Potts model, starting from a random (hot) configuration
   COLOR==2: just random pixels (not good)
*/
{
#if COLOR==2
int i,j,k;
long kl,mcl=(long)mc*xpat*ypat; /* because of DOS */
rgb p;

loop (i,0,xpat) loop (j,0,ypat) loop (k,0,3) pat[i][j][k]=irnd(256);
fprintf(stderr,"random pattern ...");
loop (kl,0,mcl) {
  i=irnd(xpat);
  j=irnd(xpat);
  copy(p,pat[i][j],3);
  switch (irnd(4)) {
    case 0: if (++i>=xpat) i=0; break;
    case 1: if (--i<0) i=xpat-1; break;
    case 2: if (++j>=ypat) j=0; break;
    case 3: if (--j<0) j=ypat-1; break; }
  copy(pat[i][j],p,3); }
fprintf(stderr," done\n");
#else
int i,j,i0,i1,j0,j1,l,ll,k;
long kl,mcl; /* because of DOS */
int nnbr[NPAT];
int s[NPAT];
#if COLOR==0
char *info[NPAT]={"white","black","1/4 ","1/13"};
#endif

ll=0;
loop (i,0,NPAT) {
  ll=ll+p[i];
  s[i]=ll; }

/* initial random configuration */
loop (i,0,xpat) loop (j,0,ypat) {
  l=irnd(ll);
  loop (i0,0,NPAT) if (l<s[i0]) { pat[i][j]=i0; goto ex; }
  ex:; }

mcl=(long)xpat*ypat;

/* MC energy minimization leading to domains of the same spins */
loop (k,0,mc) {
  fprintf(stderr,"generating random pattern %d:%d\r",k,mc);
  loop (kl,0,mcl) {
    i=irnd(xpat);
    i0=i-1; if (i0<0) i0=i0+xpat;
    i1=i+1; if (i1>=xpat) i1=i1-xpat;
    j=irnd(ypat);
    j0=j-1; if (j0<0) j0=j0+ypat;
    j1=j+1; if (j1>=ypat) j1=j1-ypat;
    memset(nnbr,0,sizeof(nnbr));
    inc(nnbr[pat[i][j0]]);
    inc(nnbr[pat[i][j1]]);
    inc(nnbr[pat[i0][j]]);
    inc(nnbr[pat[i1][j]]);
    j0=-1; j1=-1;
    loop (l,0,NPAT) if (nnbr[l]>=j0) {
      j1=j0; ll=l;
      j0=nnbr[l]; }
    if (j0>j1) pat[i][j]=ll;
    else pat[i][j]=irnd(NPAT);
    }
  }

prt("pattern generated       ");
memset(nnbr,0,sizeof(nnbr));
loop (i,0,xpat) loop (j,0,ypat) inc(nnbr[pat[i][j]]);

#if COLOR==1
{
int less=0;
loop (i,0,NPAT) if (nnbr[i]>mcl/(216*3))
  if (nnbr[i]>mcl/200)
    prt("frac of C%d Y%d M%d = %6.3f",i/36,(i/6)%6,i%6,nnbr[i]/(double)mcl);
  else less++;
prt("%d with fraction < 0.005",less);
}
#else
loop (i,0,NPAT) prt("frac of %s = %6.3f",info[i],nnbr[i]/(double)mcl);
#endif
#endif
}

int line=2;

void filepat(int narg,int rot) /* ================================== filepat */
/*
   reading pattern from file,
     superposing with previous (random) pattern (COLOR<2)
   COLOR==2: reading PPM pattern file, rot=1: rotate 90deg
   narg>1: pattern file=pattern.ppm|pattern.asc, otherwise asks for
*/
{
FILE *f;
int i,j;
int c;
#if COLOR<2
int mi,mj,m=abs(line);
byte pt;
#endif

again:
if (narg<2) {
  prt_("pattern file = ");
  scanf("%s",fn); }
else strcpy(fn,COLOR==2?"pattern.ppm":"pattern.asc");
if (!strcmp(fn,".")) return;
f=fopen(fn,COLOR==2?"rb":"rt");
if (f==NULL) {
  ERROR(("cannot open %s",fn))
  goto again; }

#if COLOR==2
/* ppm pattern file */
{
int maxxn,maxyn,depth,jj,k;
char Line[64];
rgb *row;

fgets(Line,64,f);
fgets(Line,64,f);
sscanf(Line,"%d%d",&maxxn,&maxyn);
if (xpat!=maxxn || ypat!=maxyn)
  prt("WARNING: the size of pattern file %s, %dx%d,\n\
         does not match xpat=%d,ypat=%d. Rescaling will occur.",
      fn,maxxn,maxyn,xpat,ypat);
fgets(Line,64,f);
sscanf(Line,"%d",&depth);
if (depth!=255) ERROR(("%d bad depth",depth))

alloczero(row,sizeof(row[0])*maxxn);

loop (j,0,maxyn) {

  loop (i,0,maxxn) {
    row[i][0]=getc(f);
    row[i][1]=getc(f);
    c=getc(f); if (c<0) ERROR(("unexpected EOF on %s",fn))
    row[i][2]=c; }

/*
  loop (i,0,maxxn) {
    row[i][0]=i;
    row[i][1]=j;
    row[i][2]=100; }
*/
  
  if (rot)
    loop (jj,j*xpat/maxyn,(j+1)*xpat/maxyn)
      loop (i,0,ypat) loop (k,0,3)
        pat[jj][i][k]=row[i*maxxn/ypat][k];
  else
    loop (jj,j*ypat/maxyn,(j+1)*ypat/maxyn)
      loop (i,0,xpat) loop (k,0,3)
         pat[i][jj][k]=row[i*maxxn/xpat][k];
  }
free(row);
}

#else
#if COLOR==1
ERROR(("pattern file for COLOR==1 not supported"))
#endif
/* BW patern file: special text format */
j=ypat; while (j--) {
  loop (i,0,xpat) {
    c=getc(f);
    switch(c) {
      case ' ': pt=0; /* blank */ break;
      case '.': pt=3; break;
      case ':': pt=2; break;
      default: pt=1; /* black */
      }
    if (line>0) {
      if (pt) loop (mi,0,m) loop (mj,0,m)
        pat[(i+mi) % xpat][(j+mj) % ypat]=pt; }
    else if (pt==1)
      loop (mi,0,m) loop (mj,0,m)
        pat[(i+mi) % xpat][(j+mj) % ypat]=0;
    }
  do {
    c=getc(f);
    if (c<0) ERROR(("%s: unexpected EOF",fn))
    } while (c!='\n');
  }
#endif
}

int resolution=300;
int xmargin=60,ymargin=60;
int phase,rightedge=6;
int *x,*y; /* array[-xpat..xpaper-1] */

void row(int aty, byte *z) /* ========================================= row */
/*
   basic stereo algorithm for one row of pixels
*/
{
int i,j,k,xshift,yshift,zlast,jx,jj;

if (aty%100==0) fprintf(stderr,"calculating stereogram %d:%d\r",aty,ypaper);

xshift=0;
if (phase) loop (i,0,xpaper) if (z[i]>xshift) xshift=z[i];
xshift=xpat*3-xshift; /* shift,phase at the max hill onset */
yshift=aty % ypat;
loop (i,-xpat,0) {
  x[i]=(i+xshift) % xpat;
  y[i]=yshift; }

zlast=0;
i=0;
do {

  if (z[i]<zlast) {
    loop (k,i,xpaper)
      if (z[k]+k-i>=zlast) goto kfound;
    k=xpaper;
    kfound: k--;
    if (up) {
      jx=max(0,2*i-k);
      for (j=i-1; j>=jx; j--) if (z[j]<zlast-(i-j)) { k=i; break; } }
    if (k-i>rightedge) {
      /* now the pattern in <i,k> will be seen by the right eye only
      => we generate it by shifting the normal pattern in y-dir */
      xshift=xpat-z[i];
      yshift=(aty + ypat/6 + 2*ypat*i/(3*xpaper)) % ypat;
      loopto (j,i,k) {
        if (rnd()<rightedgep) x[j]=irnd(xpat);
        else x[j]=x[j-xshift];
        y[j]=yshift; }
      i=k;
      goto edged; }
    }

  k=i-xpat+z[i];
  if (k<-xpat || k>=xpaper) DISASTER(("x subrange"))
  x[i]=x[k];
  y[i]=y[k];

  edged:

  zlast=z[i];
  inc(i);
  } while (i<xpaper);

#if COLOR==0
loop (i,0,xpaper) if (pat[x[i]][y[i]]) lputpix(i,aty);
#elif COLOR==1
loop (i,0,xpaper) loop (jj,0,3) if (cympat[x[i]][y[i]][jj]) lputpix(jj,i,aty);
#else
loop (i,0,xpaper) lputpix(i,aty,pat[x[i]][y[i]]);
#endif
}

void printHP(void) /* ============================================= printHP */
{
#if COLOR==2
WARNING(("HP output not supported for COLOR=2"))
#else
static char name[12]="stereo.hp";
FILE *hp=fopen(name,"wb");
int i,jj;

fprintf(stderr,"writing %s\r",name);

fprintf(hp,
  "\33E\33&l0O\33&l6D\33&l66P\33&l0E\33&k0G\33&l0L%s\33*t%dR\33*p%dY\33*p%dX\33*r1A",
  paper==4?"\33&l26A":paper==0?"":"\33&l2A",resolution,xmargin,ymargin);

#if COLOR==1
loop (i,0,xpaper) {
  fprintf(hp,"\33*r-3U"); /* CYM system */
  loop (jj,0,3) {
    fprintf(hp,"\33*b%d%c",ypaper/8,'V'+(jj==2));
    fwrite(sheet[jj][i],ypaper/8,1,hp); }
  }
#else
loop (i,0,xpaper) {
  fprintf(hp,"\33*b%dW",ypaper/8);
  fwrite(sheet[i],ypaper/8,1,hp);
  }
#endif
fprintf(hp,"\33*rB\14"); /* end graphics, FF */
fclose(hp);
prt("%s written",name);
if (name[9]==0) name[9]='`';
name[9]++;
#endif
}

void printPPM(void) /* ============================================ printPPM */
{
static char name[12]="stereo.ppm";
FILE *f=fopen(name,"wb");
int i,j,k,ii;
int (*row)[3],xxx,yyy;

if (ppm<0 || ppm>16) WARNING(("%d is strange jitter square",ppm))
ppm=abs(ppm);

fprintf(stderr,"writing %s\r",name);

fprintf(f,"P6\n%d %d\n255\n",xxx=xpaper/ppm,yyy=ypaper/ppm);
alloczero(row,xxx*sizeof(row[0]));

loop (j,0,ypaper) {
  if (j%ppm==0) memset(row,0,xxx*sizeof(row[0]));
  loop (i,0,xpaper) loop (k,0,3)
    row[i/ppm][k]+=
#if COLOR==0
      255*!(sheet[i][j/8]&bit[j&7]);
#elif COLOR==1
      255*!(sheet[k][i][j/8]&bit[j&7]);
#else
      sheet[i][j][k];
#endif
  if (j%ppm==ppm-1) loop (ii,0,xxx) {
    unsigned char RGB[3];
    loop (k,0,3) RGB[k]=row[ii][k]/Sqr(ppm);
    fwrite(&RGB,1,3,f); }
  }
fclose(f);
prt("%s written",name);
if (name[9]==0) name[9]='`';
name[9]++;
free(row);
}

void printHEX(void) /* ============================================ printHEX */
{
#if COLOR>0
WARNING(("HEX output not supported for COLOR>0"))
#else
static char name[12]="stereo.hex";
FILE *hex=fopen(name,"wt");
/*static char dig[17]="0123456789ABCDEF";*/
static char dig[17]="FEDCBA9876543210";
int i,j;
int p;

fprintf(stderr,"writing %s\r",name);

loop (i,0,xpaper) {
  loop (j,0,ypaper/8) {
    if ((j&31)==0) fputc('\n',hex);
    p=sheet[i][j];
    fputc(dig[p/16],hex);
    fputc(dig[p&15],hex); }
  }
fclose(hex);
prt("%s written",name);
if (name[9]==0) name[9]='`';
name[9]++;
#endif
}

int wave=5,maxk=12;
#if COLOR==0
int wcol=1;
#elif COLOR==1
int wcol=3*36;
#else
int wcol[3]={-256,-256,-256}; /*?*/
#endif
#define NW 3

void waves(void) /* ================================================ waves */
/*
   periodic waved lines (sum of random cos terms) added to the pattern
*/
{
int i,n,m,x,y;
double d;
double ph[NW],am[NW],k[NW];

#if COLOR==0
int col=wcol;
if (wcol<-2) wcol=-2;
#elif COLOR==1
int col=wcol;
if (wcol<-216) wcol=-216;
#else
int iw;
rgb col;
loop (iw,0,3) if (wcol[iw]<-256) wcol[iw]=-256;
#endif

loop (i,0,wave) {
#if COLOR==2
  loop (iw,0,3) if (wcol[iw]<0) col[iw]=irnd(-wcol[iw]);
#else
  if (wcol<0) col=irnd(-wcol);
#endif

  loop (n,0,NW) {
    ph[n]=rndcos()*PI;
    am[n]=amplitude*(double)xpat/pow((double)NW,1.5)/wave;
    k[n]=irnd(maxk)+1; }

  loop (y,0,ypat) {
    d=0;
    loop (n,0,NW) d+=am[n]*cos(k[n]*(2*PI/ypat*y+ph[n]));
    x=(int)(d+(double)xpat/wave*i)+3*xpat;
    loop (n,0,line) loop (m,0,line)
#if COLOR==2
      loop (iw,0,3) pat[(x+n)%xpat][(y+m)%ypat][iw]=col[iw];
#else
      pat[(x+n)%xpat][(y+m)%ypat]=col;
#endif
    }
  }
}

int main(int narg,char **arg) /* ===================================== main */
{ 

byte **zbuf; /* [yzbuf][xzbuf]: note x:y order! */
int xzbuf=320,yzbuf=200;
int textzbuf=0; /* zbuf is in text format: 
                   after line of xzbuf chars scans until lf */
int mirror=0;  /* y-coord mirror */
int invert=0;  /* invert z-buf */
FILE *zb;
float zbufd=2,scale,ppmscale=1;
int i,j,jj,iin,dark,pattern=5,ncol=6;
int palette=0;
byte *z;
long rndseed=0;
int iarg=0;

initscroll(6*1024);

loop (i,0,1+(narg<2)) {

  if (i) in=stdin; 
  else in=fopen("stereo.def","rt");

  if (!in) prt("stereo.def not found");
  else getdata
    get(pattern)
#if COLOR==0
    getvec(p,,4)
#elif COLOR==1
    get(ncol)
#endif
    get(mc)
    get(amplitude)
    get(maxk)
    get(wave)
    get(line) get(up)
    get(xpat) get(ypat)
    get(paper)
    get(ypaper)
    get(xzbuf) get(yzbuf) get(textzbuf)
    get(mirror) get(invert)
    get(rightedge) get(zbufd)
    get(rightedgep)
    get(phase)
    get(resolution) get(aspect)
    get(xmargin) get(ymargin)
    get(hex) get(hp) get(ppm)
#if COLOR==2
    getvec(wcol,,3)
#else
    get(wcol)
#endif
    get(palette)
    get(rndseed)
    checkdata enddata
  }

if (palette) { mc=0; zbufd=1e9; }
rndinit(0,rndseed);

#if COLOR==1
{
  int C,Y,M,prob,ind;
  FILE *f=fopen("stereo.col","rt");
  char line[80];

  if (!f) ERROR(("missing stereo.col"))
  fgets(line,80,f);
  loop (i,0,ncol) {
    fgets(line,80,f);
    C=Y=M=prob=0;
    sscanf(line,"%d%d%d%d",&C,&Y,&M,&prob);
    ind=36*C+6*Y+M;
    if (ind<0 || ind>=216) { ERROR(("bad C Y M in stereo.col")) ind=0; }
    p[ind]+=prob; }
  fclose(f);
}
#endif

if (ypaper&15) {
  WARNING(("ypaper should be multiple of 16"))
  ypaper=(ypaper+15)/16*16;
  put(ypaper) }

if (hp) {
  for (i=75; i<=600; i*=2) if (resolution==i) goto resolutionOK;
    ERROR(("resolution=%d incompatible with HP LaserJet: use 75 150 300 600",resolution))
    resolution=300;
    put(resolution)
  resolutionOK:; }

if (ppm) ppmscale=1./abs(ppm);
prt("eye distance at background=%.2fcm",scale=2.54*xpat/resolution*ppmscale);
if (scale<1.7 || scale>7) WARNING(("strange eye distance (%g)",scale))

scale=(double)ypaper/yzbuf;
xpaper=scale/aspect*xzbuf+xpat+0.999;
put2(scale,xpaper)
prt("picture size (landscape) = %.2fcm x %.2fcm",
     xpaper*2.54/resolution*ppmscale,ypaper*2.54/resolution*ppmscale);

alloc(pat,sizeof(pat[0])*xpat);
loop (i,0,xpat)
  alloczero(pat[i],sizeof(pat[0][0])*ypat);

#if COLOR==1
alloc(cympat,sizeof(cympat[0])*xpat);
loop (i,0,xpat)
  alloczero(cympat[i],sizeof(cympat[0][0])*ypat);
#endif

if (abs(pattern)&1) rndpat();
if (abs(pattern)&2) filepat(narg,pattern<0);
if (abs(pattern)&4) waves();

#if COLOR==1
if (palette)
  /* palette test */
  loop (i,0,xpat) loop (j,0,ypat) {
    int C=i*6/xpat;
    int Y=(i*36/xpat)%6;
    int M=j*6/ypat;
    pat[i][j]=C*36+Y*6+M; }

loop (i,0,xpat) loop (j,0,ypat) {
  int black,sh;
  dark=pat[i][j];
  black=dark/36 && (dark/6)%6 && dark%6;
  loop (jj,0,3) {
    sh=jj*!black;
    switch (dark%6) {
      case 0: cympat[i][j][jj]=0; break;              /* 0 */
      case 1: cympat[i][j][jj]=!((i+sh+7*j) % 17); break; /* 1/17 */
      case 2: cympat[i][j][jj]=!((i+sh+4*j) % 11); break; /* 1/11 */
      case 3: cympat[i][j][jj]=!((i+sh+2*j) % 7);  break; /* 1/7 */
      case 4: cympat[i][j][jj]=(i+sh+((j+sh/2)/2)) && (j+sh/2) && 1; 
                                                        /* 1/4 */ break;
      case 5: cympat[i][j][jj]=(i+sh/2+j); /* 1/2 */ break;
/*      case 5: cympat[i][j][jj]=1;  1-too dark=removed */
      }
   dark/=6; }
  }
#elif COLOR==0
/* change `colored" pattern,black && white */
/* 0=white (background), 1=black */
loop (i,0,xpat) loop (j,0,ypat) switch (pat[i][j]) {
  case 2: pat[i][j]=(i+(j / 2)) && j && 1; /* 1/4 */ break;
  case 3: pat[i][j]=!((i+5*j) % 13);      /* 1/13 */
  }

/* show pattern */
loop (j,0,24) {
  loop (i,0,79) switch(pat[i][2*j]+2*pat[i][2*j+1]) {
    case 0: prtc(' '); break;
    case 1: prtc('_'); break;
    case 2: prtc('\"'); break;
    case 3: prtc('I'); }
  _n }
#endif

/* === z-buffer === */
alloc(zbuf,sizeof(zbuf[0])*yzbuf);
loop (i,0,yzbuf)
  alloc(zbuf[i],sizeof(zbuf[0][0])*xzbuf);

#if COLOR==1
loop (jj,0,3) {
  alloc(sheet[jj],sizeof(sheet[0][0])*xpaper);
  loop (i,0,xpaper) alloc(sheet[jj][i],sizeof(sheet[0][0][0])*ypaper/8); }
#else
alloc(sheet,sizeof(sheet[0])*xpaper);
loop (i,0,xpaper) alloc(sheet[i],sizeof(sheet[0][0])*ypaper/(COLOR==2?1:8));
#endif

alloc(z,xpaper*sizeof(z[0]));
alloc(x,(xpaper+xpat)*sizeof(x[0])); x+=xpat;
alloc(y,(xpaper+xpat)*sizeof(y[0])); y+=xpat;

for (;;) {

#if COLOR==1
loop (jj,0,3) loop (i,0,xpaper)
  memset(sheet[jj][i],0,sizeof(sheet[0][0][0])*ypaper/8);
#else
loop (i,0,xpaper) memset(sheet[i],0,sizeof(sheet[0][0])*ypaper/(COLOR==2?1:8));
#endif

again:
if (narg>1) {
  iarg++;
  if (iarg<narg) strcpy(fn,arg[iarg]);
  else exit(0); }
else {
  prt_("z-buffer file (\".\"=exit) = ");
  scanf("%s",fn);
  if (!strcmp(fn,".")) exit(0); }

zb=fopen(fn,textzbuf?"rt":"rb");
if (zb==NULL) {
  ERROR(("no such file"))
  goto again; }

/* reading the z-buffer file */
/* note that in the z-buf file y=0 is the top line - mirror! */
{
int minz=255,maxz=0,j,ii;
loop (ii,0,yzbuf) {
  if (mirror) i=ii; else i=yzbuf-1-ii;
  if (fread(zbuf[i],xzbuf,1,zb)!=1) ERROR(("z-buf file too short"))
  if (invert) loop (j,0,xzbuf) zbuf[i][j]=~zbuf[i][j];
  if (textzbuf) do {
    j=fgetc(zb);
    if (j<0) ERROR(("z-buf file too short"))
    } while (j!='\n');
  loop (j,0,xzbuf) {
    Min(minz,zbuf[i][j]); 
    Max(maxz,zbuf[i][j]) } }
fclose(zb);
loop (i,0,yzbuf) loop (j,0,xzbuf) zbuf[i][j]-=minz;
prt_("%s: ",fn); put2(minz,maxz)
}

/* generating the stereogram by rows of pixels */
loop (j,0,ypaper) {
  jj=j / scale;
  if (jj>=yzbuf) ERROR(("zbuf y range"))
  loop (i,0,xpat) z[i]=0;
  loop (i,xpat,xpaper) {
    iin=(i-xpat) / scale*aspect;
    if (iin<xzbuf) z[i]=zbuf[jj][iin]/zbufd;
    else z[i]=0; }
  row(j,z);
  }

prt("stereogram calculated     ");

if (ppm) printPPM();
if (hex) printHEX(); 
if (hp) printHP();
}
}
